@@ -134,6 +134,60 @@ snaps:
     presence: optional
 OTHERgrade: secured
 storage-safety: encrypted
+` + "TSLINE" +
+		"body-length: 0\n" +
+		"sign-key-sha3-384: Jv8_JiHiIzJVcO9M55pPdqSDWUvuhfDIBJUS-3VW7F_idjix7Ffn5qMxB21ZQuij" +
+		"\n\n" +
+		"AXNpZw=="
+
+	classicModelWithSnapsExample = `type: model
+authority-id: brand-id1
+series: 16
+brand-id: brand-id1
+model: baz-3000
+display-name: Baz 3000
+architecture: amd64
+system-user-authority: *
+base: core20
+classic: true
+distribution: ubuntu
+store: brand-store
+snaps:
+  -
+    name: baz-linux
+    id: bazlinuxidididididididididididid
+    type: kernel
+    default-channel: 20
+  -
+    name: brand-gadget
+    id: brandgadgetdidididididididididid
+    type: gadget
+  -
+    name: other-base
+    id: otherbasedididididididididididid
+    type: base
+    modes:
+      - run
+    presence: required
+  -
+    name: nm
+    id: nmididididididididididididididid
+    modes:
+      - ephemeral
+      - run
+    default-channel: 1.0
+  -
+    name: myapp
+    id: myappdididididididididididididid
+    type: app
+    default-channel: 2.0
+  -
+    name: myappopt
+    id: myappoptidididididididididididid
+    type: app
+    presence: optional
+OTHERgrade: secured
+storage-safety: encrypted
 ` + "TSLINE" +
 		"body-length: 0\n" +
 		"sign-key-sha3-384: Jv8_JiHiIzJVcO9M55pPdqSDWUvuhfDIBJUS-3VW7F_idjix7Ffn5qMxB21ZQuij" +
@@ -521,7 +575,7 @@ func (mods *modelSuite) TestModelCheckInconsistentTimestamp(c *C) {
 	c.Assert(err, IsNil)
 
 	err = db.Check(model)
-	c.Assert(err, ErrorMatches, `model assertion timestamp outside of signing key validity \(key valid since.*\)`)
+	c.Assert(err, ErrorMatches, `model assertion timestamp "2011-01-01 14:00:00 \+0000 UTC" outside of signing key validity \(key valid since.*\)`)
 }
 
 func (mods *modelSuite) TestClassicDecodeOK(c *C) {
@@ -537,6 +591,7 @@ func (mods *modelSuite) TestClassicDecodeOK(c *C) {
 	c.Check(model.Model(), Equals, "baz-3000")
 	c.Check(model.DisplayName(), Equals, "Baz 3000")
 	c.Check(model.Classic(), Equals, true)
+	c.Check(model.Distribution(), Equals, "")
 	c.Check(model.Architecture(), Equals, "amd64")
 	c.Check(model.GadgetSnap(), DeepEquals, &asserts.ModelSnap{
 		Name:     "brand-gadget",
@@ -595,8 +650,8 @@ func (mods *modelSuite) TestClassicDecodeInvalid(c *C) {
 		{"classic: true\n", "classic: foo\n", `"classic" header must be 'true' or 'false'`},
 		{"architecture: amd64\n", "architecture:\n  - foo\n", `"architecture" header must be a string`},
 		{"gadget: brand-gadget\n", "gadget:\n  - foo\n", `"gadget" header must be a string`},
-		{"gadget: brand-gadget\n", "kernel: brand-kernel\n", `cannot specify a kernel with a classic model`},
-		{"gadget: brand-gadget\n", "base: some-base\n", `cannot specify a base with a classic model`},
+		{"gadget: brand-gadget\n", "kernel: brand-kernel\n", `cannot specify a kernel with a non-extended classic model`},
+		{"gadget: brand-gadget\n", "base: some-base\n", `cannot specify a base with a non-extended classic model`},
 		{"gadget: brand-gadget\n", "gadget:\n  - xyz\n", `"gadget" header must be a string`},
 	}
 
@@ -616,14 +671,29 @@ func (mods *modelSuite) TestClassicDecodeGadgetAndArchOptional(c *C) {
 	c.Check(a.Type(), Equals, asserts.ModelType)
 	model := a.(*asserts.Model)
 	c.Check(model.Classic(), Equals, true)
+	c.Check(model.Distribution(), Equals, "")
 	c.Check(model.Architecture(), Equals, "")
 	c.Check(model.GadgetSnap(), IsNil)
 	c.Check(model.Gadget(), Equals, "")
 	c.Check(model.GadgetTrack(), Equals, "")
 }
 
-func (mods *modelSuite) TestCore20DecodeOK(c *C) {
-	encoded := strings.Replace(core20ModelExample, "TSLINE", mods.tsLine, 1)
+func (mods *modelSuite) TestWithSnapsDecodeOK(c *C) {
+	tt := []struct {
+		modelRaw  string
+		isClassic bool
+	}{
+		{modelRaw: core20ModelExample, isClassic: false},
+		{modelRaw: classicModelWithSnapsExample, isClassic: true},
+	}
+
+	for _, t := range tt {
+		mods.testWithSnapsDecodeOK(c, t.modelRaw, t.isClassic)
+	}
+}
+
+func (mods *modelSuite) testWithSnapsDecodeOK(c *C, modelRaw string, isClassic bool) {
+	encoded := strings.Replace(modelRaw, "TSLINE", mods.tsLine, 1)
 	encoded = strings.Replace(encoded, "OTHER", "", 1)
 	a, err := asserts.Decode([]byte(encoded))
 	c.Assert(err, IsNil)
@@ -636,6 +706,12 @@ func (mods *modelSuite) TestCore20DecodeOK(c *C) {
 	c.Check(model.Model(), Equals, "baz-3000")
 	c.Check(model.DisplayName(), Equals, "Baz 3000")
 	c.Check(model.Architecture(), Equals, "amd64")
+	c.Check(model.Classic(), Equals, isClassic)
+	if isClassic {
+		c.Check(model.Distribution(), Equals, "ubuntu")
+	} else {
+		c.Check(model.Distribution(), Equals, "")
+	}
 	c.Check(model.GadgetSnap(), DeepEquals, &asserts.ModelSnap{
 		Name:           "brand-gadget",
 		SnapID:         "brandgadgetdidididididididididid",
@@ -892,15 +968,28 @@ func (mods *modelSuite) TestCore20DefaultStorageSafetySignedDangerous(c *C) {
 	}
 }
 
-func (mods *modelSuite) TestCore20DecodeInvalid(c *C) {
-	encoded := strings.Replace(core20ModelExample, "TSLINE", mods.tsLine, 1)
+func (mods *modelSuite) TestWithSnapsDecodeInvalid(c *C) {
+	tt := []struct {
+		modelRaw  string
+		isClassic bool
+	}{
+		{modelRaw: core20ModelExample, isClassic: false},
+		{modelRaw: classicModelWithSnapsExample, isClassic: true},
+	}
+
+	for _, t := range tt {
+		mods.testWithSnapsDecodeInvalid(c, t.modelRaw, t.isClassic)
+	}
+}
+
+func (mods *modelSuite) testWithSnapsDecodeInvalid(c *C, modelRaw string, isClassic bool) {
+	encoded := strings.Replace(modelRaw, "TSLINE", mods.tsLine, 1)
 
 	snapsStanza := encoded[strings.Index(encoded, "snaps:"):strings.Index(encoded, "grade:")]
 
 	invalidTests := []struct{ original, invalid, expectedErr string }{
 		{"base: core20\n", "", `"base" header is mandatory`},
 		{"base: core20\n", "base: alt-base\n", `cannot specify not well-known base "alt-base" without a corresponding "snaps" header entry`},
-		{"OTHER", "classic: true\n", `cannot use extended snaps header for a classic model \(yet\)`},
 		{snapsStanza, "snaps: snap\n", `"snaps" header must be a list of maps`},
 		{snapsStanza, "snaps:\n  - snap\n", `"snaps" header must be a list of maps`},
 		{"name: myapp\n", "other: 1\n", `"name" of snap is mandatory`},
@@ -933,6 +1022,19 @@ func (mods *modelSuite) TestCore20DecodeInvalid(c *C) {
 		{"storage-safety: encrypted\n", "storage-safety: foo\n", `storage-safety for model must be encrypted\|prefer-encrypted\|prefer-unencrypted, not "foo"`},
 		{"storage-safety: encrypted\n", "storage-safety: prefer-unencrypted\n", `secured grade model must not have storage-safety overridden, only "encrypted" is valid`},
 	}
+	if isClassic {
+		classicInvalid := []struct{ original, invalid, expectedErr string }{
+			{"distribution: ubuntu\n", "", `"distribution" header is mandatory, see distribution ID in os-release spec`},
+			{"distribution: ubuntu\n", "distribution: Ubuntu\n", `"distribution" header contains invalid characters: "Ubuntu", see distribution ID in os-release spec`},
+			{"distribution: ubuntu\n", "distribution: *buntu\n", `"distribution" header contains invalid characters: "\*buntu", see distribution ID in os-release spec`},
+		}
+		invalidTests = append(invalidTests, classicInvalid...)
+	} else {
+		coreInvalid := []struct{ original, invalid, expectedErr string }{
+			{"OTHER", "distribution: ubuntu\n", `cannot specify distribution for model unless it is classic and has an extended snaps header`},
+		}
+		invalidTests = append(invalidTests, coreInvalid...)
+	}
 	for _, test := range invalidTests {
 		invalid := strings.Replace(encoded, test.original, test.invalid, 1)
 		invalid = strings.Replace(invalid, "OTHER", "", 1)
