@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2015-2017 Canonical Ltd
+ * Copyright (C) 2015-2022 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -106,6 +106,110 @@ func (sds *snapDeclSuite) TestDecodeOK(c *C) {
 		"Cmd-3": "cmd-3",
 		"CMD.4": "cmd-4",
 	})
+	c.Check(snapDecl.RevisionAuthority(""), IsNil)
+}
+
+func (sds *snapDeclSuite) TestDecodeOKWithRevisionAuthority(c *C) {
+	encoded := "type: snap-declaration\n" +
+		"authority-id: canonical\n" +
+		"series: 16\n" +
+		"snap-id: snap-id-1\n" +
+		"snap-name: first\n" +
+		"publisher-id: dev-id1\n" +
+		"refresh-control:\n  - foo\n  - bar\n" +
+		sds.tsLine +
+		`revision-authority:
+  -
+    account-id: delegated-acc-id
+    provenance:
+      - prov1
+      - prov2
+    min-revision: 100
+    max-revision: 1000000
+    on-store:
+      - store1
+` +
+		"body-length: 0\n" +
+		"sign-key-sha3-384: Jv8_JiHiIzJVcO9M55pPdqSDWUvuhfDIBJUS-3VW7F_idjix7Ffn5qMxB21ZQuij" +
+		"\n\n" +
+		"AXNpZw=="
+	a, err := asserts.Decode([]byte(encoded))
+	c.Assert(err, IsNil)
+	c.Check(a.Type(), Equals, asserts.SnapDeclarationType)
+	snapDecl := a.(*asserts.SnapDeclaration)
+	c.Check(snapDecl.AuthorityID(), Equals, "canonical")
+	c.Check(snapDecl.Timestamp(), Equals, sds.ts)
+	c.Check(snapDecl.Series(), Equals, "16")
+	c.Check(snapDecl.SnapID(), Equals, "snap-id-1")
+	c.Check(snapDecl.SnapName(), Equals, "first")
+	c.Check(snapDecl.PublisherID(), Equals, "dev-id1")
+	c.Check(snapDecl.RefreshControl(), DeepEquals, []string{"foo", "bar"})
+	ras := snapDecl.RevisionAuthority("prov1")
+	c.Check(ras, DeepEquals, []*asserts.RevisionAuthority{
+		{
+			AccountID:   "delegated-acc-id",
+			Provenance:  []string{"prov1", "prov2"},
+			MinRevision: 100,
+			MaxRevision: 1000000,
+			DeviceScope: &asserts.DeviceScopeConstraint{
+				Store: []string{"store1"},
+			},
+		},
+	})
+}
+
+func (sds *snapDeclSuite) TestDecodeOKWithRevisionAuthorityDefaults(c *C) {
+	initial := "type: snap-declaration\n" +
+		"authority-id: canonical\n" +
+		"series: 16\n" +
+		"snap-id: snap-id-1\n" +
+		"snap-name: first\n" +
+		"publisher-id: dev-id1\n" +
+		"refresh-control:\n  - foo\n  - bar\n" +
+		sds.tsLine +
+		`revision-authority:
+  -
+    account-id: delegated-acc-id
+    provenance:
+      - prov1
+      - prov2
+    min-revision: 100
+` +
+		"body-length: 0\n" +
+		"sign-key-sha3-384: Jv8_JiHiIzJVcO9M55pPdqSDWUvuhfDIBJUS-3VW7F_idjix7Ffn5qMxB21ZQuij" +
+		"\n\n" +
+		"AXNpZw=="
+	tests := []struct {
+		original, replaced string
+		revAuth            asserts.RevisionAuthority
+	}{
+		{"min", "min", asserts.RevisionAuthority{
+			AccountID:   "delegated-acc-id",
+			Provenance:  []string{"prov1", "prov2"},
+			MinRevision: 100,
+		}},
+		{"min", "max", asserts.RevisionAuthority{
+			AccountID:   "delegated-acc-id",
+			Provenance:  []string{"prov1", "prov2"},
+			MinRevision: 1,
+			MaxRevision: 100,
+		}},
+		{"    min-revision: 100\n", "", asserts.RevisionAuthority{
+			AccountID:   "delegated-acc-id",
+			Provenance:  []string{"prov1", "prov2"},
+			MinRevision: 1,
+		}},
+	}
+
+	for _, t := range tests {
+		encoded := strings.Replace(initial, t.original, t.replaced, 1)
+		a, err := asserts.Decode([]byte(encoded))
+		c.Assert(err, IsNil)
+		snapDecl := a.(*asserts.SnapDeclaration)
+		ras := snapDecl.RevisionAuthority("prov2")
+		c.Check(ras, HasLen, 1)
+		c.Check(*ras[0], DeepEquals, t.revAuth)
+	}
 }
 
 func (sds *snapDeclSuite) TestEmptySnapName(c *C) {
@@ -207,6 +311,52 @@ func (sds *snapDeclSuite) TestDecodeInvalid(c *C) {
 
 }
 
+func (sds *snapDeclSuite) TestDecodeInvalidWithRevisionAuthority(c *C) {
+	const revAuth = `revision-authority:
+  -
+    account-id: delegated-acc-id
+    provenance:
+      - prov1
+      - prov2
+    min-revision: 100
+    max-revision: 1000000
+    on-store:
+      - store1
+`
+	encoded := "type: snap-declaration\n" +
+		"authority-id: canonical\n" +
+		"series: 16\n" +
+		"snap-id: snap-id-1\n" +
+		"snap-name: first\n" +
+		"publisher-id: dev-id1\n" +
+		"refresh-control:\n  - foo\n  - bar\n" +
+		sds.tsLine +
+		revAuth +
+		"body-length: 0\n" +
+		"sign-key-sha3-384: Jv8_JiHiIzJVcO9M55pPdqSDWUvuhfDIBJUS-3VW7F_idjix7Ffn5qMxB21ZQuij" +
+		"\n\n" +
+		"AXNpZw=="
+
+	invalidTests := []struct{ original, invalid, expectedErr string }{
+		{revAuth, "revision-authority: x\n", `revision-authority stanza must be a list of maps`},
+		{revAuth, "revision-authority:\n  - x\n", `revision-authority stanza must be a list of maps`},
+		{"    account-id: delegated-acc-id\n", "", `"account-id" in revision authority is mandatory`},
+		{"account-id: delegated-acc-id\n", "account-id: *\n", `"account-id" in revision authority contains invalid characters: "\*"`},
+		{"    provenance:\n      - prov1\n      - prov2\n", "    provenance: \n", `provenance in revision authority must be a list of strings`},
+		{"prov2\n", "*\n", `provenance in revision authority contains an invalid element: "\*"`},
+		{"    min-revision: 100\n", "    min-revision: 0\n", `"min-revision" in revision authority must be >=1: 0`},
+		{"    max-revision: 1000000\n", "    max-revision: 0\n", `"max-revision" in revision authority must be >=1: 0`},
+		{"    max-revision: 1000000\n", "    max-revision: 10\n", `optional max-revision cannot be less than min-revision in revision-authority`},
+		{"    on-store:\n      - store1\n", "    on-store: foo", `on-store in revision-authority must be a list of strings`},
+	}
+
+	for _, test := range invalidTests {
+		invalid := strings.Replace(encoded, test.original, test.invalid, 1)
+		_, err := asserts.Decode([]byte(invalid))
+		c.Check(err, ErrorMatches, snapDeclErrPrefix+test.expectedErr)
+	}
+}
+
 func (sds *snapDeclSuite) TestDecodePlugsAndSlots(c *C) {
 	encoded := `type: snap-declaration
 format: 1
@@ -546,6 +696,24 @@ func (sds *snapDeclSuite) TestSuggestedFormat(c *C) {
 			c.Check(fmtnum, Equals, 4)
 		}
 	}
+
+	// alt matcher (so far unused) => format 5
+	for _, sidePrefix := range []string{"plug", "slot"} {
+		headers = map[string]interface{}{
+			sidePrefix + "s": map[string]interface{}{
+				"interface5": map[string]interface{}{
+					"allow-auto-connection": map[string]interface{}{
+						sidePrefix + "-attributes": map[string]interface{}{
+							"x": []interface{}{"alt1", "alt2"}, // alt matcher
+						},
+					},
+				},
+			},
+		}
+		fmtnum, err = asserts.SuggestFormat(asserts.SnapDeclarationType, headers, nil)
+		c.Assert(err, IsNil)
+		c.Check(fmtnum, Equals, 5)
+	}
 }
 
 func prereqDevAccount(c *C, storeDB assertstest.SignerDB, db *asserts.Database) {
@@ -805,7 +973,7 @@ func (sbs *snapBuildSuite) TestSnapBuildCheckInconsistentTimestamp(c *C) {
 	c.Assert(err, IsNil)
 
 	err = db.Check(snapBuild)
-	c.Assert(err, ErrorMatches, `snap-build assertion timestamp outside of signing key validity \(key valid since.*\)`)
+	c.Assert(err, ErrorMatches, `snap-build assertion timestamp "2013-01-01 14:00:00 \+0000 UTC" outside of signing key validity \(key valid since.*\)`)
 }
 
 type snapRevSuite struct {
@@ -834,7 +1002,7 @@ func (srs *snapRevSuite) makeValidEncoded() string {
 		"AXNpZw=="
 }
 
-func (srs *snapRevSuite) makeHeaders(overrides map[string]interface{}) map[string]interface{} {
+func makeSnapRevisionHeaders(overrides map[string]interface{}) map[string]interface{} {
 	headers := map[string]interface{}{
 		"authority-id":  "canonical",
 		"snap-sha3-384": blobSHA3_384,
@@ -851,6 +1019,10 @@ func (srs *snapRevSuite) makeHeaders(overrides map[string]interface{}) map[strin
 	return headers
 }
 
+func (srs *snapRevSuite) makeHeaders(overrides map[string]interface{}) map[string]interface{} {
+	return makeSnapRevisionHeaders(overrides)
+}
+
 func (srs *snapRevSuite) TestDecodeOK(c *C) {
 	encoded := srs.makeValidEncoded()
 	a, err := asserts.Decode([]byte(encoded))
@@ -865,6 +1037,25 @@ func (srs *snapRevSuite) TestDecodeOK(c *C) {
 	c.Check(snapRev.SnapRevision(), Equals, 1)
 	c.Check(snapRev.DeveloperID(), Equals, "dev-id1")
 	c.Check(snapRev.Revision(), Equals, 1)
+	c.Check(snapRev.Provenance(), Equals, "global-upload")
+}
+
+func (srs *snapRevSuite) TestDecodeOKWithProvenance(c *C) {
+	encoded := srs.makeValidEncoded()
+	encoded = strings.Replace(encoded, "snap-id: snap-id-1", "provenance: foo\nsnap-id: snap-id-1", 1)
+	a, err := asserts.Decode([]byte(encoded))
+	c.Assert(err, IsNil)
+	c.Check(a.Type(), Equals, asserts.SnapRevisionType)
+	snapRev := a.(*asserts.SnapRevision)
+	c.Check(snapRev.AuthorityID(), Equals, "store-id1")
+	c.Check(snapRev.Timestamp(), Equals, srs.ts)
+	c.Check(snapRev.SnapID(), Equals, "snap-id-1")
+	c.Check(snapRev.SnapSHA3_384(), Equals, blobSHA3_384)
+	c.Check(snapRev.SnapSize(), Equals, uint64(123))
+	c.Check(snapRev.SnapRevision(), Equals, 1)
+	c.Check(snapRev.DeveloperID(), Equals, "dev-id1")
+	c.Check(snapRev.Revision(), Equals, 1)
+	c.Check(snapRev.Provenance(), Equals, "foo")
 }
 
 const (
@@ -882,6 +1073,8 @@ func (srs *snapRevSuite) TestDecodeInvalid(c *C) {
 		{digestHdr, "snap-sha3-384: \n", `"snap-sha3-384" header should not be empty`},
 		{digestHdr, "snap-sha3-384: #\n", `"snap-sha3-384" header cannot be decoded:.*`},
 		{digestHdr, "snap-sha3-384: eHl6\n", `"snap-sha3-384" header does not have the expected bit length: 24`},
+		{"snap-id: snap-id-1\n", "provenance: \nsnap-id: snap-id-1\n", `"provenance" header should not be empty`},
+		{"snap-id: snap-id-1\n", "provenance: *\nsnap-id: snap-id-1\n", `"provenance" header contains invalid characters: "\*"`},
 		{"snap-size: 123\n", "", `"snap-size" header is mandatory`},
 		{"snap-size: 123\n", "snap-size: \n", `"snap-size" header should not be empty`},
 		{"snap-size: 123\n", "snap-size: -1\n", `"snap-size" header is not an unsigned integer: -1`},
@@ -942,7 +1135,7 @@ func (srs *snapRevSuite) TestSnapRevisionCheckInconsistentTimestamp(c *C) {
 	c.Assert(err, IsNil)
 
 	err = db.Check(snapRev)
-	c.Assert(err, ErrorMatches, `snap-revision assertion timestamp outside of signing key validity \(key valid since.*\)`)
+	c.Assert(err, ErrorMatches, `snap-revision assertion timestamp "2013-01-01 14:00:00 \+0000 UTC" outside of signing key validity \(key valid since.*\)`)
 }
 
 func (srs *snapRevSuite) TestSnapRevisionCheckUntrustedAuthority(c *C) {
@@ -984,6 +1177,329 @@ func (srs *snapRevSuite) TestSnapRevisionCheckMissingDeclaration(c *C) {
 	c.Assert(err, ErrorMatches, `snap-revision assertion for snap id "snap-id-1" does not have a matching snap-declaration assertion`)
 }
 
+func (srs *snapRevSuite) TestRevisionAuthorityCheck(c *C) {
+	storeDB, db := makeStoreAndCheckDB(c)
+
+	delegatedDB := setup3rdPartySigning(c, "delegated-id", storeDB, db)
+	headers := srs.makeHeaders(map[string]interface{}{
+		"authority-id":  "delegated-id",
+		"developer-id":  "delegated-id",
+		"snap-revision": "200",
+		"provenance":    "prov1",
+	})
+	a, err := delegatedDB.Sign(asserts.SnapRevisionType, headers, nil, "")
+	c.Assert(err, IsNil)
+	snapRev := a.(*asserts.SnapRevision)
+
+	tests := []struct {
+		revAuth asserts.RevisionAuthority
+		err     string
+	}{
+		{asserts.RevisionAuthority{
+			AccountID:   "delegated-id",
+			Provenance:  []string{"prov1", "prov2"},
+			MinRevision: 1,
+		}, ""},
+		{asserts.RevisionAuthority{
+			AccountID:   "delegated-id",
+			Provenance:  []string{"prov1", "prov2"},
+			MinRevision: 1,
+			MaxRevision: 1000,
+		}, ""},
+		{asserts.RevisionAuthority{
+			AccountID:   "delegated-id",
+			Provenance:  []string{"prov2"},
+			MinRevision: 1,
+			MaxRevision: 1000,
+		}, "provenance mismatch"},
+		{asserts.RevisionAuthority{
+			AccountID:   "delegated-id-2",
+			Provenance:  []string{"prov1", "prov2"},
+			MinRevision: 1,
+			MaxRevision: 1000,
+		}, "authority-id mismatch"},
+		{asserts.RevisionAuthority{
+			AccountID:   "delegated-id",
+			Provenance:  []string{"prov1", "prov2"},
+			MinRevision: 1000,
+		}, "snap revision 200 is less than min-revision 1000"},
+		{asserts.RevisionAuthority{
+			AccountID:   "delegated-id",
+			Provenance:  []string{"prov1", "prov2"},
+			MinRevision: 10,
+			MaxRevision: 110,
+		}, "snap revision 200 is greater than max-revision 110"},
+	}
+
+	for _, t := range tests {
+		err := t.revAuth.Check(snapRev, nil, nil)
+		if t.err == "" {
+			c.Check(err, IsNil)
+		} else {
+			c.Check(err, ErrorMatches, t.err)
+		}
+	}
+}
+
+func (srs *snapRevSuite) TestRevisionAuthorityCheckDeviceScope(c *C) {
+	a, err := asserts.Decode([]byte(`type: model
+authority-id: my-brand
+series: 16
+brand-id: my-brand
+model: my-model
+store: substore
+architecture: armhf
+kernel: krnl
+gadget: gadget
+timestamp: 2018-09-12T12:00:00Z
+sign-key-sha3-384: Jv8_JiHiIzJVcO9M55pPdqSDWUvuhfDIBJUS-3VW7F_idjix7Ffn5qMxB21ZQuij
+
+AXNpZw==`))
+	c.Assert(err, IsNil)
+	myModel := a.(*asserts.Model)
+
+	a, err = asserts.Decode([]byte(`type: store
+store: substore
+authority-id: canonical
+operator-id: canonical
+friendly-stores:
+  - a-store
+  - store1
+  - store2
+timestamp: 2018-09-12T12:00:00Z
+sign-key-sha3-384: Jv8_JiHiIzJVcO9M55pPdqSDWUvuhfDIBJUS-3VW7F_idjix7Ffn5qMxB21ZQuij
+
+AXNpZw==`))
+	c.Assert(err, IsNil)
+	substore := a.(*asserts.Store)
+
+	storeDB, db := makeStoreAndCheckDB(c)
+
+	delegatedDB := setup3rdPartySigning(c, "my-brand", storeDB, db)
+	headers := srs.makeHeaders(map[string]interface{}{
+		"authority-id":  "my-brand",
+		"developer-id":  "my-brand",
+		"snap-revision": "200",
+		"provenance":    "prov1",
+	})
+	a, err = delegatedDB.Sign(asserts.SnapRevisionType, headers, nil, "")
+	c.Assert(err, IsNil)
+	snapRev := a.(*asserts.SnapRevision)
+
+	tests := []struct {
+		revAuth  asserts.RevisionAuthority
+		substore *asserts.Store
+		err      string
+	}{
+		{asserts.RevisionAuthority{
+			AccountID:   "my-brand",
+			Provenance:  []string{"prov1", "prov2"},
+			MinRevision: 1,
+		}, nil, ""},
+		{asserts.RevisionAuthority{
+			AccountID:   "my-brand",
+			Provenance:  []string{"prov1", "prov2"},
+			MinRevision: 1,
+			DeviceScope: &asserts.DeviceScopeConstraint{
+				Store: []string{"other-store"},
+			},
+		}, nil, "on-store mismatch"},
+		{asserts.RevisionAuthority{
+			AccountID:   "my-brand",
+			Provenance:  []string{"prov1", "prov2"},
+			MinRevision: 1,
+			DeviceScope: &asserts.DeviceScopeConstraint{
+				Store: []string{"substore"},
+			},
+		}, nil, ""},
+		{asserts.RevisionAuthority{
+			AccountID:   "my-brand",
+			Provenance:  []string{"prov1", "prov2"},
+			MinRevision: 1,
+			DeviceScope: &asserts.DeviceScopeConstraint{
+				Store: []string{"substore"},
+			},
+		}, substore, ""},
+		{asserts.RevisionAuthority{
+			AccountID:   "my-brand",
+			Provenance:  []string{"prov1", "prov2"},
+			MinRevision: 1,
+			DeviceScope: &asserts.DeviceScopeConstraint{
+				Store: []string{"a-store"},
+			},
+		}, substore, ""},
+		{asserts.RevisionAuthority{
+			AccountID:   "my-brand",
+			Provenance:  []string{"prov1", "prov2"},
+			MinRevision: 1,
+			DeviceScope: &asserts.DeviceScopeConstraint{
+				Store: []string{"store1"},
+			},
+		}, nil, "on-store mismatch"},
+		{asserts.RevisionAuthority{
+			AccountID:   "my-brand",
+			Provenance:  []string{"prov1", "prov2"},
+			MinRevision: 1,
+			DeviceScope: &asserts.DeviceScopeConstraint{
+				Store: []string{"store1", "other-store"},
+			},
+		}, substore, ""},
+	}
+
+	for _, t := range tests {
+		err := t.revAuth.Check(snapRev, myModel, t.substore)
+		if t.err == "" {
+			c.Check(err, IsNil)
+		} else {
+			c.Check(err, ErrorMatches, t.err)
+		}
+	}
+}
+
+func (srs *snapRevSuite) TestSnapRevisionDelegation(c *C) {
+	storeDB, db := makeStoreAndCheckDB(c)
+
+	delegatedDB := setup3rdPartySigning(c, "delegated-id", storeDB, db)
+
+	snapDecl, err := storeDB.Sign(asserts.SnapDeclarationType, map[string]interface{}{
+		"series":       "16",
+		"snap-id":      "snap-id-1",
+		"snap-name":    "foo",
+		"publisher-id": "delegated-id",
+		"timestamp":    time.Now().Format(time.RFC3339),
+	}, nil, "")
+	c.Assert(err, IsNil)
+	err = db.Add(snapDecl)
+	c.Assert(err, IsNil)
+
+	headers := srs.makeHeaders(map[string]interface{}{
+		"authority-id": "delegated-id",
+		"developer-id": "delegated-id",
+		"provenance":   "prov1",
+	})
+	snapRev, err := delegatedDB.Sign(asserts.SnapRevisionType, headers, nil, "")
+	c.Assert(err, IsNil)
+
+	err = db.Check(snapRev)
+	c.Check(err, ErrorMatches, `snap-revision assertion with provenance "prov1" for snap id "snap-id-1" is not signed by an authorized authority: delegated-id`)
+
+	// establish delegation
+	snapDecl, err = storeDB.Sign(asserts.SnapDeclarationType, map[string]interface{}{
+		"series":       "16",
+		"snap-id":      "snap-id-1",
+		"snap-name":    "foo",
+		"publisher-id": "delegated-id",
+		"revision":     "1",
+		"revision-authority": []interface{}{
+			map[string]interface{}{
+				"account-id": "delegated-id",
+				"provenance": []interface{}{
+					"prov1",
+				},
+				// present but not checked at this level
+				"on-store": []interface{}{
+					"store1",
+				},
+			},
+		},
+		"timestamp": time.Now().Format(time.RFC3339),
+	}, nil, "")
+	c.Assert(err, IsNil)
+	err = db.Add(snapDecl)
+	c.Assert(err, IsNil)
+
+	// now revision should be accepted
+	err = db.Check(snapRev)
+	c.Check(err, IsNil)
+}
+
+func (srs *snapRevSuite) TestSnapRevisionDelegationRevisionOutOfRange(c *C) {
+	storeDB, db := makeStoreAndCheckDB(c)
+
+	delegatedDB := setup3rdPartySigning(c, "delegated-id", storeDB, db)
+
+	// establish delegation
+	snapDecl, err := storeDB.Sign(asserts.SnapDeclarationType, map[string]interface{}{
+		"series":       "16",
+		"snap-id":      "snap-id-1",
+		"snap-name":    "foo",
+		"publisher-id": "delegated-id",
+		"revision-authority": []interface{}{
+			map[string]interface{}{
+				"account-id": "delegated-id",
+				"provenance": []interface{}{
+					"prov1",
+				},
+				// present but not checked at this level
+				"on-store": []interface{}{
+					"store1",
+				},
+				"max-revision": "200",
+			},
+		},
+		"timestamp": time.Now().Format(time.RFC3339),
+	}, nil, "")
+	c.Assert(err, IsNil)
+	err = db.Add(snapDecl)
+	c.Assert(err, IsNil)
+
+	headers := srs.makeHeaders(map[string]interface{}{
+		"authority-id":  "delegated-id",
+		"developer-id":  "delegated-id",
+		"provenance":    "prov1",
+		"snap-revision": "1000",
+	})
+	snapRev, err := delegatedDB.Sign(asserts.SnapRevisionType, headers, nil, "")
+	c.Assert(err, IsNil)
+
+	err = db.Check(snapRev)
+	c.Check(err, ErrorMatches, `snap-revision assertion with provenance "prov1" for snap id "snap-id-1" is not signed by an authorized authority: delegated-id`)
+}
+
+func (srs *snapRevSuite) TestSnapRevisionDelegationInconsistentTimestamp(c *C) {
+	c.Skip("authority-delegation disabled")
+
+	storeDB, db := makeStoreAndCheckDB(c)
+
+	prereqDevAccount(c, storeDB, db)
+	prereqSnapDecl(c, storeDB, db)
+
+	otherDB := setup3rdPartySigning(c, "other", storeDB, db)
+
+	headers := srs.makeHeaders(map[string]interface{}{
+		"authority-id": "canonical",
+		"signatory-id": "other",
+		"timestamp":    time.Now().Format(time.RFC3339),
+	})
+	snapRev, err := otherDB.Sign(asserts.SnapRevisionType, headers, nil, "")
+	c.Assert(err, IsNil)
+
+	// move forward in time
+	time.Sleep(100 * time.Millisecond)
+
+	// now add authority-delegation
+	headers = map[string]interface{}{
+		"authority-id": "canonical",
+		"account-id":   "canonical",
+		"delegate-id":  "other",
+		"assertions": []interface{}{
+			map[string]interface{}{
+				"type": "snap-revision",
+				"headers": map[string]interface{}{
+					"snap-id": "snap-id-1",
+				},
+				"since": time.Now().Format(time.RFC3339Nano),
+			},
+		},
+	}
+	ad, err := storeDB.Sign(asserts.AuthorityDelegationType, headers, nil, "")
+	c.Assert(err, IsNil)
+	c.Check(db.Add(ad), IsNil)
+
+	err = db.Check(snapRev)
+	c.Check(err, ErrorMatches, `delegated snap-revision assertion from "canonical" to "other" timestamp ".*" is outside of all supporting delegation constraints validity`)
+}
+
 func (srs *snapRevSuite) TestPrimaryKey(c *C) {
 	storeDB, db := makeStoreAndCheckDB(c)
 
