@@ -29,56 +29,19 @@ import (
 
 	"github.com/snapcore/snapd/client"
 	. "github.com/snapcore/snapd/cmd/snap"
+	"github.com/snapcore/snapd/testutil"
 )
 
-func (s *SnapSuite) TestInterfacesHelp(c *C) {
-	msg := `Usage:
-  snap.test [OPTIONS] interfaces [interfaces-OPTIONS] [<snap>:<slot or plug>]
-
-The interfaces command lists interfaces available in the system.
-
-By default all slots and plugs, used and offered by all snaps, are displayed.
-
-$ snap interfaces <snap>:<slot or plug>
-
-Lists only the specified slot or plug.
-
-$ snap interfaces <snap>
-
-Lists the slots offered and plugs used by the specified snap.
-
-$ snap interfaces -i=<interface> [<snap>]
-
-Filters the complete output so only plugs and/or slots matching the provided
-details are listed.
-
-Application Options:
-      --version                    Print the version and exit
-
-Help Options:
-  -h, --help                       Show this help message
-
-[interfaces command options]
-      -i=                          Constrain listing to specific interfaces
-
-[interfaces command arguments]
-  <snap>:<slot or plug>:           Constrain listing to a specific snap or snap:name
-`
-	rest, err := Parser().ParseArgs([]string{"interfaces", "--help"})
-	c.Assert(err.Error(), Equals, msg)
-	c.Assert(rest, DeepEquals, []string{})
-}
-
 func (s *SnapSuite) TestInterfacesZeroSlotsOnePlug(c *C) {
 	s.RedirectClientToTestServer(func(w http.ResponseWriter, r *http.Request) {
 		c.Check(r.Method, Equals, "GET")
-		c.Check(r.URL.Path, Equals, "/v2/interfaces")
+		c.Check(r.URL.Path, Equals, "/v2/connections")
 		body, err := ioutil.ReadAll(r.Body)
 		c.Check(err, IsNil)
 		c.Check(body, DeepEquals, []byte{})
 		EncodeResponseBody(c, w, map[string]interface{}{
 			"type": "sync",
-			"result": client.Interfaces{
+			"result": client.Connections{
 				Plugs: []client.Plug{
 					{
 						Snap: "keyboard-lights",
@@ -88,26 +51,26 @@ func (s *SnapSuite) TestInterfacesZeroSlotsOnePlug(c *C) {
 			},
 		})
 	})
-	rest, err := Parser().ParseArgs([]string{"interfaces"})
+	rest, err := Parser(Client()).ParseArgs([]string{"interfaces"})
 	c.Assert(err, IsNil)
 	c.Assert(rest, DeepEquals, []string{})
 	expectedStdout := "" +
 		"Slot  Plug\n" +
 		"-     keyboard-lights:capslock-led\n"
 	c.Assert(s.Stdout(), Equals, expectedStdout)
-	c.Assert(s.Stderr(), Equals, "")
+	c.Assert(s.Stderr(), testutil.EqualsWrapped, InterfacesDeprecationNotice)
 }
 
 func (s *SnapSuite) TestInterfacesZeroPlugsOneSlot(c *C) {
 	s.RedirectClientToTestServer(func(w http.ResponseWriter, r *http.Request) {
 		c.Check(r.Method, Equals, "GET")
-		c.Check(r.URL.Path, Equals, "/v2/interfaces")
+		c.Check(r.URL.Path, Equals, "/v2/connections")
 		body, err := ioutil.ReadAll(r.Body)
 		c.Check(err, IsNil)
 		c.Check(body, DeepEquals, []byte{})
 		EncodeResponseBody(c, w, map[string]interface{}{
 			"type": "sync",
-			"result": client.Interfaces{
+			"result": client.Connections{
 				Slots: []client.Slot{
 					{
 						Snap:      "canonical-pi2",
@@ -119,26 +82,26 @@ func (s *SnapSuite) TestInterfacesZeroPlugsOneSlot(c *C) {
 			},
 		})
 	})
-	rest, err := Parser().ParseArgs([]string{"interfaces"})
+	rest, err := Parser(Client()).ParseArgs([]string{"interfaces"})
 	c.Assert(err, IsNil)
 	c.Assert(rest, DeepEquals, []string{})
 	expectedStdout := "" +
 		"Slot                  Plug\n" +
 		"canonical-pi2:pin-13  -\n"
 	c.Assert(s.Stdout(), Equals, expectedStdout)
-	c.Assert(s.Stderr(), Equals, "")
+	c.Assert(s.Stderr(), testutil.EqualsWrapped, InterfacesDeprecationNotice)
 }
 
 func (s *SnapSuite) TestInterfacesOneSlotOnePlug(c *C) {
 	s.RedirectClientToTestServer(func(w http.ResponseWriter, r *http.Request) {
 		c.Check(r.Method, Equals, "GET")
-		c.Check(r.URL.Path, Equals, "/v2/interfaces")
+		c.Check(r.URL.Path, Equals, "/v2/connections")
 		body, err := ioutil.ReadAll(r.Body)
 		c.Check(err, IsNil)
 		c.Check(body, DeepEquals, []byte{})
 		EncodeResponseBody(c, w, map[string]interface{}{
 			"type": "sync",
-			"result": client.Interfaces{
+			"result": client.Connections{
 				Slots: []client.Slot{
 					{
 						Snap:      "canonical-pi2",
@@ -170,42 +133,41 @@ func (s *SnapSuite) TestInterfacesOneSlotOnePlug(c *C) {
 			},
 		})
 	})
-	rest, err := Parser().ParseArgs([]string{"interfaces"})
+	rest, err := Parser(Client()).ParseArgs([]string{"interfaces"})
 	c.Assert(err, IsNil)
 	c.Assert(rest, DeepEquals, []string{})
 	expectedStdout := "" +
 		"Slot                  Plug\n" +
 		"canonical-pi2:pin-13  keyboard-lights:capslock-led\n"
 	c.Assert(s.Stdout(), Equals, expectedStdout)
-	c.Assert(s.Stderr(), Equals, "")
+	c.Assert(s.Stderr(), testutil.EqualsWrapped, InterfacesDeprecationNotice)
 
-	s.SetUpTest(c)
+	s.ResetStdStreams()
 	// should be the same
-	rest, err = Parser().ParseArgs([]string{"interfaces", "canonical-pi2"})
+	rest, err = Parser(Client()).ParseArgs([]string{"interfaces", "canonical-pi2"})
 	c.Assert(err, IsNil)
 	c.Assert(rest, DeepEquals, []string{})
 	c.Assert(s.Stdout(), Equals, expectedStdout)
-	c.Assert(s.Stderr(), Equals, "")
-
-	s.SetUpTest(c)
+	c.Assert(s.Stderr(), testutil.EqualsWrapped, InterfacesDeprecationNotice)
+	s.ResetStdStreams()
 	// and the same again
-	rest, err = Parser().ParseArgs([]string{"interfaces", "keyboard-lights"})
+	rest, err = Parser(Client()).ParseArgs([]string{"interfaces", "keyboard-lights"})
 	c.Assert(err, IsNil)
 	c.Assert(rest, DeepEquals, []string{})
 	c.Assert(s.Stdout(), Equals, expectedStdout)
-	c.Assert(s.Stderr(), Equals, "")
+	c.Assert(s.Stderr(), testutil.EqualsWrapped, InterfacesDeprecationNotice)
 }
 
 func (s *SnapSuite) TestInterfacesTwoPlugs(c *C) {
 	s.RedirectClientToTestServer(func(w http.ResponseWriter, r *http.Request) {
 		c.Check(r.Method, Equals, "GET")
-		c.Check(r.URL.Path, Equals, "/v2/interfaces")
+		c.Check(r.URL.Path, Equals, "/v2/connections")
 		body, err := ioutil.ReadAll(r.Body)
 		c.Check(err, IsNil)
 		c.Check(body, DeepEquals, []byte{})
 		EncodeResponseBody(c, w, map[string]interface{}{
 			"type": "sync",
-			"result": client.Interfaces{
+			"result": client.Connections{
 				Slots: []client.Slot{
 					{
 						Snap:      "canonical-pi2",
@@ -227,26 +189,26 @@ func (s *SnapSuite) TestInterfacesTwoPlugs(c *C) {
 			},
 		})
 	})
-	rest, err := Parser().ParseArgs([]string{"interfaces"})
+	rest, err := Parser(Client()).ParseArgs([]string{"interfaces"})
 	c.Assert(err, IsNil)
 	c.Assert(rest, DeepEquals, []string{})
 	expectedStdout := "" +
 		"Slot                  Plug\n" +
 		"canonical-pi2:pin-13  keyboard-lights:capslock-led,keyboard-lights:scrollock-led\n"
 	c.Assert(s.Stdout(), Equals, expectedStdout)
-	c.Assert(s.Stderr(), Equals, "")
+	c.Assert(s.Stderr(), testutil.EqualsWrapped, InterfacesDeprecationNotice)
 }
 
 func (s *SnapSuite) TestInterfacesPlugsWithCommonName(c *C) {
 	s.RedirectClientToTestServer(func(w http.ResponseWriter, r *http.Request) {
 		c.Check(r.Method, Equals, "GET")
-		c.Check(r.URL.Path, Equals, "/v2/interfaces")
+		c.Check(r.URL.Path, Equals, "/v2/connections")
 		body, err := ioutil.ReadAll(r.Body)
 		c.Check(err, IsNil)
 		c.Check(body, DeepEquals, []byte{})
 		EncodeResponseBody(c, w, map[string]interface{}{
 			"type": "sync",
-			"result": client.Interfaces{
+			"result": client.Connections{
 				Slots: []client.Slot{
 					{
 						Snap:      "canonical-pi2",
@@ -294,29 +256,29 @@ func (s *SnapSuite) TestInterfacesPlugsWithCommonName(c *C) {
 			},
 		})
 	})
-	rest, err := Parser().ParseArgs([]string{"interfaces"})
+	rest, err := Parser(Client()).ParseArgs([]string{"interfaces"})
 	c.Assert(err, IsNil)
 	c.Assert(rest, DeepEquals, []string{})
 	expectedStdout := "" +
 		"Slot                             Plug\n" +
 		"canonical-pi2:network-listening  paste-daemon,time-daemon\n"
 	c.Assert(s.Stdout(), Equals, expectedStdout)
-	c.Assert(s.Stderr(), Equals, "")
+	c.Assert(s.Stderr(), testutil.EqualsWrapped, InterfacesDeprecationNotice)
 }
 
 func (s *SnapSuite) TestInterfacesOsSnapSlots(c *C) {
 	s.RedirectClientToTestServer(func(w http.ResponseWriter, r *http.Request) {
 		c.Check(r.Method, Equals, "GET")
-		c.Check(r.URL.Path, Equals, "/v2/interfaces")
+		c.Check(r.URL.Path, Equals, "/v2/connections")
 		body, err := ioutil.ReadAll(r.Body)
 		c.Check(err, IsNil)
 		c.Check(body, DeepEquals, []byte{})
 		EncodeResponseBody(c, w, map[string]interface{}{
 			"type": "sync",
-			"result": client.Interfaces{
+			"result": client.Connections{
 				Slots: []client.Slot{
 					{
-						Snap:      "core",
+						Snap:      "system",
 						Name:      "network-listening",
 						Interface: "network-listening",
 						Label:     "Ability to be a network service",
@@ -340,7 +302,7 @@ func (s *SnapSuite) TestInterfacesOsSnapSlots(c *C) {
 						Label:     "Ability to be a network service",
 						Connections: []client.SlotRef{
 							{
-								Snap: "core",
+								Snap: "system",
 								Name: "network-listening",
 							},
 						},
@@ -352,7 +314,7 @@ func (s *SnapSuite) TestInterfacesOsSnapSlots(c *C) {
 						Label:     "Ability to be a network service",
 						Connections: []client.SlotRef{
 							{
-								Snap: "core",
+								Snap: "system",
 								Name: "network-listening",
 							},
 						},
@@ -361,26 +323,26 @@ func (s *SnapSuite) TestInterfacesOsSnapSlots(c *C) {
 			},
 		})
 	})
-	rest, err := Parser().ParseArgs([]string{"interfaces"})
+	rest, err := Parser(Client()).ParseArgs([]string{"interfaces"})
 	c.Assert(err, IsNil)
 	c.Assert(rest, DeepEquals, []string{})
 	expectedStdout := "" +
 		"Slot                Plug\n" +
 		":network-listening  paste-daemon,time-daemon\n"
 	c.Assert(s.Stdout(), Equals, expectedStdout)
-	c.Assert(s.Stderr(), Equals, "")
+	c.Assert(s.Stderr(), testutil.EqualsWrapped, InterfacesDeprecationNotice)
 }
 
 func (s *SnapSuite) TestInterfacesTwoSlotsAndFiltering(c *C) {
 	s.RedirectClientToTestServer(func(w http.ResponseWriter, r *http.Request) {
 		c.Check(r.Method, Equals, "GET")
-		c.Check(r.URL.Path, Equals, "/v2/interfaces")
+		c.Check(r.URL.Path, Equals, "/v2/connections")
 		body, err := ioutil.ReadAll(r.Body)
 		c.Check(err, IsNil)
 		c.Check(body, DeepEquals, []byte{})
 		EncodeResponseBody(c, w, map[string]interface{}{
 			"type": "sync",
-			"result": client.Interfaces{
+			"result": client.Connections{
 				Slots: []client.Slot{
 					{
 						Snap:      "canonical-pi2",
@@ -410,26 +372,26 @@ func (s *SnapSuite) TestInterfacesTwoSlotsAndFiltering(c *C) {
 			},
 		})
 	})
-	rest, err := Parser().ParseArgs([]string{"interfaces", "-i=serial-port"})
+	rest, err := Parser(Client()).ParseArgs([]string{"interfaces", "-i=serial-port"})
 	c.Assert(err, IsNil)
 	c.Assert(rest, DeepEquals, []string{})
 	expectedStdout := "" +
 		"Slot                         Plug\n" +
 		"canonical-pi2:debug-console  core\n"
 	c.Assert(s.Stdout(), Equals, expectedStdout)
-	c.Assert(s.Stderr(), Equals, "")
+	c.Assert(s.Stderr(), testutil.EqualsWrapped, InterfacesDeprecationNotice)
 }
 
 func (s *SnapSuite) TestInterfacesOfSpecificSnap(c *C) {
 	s.RedirectClientToTestServer(func(w http.ResponseWriter, r *http.Request) {
 		c.Check(r.Method, Equals, "GET")
-		c.Check(r.URL.Path, Equals, "/v2/interfaces")
+		c.Check(r.URL.Path, Equals, "/v2/connections")
 		body, err := ioutil.ReadAll(r.Body)
 		c.Check(err, IsNil)
 		c.Check(body, DeepEquals, []byte{})
 		EncodeResponseBody(c, w, map[string]interface{}{
 			"type": "sync",
-			"result": client.Interfaces{
+			"result": client.Connections{
 				Slots: []client.Slot{
 					{
 						Snap:      "cheese",
@@ -453,7 +415,7 @@ func (s *SnapSuite) TestInterfacesOfSpecificSnap(c *C) {
 			},
 		})
 	})
-	rest, err := Parser().ParseArgs([]string{"interfaces", "wake-up-alarm"})
+	rest, err := Parser(Client()).ParseArgs([]string{"interfaces", "wake-up-alarm"})
 	c.Assert(err, IsNil)
 	c.Assert(rest, DeepEquals, []string{})
 	expectedStdout := "" +
@@ -461,19 +423,74 @@ func (s *SnapSuite) TestInterfacesOfSpecificSnap(c *C) {
 		"wake-up-alarm:toggle  -\n" +
 		"wake-up-alarm:snooze  -\n"
 	c.Assert(s.Stdout(), Equals, expectedStdout)
-	c.Assert(s.Stderr(), Equals, "")
+	c.Assert(s.Stderr(), testutil.EqualsWrapped, InterfacesDeprecationNotice)
 }
 
-func (s *SnapSuite) TestInterfacesOfSpecificSnapAndSlot(c *C) {
+func (s *SnapSuite) TestInterfacesOfSystemNicknameSnap(c *C) {
 	s.RedirectClientToTestServer(func(w http.ResponseWriter, r *http.Request) {
 		c.Check(r.Method, Equals, "GET")
-		c.Check(r.URL.Path, Equals, "/v2/interfaces")
+		c.Check(r.URL.Path, Equals, "/v2/connections")
 		body, err := ioutil.ReadAll(r.Body)
 		c.Check(err, IsNil)
 		c.Check(body, DeepEquals, []byte{})
 		EncodeResponseBody(c, w, map[string]interface{}{
 			"type": "sync",
-			"result": client.Interfaces{
+			"result": client.Connections{
+				Slots: []client.Slot{
+					{
+						Snap:        "system",
+						Name:        "core-support",
+						Interface:   "some-iface",
+						Connections: []client.PlugRef{{Snap: "core", Name: "core-support-plug"}},
+					}, {
+						Snap:      "foo",
+						Name:      "foo-slot",
+						Interface: "foo-slot-iface",
+					},
+				},
+				Plugs: []client.Plug{
+					{
+						Snap:        "core",
+						Name:        "core-support-plug",
+						Interface:   "some-iface",
+						Connections: []client.SlotRef{{Snap: "system", Name: "core-support"}},
+					},
+				},
+			},
+		})
+	})
+	rest, err := Parser(Client()).ParseArgs([]string{"interfaces", "system"})
+	c.Assert(err, IsNil)
+	c.Assert(rest, DeepEquals, []string{})
+	expectedStdout := "" +
+		"Slot           Plug\n" +
+		":core-support  core:core-support-plug\n"
+	c.Assert(s.Stdout(), Equals, expectedStdout)
+	c.Assert(s.Stderr(), testutil.EqualsWrapped, InterfacesDeprecationNotice)
+
+	s.ResetStdStreams()
+
+	// when called with system nickname we get the same output
+	rest, err = Parser(Client()).ParseArgs([]string{"interfaces", "system"})
+	c.Assert(err, IsNil)
+	c.Assert(rest, DeepEquals, []string{})
+	expectedStdoutSystem := "" +
+		"Slot           Plug\n" +
+		":core-support  core:core-support-plug\n"
+	c.Assert(s.Stdout(), Equals, expectedStdoutSystem)
+	c.Assert(s.Stderr(), testutil.EqualsWrapped, InterfacesDeprecationNotice)
+}
+
+func (s *SnapSuite) TestInterfacesOfSpecificSnapAndSlot(c *C) {
+	s.RedirectClientToTestServer(func(w http.ResponseWriter, r *http.Request) {
+		c.Check(r.Method, Equals, "GET")
+		c.Check(r.URL.Path, Equals, "/v2/connections")
+		body, err := ioutil.ReadAll(r.Body)
+		c.Check(err, IsNil)
+		c.Check(body, DeepEquals, []byte{})
+		EncodeResponseBody(c, w, map[string]interface{}{
+			"type": "sync",
+			"result": client.Connections{
 				Slots: []client.Slot{
 					{
 						Snap:      "cheese",
@@ -497,29 +514,29 @@ func (s *SnapSuite) TestInterfacesOfSpecificSnapAndSlot(c *C) {
 			},
 		})
 	})
-	rest, err := Parser().ParseArgs([]string{"interfaces", "wake-up-alarm:snooze"})
+	rest, err := Parser(Client()).ParseArgs([]string{"interfaces", "wake-up-alarm:snooze"})
 	c.Assert(err, IsNil)
 	c.Assert(rest, DeepEquals, []string{})
 	expectedStdout := "" +
 		"Slot                  Plug\n" +
 		"wake-up-alarm:snooze  -\n"
 	c.Assert(s.Stdout(), Equals, expectedStdout)
-	c.Assert(s.Stderr(), Equals, "")
+	c.Assert(s.Stderr(), testutil.EqualsWrapped, InterfacesDeprecationNotice)
 }
 
 func (s *SnapSuite) TestInterfacesNothingAtAll(c *C) {
 	s.RedirectClientToTestServer(func(w http.ResponseWriter, r *http.Request) {
 		c.Check(r.Method, Equals, "GET")
-		c.Check(r.URL.Path, Equals, "/v2/interfaces")
+		c.Check(r.URL.Path, Equals, "/v2/connections")
 		body, err := ioutil.ReadAll(r.Body)
 		c.Check(err, IsNil)
 		c.Check(body, DeepEquals, []byte{})
 		EncodeResponseBody(c, w, map[string]interface{}{
 			"type":   "sync",
-			"result": client.Interfaces{},
+			"result": client.Connections{},
 		})
 	})
-	rest, err := Parser().ParseArgs([]string{"interfaces"})
+	rest, err := Parser(Client()).ParseArgs([]string{"interfaces"})
 	c.Assert(err, ErrorMatches, "no interfaces found")
 	// XXX: not sure why this is returned, I guess that's what happens when a
 	// command Execute returns an error.
@@ -531,13 +548,13 @@ func (s *SnapSuite) TestInterfacesNothingAtAll(c *C) {
 func (s *SnapSuite) TestInterfacesOfSpecificType(c *C) {
 	s.RedirectClientToTestServer(func(w http.ResponseWriter, r *http.Request) {
 		c.Check(r.Method, Equals, "GET")
-		c.Check(r.URL.Path, Equals, "/v2/interfaces")
+		c.Check(r.URL.Path, Equals, "/v2/connections")
 		body, err := ioutil.ReadAll(r.Body)
 		c.Check(err, IsNil)
 		c.Check(body, DeepEquals, []byte{})
 		EncodeResponseBody(c, w, map[string]interface{}{
 			"type": "sync",
-			"result": client.Interfaces{
+			"result": client.Connections{
 				Slots: []client.Slot{
 					{
 						Snap:      "cheese",
@@ -561,7 +578,7 @@ func (s *SnapSuite) TestInterfacesOfSpecificType(c *C) {
 			},
 		})
 	})
-	rest, err := Parser().ParseArgs([]string{"interfaces", "-i", "bool-file"})
+	rest, err := Parser(Client()).ParseArgs([]string{"interfaces", "-i", "bool-file"})
 	c.Assert(err, IsNil)
 	c.Assert(rest, DeepEquals, []string{})
 	expectedStdout := "" +
@@ -570,17 +587,17 @@ func (s *SnapSuite) TestInterfacesOfSpecificType(c *C) {
 		"wake-up-alarm:toggle  -\n" +
 		"wake-up-alarm:snooze  -\n"
 	c.Assert(s.Stdout(), Equals, expectedStdout)
-	c.Assert(s.Stderr(), Equals, "")
+	c.Assert(s.Stderr(), testutil.EqualsWrapped, InterfacesDeprecationNotice)
 }
 
 func (s *SnapSuite) TestInterfacesCompletion(c *C) {
 	s.RedirectClientToTestServer(func(w http.ResponseWriter, r *http.Request) {
 		switch r.URL.Path {
-		case "/v2/interfaces":
+		case "/v2/connections":
 			c.Assert(r.Method, Equals, "GET")
 			EncodeResponseBody(c, w, map[string]interface{}{
 				"type":   "sync",
-				"result": fortestingInterfaceList,
+				"result": fortestingConnectionList,
 			})
 		default:
 			c.Fatalf("unexpected path %q", r.URL.Path)
@@ -590,7 +607,7 @@ func (s *SnapSuite) TestInterfacesCompletion(c *C) {
 	defer os.Unsetenv("GO_FLAGS_COMPLETION")
 
 	expected := []flags.Completion{}
-	parser := Parser()
+	parser := Parser(Client())
 	parser.CompletionHandler = func(obtained []flags.Completion) {
 		c.Check(obtained, DeepEquals, expected)
 	}
@@ -610,3 +627,48 @@ func (s *SnapSuite) TestInterfacesCompletion(c *C) {
 	c.Assert(s.Stdout(), Equals, "")
 	c.Assert(s.Stderr(), Equals, "")
 }
+
+func (s *SnapSuite) TestInterfacesCoreNicknamedSystem(c *C) {
+	s.checkConnectionsSystemCoreRemapping(c, "core", "system")
+}
+
+func (s *SnapSuite) TestInterfacesSnapdNicknamedSystem(c *C) {
+	s.checkConnectionsSystemCoreRemapping(c, "snapd", "system")
+}
+
+func (s *SnapSuite) TestInterfacesSnapdNicknamedCore(c *C) {
+	s.checkConnectionsSystemCoreRemapping(c, "snapd", "core")
+}
+
+func (s *SnapSuite) TestInterfacesCoreSnap(c *C) {
+	s.checkConnectionsSystemCoreRemapping(c, "core", "core")
+}
+
+func (s *SnapSuite) checkConnectionsSystemCoreRemapping(c *C, apiSnapName, cliSnapName string) {
+	s.RedirectClientToTestServer(func(w http.ResponseWriter, r *http.Request) {
+		c.Check(r.Method, Equals, "GET")
+		c.Check(r.URL.Path, Equals, "/v2/connections")
+		body, err := ioutil.ReadAll(r.Body)
+		c.Check(err, IsNil)
+		c.Check(body, DeepEquals, []byte{})
+		EncodeResponseBody(c, w, map[string]interface{}{
+			"type": "sync",
+			"result": client.Connections{
+				Slots: []client.Slot{
+					{
+						Snap: apiSnapName,
+						Name: "network",
+					},
+				},
+			},
+		})
+	})
+	rest, err := Parser(Client()).ParseArgs([]string{"interfaces", cliSnapName})
+	c.Assert(err, IsNil)
+	c.Assert(rest, DeepEquals, []string{})
+	expectedStdout := "" +
+		"Slot      Plug\n" +
+		":network  -\n"
+	c.Assert(s.Stdout(), Equals, expectedStdout)
+	c.Assert(s.Stderr(), testutil.EqualsWrapped, InterfacesDeprecationNotice)
+}
