@@ -26,8 +26,9 @@ import (
 	"github.com/jessevdk/go-flags"
 
 	"github.com/snapcore/snapd/client"
-	"github.com/snapcore/snapd/cmd"
+	"github.com/snapcore/snapd/client/clientutil"
 	"github.com/snapcore/snapd/i18n"
+	"github.com/snapcore/snapd/snap"
 )
 
 type svcStatus struct {
@@ -39,6 +40,7 @@ type svcStatus struct {
 
 type svcLogs struct {
 	clientMixin
+	timeMixin
 	N          string `short:"n" default:"10"`
 	Follow     bool   `short:"f"`
 	Positional struct {
@@ -83,12 +85,12 @@ func init() {
 	}}
 	addCommand("services", shortServicesHelp, longServicesHelp, func() flags.Commander { return &svcStatus{} }, nil, argdescs)
 	addCommand("logs", shortLogsHelp, longLogsHelp, func() flags.Commander { return &svcLogs{} },
-		map[string]string{
+		timeDescs.also(map[string]string{
 			// TRANSLATORS: This should not start with a lowercase letter.
 			"n": i18n.G("Show only the given number of lines, or 'all'."),
 			// TRANSLATORS: This should not start with a lowercase letter.
 			"f": i18n.G("Wait for new lines and print them as they come in."),
-		}, argdescs)
+		}), argdescs)
 
 	addCommand("start", shortStartHelp, longStartHelp, func() flags.Commander { return &svcStart{} },
 		waitDescs.also(map[string]string{
@@ -141,10 +143,12 @@ func (s *svcStatus) Execute(args []string) error {
 			startup = i18n.G("enabled")
 		}
 		current := i18n.G("inactive")
-		if svc.Active {
+		if svc.DaemonScope == snap.UserDaemon {
+			current = "-"
+		} else if svc.Active {
 			current = i18n.G("active")
 		}
-		fmt.Fprintf(w, "%s.%s\t%s\t%s\t%s\n", svc.Snap, svc.Name, startup, current, cmd.ClientAppInfoNotes(svc))
+		fmt.Fprintf(w, "%s.%s\t%s\t%s\t%s\n", svc.Snap, svc.Name, startup, current, clientutil.ClientAppInfoNotes(svc))
 	}
 
 	return nil
@@ -170,7 +174,11 @@ func (s *svcLogs) Execute(args []string) error {
 	}
 
 	for log := range logs {
-		fmt.Fprintln(Stdout, log)
+		if s.AbsTime {
+			fmt.Fprintln(Stdout, log.StringInUTC())
+		} else {
+			fmt.Fprintln(Stdout, log)
+		}
 	}
 
 	return nil
@@ -200,7 +208,7 @@ func (s *svcStart) Execute(args []string) error {
 		return err
 	}
 
-	fmt.Fprintf(Stdout, i18n.G("Started.\n"))
+	fmt.Fprintln(Stdout, i18n.G("Started."))
 
 	return nil
 }
