@@ -32,8 +32,12 @@ import (
 	"github.com/jessevdk/go-flags"
 
 	"github.com/snapcore/snapd/client"
+	"github.com/snapcore/snapd/dirs"
 	"github.com/snapcore/snapd/i18n"
 	"github.com/snapcore/snapd/osutil"
+	"github.com/snapcore/snapd/release"
+	"github.com/snapcore/snapd/snap/channel"
+	"github.com/snapcore/snapd/strutil"
 )
 
 var (
@@ -52,13 +56,14 @@ To install multiple instances of the same snap, append an underscore and a
 unique identifier (for each instance) to a snap's name.
 
 With no further options, the snaps are installed tracking the stable channel,
-with strict security confinement.
+with strict security confinement. All available channels of a snap are listed in
+its 'snap info' output.
 
-Revision choice via the --revision override requires the the user to
+Revision choice via the --revision override requires the user to
 have developer access to the snap, either directly or through the
 store's collaboration feature, and to be logged in (see 'snap help login').
 
-Note a later refresh will typically undo a revision override, taking the snap
+Note that a later refresh will typically undo a revision override, taking the snap
 back to the current revision of the channel it's tracking.
 
 Use --name to set the instance name when installing from snap file.
@@ -70,6 +75,10 @@ The remove command removes the named snap instance from the system.
 By default all the snap revisions are removed, including their data and the
 common data directory. When a --revision option is passed only the specified
 revision is removed.
+
+Unless automatic snapshots are disabled, a snapshot of all data for the snap is 
+saved upon removal, which is then available for future restoration with snap
+restore. The --purge option disables automatically creating snapshots.
 `)
 
 var longRefreshHelp = i18n.G(`
@@ -77,9 +86,10 @@ The refresh command updates the specified snaps, or all snaps in the system if
 none are specified.
 
 With no further options, the snaps are refreshed to the current revision of the
-channel they're tracking, preserving their confinement options.
+channel they're tracking, preserving their confinement options. All available
+channels of a snap are listed in its 'snap info' output.
 
-Revision choice via the --revision override requires the the user to
+Revision choice via the --revision override requires the user to
 have developer access to the snap, either directly or through the
 store's collaboration feature, and to be logged in (see 'snap help login').
 
@@ -111,6 +121,7 @@ type cmdRemove struct {
 	waitMixin
 
 	Revision   string `long:"revision"`
+	Purge      bool   `long:"purge"`
 	Positional struct {
 		Snaps []installedSnapName `positional-arg-name:"<snap>" required:"1"`
 	} `positional-args:"yes" required:"yes"`
@@ -121,7 +132,7 @@ func (x *cmdRemove) removeOne(opts *client.SnapOptions) error {
 
 	changeID, err := x.client.Remove(name, opts)
 	if err != nil {
-		msg, err := errorToCmdMessage(name, err, opts)
+		msg, err := errorToCmdMessage(name, "remove", err, opts)
 		if err != nil {
 			return err
 		}
@@ -148,7 +159,19 @@ func (x *cmdRemove) removeMany(opts *client.SnapOptions) error {
 	names := installedSnapNames(x.Positional.Snaps)
 	changeID, err := x.client.RemoveMany(names, opts)
 	if err != nil {
-		return err
+		var name string
+		if cerr, ok := err.(*client.Error); ok {
+			if snapName, ok := cerr.Value.(string); ok {
+				name = snapName
+			}
+		}
+
+		msg, err := errorToCmdMessage(name, "remove", err, opts)
+		if err != nil {
+			return err
+		}
+		fmt.Fprintln(Stderr, msg)
+		return nil
 	}
 
 	chg, err := x.wait(changeID)
@@ -182,13 +205,13 @@ func (x *cmdRemove) removeMany(opts *client.SnapOptions) error {
 }
 
 func (x *cmdRemove) Execute([]string) error {
-	opts := &client.SnapOptions{Revision: x.Revision}
+	opts := &client.SnapOptions{Revision: x.Revision, Purge: x.Purge}
 	if len(x.Positional.Snaps) == 1 {
 		return x.removeOne(opts)
 	}
 
-	if x.Revision != "" {
-		return errors.New(i18n.G("a single snap name is needed to specify the revision"))
+	if x.Purge || x.Revision != "" {
+		return errors.New(i18n.G("a single snap name is needed to specify options"))
 	}
 	return x.removeMany(nil)
 }
@@ -248,15 +271,74 @@ func (mx *channelMixin) setChannelFromCommandline() error {
 		mx.Channel = ch.chName
 	}
 
-	if !strings.Contains(mx.Channel, "/") && mx.Channel != "" && mx.Channel != "edge" && mx.Channel != "beta" && mx.Channel != "candidate" && mx.Channel != "stable" {
-		// shortcut to jump to a different track, e.g.
-		// snap install foo --channel=3.4 # implies 3.4/stable
-		mx.Channel += "/stable"
+	if mx.Channel != "" {
+		if _, err := channel.Parse(mx.Channel, ""); err != nil {
+			full, er := channel.Full(mx.Channel)
+			if er != nil {
+				// the parse error has more detailed info
+				return err
+			}
+
+			// TODO: get escapes in here so we can bold the Warning
+			head := i18n.G("Warning:")
+			msg := i18n.G("Specifying a channel %q is relying on undefined behaviour. Interpreting it as %q for now, but this will be an error later.\n")
+			warn := fill(fmt.Sprintf(msg, mx.Channel, full), utf8.RuneCountInString(head)+1) // +1 for the space
+			fmt.Fprint(Stderr, head, " ", warn, "\n\n")
+			mx.Channel = full // so a malformed-but-eh channel will always be full, i.e. //stable// -> latest/stable
+		}
 	}
 
 	return nil
 }
 
+// isSnapInPath checks whether the snap binaries dir (e.g. /snap/bin)
+// is in $PATH.
+//
+// TODO: consider symlinks
+func isSnapInPath() bool {
+	paths := filepath.SplitList(os.Getenv("PATH"))
+	for _, path := range paths {
+		if filepath.Clean(path) == dirs.SnapBinariesDir {
+			return true
+		}
+	}
+	return false
+}
+
+func isSameRisk(tracking, current string) (bool, error) {
+	if tracking == current {
+		return true, nil
+	}
+	var trackingRisk, currentRisk string
+	if tracking != "" {
+		traCh, err := channel.Parse(tracking, "")
+		if err != nil {
+			return false, err
+		}
+		trackingRisk = traCh.Risk
+	}
+	if current != "" {
+		curCh, err := channel.Parse(current, "")
+		if err != nil {
+			return false, err
+		}
+		currentRisk = curCh.Risk
+	}
+	return trackingRisk == currentRisk, nil
+}
+
+func maybeWithSudoSecurePath() bool {
+	// Some distributions set secure_path to a known list of paths in
+	// /etc/sudoers. The snapd package currently has no means of overwriting
+	// or extending that setting.
+	if _, isSet := os.LookupEnv("SUDO_UID"); !isSet {
+		return false
+	}
+	// Known distros setting secure_path that does not include
+	// $SNAP_MOUNT_DIR/bin:
+	return release.DistroLike("fedora", "opensuse", "debian")
+}
+
 // show what has been done
 func showDone(cli *client.Client, names []string, op string, opts *client.SnapOptions, esc *escapes) error {
 	snaps, err := cli.List(names, nil)
@@ -264,19 +346,33 @@ func showDone(cli *client.Client, names []string, op string, opts *client.SnapOp
 		return err
 	}
 
+	needsPathWarning := !isSnapInPath() && !maybeWithSudoSecurePath()
 	for _, snap := range snaps {
 		channelStr := ""
-		if snap.Channel != "" && snap.Channel != "stable" {
-			channelStr = fmt.Sprintf(" (%s)", snap.Channel)
+		if snap.Channel != "" {
+			ch, err := channel.Parse(snap.Channel, "")
+			if err != nil {
+				return err
+			}
+			if ch.Name != "stable" {
+				channelStr = fmt.Sprintf(" (%s)", ch.Name)
+			}
 		}
 		switch op {
 		case "install":
+			if needsPathWarning {
+				head := i18n.G("Warning:")
+				warn := fill(fmt.Sprintf(i18n.G("%s was not found in your $PATH. If you've not restarted your session since you installed snapd, try doing that. Please see https://forum.snapcraft.io/t/9469 for more details."), dirs.SnapBinariesDir), utf8.RuneCountInString(head)+1) // +1 for the space
+				fmt.Fprint(Stderr, esc.bold, head, esc.end, " ", warn, "\n\n")
+				needsPathWarning = false
+			}
+
 			if opts != nil && opts.Classic && snap.Confinement != client.ClassicConfinement {
 				// requested classic but the snap is not classic
 				head := i18n.G("Warning:")
 				// TRANSLATORS: the arg is a snap name (e.g. "some-snap")
 				warn := fill(fmt.Sprintf(i18n.G("flag --classic ignored for strictly confined snap %s"), snap.Name), utf8.RuneCountInString(head)+1) // +1 for the space
-				fmt.Fprint(Stderr, head, " ", warn, "\n\n")
+				fmt.Fprint(Stderr, esc.bold, head, esc.end, " ", warn, "\n\n")
 			}
 
 			if snap.Publisher != nil {
@@ -297,12 +393,40 @@ func showDone(cli *client.Client, names []string, op string, opts *client.SnapOp
 		case "revert":
 			// TRANSLATORS: first %s is a snap name, second %s is a revision
 			fmt.Fprintf(Stdout, i18n.G("%s reverted to %s\n"), snap.Name, snap.Version)
+		case "switch":
+			switchCohort := opts.CohortKey != ""
+			switchChannel := opts.Channel != ""
+			var msg string
+			// we have three boolean things to check, meaning 2³=8 possibilities,
+			// minus 3 error cases which are handled before the call to showDone.
+			switch {
+			case switchCohort && !opts.LeaveCohort && !switchChannel:
+				// TRANSLATORS: the first %q will be the (quoted) snap name, the second an ellipted cohort string
+				msg = fmt.Sprintf(i18n.G("%q switched to the %q cohort\n"), snap.Name, strutil.ElliptLeft(opts.CohortKey, 10))
+			case switchCohort && !opts.LeaveCohort && switchChannel:
+				// TRANSLATORS: the first %q will be the (quoted) snap name, the second a channel, the third an ellipted cohort string
+				msg = fmt.Sprintf(i18n.G("%q switched to the %q channel and the %q cohort\n"), snap.Name, snap.TrackingChannel, strutil.ElliptLeft(opts.CohortKey, 10))
+			case !switchCohort && !opts.LeaveCohort && switchChannel:
+				// TRANSLATORS: the first %q will be the (quoted) snap name, the second a channel
+				msg = fmt.Sprintf(i18n.G("%q switched to the %q channel\n"), snap.Name, snap.TrackingChannel)
+			case !switchCohort && opts.LeaveCohort && switchChannel:
+				// TRANSLATORS: the first %q will be the (quoted) snap name, the second a channel
+				msg = fmt.Sprintf(i18n.G("%q left the cohort, and switched to the %q channel"), snap.Name, snap.TrackingChannel)
+			case !switchCohort && opts.LeaveCohort && !switchChannel:
+				// TRANSLATORS: %q will be the (quoted) snap name
+				msg = fmt.Sprintf(i18n.G("%q left the cohort"), snap.Name)
+			}
+			fmt.Fprintln(Stdout, msg)
 		default:
 			fmt.Fprintf(Stdout, "internal error: unknown op %q", op)
 		}
-		if snap.TrackingChannel != snap.Channel && snap.Channel != "" {
-			// TRANSLATORS: first %s is a channel name, following %s is a snap name, last %s is a channel name again.
-			fmt.Fprintf(Stdout, i18n.G("Channel %s for %s is closed; temporarily forwarding to %s.\n"), snap.TrackingChannel, snap.Name, snap.Channel)
+		if op == "install" || op == "refresh" {
+			if snap.TrackingChannel != snap.Channel && snap.Channel != "" {
+				if sameRisk, err := isSameRisk(snap.TrackingChannel, snap.Channel); err == nil && !sameRisk {
+					// TRANSLATORS: first %s is a channel name, following %s is a snap name, last %s is a channel name again.
+					fmt.Fprintf(Stdout, i18n.G("Channel %s for %s is closed; temporarily forwarding to %s.\n"), snap.TrackingChannel, snap.Name, snap.Channel)
+				}
+			}
 		}
 	}
 
@@ -364,8 +488,12 @@ type cmdInstall struct {
 
 	Name string `long:"name"`
 
-	Positional struct {
-		Snaps []remoteSnapName `positional-arg-name:"<snap>"`
+	Cohort           string                 `long:"cohort"`
+	IgnoreValidation bool                   `long:"ignore-validation"`
+	IgnoreRunning    bool                   `long:"ignore-running" hidden:"yes"`
+	Transaction      client.TransactionType `long:"transaction" default:"per-snap" choice:"all-snaps" choice:"per-snap"`
+	Positional       struct {
+		Snaps []remoteSnapName `positional-arg-name:"<snap>" required:"1"`
 	} `positional-args:"yes" required:"yes"`
 }
 
@@ -375,7 +503,7 @@ func (x *cmdInstall) installOne(nameOrPath, desiredName string, opts *client.Sna
 	var snapName string
 	var path string
 
-	if strings.Contains(nameOrPath, "/") || strings.HasSuffix(nameOrPath, ".snap") || strings.Contains(nameOrPath, ".snap.") {
+	if isLocalSnap(nameOrPath) {
 		path = nameOrPath
 		changeID, err = x.client.InstallPath(path, x.Name, opts)
 	} else {
@@ -386,7 +514,7 @@ func (x *cmdInstall) installOne(nameOrPath, desiredName string, opts *client.Sna
 		changeID, err = x.client.Install(snapName, opts)
 	}
 	if err != nil {
-		msg, err := errorToCmdMessage(nameOrPath, err, opts)
+		msg, err := errorToCmdMessage(nameOrPath, "install", err, opts)
 		if err != nil {
 			return err
 		}
@@ -409,24 +537,41 @@ func (x *cmdInstall) installOne(nameOrPath, desiredName string, opts *client.Sna
 		}
 	}
 
+	// TODO: mention details of the install (e.g. like switch does)
 	return showDone(x.client, []string{snapName}, "install", opts, x.getEscapes())
 }
 
+func isLocalSnap(name string) bool {
+	return strings.Contains(name, "/") || strings.HasSuffix(name, ".snap") || strings.Contains(name, ".snap.")
+}
+
 func (x *cmdInstall) installMany(names []string, opts *client.SnapOptions) error {
-	// sanity check
+	isLocal := isLocalSnap(names[0])
 	for _, name := range names {
-		if strings.Contains(name, "/") || strings.HasSuffix(name, ".snap") || strings.Contains(name, ".snap.") {
-			return fmt.Errorf("only one snap file can be installed at a time")
+		if isLocalSnap(name) != isLocal {
+			return fmt.Errorf(i18n.G("cannot install local and store snaps at the same time"))
 		}
 	}
 
-	changeID, err := x.client.InstallMany(names, opts)
+	var changeID string
+	var err error
+
+	if isLocal {
+		changeID, err = x.client.InstallPathMany(names, opts)
+	} else {
+		if x.asksForMode() {
+			return errors.New(i18n.G("cannot specify mode for multiple store snaps (only for one store snap or several local ones)"))
+		}
+
+		changeID, err = x.client.InstallMany(names, opts)
+	}
+
 	if err != nil {
 		var snapName string
 		if err, ok := err.(*client.Error); ok {
 			snapName, _ = err.Value.(string)
 		}
-		msg, err := errorToCmdMessage(snapName, err, opts)
+		msg, err := errorToCmdMessage(snapName, "install", err, opts)
 		if err != nil {
 			return err
 		}
@@ -453,6 +598,11 @@ func (x *cmdInstall) installMany(names []string, opts *client.SnapOptions) error
 		}
 	}
 
+	// local installs aren't skipped if the snap is installed
+	if isLocal {
+		return nil
+	}
+
 	// show skipped
 	seen := make(map[string]bool)
 	for _, name := range installed {
@@ -479,17 +629,18 @@ func (x *cmdInstall) Execute([]string) error {
 
 	dangerous := x.Dangerous || x.ForceDangerous
 	opts := &client.SnapOptions{
-		Channel:   x.Channel,
-		Revision:  x.Revision,
-		Dangerous: dangerous,
-		Unaliased: x.Unaliased,
+		Channel:          x.Channel,
+		Revision:         x.Revision,
+		Dangerous:        dangerous,
+		Unaliased:        x.Unaliased,
+		CohortKey:        x.Cohort,
+		IgnoreValidation: x.IgnoreValidation,
+		IgnoreRunning:    x.IgnoreRunning,
+		Transaction:      x.Transaction,
 	}
 	x.setModes(opts)
 
 	names := remoteSnapNames(x.Positional.Snaps)
-	if len(names) == 0 {
-		return errors.New(i18n.G("cannot install zero snaps"))
-	}
 	for _, name := range names {
 		if len(name) == 0 {
 			return errors.New(i18n.G("cannot install snap with empty name"))
@@ -500,14 +651,17 @@ func (x *cmdInstall) Execute([]string) error {
 		return x.installOne(names[0], x.Name, opts)
 	}
 
-	if x.asksForMode() || x.asksForChannel() {
-		return errors.New(i18n.G("a single snap name is needed to specify mode or channel flags"))
+	if x.asksForChannel() {
+		return errors.New(i18n.G("a single snap name is needed to specify channel flags"))
+	}
+	if x.IgnoreValidation {
+		return errors.New(i18n.G("a single snap name must be specified when ignoring validation"))
 	}
 
 	if x.Name != "" {
 		return errors.New(i18n.G("cannot use instance name when installing multiple snaps"))
 	}
-	return x.installMany(names, nil)
+	return x.installMany(names, opts)
 }
 
 type cmdRefresh struct {
@@ -517,11 +671,15 @@ type cmdRefresh struct {
 	channelMixin
 	modeMixin
 
-	Amend            bool   `long:"amend"`
-	Revision         string `long:"revision"`
-	List             bool   `long:"list"`
-	Time             bool   `long:"time"`
-	IgnoreValidation bool   `long:"ignore-validation"`
+	Amend            bool                   `long:"amend"`
+	Revision         string                 `long:"revision"`
+	Cohort           string                 `long:"cohort"`
+	LeaveCohort      bool                   `long:"leave-cohort"`
+	List             bool                   `long:"list"`
+	Time             bool                   `long:"time"`
+	IgnoreValidation bool                   `long:"ignore-validation"`
+	IgnoreRunning    bool                   `long:"ignore-running" hidden:"yes"`
+	Transaction      client.TransactionType `long:"transaction" default:"per-snap" choice:"all-snaps" choice:"per-snap"`
 	Positional       struct {
 		Snaps []installedSnapName `positional-arg-name:"<snap>"`
 	} `positional-args:"yes"`
@@ -558,7 +716,7 @@ func (x *cmdRefresh) refreshMany(snaps []string, opts *client.SnapOptions) error
 func (x *cmdRefresh) refreshOne(name string, opts *client.SnapOptions) error {
 	changeID, err := x.client.Refresh(name, opts)
 	if err != nil {
-		msg, err := errorToCmdMessage(name, err, opts)
+		msg, err := errorToCmdMessage(name, "refresh", err, opts)
 		if err != nil {
 			return err
 		}
@@ -573,6 +731,8 @@ func (x *cmdRefresh) refreshOne(name string, opts *client.SnapOptions) error {
 		return err
 	}
 
+	// TODO: this doesn't really tell about all the things you
+	// could set while refreshing (something switch does)
 	return showDone(x.client, []string{name}, "refresh", opts, x.getEscapes())
 }
 
@@ -643,9 +803,9 @@ func (x *cmdRefresh) listRefresh() error {
 	defer w.Flush()
 
 	// TRANSLATORS: the %s is to insert a filler escape sequence (please keep it flush to the column header, with no extra spaces)
-	fmt.Fprintf(w, i18n.G("Name\tVersion\tRev\tPublisher%s\tNotes\n"), fillerPublisher(esc))
+	fmt.Fprintf(w, i18n.G("Name\tVersion\tRev\tSize\tPublisher%s\tNotes\n"), fillerPublisher(esc))
 	for _, snap := range snaps {
-		fmt.Fprintf(w, "%s\t%s\t%s\t%s\t%s\n", snap.Name, snap.Version, snap.Revision, shortPublisher(esc, snap.Publisher), NotesFromRemote(snap, nil))
+		fmt.Fprintf(w, "%s\t%s\t%s\t%s\t%s\t%s\n", snap.Name, snap.Version, snap.Revision, strutil.SizeToStr(snap.DownloadSize), shortPublisher(esc, snap.Publisher), NotesFromRemote(snap, nil))
 	}
 
 	return nil
@@ -685,11 +845,21 @@ func (x *cmdRefresh) Execute([]string) error {
 			Amend:            x.Amend,
 			Channel:          x.Channel,
 			IgnoreValidation: x.IgnoreValidation,
+			IgnoreRunning:    x.IgnoreRunning,
 			Revision:         x.Revision,
+			CohortKey:        x.Cohort,
+			LeaveCohort:      x.LeaveCohort,
+			Transaction:      x.Transaction,
 		}
 		x.setModes(opts)
 		return x.refreshOne(names[0], opts)
 	}
+	// transaction flag and ignore-running flags are the only ones with meaning when
+	// refreshing many snaps
+	opts := &client.SnapOptions{
+		IgnoreRunning: x.IgnoreRunning,
+		Transaction:   x.Transaction,
+	}
 
 	if x.asksForMode() || x.asksForChannel() {
 		return errors.New(i18n.G("a single snap name is needed to specify mode or channel flags"))
@@ -699,7 +869,7 @@ func (x *cmdRefresh) Execute([]string) error {
 		return errors.New(i18n.G("a single snap name must be specified when ignoring validation"))
 	}
 
-	return x.refreshMany(names, nil)
+	return x.refreshMany(names, opts)
 }
 
 type cmdTry struct {
@@ -754,7 +924,7 @@ func (x *cmdTry) Execute([]string) error {
 
 	changeID, err := x.client.Try(path, opts)
 	if err != nil {
-		msg, err := errorToCmdMessage(name, err, opts)
+		msg, err := errorToCmdMessage(name, "try", err, opts)
 		if err != nil {
 			return err
 		}
@@ -851,8 +1021,9 @@ type cmdRevert struct {
 	waitMixin
 
 	modeMixin
-	Revision   string `long:"revision"`
-	Positional struct {
+	Revision      string `long:"revision"`
+	IgnoreRunning bool   `long:"ignore-running" hidden:"yes"`
+	Positional    struct {
 		Snap installedSnapName `positional-arg-name:"<snap>"`
 	} `positional-args:"yes" required:"yes"`
 }
@@ -877,7 +1048,10 @@ func (x *cmdRevert) Execute(args []string) error {
 	}
 
 	name := string(x.Positional.Snap)
-	opts := &client.SnapOptions{Revision: x.Revision}
+	opts := &client.SnapOptions{
+		Revision:      x.Revision,
+		IgnoreRunning: x.IgnoreRunning,
+	}
 	x.setModes(opts)
 	changeID, err := x.client.Revert(name, opts)
 	if err != nil {
@@ -897,13 +1071,17 @@ func (x *cmdRevert) Execute(args []string) error {
 var shortSwitchHelp = i18n.G("Switches snap to a different channel")
 var longSwitchHelp = i18n.G(`
 The switch command switches the given snap to a different channel without
-doing a refresh.
+doing a refresh. All available channels of a snap are listed in
+its 'snap info' output.
 `)
 
 type cmdSwitch struct {
 	waitMixin
 	channelMixin
 
+	Cohort      string `long:"cohort"`
+	LeaveCohort bool   `long:"leave-cohort"`
+
 	Positional struct {
 		Snap installedSnapName `positional-arg-name:"<snap>" required:"1"`
 	} `positional-args:"yes" required:"yes"`
@@ -913,14 +1091,28 @@ func (x cmdSwitch) Execute(args []string) error {
 	if err := x.setChannelFromCommandline(); err != nil {
 		return err
 	}
-	if x.Channel == "" {
-		return fmt.Errorf("missing --channel=<channel-name> parameter")
-	}
 
 	name := string(x.Positional.Snap)
 	channel := string(x.Channel)
+
+	switchCohort := x.Cohort != ""
+	switchChannel := x.Channel != ""
+
+	// we have three boolean things to check, meaning 2³=8 possibilities
+	// of which 3 are errors (which is why we look at the errors first).
+	// the 5 valid cases are handled by showDone.
+	if switchCohort && x.LeaveCohort {
+		// this one counts as two (no channel filter)
+		return fmt.Errorf(i18n.G("cannot specify both --cohort and --leave-cohort"))
+	}
+	if !switchCohort && !x.LeaveCohort && !switchChannel {
+		return fmt.Errorf(i18n.G("nothing to switch; specify --channel (and/or one of --cohort/--leave-cohort)"))
+	}
+
 	opts := &client.SnapOptions{
-		Channel: channel,
+		Channel:     channel,
+		CohortKey:   x.Cohort,
+		LeaveCohort: x.LeaveCohort,
 	}
 	changeID, err := x.client.Switch(name, opts)
 	if err != nil {
@@ -934,8 +1126,7 @@ func (x cmdSwitch) Execute(args []string) error {
 		return err
 	}
 
-	fmt.Fprintf(Stdout, i18n.G("%q switched to the %q channel\n"), name, channel)
-	return nil
+	return showDone(x.client, []string{name}, "switch", opts, nil)
 }
 
 func init() {
@@ -943,6 +1134,8 @@ func init() {
 		waitDescs.also(map[string]string{
 			// TRANSLATORS: This should not start with a lowercase letter.
 			"revision": i18n.G("Remove only the given revision"),
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"purge": i18n.G("Remove the snap without saving a snapshot of its data"),
 		}), nil)
 	addCommand("install", shortInstallHelp, longInstallHelp, func() flags.Commander { return &cmdInstall{} },
 		colorDescs.also(waitDescs).also(channelDescs).also(modeDescs).also(map[string]string{
@@ -956,6 +1149,14 @@ func init() {
 			"unaliased": i18n.G("Install the given snap without enabling its automatic aliases"),
 			// TRANSLATORS: This should not start with a lowercase letter.
 			"name": i18n.G("Install the snap file under the given instance name"),
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"cohort": i18n.G("Install the snap in the given cohort"),
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"ignore-validation": i18n.G("Ignore validation by other snaps blocking the installation"),
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"ignore-running": i18n.G("Ignore running hooks or applications blocking the installation"),
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"transaction": i18n.G("Have one transaction per-snap or one for all the specified snaps"),
 		}), nil)
 	addCommand("refresh", shortRefreshHelp, longRefreshHelp, func() flags.Commander { return &cmdRefresh{} },
 		colorDescs.also(waitDescs).also(channelDescs).also(modeDescs).also(timeDescs).also(map[string]string{
@@ -969,6 +1170,14 @@ func init() {
 			"time": i18n.G("Show auto refresh information but do not perform a refresh"),
 			// TRANSLATORS: This should not start with a lowercase letter.
 			"ignore-validation": i18n.G("Ignore validation by other snaps blocking the refresh"),
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"ignore-running": i18n.G("Ignore running hooks or applications blocking the refresh"),
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"cohort": i18n.G("Refresh the snap into the given cohort"),
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"leave-cohort": i18n.G("Refresh the snap out of its cohort"),
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"transaction": i18n.G("Have one transaction per-snap or one for all the specified snaps"),
 		}), nil)
 	addCommand("try", shortTryHelp, longTryHelp, func() flags.Commander { return &cmdTry{} }, waitDescs.also(modeDescs), nil)
 	addCommand("enable", shortEnableHelp, longEnableHelp, func() flags.Commander { return &cmdEnable{} }, waitDescs, nil)
@@ -976,6 +1185,13 @@ func init() {
 	addCommand("revert", shortRevertHelp, longRevertHelp, func() flags.Commander { return &cmdRevert{} }, waitDescs.also(modeDescs).also(map[string]string{
 		// TRANSLATORS: This should not start with a lowercase letter.
 		"revision": i18n.G("Revert to the given revision"),
+		// TRANSLATORS: This should not start with a lowercase letter.
+		"ignore-running": i18n.G("Ignore running hooks or applications blocking the revert"),
+	}), nil)
+	addCommand("switch", shortSwitchHelp, longSwitchHelp, func() flags.Commander { return &cmdSwitch{} }, waitDescs.also(channelDescs).also(map[string]string{
+		// TRANSLATORS: This should not start with a lowercase letter.
+		"cohort": i18n.G("Switch the snap into the given cohort"),
+		// TRANSLATORS: This should not start with a lowercase letter.
+		"leave-cohort": i18n.G("Switch the snap out of its cohort"),
 	}), nil)
-	addCommand("switch", shortSwitchHelp, longSwitchHelp, func() flags.Commander { return &cmdSwitch{} }, waitDescs.also(channelDescs), nil)
 }
