@@ -64,7 +64,7 @@ func (s *generalSuite) TestRoot(c *check.C) {
 	c.Check(rec.HeaderMap.Get("Content-Type"), check.Equals, "application/json")
 
 	expected := []interface{}{"TBD"}
-	var rsp daemon.Resp
+	var rsp daemon.RespJSON
 	c.Assert(json.Unmarshal(rec.Body.Bytes(), &rsp), check.IsNil)
 	c.Check(rsp.Status, check.Equals, 200)
 	c.Check(rsp.Result, check.DeepEquals, expected)
@@ -133,7 +133,7 @@ func (s *generalSuite) TestSysInfo(c *check.C) {
 		"virtualization":   "magic",
 		"system-mode":      "run",
 	}
-	var rsp daemon.Resp
+	var rsp daemon.RespJSON
 	c.Assert(json.Unmarshal(rec.Body.Bytes(), &rsp), check.IsNil)
 	c.Check(rsp.Status, check.Equals, 200)
 	c.Check(rsp.Type, check.Equals, daemon.ResponseTypeSync)
@@ -214,7 +214,7 @@ func (s *generalSuite) TestSysInfoLegacyRefresh(c *check.C) {
 		"virtualization": "kvm",
 		"system-mode":    "run",
 	}
-	var rsp daemon.Resp
+	var rsp daemon.RespJSON
 	c.Assert(json.Unmarshal(rec.Body.Bytes(), &rsp), check.IsNil)
 	c.Check(rsp.Status, check.Equals, 200)
 	c.Check(rsp.Type, check.Equals, daemon.ResponseTypeSync)
@@ -292,7 +292,7 @@ func (s *generalSuite) testSysInfoSystemMode(c *check.C, mode string) {
 		"architecture": arch.DpkgArchitecture(),
 		"system-mode":  mode,
 	}
-	var rsp daemon.Resp
+	var rsp daemon.RespJSON
 	c.Assert(json.Unmarshal(rec.Body.Bytes(), &rsp), check.IsNil)
 	c.Check(rsp.Status, check.Equals, 200)
 	c.Check(rsp.Type, check.Equals, daemon.ResponseTypeSync)
@@ -324,9 +324,7 @@ func (s *generalSuite) TestSysInfoIsManaged(c *check.C) {
 	req, err := http.NewRequest("GET", "/v2/system-info", nil)
 	c.Assert(err, check.IsNil)
 
-	rsp := s.req(c, req, nil).(*daemon.Resp)
-
-	c.Check(rsp.Type, check.Equals, daemon.ResponseTypeSync)
+	rsp := s.syncReq(c, req, nil)
 	c.Check(rsp.Result.(map[string]interface{})["managed"], check.Equals, true)
 }
 
@@ -338,7 +336,7 @@ func (s *generalSuite) TestSysInfoWorksDegraded(c *check.C) {
 	req, err := http.NewRequest("GET", "/v2/system-info", nil)
 	c.Assert(err, check.IsNil)
 
-	rsp := s.req(c, req, nil).(*daemon.Resp)
+	rsp := s.syncReq(c, req, nil)
 	c.Check(rsp.Status, check.Equals, 200)
 }
 
@@ -373,15 +371,16 @@ func (s *generalSuite) TestStateChangesDefaultToInProgress(c *check.C) {
 	// Execute
 	req, err := http.NewRequest("GET", "/v2/changes", nil)
 	c.Assert(err, check.IsNil)
-	rsp := s.req(c, req, nil).(*daemon.Resp)
+	rsp := s.syncReq(c, req, nil)
 
 	// Verify
-	c.Check(rsp.Type, check.Equals, daemon.ResponseTypeSync)
 	c.Check(rsp.Status, check.Equals, 200)
 	c.Assert(rsp.Result, check.HasLen, 1)
 
-	res, err := rsp.MarshalJSON()
-	c.Assert(err, check.IsNil)
+	rec := httptest.NewRecorder()
+	rsp.ServeHTTP(rec, nil)
+	c.Assert(rec.Code, check.Equals, 200)
+	res := rec.Body.Bytes()
 
 	c.Check(string(res), check.Matches, `.*{"id":"\w+","kind":"install","summary":"install...","status":"Do","tasks":\[{"id":"\w+","kind":"download","summary":"1...","status":"Do","log":\["2016-04-21T01:02:03Z INFO l11","2016-04-21T01:02:03Z INFO l12"],"progress":{"label":"","done":0,"total":1},"spawn-time":"2016-04-21T01:02:03Z"}.*`)
 }
@@ -400,15 +399,16 @@ func (s *generalSuite) TestStateChangesInProgress(c *check.C) {
 	// Execute
 	req, err := http.NewRequest("GET", "/v2/changes?select=in-progress", nil)
 	c.Assert(err, check.IsNil)
-	rsp := s.req(c, req, nil).(*daemon.Resp)
+	rsp := s.syncReq(c, req, nil)
 
 	// Verify
-	c.Check(rsp.Type, check.Equals, daemon.ResponseTypeSync)
 	c.Check(rsp.Status, check.Equals, 200)
 	c.Assert(rsp.Result, check.HasLen, 1)
 
-	res, err := rsp.MarshalJSON()
-	c.Assert(err, check.IsNil)
+	rec := httptest.NewRecorder()
+	rsp.ServeHTTP(rec, nil)
+	c.Assert(rec.Code, check.Equals, 200)
+	res := rec.Body.Bytes()
 
 	c.Check(string(res), check.Matches, `.*{"id":"\w+","kind":"install","summary":"install...","status":"Do","tasks":\[{"id":"\w+","kind":"download","summary":"1...","status":"Do","log":\["2016-04-21T01:02:03Z INFO l11","2016-04-21T01:02:03Z INFO l12"],"progress":{"label":"","done":0,"total":1},"spawn-time":"2016-04-21T01:02:03Z"}.*],"ready":false,"spawn-time":"2016-04-21T01:02:03Z"}.*`)
 }
@@ -427,14 +427,16 @@ func (s *generalSuite) TestStateChangesAll(c *check.C) {
 	// Execute
 	req, err := http.NewRequest("GET", "/v2/changes?select=all", nil)
 	c.Assert(err, check.IsNil)
-	rsp := s.req(c, req, nil).(*daemon.Resp)
+	rsp := s.syncReq(c, req, nil)
 
 	// Verify
 	c.Check(rsp.Status, check.Equals, 200)
 	c.Assert(rsp.Result, check.HasLen, 2)
 
-	res, err := rsp.MarshalJSON()
-	c.Assert(err, check.IsNil)
+	rec := httptest.NewRecorder()
+	rsp.ServeHTTP(rec, nil)
+	c.Assert(rec.Code, check.Equals, 200)
+	res := rec.Body.Bytes()
 
 	c.Check(string(res), check.Matches, `.*{"id":"\w+","kind":"install","summary":"install...","status":"Do","tasks":\[{"id":"\w+","kind":"download","summary":"1...","status":"Do","log":\["2016-04-21T01:02:03Z INFO l11","2016-04-21T01:02:03Z INFO l12"],"progress":{"label":"","done":0,"total":1},"spawn-time":"2016-04-21T01:02:03Z"}.*],"ready":false,"spawn-time":"2016-04-21T01:02:03Z"}.*`)
 	c.Check(string(res), check.Matches, `.*{"id":"\w+","kind":"remove","summary":"remove..","status":"Error","tasks":\[{"id":"\w+","kind":"unlink","summary":"1...","status":"Error","log":\["2016-04-21T01:02:03Z ERROR rm failed"],"progress":{"label":"","done":1,"total":1},"spawn-time":"2016-04-21T01:02:03Z","ready-time":"2016-04-21T01:02:03Z"}.*],"ready":true,"err":"[^"]+".*`)
@@ -454,14 +456,16 @@ func (s *generalSuite) TestStateChangesReady(c *check.C) {
 	// Execute
 	req, err := http.NewRequest("GET", "/v2/changes?select=ready", nil)
 	c.Assert(err, check.IsNil)
-	rsp := s.req(c, req, nil).(*daemon.Resp)
+	rsp := s.syncReq(c, req, nil)
 
 	// Verify
 	c.Check(rsp.Status, check.Equals, 200)
 	c.Assert(rsp.Result, check.HasLen, 1)
 
-	res, err := rsp.MarshalJSON()
-	c.Assert(err, check.IsNil)
+	rec := httptest.NewRecorder()
+	rsp.ServeHTTP(rec, nil)
+	c.Assert(rec.Code, check.Equals, 200)
+	res := rec.Body.Bytes()
 
 	c.Check(string(res), check.Matches, `.*{"id":"\w+","kind":"remove","summary":"remove..","status":"Error","tasks":\[{"id":"\w+","kind":"unlink","summary":"1...","status":"Error","log":\["2016-04-21T01:02:03Z ERROR rm failed"],"progress":{"label":"","done":1,"total":1},"spawn-time":"2016-04-21T01:02:03Z","ready-time":"2016-04-21T01:02:03Z"}.*],"ready":true,"err":"[^"]+".*`)
 }
@@ -480,10 +484,9 @@ func (s *generalSuite) TestStateChangesForSnapName(c *check.C) {
 	// Execute
 	req, err := http.NewRequest("GET", "/v2/changes?for=funky-snap-name&select=all", nil)
 	c.Assert(err, check.IsNil)
-	rsp := s.req(c, req, nil).(*daemon.Resp)
+	rsp := s.syncReq(c, req, nil)
 
 	// Verify
-	c.Check(rsp.Type, check.Equals, daemon.ResponseTypeSync)
 	c.Check(rsp.Status, check.Equals, 200)
 	c.Assert(rsp.Result, check.FitsTypeOf, []*daemon.ChangeInfo(nil))
 
@@ -491,8 +494,9 @@ func (s *generalSuite) TestStateChangesForSnapName(c *check.C) {
 	c.Assert(res, check.HasLen, 1)
 	c.Check(res[0].Kind, check.Equals, `install`)
 
-	_, err = rsp.MarshalJSON()
-	c.Assert(err, check.IsNil)
+	rec := httptest.NewRecorder()
+	rsp.ServeHTTP(rec, nil)
+	c.Assert(rec.Code, check.Equals, 200)
 }
 
 func (s *generalSuite) TestStateChangesForSnapNameWithApp(c *check.C) {
@@ -515,10 +519,9 @@ func (s *generalSuite) TestStateChangesForSnapNameWithApp(c *check.C) {
 	// Execute
 	req, err := http.NewRequest("GET", "/v2/changes?for=lxd&select=all", nil)
 	c.Assert(err, check.IsNil)
-	rsp := s.req(c, req, nil).(*daemon.Resp)
+	rsp := s.syncReq(c, req, nil)
 
 	// Verify
-	c.Check(rsp.Type, check.Equals, daemon.ResponseTypeSync)
 	c.Check(rsp.Status, check.Equals, 200)
 	c.Assert(rsp.Result, check.FitsTypeOf, []*daemon.ChangeInfo(nil))
 
@@ -526,8 +529,9 @@ func (s *generalSuite) TestStateChangesForSnapNameWithApp(c *check.C) {
 	c.Assert(res, check.HasLen, 1)
 	c.Check(res[0].Kind, check.Equals, `service-control`)
 
-	_, err = rsp.MarshalJSON()
-	c.Assert(err, check.IsNil)
+	rec := httptest.NewRecorder()
+	rsp.ServeHTTP(rec, nil)
+	c.Assert(rec.Code, check.Equals, 200)
 }
 
 func (s *generalSuite) TestStateChange(c *check.C) {
@@ -546,14 +550,13 @@ func (s *generalSuite) TestStateChange(c *check.C) {
 	// Execute
 	req, err := http.NewRequest("GET", "/v2/changes/"+ids[0], nil)
 	c.Assert(err, check.IsNil)
-	rsp := s.req(c, req, nil).(*daemon.Resp)
+	rsp := s.syncReq(c, req, nil)
 	rec := httptest.NewRecorder()
 	rsp.ServeHTTP(rec, req)
 
 	// Verify
 	c.Check(rec.Code, check.Equals, 200)
 	c.Check(rsp.Status, check.Equals, 200)
-	c.Check(rsp.Type, check.Equals, daemon.ResponseTypeSync)
 	c.Check(rsp.Result, check.NotNil)
 
 	var body map[string]interface{}
@@ -591,6 +594,10 @@ func (s *generalSuite) TestStateChange(c *check.C) {
 	})
 }
 
+func (s *generalSuite) expectManageAccess() {
+	s.expectWriteAccess(daemon.AuthenticatedAccess{Polkit: "io.snapcraft.snapd.manage"})
+}
+
 func (s *generalSuite) TestStateChangeAbort(c *check.C) {
 	restore := state.MockTime(time.Date(2016, 04, 21, 1, 2, 3, 0, time.UTC))
 	defer restore()
@@ -608,12 +615,14 @@ func (s *generalSuite) TestStateChangeAbort(c *check.C) {
 	ids := setupChanges(st)
 	st.Unlock()
 
+	s.expectManageAccess()
+
 	buf := bytes.NewBufferString(`{"action": "abort"}`)
 
 	// Execute
 	req, err := http.NewRequest("POST", "/v2/changes/"+ids[0], buf)
 	c.Assert(err, check.IsNil)
-	rsp := s.req(c, req, nil).(*daemon.Resp)
+	rsp := s.syncReq(c, req, nil)
 	rec := httptest.NewRecorder()
 	rsp.ServeHTTP(rec, req)
 
@@ -623,7 +632,6 @@ func (s *generalSuite) TestStateChangeAbort(c *check.C) {
 	// Verify
 	c.Check(rec.Code, check.Equals, 200)
 	c.Check(rsp.Status, check.Equals, 200)
-	c.Check(rsp.Type, check.Equals, daemon.ResponseTypeSync)
 	c.Check(rsp.Result, check.NotNil)
 
 	var body map[string]interface{}
@@ -673,20 +681,20 @@ func (s *generalSuite) TestStateChangeAbortIsReady(c *check.C) {
 	st.Change(ids[0]).SetStatus(state.DoneStatus)
 	st.Unlock()
 
+	s.expectManageAccess()
+
 	buf := bytes.NewBufferString(`{"action": "abort"}`)
 
 	// Execute
 	req, err := http.NewRequest("POST", "/v2/changes/"+ids[0], buf)
 	c.Assert(err, check.IsNil)
-	rsp := s.req(c, req, nil).(*daemon.Resp)
+	rspe := s.errorReq(c, req, nil)
 	rec := httptest.NewRecorder()
-	rsp.ServeHTTP(rec, req)
+	rspe.ServeHTTP(rec, req)
 
 	// Verify
 	c.Check(rec.Code, check.Equals, 400)
-	c.Check(rsp.Status, check.Equals, 400)
-	c.Check(rsp.Type, check.Equals, daemon.ResponseTypeError)
-	c.Check(rsp.Result, check.NotNil)
+	c.Check(rspe.Status, check.Equals, 400)
 
 	var body map[string]interface{}
 	err = json.Unmarshal(rec.Body.Bytes(), &body)
@@ -699,6 +707,8 @@ func (s *generalSuite) TestStateChangeAbortIsReady(c *check.C) {
 func (s *generalSuite) testWarnings(c *check.C, all bool, body io.Reader) (calls string, result interface{}) {
 	s.daemon(c)
 
+	s.expectManageAccess()
+
 	okayWarns := func(*state.State, time.Time) int { calls += "ok"; return 0 }
 	allWarns := func(*state.State) []*state.Warning { calls += "all"; return nil }
 	pendingWarns := func(*state.State) ([]*state.Warning, time.Time) { calls += "show"; return nil, time.Time{} }
@@ -716,10 +726,8 @@ func (s *generalSuite) testWarnings(c *check.C, all bool, body io.Reader) (calls
 	req, err := http.NewRequest(method, "/v2/warnings?"+q.Encode(), body)
 	c.Assert(err, check.IsNil)
 
-	rsp, ok := s.req(c, req, nil).(*daemon.Resp)
-	c.Assert(ok, check.Equals, true)
+	rsp := s.syncReq(c, req, nil)
 
-	c.Check(rsp.Type, check.Equals, daemon.ResponseTypeSync)
 	c.Check(rsp.Status, check.Equals, 200)
 	c.Assert(rsp.Result, check.NotNil)
 	return calls, rsp.Result
