@@ -20,6 +20,9 @@
 package gadget
 
 import (
+	"bufio"
+	"bytes"
+	"encoding/json"
 	"errors"
 	"fmt"
 	"io/ioutil"
@@ -35,8 +38,11 @@ import (
 	"github.com/snapcore/snapd/gadget/edition"
 	"github.com/snapcore/snapd/gadget/quantity"
 	"github.com/snapcore/snapd/metautil"
+	"github.com/snapcore/snapd/osutil"
+	"github.com/snapcore/snapd/osutil/disks"
 	"github.com/snapcore/snapd/snap"
 	"github.com/snapcore/snapd/snap/naming"
+	"github.com/snapcore/snapd/snap/snapfile"
 	"github.com/snapcore/snapd/strutil"
 )
 
@@ -108,12 +114,16 @@ type Volume struct {
 	ID string `yaml:"id"`
 	// Structure describes the structures that are part of the volume
 	Structure []VolumeStructure `yaml:"structure"`
+	// Name is the name of the volume from the gadget.yaml
+	Name string
 }
 
 // VolumeStructure describes a single structure inside a volume. A structure can
 // represent a partition, Master Boot Record, or any other contiguous range
 // within the volume.
 type VolumeStructure struct {
+	// VolumeName is the name of the volume that this structure belongs to.
+	VolumeName string
 	// Name, when non empty, provides the name of the structure
 	Name string `yaml:"name"`
 	// Label provides the filesystem label
@@ -139,7 +149,8 @@ type VolumeStructure struct {
 	// 'system-boot-select' or 'system-recovery-select'. Structures of type 'mbr', must have a
 	// size of 446 bytes and must start at 0 offset.
 	Role string `yaml:"role"`
-	// ID is the GPT partition ID
+	// ID is the GPT partition ID, this should always be made upper case for
+	// comparison purposes.
 	ID string `yaml:"id"`
 	// Filesystem used for the partition, 'vfat', 'ext4' or 'none' for
 	// structures of type 'bare'
@@ -184,20 +195,6 @@ type VolumeContent struct {
 	Size quantity.Size `yaml:"size"`
 
 	Unpack bool `yaml:"unpack"`
-
-	// resolvedSource is the absolute path of the Source after resolving
-	// any references (e.g. to a "$kernel:" snap)
-	resolvedSource string
-	// TODO: provide resolvedImage too
-}
-
-func (vc VolumeContent) ResolvedSource() string {
-	// TODO: ensure that sources are always resolved and only return
-	//       vc.resolvedSource(). This will come in the next PR.
-	if vc.resolvedSource != "" {
-		return vc.resolvedSource
-	}
-	return vc.UnresolvedSource
 }
 
 func (vc VolumeContent) String() string {
@@ -212,6 +209,239 @@ type VolumeUpdate struct {
 	Preserve []string       `yaml:"preserve"`
 }
 
+// DiskVolumeDeviceTraits is a set of traits about a disk that were measured at
+// a previous point in time on the same device, and is used primarily to try and
+// map a volume in the gadget.yaml to a physical device on the system after the
+// initial installation is done. We don't have a steadfast and predictable way
+// to always find the device again, so we need to do a search, trying to find a
+// device which matches each trait in turn, and verify it matches the  physical
+// structure layout and if not move on to using the next trait.
+type DiskVolumeDeviceTraits struct {
+	// each member here is presented in descending order of certainty about the
+	// likelihood of being compatible if a candidate physical device matches the
+	// member. I.e. OriginalDevicePath is more trusted than OriginalKernelPath is
+	// more trusted than DiskID is more trusted than using the MappedStructures
+
+	// OriginalDevicePath is the device path in sysfs and in /dev/disk/by-path
+	// the volume was measured and observed at during UC20+ install mode.
+	OriginalDevicePath string `json:"device-path"`
+
+	// OriginalKernelPath is the device path like /dev/vda the volume was
+	// measured and observed at during UC20+ install mode.
+	OriginalKernelPath string `json:"kernel-path"`
+
+	// DiskID is the disk's identifier, it is a UUID for GPT disks or an
+	// unsigned integer for DOS disks encoded as a string in hexadecimal as in
+	// "0x1212e868".
+	DiskID string `json:"disk-id"`
+
+	// Size is the physical size of the disk, regardless of usable space
+	// considerations.
+	Size quantity.Size `json:"size"`
+
+	// SectorSize is the physical sector size of the disk, typically 512 or
+	// 4096.
+	SectorSize quantity.Size `json:"sector-size"`
+
+	// Schema is the disk schema, either dos or gpt in lowercase.
+	Schema string `json:"schema"`
+
+	// Structure contains trait information about each individual structure in
+	// the volume that may be useful in identifying whether a disk matches a
+	// volume or not.
+	Structure []DiskStructureDeviceTraits `json:"structure"`
+
+	// StructureEncryption is the set of partitions that are encrypted on the
+	// volume - this should only ever have ubuntu-data or ubuntu-save keys for
+	// now in the map. The value indicates parameters of the encryption present
+	// that enable matching/identifying encrypted structures with their laid out
+	// counterparts in the gadget.yaml.
+	StructureEncryption map[string]StructureEncryptionParameters `json:"structure-encryption"`
+}
+
+// StructureEncryptionParameters contains information about an encrypted
+// structure, used to match encrypted structures on disk with their abstract,
+// laid out counterparts in the gadget.yaml.
+type StructureEncryptionParameters struct {
+	// Method is the method of encryption used, currently only EncryptionLUKS is
+	// recognized.
+	Method DiskEncryptionMethod `json:"method"`
+
+	// unknownKeys is used to log messages about unknown, unrecognized keys that
+	// we may encounter and may be used in the future
+	unknownKeys map[string]string
+}
+
+func (s *StructureEncryptionParameters) UnmarshalJSON(b []byte) error {
+	m := map[string]string{}
+
+	if err := json.Unmarshal(b, &m); err != nil {
+		return err
+	}
+
+	for key, val := range m {
+		if key == "method" {
+			s.Method = DiskEncryptionMethod(val)
+		} else {
+			if s.unknownKeys == nil {
+				s.unknownKeys = make(map[string]string)
+			}
+			s.unknownKeys[key] = val
+		}
+	}
+
+	return nil
+}
+
+// DiskStructureDeviceTraits is a similar to DiskVolumeDeviceTraits, but is a
+// set of traits for a specific structure on a disk rather than the full disk
+// itself. Structures can be full partitions or just raw slices on a disk like
+// the "BIOS Boot" structure on default amd64 grub Ubuntu Core systems.
+type DiskStructureDeviceTraits struct {
+	// OriginalDevicePath is the device path in sysfs and in /dev/disk/by-path the
+	// partition was measured and observed at during UC20+ install mode.
+	OriginalDevicePath string `json:"device-path"`
+	// OriginalKernelPath is the device path like /dev/vda1 the partition was
+	// measured and observed at during UC20+ install mode.
+	OriginalKernelPath string `json:"kernel-path"`
+	// PartitionUUID is the partuuid as defined by i.e. /dev/disk/by-partuuid
+	PartitionUUID string `json:"partition-uuid"`
+	// PartitionLabel is the label of the partition for GPT disks, i.e.
+	// /dev/disk/by-partlabel
+	PartitionLabel string `json:"partition-label"`
+	// PartitionType is the type of the partition i.e. 0x83 for a
+	// Linux native partition on DOS, or
+	// 0FC63DAF-8483-4772-8E79-3D69D8477DE4 for a Linux filesystem
+	// data partition on GPT.
+	PartitionType string `json:"partition-type"`
+	// FilesystemUUID is the UUID of the filesystem on the partition, i.e.
+	// /dev/disk/by-uuid
+	FilesystemUUID string `json:"filesystem-uuid"`
+	// FilesystemLabel is the label of the filesystem for structures that have
+	// filesystems, i.e. /dev/disk/by-label
+	FilesystemLabel string `json:"filesystem-label"`
+	// FilesystemType is the type of the filesystem, i.e. vfat or ext4, etc.
+	FilesystemType string `json:"filesystem-type"`
+	// Offset is the offset of the structure
+	Offset quantity.Offset `json:"offset"`
+	// Size is the size of the structure
+	Size quantity.Size `json:"size"`
+}
+
+// SaveDiskVolumesDeviceTraits saves the mapping of volume names to volume /
+// device traits to a file inside the provided directory on disk for
+// later loading and verification.
+func SaveDiskVolumesDeviceTraits(dir string, mapping map[string]DiskVolumeDeviceTraits) error {
+	b, err := json.Marshal(mapping)
+	if err != nil {
+		return err
+	}
+
+	filename := filepath.Join(dir, "disk-mapping.json")
+
+	if err := os.MkdirAll(filepath.Dir(filename), 0755); err != nil {
+		return err
+	}
+	return osutil.AtomicWriteFile(filename, b, 0644, 0)
+}
+
+// LoadDiskVolumesDeviceTraits loads the mapping of volumes to disk traits if
+// there is any. If there is no file with the mapping available, nil is
+// returned.
+func LoadDiskVolumesDeviceTraits(dir string) (map[string]DiskVolumeDeviceTraits, error) {
+	var mapping map[string]DiskVolumeDeviceTraits
+
+	filename := filepath.Join(dir, "disk-mapping.json")
+	if !osutil.FileExists(filename) {
+		return nil, nil
+	}
+
+	b, err := ioutil.ReadFile(filename)
+	if err != nil {
+		return nil, err
+	}
+
+	if err := json.Unmarshal(b, &mapping); err != nil {
+		return nil, err
+	}
+
+	return mapping, nil
+}
+
+// AllDiskVolumeDeviceTraits takes a mapping of volume name to LaidOutVolume and
+// produces a map of volume name to DiskVolumeDeviceTraits. Since doing so uses
+// DiskVolumeDeviceTraitsForDevice, it will also validate that disk devices
+// identified for the laid out volume are compatible and matching before
+// returning.
+func AllDiskVolumeDeviceTraits(allLaidOutVols map[string]*LaidOutVolume, optsPerVolume map[string]*DiskVolumeValidationOptions) (map[string]DiskVolumeDeviceTraits, error) {
+	// build up the mapping of volumes to disk device traits
+
+	allVols := map[string]DiskVolumeDeviceTraits{}
+
+	// find all devices which map to volumes to save the current state of the
+	// system
+	for name, vol := range allLaidOutVols {
+		// try to find a device for a structure inside the volume, we have a
+		// loop to attempt to use all structures in the volume in case there are
+		// partitions we can't map to a device directly at first using the
+		// device symlinks that FindDeviceForStructure uses
+		dev := ""
+		for _, vs := range vol.LaidOutStructure {
+			// TODO: This code works for volumes that have at least one
+			// partition (i.e. not type: bare structure), but does not work for
+			// volumes which contain only type: bare structures with no other
+			// structures on them. It is entirely unclear how to identify such
+			// a volume, since there is no information on the disk about where
+			// such raw structures begin and end and thus no way to validate
+			// that a given disk "has" such raw structures at particular
+			// locations, aside from potentially reading and comparing the bytes
+			// at the expected locations, but that is probably fragile and very
+			// non-performant.
+
+			if !vs.IsPartition() {
+				// skip trying to find non-partitions on disk, it won't work
+				continue
+			}
+
+			structureDevice, err := FindDeviceForStructure(&vs)
+			if err != nil && err != ErrDeviceNotFound {
+				return nil, err
+			}
+			if structureDevice != "" {
+				// we found a device for this structure, get the parent disk
+				// and save that as the device for this volume
+				disk, err := disks.DiskFromPartitionDeviceNode(structureDevice)
+				if err != nil {
+					return nil, err
+				}
+
+				dev = disk.KernelDeviceNode()
+				break
+			}
+		}
+
+		if dev == "" {
+			return nil, fmt.Errorf("cannot find disk for volume %s from gadget", name)
+		}
+
+		// now that we have a candidate device for this disk, build up the
+		// traits for it, this will also validate concretely that the
+		// device we picked and the volume are compatible
+		opts := optsPerVolume[name]
+		if opts == nil {
+			opts = &DiskVolumeValidationOptions{}
+		}
+		traits, err := DiskTraitsFromDeviceAndValidate(vol, dev, opts)
+		if err != nil {
+			return nil, fmt.Errorf("cannot gather disk traits for device %s to use with volume %s: %v", dev, name, err)
+		}
+
+		allVols[name] = traits
+	}
+
+	return allVols, nil
+}
+
 // GadgetConnect describes an interface connection requested by the gadget
 // between seeded snaps. The syntax is of a mapping like:
 //
@@ -306,6 +536,10 @@ func wantsSystemSeed(m Model) bool {
 	return m != nil && m.Grade() != asserts.ModelGradeUnset
 }
 
+func compatWithPibootOrIndeterminate(m Model) bool {
+	return m == nil || m.Grade() != asserts.ModelGradeUnset
+}
+
 // InfoFromGadgetYaml parses the provided gadget metadata.
 // If model is nil only self-consistency checks are performed.
 // If model is not nil implied values for filesystem labels will be set
@@ -348,8 +582,14 @@ func InfoFromGadgetYaml(gadgetYaml []byte, model Model) (*Info, error) {
 	// basic validation
 	var bootloadersFound int
 	knownFsLabelsPerVolume := make(map[string]map[string]bool, len(gi.Volumes))
-	for name, v := range gi.Volumes {
-		if err := validateVolume(name, v, model, knownFsLabelsPerVolume); err != nil {
+	for name := range gi.Volumes {
+		v := gi.Volumes[name]
+		if v == nil {
+			return nil, fmt.Errorf("volume %q stanza is empty", name)
+		}
+		// set the VolumeName for the volume
+		v.Name = name
+		if err := validateVolume(v, knownFsLabelsPerVolume); err != nil {
 			return nil, fmt.Errorf("invalid volume %q: %v", name, err)
 		}
 
@@ -358,8 +598,13 @@ func InfoFromGadgetYaml(gadgetYaml []byte, model Model) (*Info, error) {
 			// pass
 		case "grub", "u-boot", "android-boot", "lk":
 			bootloadersFound += 1
+		case "piboot":
+			if !compatWithPibootOrIndeterminate(model) {
+				return nil, errors.New("piboot bootloader valid only for UC20 onwards")
+			}
+			bootloadersFound += 1
 		default:
-			return nil, errors.New("bootloader must be one of grub, u-boot, android-boot or lk")
+			return nil, errors.New("bootloader must be one of grub, u-boot, android-boot, piboot or lk")
 		}
 	}
 	switch {
@@ -370,7 +615,7 @@ func InfoFromGadgetYaml(gadgetYaml []byte, model Model) (*Info, error) {
 	}
 
 	for name, v := range gi.Volumes {
-		if err := setImplicitForVolume(name, v, model, knownFsLabelsPerVolume[name]); err != nil {
+		if err := setImplicitForVolume(v, model, knownFsLabelsPerVolume[name]); err != nil {
 			return nil, fmt.Errorf("invalid volume %q: %v", name, err)
 		}
 	}
@@ -396,13 +641,15 @@ func whichVolRuleset(model Model) volRuleset {
 	return volRuleset16
 }
 
-func setImplicitForVolume(name string, vol *Volume, model Model, knownFsLabels map[string]bool) error {
+func setImplicitForVolume(vol *Volume, model Model, knownFsLabels map[string]bool) error {
 	rs := whichVolRuleset(model)
 	if vol.Schema == "" {
 		// default for schema is gpt
 		vol.Schema = schemaGPT
 	}
 	for i := range vol.Structure {
+		// set the VolumeName for the structure from the volume itself
+		vol.Structure[i].VolumeName = vol.Name
 		if err := setImplicitForVolumeStructure(&vol.Structure[i], rs, knownFsLabels); err != nil {
 			return err
 		}
@@ -523,8 +770,8 @@ func fmtIndexAndName(idx int, name string) string {
 	return fmt.Sprintf("#%v", idx)
 }
 
-func validateVolume(name string, vol *Volume, model Model, knownFsLabelsPerVolume map[string]map[string]bool) error {
-	if !validVolumeName.MatchString(name) {
+func validateVolume(vol *Volume, knownFsLabelsPerVolume map[string]map[string]bool) error {
+	if !validVolumeName.MatchString(vol.Name) {
 		return errors.New("invalid name")
 	}
 	if vol.Schema != "" && vol.Schema != schemaGPT && vol.Schema != schemaMBR {
@@ -539,7 +786,7 @@ func validateVolume(name string, vol *Volume, model Model, knownFsLabelsPerVolum
 	structures := make([]LaidOutStructure, len(vol.Structure))
 
 	if knownFsLabelsPerVolume != nil {
-		knownFsLabelsPerVolume[name] = knownFsLabels
+		knownFsLabelsPerVolume[vol.Name] = knownFsLabels
 	}
 
 	previousEnd := quantity.Offset(0)
@@ -560,7 +807,7 @@ func validateVolume(name string, vol *Volume, model Model, knownFsLabelsPerVolum
 		ps := LaidOutStructure{
 			VolumeStructure: &vol.Structure[idx],
 			StartOffset:     start,
-			Index:           idx,
+			YamlIndex:       idx,
 		}
 		structures[idx] = ps
 		if s.Name != "" {
@@ -649,7 +896,7 @@ func validateVolumeStructure(vs *VolumeStructure, vol *Volume) error {
 	if err := validateStructureType(vs.Type, vol); err != nil {
 		return fmt.Errorf("invalid type %q: %v", vs.Type, err)
 	}
-	if err := validateRole(vs, vol); err != nil {
+	if err := validateRole(vs); err != nil {
 		var what string
 		if vs.Role != "" {
 			what = fmt.Sprintf("role %q", vs.Role)
@@ -675,7 +922,7 @@ func validateVolumeStructure(vs *VolumeStructure, vol *Volume) error {
 		}
 	}
 
-	if err := validateStructureUpdate(&vs.Update, vs); err != nil {
+	if err := validateStructureUpdate(vs); err != nil {
 		return err
 	}
 
@@ -748,7 +995,7 @@ func validateStructureType(s string, vol *Volume) error {
 	return nil
 }
 
-func validateRole(vs *VolumeStructure, vol *Volume) error {
+func validateRole(vs *VolumeStructure) error {
 	if vs.Type == "bare" {
 		if vs.Role != "" && vs.Role != schemaMBR {
 			return fmt.Errorf("conflicting type: %q", vs.Type)
@@ -806,13 +1053,16 @@ func validateFilesystemContent(vc *VolumeContent) error {
 	if vc.Image != "" || vc.Offset != nil || vc.OffsetWrite != nil || vc.Size != 0 {
 		return fmt.Errorf("cannot use image content for non-bare file system")
 	}
-	if vc.UnresolvedSource == "" || vc.Target == "" {
-		return fmt.Errorf("missing source or target")
+	if vc.UnresolvedSource == "" {
+		return fmt.Errorf("missing source")
+	}
+	if vc.Target == "" {
+		return fmt.Errorf("missing target")
 	}
 	return nil
 }
 
-func validateStructureUpdate(up *VolumeUpdate, vs *VolumeStructure) error {
+func validateStructureUpdate(vs *VolumeStructure) error {
 	if !vs.HasFilesystem() && len(vs.Update.Preserve) > 0 {
 		return errors.New("preserving files during update is not supported for non-filesystem structures")
 	}
@@ -901,6 +1151,7 @@ func (s *RelativeOffset) UnmarshalYAML(unmarshal func(interface{}) error) error
 
 // IsCompatible checks whether the current and an update are compatible. Returns
 // nil or an error describing the incompatibility.
+// TODO: make this reasonably consistent with Update for multi-volume scenarios
 func IsCompatible(current, new *Info) error {
 	// XXX: the only compatibility we have now is making sure that the new
 	// layout can be used on an existing volume
@@ -923,11 +1174,11 @@ func IsCompatible(current, new *Info) error {
 	// layout both volumes partially, without going deep into the layout of
 	// structure content, we only want to make sure that structures are
 	// comapatible
-	pCurrent, err := LayoutVolumePartially(currentVol, defaultConstraints)
+	pCurrent, err := LayoutVolumePartially(currentVol, DefaultConstraints)
 	if err != nil {
 		return fmt.Errorf("cannot lay out the current volume: %v", err)
 	}
-	pNew, err := LayoutVolumePartially(newVol, defaultConstraints)
+	pNew, err := LayoutVolumePartially(newVol, DefaultConstraints)
 	if err != nil {
 		return fmt.Errorf("cannot lay out the new volume: %v", err)
 	}
@@ -937,32 +1188,60 @@ func IsCompatible(current, new *Info) error {
 	return nil
 }
 
-// LaidOutVolumeFromGadget takes a gadget rootdir and lays out the
-// partitions as specified.
-func LaidOutVolumeFromGadget(gadgetRoot string, model Model) (*LaidOutVolume, error) {
-	info, err := ReadInfo(gadgetRoot, model)
-	if err != nil {
-		return nil, err
+// LaidOutVolumesFromGadget takes a gadget rootdir and lays out the partitions
+// on all volumes as specified. It returns the specific volume on which system-*
+// roles/partitions exist, as well as all volumes mentioned in the gadget.yaml
+// and their laid out representations. Those volumes are assumed to already be
+// flashed and managed separately at image build/flash time, while the system
+// volume with all the system-* roles on it can be manipulated during install
+// mode.
+func LaidOutVolumesFromGadget(gadgetRoot, kernelRoot string, model Model) (system *LaidOutVolume, all map[string]*LaidOutVolume, err error) {
+	all = make(map[string]*LaidOutVolume)
+	// model should never be nil here
+	if model == nil {
+		return nil, nil, fmt.Errorf("internal error: must have model to lay out system volumes from a gadget")
 	}
-	// Limit ourselves to just one volume for now.
-	if len(info.Volumes) != 1 {
-		return nil, fmt.Errorf("cannot position multiple volumes yet")
+	// rely on the basic validation from ReadInfo to ensure that the system-*
+	// roles are all on the same volume for example
+	info, err := ReadInfoAndValidate(gadgetRoot, model, nil)
+	if err != nil {
+		return nil, nil, err
 	}
 
 	constraints := LayoutConstraints{
 		NonMBRStartOffset: 1 * quantity.OffsetMiB,
-		SectorSize:        512,
 	}
 
-	for _, vol := range info.Volumes {
-		pvol, err := LayoutVolume(gadgetRoot, vol, constraints)
+	// find the volume with the system-boot role on it, we already validated
+	// that the system-* roles are all on the same volume
+	for name, vol := range info.Volumes {
+		// layout all volumes saving them
+		lvol, err := LayoutVolume(gadgetRoot, kernelRoot, vol, constraints)
 		if err != nil {
-			return nil, err
+			return nil, nil, err
+		}
+		all[name] = lvol
+		// check if this volume is the boot volume using the system-boot volume
+		// to identify it
+		for _, structure := range vol.Structure {
+			if structure.Role == SystemBoot {
+				if system != nil {
+					// this should be impossible, the validation above should
+					// ensure there are not multiple volumes with the same role
+					// on them
+					return nil, nil, fmt.Errorf("internal error: gadget passed validation but duplicated system-* roles across multiple volumes")
+				}
+				system = lvol
+			}
 		}
-		// we know  info.Volumes map has size 1 so we can return here
-		return pvol, nil
 	}
-	return nil, fmt.Errorf("internal error in PositionedVolumeFromGadget: this line cannot be reached")
+
+	if system == nil {
+		// this should be impossible, the validation above should ensure this
+		return nil, nil, fmt.Errorf("internal error: gadget passed validation but does not have system-* roles on any volume")
+	}
+
+	return system, all, nil
 }
 
 func flatten(path string, cfg interface{}, out map[string]interface{}) {
@@ -990,3 +1269,107 @@ func SystemDefaults(gadgetDefaults map[string]map[string]interface{}) map[string
 	}
 	return nil
 }
+
+// See https://www.kernel.org/doc/html/latest/admin-guide/kernel-parameters.html
+var disallowedKernelArguments = []string{
+	"root", "nfsroot",
+	"init",
+}
+
+// isKernelArgumentAllowed checks whether the kernel command line argument is
+// allowed. Prohibits all arguments listed explicitly in
+// disallowedKernelArguments list and those prefixed with snapd, with exception
+// of snapd.debug. All other arguments are allowed.
+func isKernelArgumentAllowed(arg string) bool {
+	if strings.HasPrefix(arg, "snapd") && arg != "snapd.debug" {
+		return false
+	}
+	if strutil.ListContains(disallowedKernelArguments, arg) {
+		return false
+	}
+	return true
+}
+
+var ErrNoKernelCommandline = errors.New("no kernel command line in the gadget")
+
+// KernelCommandLineFromGadget returns the desired kernel command line provided by the
+// gadget. The full flag indicates whether the gadget provides a full command
+// line or just the extra parameters that will be appended to the static ones.
+// An ErrNoKernelCommandline is returned when thea gadget does not set any
+// kernel command line.
+func KernelCommandLineFromGadget(gadgetDirOrSnapPath string) (cmdline string, full bool, err error) {
+	sf, err := snapfile.Open(gadgetDirOrSnapPath)
+	if err != nil {
+		return "", false, fmt.Errorf("cannot open gadget snap: %v", err)
+	}
+	contentExtra, err := sf.ReadFile("cmdline.extra")
+	if err != nil && !os.IsNotExist(err) {
+		return "", false, err
+	}
+	// TODO: should we enforce the maximum kernel command line for cmdline.full?
+	contentFull, err := sf.ReadFile("cmdline.full")
+	if err != nil && !os.IsNotExist(err) {
+		return "", false, err
+	}
+	content := contentExtra
+	whichFile := "cmdline.extra"
+	switch {
+	case contentExtra != nil && contentFull != nil:
+		return "", false, fmt.Errorf("cannot support both extra and full kernel command lines")
+	case contentExtra == nil && contentFull == nil:
+		return "", false, ErrNoKernelCommandline
+	case contentFull != nil:
+		content = contentFull
+		whichFile = "cmdline.full"
+		full = true
+	}
+	parsed, err := parseCommandLineFromGadget(content)
+	if err != nil {
+		return "", full, fmt.Errorf("invalid kernel command line in %v: %v", whichFile, err)
+	}
+	return parsed, full, nil
+}
+
+// parseCommandLineFromGadget parses the command line file and returns a
+// reassembled kernel command line as a single string. The file can be multi
+// line, where only lines stating with # are treated as comments, eg.
+//
+// foo
+// # this is a comment
+//
+// According to https://elixir.bootlin.com/linux/latest/source/Documentation/admin-guide/kernel-parameters.txt
+// the # character can appear as part of a valid kernel command line argument,
+// specifically in the following argument:
+//   memmap=nn[KMG]#ss[KMG]
+//   memmap=100M@2G,100M#3G,1G!1024G
+// Thus a lone # or a token starting with # are treated as errors.
+func parseCommandLineFromGadget(content []byte) (string, error) {
+	s := bufio.NewScanner(bytes.NewBuffer(content))
+	filtered := &bytes.Buffer{}
+	for s.Scan() {
+		line := s.Text()
+		if len(line) > 0 && line[0] == '#' {
+			// comment
+			continue
+		}
+		filtered.WriteRune(' ')
+		filtered.WriteString(line)
+	}
+	if err := s.Err(); err != nil {
+		return "", err
+	}
+	kargs, err := osutil.KernelCommandLineSplit(filtered.String())
+	if err != nil {
+		return "", err
+	}
+	for _, argValue := range kargs {
+		if strings.HasPrefix(argValue, "#") {
+			return "", fmt.Errorf("unexpected or invalid use of # in argument %q", argValue)
+		}
+		split := strings.SplitN(argValue, "=", 2)
+		if !isKernelArgumentAllowed(split[0]) {
+			return "", fmt.Errorf("disallowed kernel argument %q", argValue)
+		}
+	}
+	return strings.Join(kargs, " "), nil
+}
