@@ -47,7 +47,6 @@ func (p *layoutTestSuite) SetUpTest(c *C) {
 
 var defaultConstraints = gadget.LayoutConstraints{
 	NonMBRStartOffset: 1 * quantity.OffsetMiB,
-	SectorSize:        512,
 }
 
 func (p *layoutTestSuite) TestVolumeSize(c *C) {
@@ -56,7 +55,7 @@ func (p *layoutTestSuite) TestVolumeSize(c *C) {
 			{Size: 2 * quantity.SizeMiB},
 		},
 	}
-	v, err := gadget.LayoutVolume(p.dir, &vol, defaultConstraints)
+	v, err := gadget.LayoutVolume(p.dir, "", &vol, defaultConstraints)
 	c.Assert(err, IsNil)
 
 	c.Assert(v, DeepEquals, &gadget.LaidOutVolume{
@@ -65,9 +64,8 @@ func (p *layoutTestSuite) TestVolumeSize(c *C) {
 				{Size: 2 * quantity.SizeMiB},
 			},
 		},
-		Size:       3 * quantity.SizeMiB,
-		SectorSize: 512,
-		RootDir:    p.dir,
+		Size:    3 * quantity.SizeMiB,
+		RootDir: p.dir,
 		LaidOutStructure: []gadget.LaidOutStructure{{
 			VolumeStructure: &gadget.VolumeStructure{Size: 2 * quantity.SizeMiB},
 			StartOffset:     1 * quantity.OffsetMiB,
@@ -98,24 +96,23 @@ volumes:
 	vol := mustParseVolume(c, gadgetYaml, "first-image")
 	c.Assert(vol.Structure, HasLen, 2)
 
-	v, err := gadget.LayoutVolume(p.dir, vol, defaultConstraints)
+	v, err := gadget.LayoutVolume(p.dir, "", vol, defaultConstraints)
 	c.Assert(err, IsNil)
 
 	c.Assert(v, DeepEquals, &gadget.LaidOutVolume{
-		Volume:     vol,
-		Size:       501 * quantity.SizeMiB,
-		SectorSize: 512,
-		RootDir:    p.dir,
+		Volume:  vol,
+		Size:    501 * quantity.SizeMiB,
+		RootDir: p.dir,
 		LaidOutStructure: []gadget.LaidOutStructure{
 			{
 				VolumeStructure: &vol.Structure[0],
 				StartOffset:     1 * quantity.OffsetMiB,
-				Index:           0,
+				YamlIndex:       0,
 			},
 			{
 				VolumeStructure: &vol.Structure[1],
 				StartOffset:     401 * quantity.OffsetMiB,
-				Index:           1,
+				YamlIndex:       1,
 			},
 		},
 	})
@@ -141,34 +138,33 @@ volumes:
 	vol := mustParseVolume(c, gadgetYaml, "first")
 	c.Assert(vol.Structure, HasLen, 4)
 
-	v, err := gadget.LayoutVolume(p.dir, vol, defaultConstraints)
+	v, err := gadget.LayoutVolume(p.dir, "", vol, defaultConstraints)
 	c.Assert(err, IsNil)
 
 	c.Assert(v, DeepEquals, &gadget.LaidOutVolume{
-		Volume:     vol,
-		Size:       1101 * quantity.SizeMiB,
-		SectorSize: 512,
-		RootDir:    p.dir,
+		Volume:  vol,
+		Size:    1101 * quantity.SizeMiB,
+		RootDir: p.dir,
 		LaidOutStructure: []gadget.LaidOutStructure{
 			{
 				VolumeStructure: &vol.Structure[0],
 				StartOffset:     1 * quantity.OffsetMiB,
-				Index:           0,
+				YamlIndex:       0,
 			},
 			{
 				VolumeStructure: &vol.Structure[1],
 				StartOffset:     401 * quantity.OffsetMiB,
-				Index:           1,
+				YamlIndex:       1,
 			},
 			{
 				VolumeStructure: &vol.Structure[2],
 				StartOffset:     901 * quantity.OffsetMiB,
-				Index:           2,
+				YamlIndex:       2,
 			},
 			{
 				VolumeStructure: &vol.Structure[3],
 				StartOffset:     1001 * quantity.OffsetMiB,
-				Index:           3,
+				YamlIndex:       3,
 			},
 		},
 	})
@@ -198,34 +194,33 @@ volumes:
 	vol := mustParseVolume(c, gadgetYaml, "first")
 	c.Assert(vol.Structure, HasLen, 4)
 
-	v, err := gadget.LayoutVolume(p.dir, vol, defaultConstraints)
+	v, err := gadget.LayoutVolume(p.dir, "", vol, defaultConstraints)
 	c.Assert(err, IsNil)
 
 	c.Assert(v, DeepEquals, &gadget.LaidOutVolume{
-		Volume:     vol,
-		Size:       1300 * quantity.SizeMiB,
-		SectorSize: 512,
-		RootDir:    p.dir,
+		Volume:  vol,
+		Size:    1300 * quantity.SizeMiB,
+		RootDir: p.dir,
 		LaidOutStructure: []gadget.LaidOutStructure{
 			{
 				VolumeStructure: &vol.Structure[3],
 				StartOffset:     1 * quantity.OffsetMiB,
-				Index:           3,
+				YamlIndex:       3,
 			},
 			{
 				VolumeStructure: &vol.Structure[1],
 				StartOffset:     200 * quantity.OffsetMiB,
-				Index:           1,
+				YamlIndex:       1,
 			},
 			{
 				VolumeStructure: &vol.Structure[0],
 				StartOffset:     800 * quantity.OffsetMiB,
-				Index:           0,
+				YamlIndex:       0,
 			},
 			{
 				VolumeStructure: &vol.Structure[2],
 				StartOffset:     1200 * quantity.OffsetMiB,
-				Index:           2,
+				YamlIndex:       2,
 			},
 		},
 	})
@@ -254,34 +249,33 @@ volumes:
 	vol := mustParseVolume(c, gadgetYaml, "first")
 	c.Assert(vol.Structure, HasLen, 4)
 
-	v, err := gadget.LayoutVolume(p.dir, vol, defaultConstraints)
+	v, err := gadget.LayoutVolume(p.dir, "", vol, defaultConstraints)
 	c.Assert(err, IsNil)
 
 	c.Assert(v, DeepEquals, &gadget.LaidOutVolume{
-		Volume:     vol,
-		Size:       1200 * quantity.SizeMiB,
-		SectorSize: 512,
-		RootDir:    p.dir,
+		Volume:  vol,
+		Size:    1200 * quantity.SizeMiB,
+		RootDir: p.dir,
 		LaidOutStructure: []gadget.LaidOutStructure{
 			{
 				VolumeStructure: &vol.Structure[3],
 				StartOffset:     1 * quantity.OffsetMiB,
-				Index:           3,
+				YamlIndex:       3,
 			},
 			{
 				VolumeStructure: &vol.Structure[1],
 				StartOffset:     200 * quantity.OffsetMiB,
-				Index:           1,
+				YamlIndex:       1,
 			},
 			{
 				VolumeStructure: &vol.Structure[2],
 				StartOffset:     700 * quantity.OffsetMiB,
-				Index:           2,
+				YamlIndex:       2,
 			},
 			{
 				VolumeStructure: &vol.Structure[0],
 				StartOffset:     800 * quantity.OffsetMiB,
-				Index:           0,
+				YamlIndex:       0,
 			},
 		},
 	})
@@ -301,7 +295,7 @@ volumes:
               - image: foo.img
 `
 	vol := mustParseVolume(c, gadgetYaml, "first")
-	v, err := gadget.LayoutVolume(p.dir, vol, defaultConstraints)
+	v, err := gadget.LayoutVolume(p.dir, "", vol, defaultConstraints)
 	c.Assert(v, IsNil)
 	c.Assert(err, ErrorMatches, `cannot lay out structure #0: content "foo.img":.*no such file or directory`)
 }
@@ -339,7 +333,7 @@ volumes:
 
 	vol := mustParseVolume(c, gadgetYaml, "first")
 
-	v, err := gadget.LayoutVolume(p.dir, vol, defaultConstraints)
+	v, err := gadget.LayoutVolume(p.dir, "", vol, defaultConstraints)
 	c.Assert(v, IsNil)
 	c.Assert(err, ErrorMatches, `cannot lay out structure #0: content "foo.img" does not fit in the structure`)
 }
@@ -364,9 +358,8 @@ volumes:
 
 	constraints := gadget.LayoutConstraints{
 		NonMBRStartOffset: 1 * quantity.OffsetMiB,
-		SectorSize:        512,
 	}
-	v, err := gadget.LayoutVolume(p.dir, vol, constraints)
+	v, err := gadget.LayoutVolume(p.dir, "", vol, constraints)
 	c.Assert(v, IsNil)
 	c.Assert(err, ErrorMatches, `cannot lay out structure #0: content "bar.img" does not fit in the structure`)
 }
@@ -389,7 +382,7 @@ volumes:
 
 	vol := mustParseVolume(c, gadgetYaml, "first")
 
-	v, err := gadget.LayoutVolume(p.dir, vol, defaultConstraints)
+	v, err := gadget.LayoutVolume(p.dir, "", vol, defaultConstraints)
 	c.Assert(v, IsNil)
 	c.Assert(err, ErrorMatches, `cannot lay out structure #0: content "foo.img" does not fit in the structure`)
 }
@@ -411,7 +404,7 @@ volumes:
 
 	vol := mustParseVolume(c, gadgetYaml, "first")
 
-	v, err := gadget.LayoutVolume(p.dir, vol, defaultConstraints)
+	v, err := gadget.LayoutVolume(p.dir, "", vol, defaultConstraints)
 	c.Assert(v, IsNil)
 	c.Assert(err, ErrorMatches, fmt.Sprintf(`cannot lay out structure #0: content "foo.img" size %v is larger than declared %v`, quantity.SizeMiB+1, quantity.SizeMiB))
 }
@@ -439,7 +432,7 @@ volumes:
 
 	vol := mustParseVolume(c, gadgetYaml, "first")
 
-	v, err := gadget.LayoutVolume(p.dir, vol, defaultConstraints)
+	v, err := gadget.LayoutVolume(p.dir, "", vol, defaultConstraints)
 	c.Assert(v, IsNil)
 	c.Assert(err, ErrorMatches, `cannot lay out structure #0: content "foo.img" overlaps with preceding image "bar.img"`)
 }
@@ -468,13 +461,12 @@ volumes:
 	c.Assert(vol.Structure, HasLen, 1)
 	c.Assert(vol.Structure[0].Content, HasLen, 2)
 
-	v, err := gadget.LayoutVolume(p.dir, vol, defaultConstraints)
+	v, err := gadget.LayoutVolume(p.dir, "", vol, defaultConstraints)
 	c.Assert(err, IsNil)
 	c.Assert(v, DeepEquals, &gadget.LaidOutVolume{
-		Volume:     vol,
-		Size:       3 * quantity.SizeMiB,
-		SectorSize: 512,
-		RootDir:    p.dir,
+		Volume:  vol,
+		Size:    3 * quantity.SizeMiB,
+		RootDir: p.dir,
 		LaidOutStructure: []gadget.LaidOutStructure{
 			{
 				VolumeStructure: &vol.Structure[0],
@@ -520,13 +512,12 @@ volumes:
 	c.Assert(vol.Structure, HasLen, 1)
 	c.Assert(vol.Structure[0].Content, HasLen, 2)
 
-	v, err := gadget.LayoutVolume(p.dir, vol, defaultConstraints)
+	v, err := gadget.LayoutVolume(p.dir, "", vol, defaultConstraints)
 	c.Assert(err, IsNil)
 	c.Assert(v, DeepEquals, &gadget.LaidOutVolume{
-		Volume:     vol,
-		Size:       3 * quantity.SizeMiB,
-		SectorSize: 512,
-		RootDir:    p.dir,
+		Volume:  vol,
+		Size:    3 * quantity.SizeMiB,
+		RootDir: p.dir,
 		LaidOutStructure: []gadget.LaidOutStructure{
 			{
 				VolumeStructure: &vol.Structure[0],
@@ -569,13 +560,12 @@ volumes:
 	c.Assert(vol.Structure, HasLen, 1)
 	c.Assert(vol.Structure[0].Content, HasLen, 1)
 
-	v, err := gadget.LayoutVolume(p.dir, vol, defaultConstraints)
+	v, err := gadget.LayoutVolume(p.dir, "", vol, defaultConstraints)
 	c.Assert(err, IsNil)
 	c.Assert(v, DeepEquals, &gadget.LaidOutVolume{
-		Volume:     vol,
-		Size:       3 * quantity.SizeMiB,
-		SectorSize: 512,
-		RootDir:    p.dir,
+		Volume:  vol,
+		Size:    3 * quantity.SizeMiB,
+		RootDir: p.dir,
 		LaidOutStructure: []gadget.LaidOutStructure{
 			{
 				VolumeStructure: &vol.Structure[0],
@@ -612,17 +602,25 @@ volumes:
 	c.Assert(vol.Structure, HasLen, 1)
 	c.Assert(vol.Structure[0].Content, HasLen, 1)
 
-	v, err := gadget.LayoutVolume(p.dir, vol, defaultConstraints)
+	v, err := gadget.LayoutVolume(p.dir, "", vol, defaultConstraints)
 	c.Assert(err, IsNil)
 	c.Assert(v, DeepEquals, &gadget.LaidOutVolume{
-		Volume:     vol,
-		Size:       3 * quantity.SizeMiB,
-		SectorSize: 512,
-		RootDir:    p.dir,
+		Volume:  vol,
+		Size:    3 * quantity.SizeMiB,
+		RootDir: p.dir,
 		LaidOutStructure: []gadget.LaidOutStructure{
 			{
-				VolumeStructure: &vol.Structure[0],
 				StartOffset:     1 * quantity.OffsetMiB,
+				VolumeStructure: &vol.Structure[0],
+				ResolvedContent: []gadget.ResolvedContent{
+					{
+						VolumeContent: &gadget.VolumeContent{
+							UnresolvedSource: "foo.txt",
+							Target:           "/boot",
+						},
+						ResolvedSource: filepath.Join(p.dir, "foo.txt"),
+					},
+				},
 			},
 		},
 	})
@@ -646,29 +644,39 @@ volumes:
               - source: foo.txt
                 target: /boot
 `
+	resolvedContent := []gadget.ResolvedContent{
+		{
+			VolumeContent: &gadget.VolumeContent{
+				UnresolvedSource: "foo.txt",
+				Target:           "/boot",
+			},
+			ResolvedSource: filepath.Join(p.dir, "foo.txt"),
+		},
+	}
+
 	makeSizedFile(c, filepath.Join(p.dir, "foo.txt"), 0, []byte("foobar\n"))
 
 	vol := mustParseVolume(c, gadgetYaml, "first")
 	c.Assert(vol.Structure, HasLen, 2)
 	c.Assert(vol.Structure[1].Content, HasLen, 1)
 
-	v, err := gadget.LayoutVolume(p.dir, vol, defaultConstraints)
+	v, err := gadget.LayoutVolume(p.dir, "", vol, defaultConstraints)
 	c.Assert(err, IsNil)
 	c.Assert(v, DeepEquals, &gadget.LaidOutVolume{
-		Volume:     vol,
-		Size:       3 * quantity.SizeMiB,
-		SectorSize: 512,
-		RootDir:    p.dir,
+		Volume:  vol,
+		Size:    3 * quantity.SizeMiB,
+		RootDir: p.dir,
 		LaidOutStructure: []gadget.LaidOutStructure{
 			{
 				VolumeStructure: &vol.Structure[0],
 				StartOffset:     0,
-				Index:           0,
+				YamlIndex:       0,
 			},
 			{
 				VolumeStructure: &vol.Structure[1],
 				StartOffset:     1 * quantity.OffsetMiB,
-				Index:           1,
+				YamlIndex:       1,
+				ResolvedContent: resolvedContent,
 			},
 		},
 	})
@@ -677,25 +685,24 @@ volumes:
 	constraints := gadget.LayoutConstraints{
 		// 512kiB
 		NonMBRStartOffset: 512 * quantity.OffsetKiB,
-		SectorSize:        512,
 	}
-	v, err = gadget.LayoutVolume(p.dir, vol, constraints)
+	v, err = gadget.LayoutVolume(p.dir, "", vol, constraints)
 	c.Assert(err, IsNil)
 	c.Assert(v, DeepEquals, &gadget.LaidOutVolume{
-		Volume:     vol,
-		Size:       2*quantity.SizeMiB + 512*quantity.SizeKiB,
-		SectorSize: 512,
-		RootDir:    p.dir,
+		Volume:  vol,
+		Size:    2*quantity.SizeMiB + 512*quantity.SizeKiB,
+		RootDir: p.dir,
 		LaidOutStructure: []gadget.LaidOutStructure{
 			{
 				VolumeStructure: &vol.Structure[0],
 				StartOffset:     0,
-				Index:           0,
+				YamlIndex:       0,
 			},
 			{
 				VolumeStructure: &vol.Structure[1],
 				StartOffset:     512 * quantity.OffsetKiB,
-				Index:           1,
+				YamlIndex:       1,
+				ResolvedContent: resolvedContent,
 			},
 		},
 	})
@@ -705,91 +712,26 @@ volumes:
 	// explicitly
 	constraintsBad := gadget.LayoutConstraints{
 		NonMBRStartOffset: 400,
-		SectorSize:        512,
 	}
-	v, err = gadget.LayoutVolume(p.dir, vol, constraintsBad)
+	v, err = gadget.LayoutVolume(p.dir, "", vol, constraintsBad)
 	c.Assert(err, IsNil)
 	c.Assert(v, DeepEquals, &gadget.LaidOutVolume{
-		Volume:     vol,
-		Size:       2*quantity.SizeMiB + 446,
-		SectorSize: 512,
-		RootDir:    p.dir,
+		Volume:  vol,
+		Size:    2*quantity.SizeMiB + 446,
+		RootDir: p.dir,
 		LaidOutStructure: []gadget.LaidOutStructure{
 			{
 				VolumeStructure: &vol.Structure[0],
-				Index:           0,
+				YamlIndex:       0,
 			},
 			{
 				VolumeStructure: &vol.Structure[1],
 				StartOffset:     446,
-				Index:           1,
+				YamlIndex:       1,
+				ResolvedContent: resolvedContent,
 			},
 		},
 	})
-
-	// sector size is properly recorded
-	constraintsSector := gadget.LayoutConstraints{
-		NonMBRStartOffset: 1 * quantity.OffsetMiB,
-		SectorSize:        1024,
-	}
-	v, err = gadget.LayoutVolume(p.dir, vol, constraintsSector)
-	c.Assert(err, IsNil)
-	c.Assert(v, DeepEquals, &gadget.LaidOutVolume{
-		Volume:     vol,
-		Size:       3 * quantity.SizeMiB,
-		SectorSize: 1024,
-		RootDir:    p.dir,
-		LaidOutStructure: []gadget.LaidOutStructure{
-			{
-				VolumeStructure: &vol.Structure[0],
-				Index:           0,
-			},
-			{
-				VolumeStructure: &vol.Structure[1],
-				StartOffset:     1 * quantity.OffsetMiB,
-				Index:           1,
-			},
-		},
-	})
-}
-
-func (p *layoutTestSuite) TestLayoutVolumeConstraintsSectorSize(c *C) {
-	gadgetYaml := `
-volumes:
-  first:
-    schema: gpt
-    bootloader: grub
-    structure:
-        - role: mbr
-          type: bare
-          size: 446
-          offset: 0
-        - type: 00000000-0000-0000-0000-0000deadbeef
-          filesystem: ext4
-          size: 2M
-          content:
-              - source: foo.txt
-                target: /boot
-`
-	makeSizedFile(c, filepath.Join(p.dir, "foo.txt"), 0, []byte("foobar\n"))
-
-	vol := mustParseVolume(c, gadgetYaml, "first")
-
-	constraintsBadSectorSize := gadget.LayoutConstraints{
-		NonMBRStartOffset: 1 * quantity.OffsetMiB,
-		SectorSize:        384,
-	}
-	_, err := gadget.LayoutVolume(p.dir, vol, constraintsBadSectorSize)
-	c.Assert(err, ErrorMatches, "cannot lay out volume, structure #1 size is not a multiple of sector size 384")
-}
-
-func (p *layoutTestSuite) TestLayoutVolumeConstraintsNeedsSectorSize(c *C) {
-	constraintsBadSectorSize := gadget.LayoutConstraints{
-		NonMBRStartOffset: 1 * quantity.OffsetMiB,
-		// SectorSize left unspecified
-	}
-	_, err := gadget.LayoutVolume(p.dir, &gadget.Volume{}, constraintsBadSectorSize)
-	c.Assert(err, ErrorMatches, "cannot lay out volume, invalid constraints: sector size cannot be 0")
 }
 
 func (p *layoutTestSuite) TestLayoutVolumeMBRImplicitConstraints(c *C) {
@@ -810,23 +752,22 @@ volumes:
 	vol := mustParseVolume(c, gadgetYaml, "first")
 	c.Assert(vol.Structure, HasLen, 2)
 
-	v, err := gadget.LayoutVolume(p.dir, vol, defaultConstraints)
+	v, err := gadget.LayoutVolume(p.dir, "", vol, defaultConstraints)
 	c.Assert(err, IsNil)
 	c.Assert(v, DeepEquals, &gadget.LaidOutVolume{
-		Volume:     vol,
-		Size:       2 * quantity.SizeMiB,
-		SectorSize: 512,
-		RootDir:    p.dir,
+		Volume:  vol,
+		Size:    2 * quantity.SizeMiB,
+		RootDir: p.dir,
 		LaidOutStructure: []gadget.LaidOutStructure{
 			{
 				// MBR
 				VolumeStructure: &vol.Structure[0],
 				StartOffset:     0,
-				Index:           0,
+				YamlIndex:       0,
 			}, {
 				VolumeStructure: &vol.Structure[1],
 				StartOffset:     1 * quantity.OffsetMiB,
-				Index:           1,
+				YamlIndex:       1,
 			},
 		},
 	})
@@ -863,24 +804,23 @@ volumes:
 	vol := mustParseVolume(c, gadgetYaml, "pc")
 	c.Assert(vol.Structure, HasLen, 3)
 
-	v, err := gadget.LayoutVolume(p.dir, vol, defaultConstraints)
+	v, err := gadget.LayoutVolume(p.dir, "", vol, defaultConstraints)
 	c.Assert(err, IsNil)
 	c.Assert(v, DeepEquals, &gadget.LaidOutVolume{
-		Volume:     vol,
-		Size:       3 * quantity.SizeMiB,
-		SectorSize: 512,
-		RootDir:    p.dir,
+		Volume:  vol,
+		Size:    3 * quantity.SizeMiB,
+		RootDir: p.dir,
 		LaidOutStructure: []gadget.LaidOutStructure{
 			{
 				// mbr
 				VolumeStructure: &vol.Structure[0],
 				StartOffset:     0,
-				Index:           0,
+				YamlIndex:       0,
 			}, {
 				// foo
 				VolumeStructure: &vol.Structure[1],
 				StartOffset:     1 * quantity.OffsetMiB,
-				Index:           1,
+				YamlIndex:       1,
 				// break for gofmt < 1.11
 				AbsoluteOffsetWrite: asOffsetPtr(92),
 				LaidOutContent: []gadget.LaidOutContent{
@@ -896,7 +836,7 @@ volumes:
 				// bar
 				VolumeStructure: &vol.Structure[2],
 				StartOffset:     2 * quantity.OffsetMiB,
-				Index:           2,
+				YamlIndex:       2,
 				// break for gofmt < 1.11
 				AbsoluteOffsetWrite: asOffsetPtr(600),
 				LaidOutContent: []gadget.LaidOutContent{
@@ -949,11 +889,11 @@ func (p *layoutTestSuite) TestLayoutVolumeOffsetWriteBadRelativeTo(c *C) {
 
 	makeSizedFile(c, filepath.Join(p.dir, "foo.img"), 200*quantity.SizeKiB, []byte(""))
 
-	v, err := gadget.LayoutVolume(p.dir, &volBadStructure, defaultConstraints)
+	v, err := gadget.LayoutVolume(p.dir, "", &volBadStructure, defaultConstraints)
 	c.Check(v, IsNil)
 	c.Check(err, ErrorMatches, `cannot resolve offset-write of structure #0 \("foo"\): refers to an unknown structure "bar"`)
 
-	v, err = gadget.LayoutVolume(p.dir, &volBadContent, defaultConstraints)
+	v, err = gadget.LayoutVolume(p.dir, "", &volBadContent, defaultConstraints)
 	c.Check(v, IsNil)
 	c.Check(err, ErrorMatches, `cannot resolve offset-write of structure #0 \("foo"\) content "foo.img": refers to an unknown structure "bar"`)
 }
@@ -977,7 +917,7 @@ volumes:
 `
 	vol := mustParseVolume(c, gadgetYamlStructure, "pc")
 
-	v, err := gadget.LayoutVolume(p.dir, vol, defaultConstraints)
+	v, err := gadget.LayoutVolume(p.dir, "", vol, defaultConstraints)
 	c.Assert(err, IsNil)
 	// offset-write is at 1GB
 	c.Check(v.Size, Equals, 1*quantity.SizeGiB+gadget.SizeLBA48Pointer)
@@ -1012,7 +952,7 @@ volumes:
 
 	vol = mustParseVolume(c, gadgetYamlContent, "pc")
 
-	v, err = gadget.LayoutVolume(p.dir, vol, defaultConstraints)
+	v, err = gadget.LayoutVolume(p.dir, "", vol, defaultConstraints)
 	c.Assert(err, IsNil)
 	// foo.img offset-write is at 3GB
 	c.Check(v.Size, Equals, 3*quantity.SizeGiB+gadget.SizeLBA48Pointer)
@@ -1036,13 +976,12 @@ volumes:
 
 	v, err := gadget.LayoutVolumePartially(vol, defaultConstraints)
 	c.Assert(v, DeepEquals, &gadget.PartiallyLaidOutVolume{
-		Volume:     vol,
-		SectorSize: 512,
+		Volume: vol,
 		LaidOutStructure: []gadget.LaidOutStructure{
 			{
 				VolumeStructure: &vol.Structure[0],
 				StartOffset:     800 * quantity.OffsetMiB,
-				Index:           0,
+				YamlIndex:       0,
 			},
 		},
 	})
@@ -1071,7 +1010,7 @@ volumes:
 
 	vol := mustParseVolume(c, gadgetYamlContent, "pc")
 
-	v, err := gadget.LayoutVolume(p.dir, vol, defaultConstraints)
+	v, err := gadget.LayoutVolume(p.dir, "", vol, defaultConstraints)
 	c.Assert(err, IsNil)
 	c.Assert(v.LaidOutStructure, HasLen, 1)
 	c.Assert(v.LaidOutStructure[0].LaidOutContent, HasLen, 2)
@@ -1082,7 +1021,7 @@ volumes:
 		// foo
 		VolumeStructure: &vol.Structure[0],
 		StartOffset:     1 * quantity.OffsetMiB,
-		Index:           0,
+		YamlIndex:       0,
 		LaidOutContent: []gadget.LaidOutContent{
 			{
 				VolumeContent: &vol.Structure[0].Content[0],
@@ -1103,7 +1042,7 @@ volumes:
 		// foo
 		VolumeStructure: &vol.Structure[0],
 		StartOffset:     0,
-		Index:           0,
+		YamlIndex:       0,
 		LaidOutContent: []gadget.LaidOutContent{
 			{
 				VolumeContent: &vol.Structure[0].Content[0],
@@ -1124,7 +1063,7 @@ volumes:
 		// foo
 		VolumeStructure: &vol.Structure[0],
 		StartOffset:     2 * quantity.OffsetMiB,
-		Index:           0,
+		YamlIndex:       0,
 		LaidOutContent: []gadget.LaidOutContent{
 			{
 				VolumeContent: &vol.Structure[0].Content[0],
@@ -1142,7 +1081,7 @@ volumes:
 }
 
 func mockKernel(c *C, kernelYaml string, filesWithContent map[string]string) string {
-	// sanity
+	// precondition
 	_, err := kernel.InfoFromKernelYaml([]byte(kernelYaml))
 	c.Assert(err, IsNil)
 
@@ -1160,6 +1099,10 @@ func mockKernel(c *C, kernelYaml string, filesWithContent map[string]string) str
 		c.Assert(err, IsNil)
 	}
 
+	// ensure we have valid kernel.yaml in our tests
+	err = kernel.Validate(kernelRootDir)
+	c.Assert(err, IsNil)
+
 	return kernelRootDir
 }
 
@@ -1184,23 +1127,15 @@ var gadgetYamlWithKernelRef = `
 `
 
 func (p *layoutTestSuite) TestResolveContentPathsNotInWantedAssets(c *C) {
-	kernelYaml := ""
-
 	vol := mustParseVolume(c, gadgetYamlWithKernelRef, "pi")
 	c.Assert(vol.Structure, HasLen, 1)
 
-	kernelSnapFiles := map[string]string{}
-	kernelSnapDir := mockKernel(c, kernelYaml, kernelSnapFiles)
-	lv, err := gadget.LayoutVolume(p.dir, vol, defaultConstraints)
-	c.Assert(err, IsNil)
-
-	err = gadget.ResolveContentPaths(lv, p.dir, kernelSnapDir)
-	c.Assert(err, ErrorMatches, `cannot find "dtbs" in kernel info from "/.*"`)
+	kernelSnapDir := c.MkDir()
+	_, err := gadget.LayoutVolume(p.dir, kernelSnapDir, vol, defaultConstraints)
+	c.Assert(err, ErrorMatches, `cannot resolve content for structure #0 at index 0: cannot find "dtbs" in kernel info from "/.*"`)
 }
 
 func (p *layoutTestSuite) TestResolveContentPathsErrorInKernelRef(c *C) {
-	kernelYaml := ""
-
 	// create invalid kernel ref
 	s := strings.Replace(gadgetYamlWithKernelRef, "$kernel:dtbs", "$kernel:-invalid-kernel-ref", -1)
 	// Note that mustParseVolume does not call ValidateContent() which
@@ -1208,13 +1143,9 @@ func (p *layoutTestSuite) TestResolveContentPathsErrorInKernelRef(c *C) {
 	vol := mustParseVolume(c, s, "pi")
 	c.Assert(vol.Structure, HasLen, 1)
 
-	kernelSnapFiles := map[string]string{}
-	kernelSnapDir := mockKernel(c, kernelYaml, kernelSnapFiles)
-	lv, err := gadget.LayoutVolume(p.dir, vol, defaultConstraints)
-	c.Assert(err, IsNil)
-
-	err = gadget.ResolveContentPaths(lv, p.dir, kernelSnapDir)
-	c.Assert(err, ErrorMatches, `cannot parse kernel ref: invalid asset name in kernel ref "\$kernel:-invalid-kernel-ref/boot-assets/"`)
+	kernelSnapDir := c.MkDir()
+	_, err := gadget.LayoutVolume(p.dir, kernelSnapDir, vol, defaultConstraints)
+	c.Assert(err, ErrorMatches, `cannot resolve content for structure #0 at index 0: cannot parse kernel ref: invalid asset name in kernel ref "\$kernel:-invalid-kernel-ref/boot-assets/"`)
 }
 
 func (p *layoutTestSuite) TestResolveContentPathsNotInWantedeContent(c *C) {
@@ -1223,19 +1154,17 @@ assets:
   dtbs:
     update: true
     content:
-      - dtbs
+      - dtbs/
 `
 
 	vol := mustParseVolume(c, gadgetYamlWithKernelRef, "pi")
 	c.Assert(vol.Structure, HasLen, 1)
 
-	kernelSnapFiles := map[string]string{}
-	kernelSnapDir := mockKernel(c, kernelYaml, kernelSnapFiles)
-	lv, err := gadget.LayoutVolume(p.dir, vol, defaultConstraints)
-	c.Assert(err, IsNil)
-
-	err = gadget.ResolveContentPaths(lv, p.dir, kernelSnapDir)
-	c.Assert(err, ErrorMatches, `cannot find wanted kernel content "boot-assets/" in "/.*"`)
+	kernelSnapDir := mockKernel(c, kernelYaml, map[string]string{
+		"dtbs/foo.dtb": "foo.dtb content",
+	})
+	_, err := gadget.LayoutVolume(p.dir, kernelSnapDir, vol, defaultConstraints)
+	c.Assert(err, ErrorMatches, `cannot resolve content for structure #0 at index 0: cannot find wanted kernel content "boot-assets/" in "/.*"`)
 }
 
 func (p *layoutTestSuite) TestResolveContentPaths(c *C) {
@@ -1252,12 +1181,14 @@ assets:
 
 	kernelSnapFiles := map[string]string{
 		"boot-assets/foo": "foo-content",
+		"some-file":       "some-file content",
 	}
 	kernelSnapDir := mockKernel(c, kernelYaml, kernelSnapFiles)
-	lv, err := gadget.LayoutVolume(p.dir, vol, defaultConstraints)
+	lv, err := gadget.LayoutVolume(p.dir, kernelSnapDir, vol, defaultConstraints)
 	c.Assert(err, IsNil)
-	content := lv.Volume.Structure[0].Content
-	c.Assert(content, DeepEquals, []gadget.VolumeContent{
+	// Volume.Content is unchanged
+	c.Assert(lv.Structure, HasLen, 1)
+	c.Check(lv.Structure[0].Content, DeepEquals, []gadget.VolumeContent{
 		{
 			UnresolvedSource: "$kernel:dtbs/boot-assets/",
 			Target:           "/",
@@ -1275,19 +1206,133 @@ assets:
 			Target:           "/",
 		},
 	})
+	// and the LaidOutSturctures ResolvedContent has the correct paths
+	c.Assert(lv.LaidOutStructure, HasLen, 1)
+	c.Check(lv.LaidOutStructure[0].ResolvedContent, DeepEquals, []gadget.ResolvedContent{
+		{
+			VolumeContent:  &lv.Structure[0].Content[0],
+			ResolvedSource: filepath.Join(kernelSnapDir, "boot-assets/") + "/",
+			KernelUpdate:   true,
+		},
+		{
+			VolumeContent:  &lv.Structure[0].Content[1],
+			ResolvedSource: filepath.Join(kernelSnapDir, "some-file"),
+			KernelUpdate:   true,
+		},
+		{
+			VolumeContent:  &lv.Structure[0].Content[2],
+			ResolvedSource: filepath.Join(p.dir, "file-from-gadget"),
+		},
+		{
+			VolumeContent:  &lv.Structure[0].Content[3],
+			ResolvedSource: filepath.Join(p.dir, "dir-from-gadget") + "/",
+		},
+	})
+}
 
-	// now resolve the kernel references
-	err = gadget.ResolveContentPaths(lv, p.dir, kernelSnapDir)
+func (p *layoutTestSuite) TestResolveContentPathsLp1907056(c *C) {
+	var gadgetYamlWithKernelRef = `
+ volumes:
+  pi:
+    schema: mbr
+    bootloader: u-boot
+    structure:
+      - name: ubuntu-seed
+        role: system-seed
+        filesystem: vfat
+        type: 0C
+        size: 1200M
+        content:
+          - source: $kernel:pidtbs/dtbs/broadcom/
+            target: /
+          - source: $kernel:pidtbs/dtbs/overlays/
+            target: /overlays
+          - source: boot-assets/
+            target: /
+`
+
+	kernelYaml := `
+assets:
+  pidtbs:
+    update: true
+    content:
+      - dtbs/
+`
+	vol := mustParseVolume(c, gadgetYamlWithKernelRef, "pi")
+	c.Assert(vol.Structure, HasLen, 1)
+
+	kernelSnapDir := mockKernel(c, kernelYaml, map[string]string{
+		"dtbs/foo.dtb": "foo.dtb content",
+	})
+	lv, err := gadget.LayoutVolume(p.dir, kernelSnapDir, vol, defaultConstraints)
 	c.Assert(err, IsNil)
+	// Volume.Content is unchanged
+	c.Assert(lv.Structure, HasLen, 1)
+	c.Check(lv.Structure[0].Content, DeepEquals, []gadget.VolumeContent{
+		{
+			UnresolvedSource: "$kernel:pidtbs/dtbs/broadcom/",
+			Target:           "/",
+		},
+		{
+			UnresolvedSource: "$kernel:pidtbs/dtbs/overlays/",
+			Target:           "/overlays",
+		},
+		{
+			UnresolvedSource: "boot-assets/",
+			Target:           "/",
+		},
+	})
+	// and the LaidOutSturctures ResolvedContent has the correct paths
+	c.Assert(lv.LaidOutStructure, HasLen, 1)
+	c.Check(lv.LaidOutStructure[0].ResolvedContent, DeepEquals, []gadget.ResolvedContent{
+		{
+			VolumeContent:  &lv.Structure[0].Content[0],
+			ResolvedSource: filepath.Join(kernelSnapDir, "dtbs/broadcom/") + "/",
+			KernelUpdate:   true,
+		},
+		{
+			VolumeContent:  &lv.Structure[0].Content[1],
+			ResolvedSource: filepath.Join(kernelSnapDir, "dtbs/overlays") + "/",
+			KernelUpdate:   true,
+		},
+		{
+			VolumeContent:  &lv.Structure[0].Content[2],
+			ResolvedSource: filepath.Join(p.dir, "boot-assets") + "/",
+		},
+	})
+}
+
+func (p *layoutTestSuite) TestResolveContentSamePrefixErrors(c *C) {
+	var gadgetYamlWithKernelRef = `
+ volumes:
+  pi:
+    schema: mbr
+    bootloader: u-boot
+    structure:
+      - name: ubuntu-seed
+        role: system-seed
+        filesystem: vfat
+        type: 0C
+        size: 1200M
+        content:
+          - source: $kernel:dtbs/a
+            target: /
+          - source: $kernel:dtbs/ab
+            target: /
+`
+	kernelYaml := `
+assets:
+  dtbs:
+    update: true
+    content:
+      - a
+`
+	vol := mustParseVolume(c, gadgetYamlWithKernelRef, "pi")
+	c.Assert(vol.Structure, HasLen, 1)
 
-	c.Assert(lv.Volume.Structure, HasLen, 1)
-	c.Check(content, HasLen, 4)
-	// note the trailing "/" here
-	c.Check(content[0].ResolvedSource(), Equals, filepath.Join(kernelSnapDir, "boot-assets/")+"/")
-	// no trailing "/" here
-	c.Check(content[1].ResolvedSource(), Equals, filepath.Join(kernelSnapDir, "some-file"))
-	// from gadget, no trailing "/" here
-	c.Check(content[2].ResolvedSource(), Equals, filepath.Join(p.dir, "file-from-gadget"))
-	// note the trailing "/" here
-	c.Check(content[3].ResolvedSource(), Equals, filepath.Join(p.dir, "dir-from-gadget")+"/")
+	kernelSnapDir := mockKernel(c, kernelYaml, map[string]string{
+		"a/foo.dtb": "foo.dtb content",
+	})
+	_, err := gadget.LayoutVolume(p.dir, kernelSnapDir, vol, defaultConstraints)
+	c.Assert(err, ErrorMatches, `.*: cannot find wanted kernel content "ab" in.*`)
 }
