@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2019-2020 Canonical Ltd
+ * Copyright (C) 2019-2021 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -30,6 +30,9 @@ import (
 	. "gopkg.in/check.v1"
 
 	"github.com/snapcore/snapd/gadget"
+	"github.com/snapcore/snapd/gadget/gadgettest"
+	"github.com/snapcore/snapd/gadget/quantity"
+	"github.com/snapcore/snapd/kernel"
 )
 
 type validateGadgetTestSuite struct {
@@ -211,7 +214,7 @@ func (s *validateGadgetTestSuite) TestVolumeRulesConsistencyNoModel(c *C) {
 	c.Assert(err, ErrorMatches, "model does not support the system-save role")
 }
 
-func (s *validateGadgetTestSuite) TestValidateConsistencyWithoutModelCharateristics(c *C) {
+func (s *validateGadgetTestSuite) TestValidateConsistencyWithoutModelCharacteristics(c *C) {
 	for i, tc := range []struct {
 		role  string
 		label string
@@ -263,7 +266,7 @@ volumes:
 	}
 }
 
-func (s *validateGadgetTestSuite) TestValidateConsistencyWithModelCharateristics(c *C) {
+func (s *validateGadgetTestSuite) TestValidateConsistencyWithModelCharacteristics(c *C) {
 	bloader := `
 volumes:
   pc:
@@ -338,9 +341,9 @@ volumes:
 
 		makeSizedFile(c, filepath.Join(s.dir, "meta/gadget.yaml"), 0, b.Bytes())
 
-		mod := &modelCharateristics{
-			classic:    false,
-			systemSeed: tc.requireSeed,
+		mod := &gadgettest.ModelCharacteristics{
+			IsClassic:  false,
+			SystemSeed: tc.requireSeed,
 		}
 		ginfo, err := gadget.ReadInfo(s.dir, mod)
 		c.Assert(err, IsNil)
@@ -355,8 +358,8 @@ volumes:
 	// test error with no volumes
 	makeSizedFile(c, filepath.Join(s.dir, "meta/gadget.yaml"), 0, []byte(bloader))
 
-	mod := &modelCharateristics{
-		systemSeed: true,
+	mod := &gadgettest.ModelCharacteristics{
+		SystemSeed: true,
 	}
 
 	ginfo, err := gadget.ReadInfo(s.dir, mod)
@@ -531,8 +534,8 @@ func (s *validateGadgetTestSuite) TestRuleValidateHybridGadget(c *C) {
         size: 750M
 `)
 
-	mod := &modelCharateristics{
-		classic: false,
+	mod := &gadgettest.ModelCharacteristics{
+		IsClassic: false,
 	}
 	giMeta, err := gadget.InfoFromGadgetYaml(hybridyGadgetYaml, mod)
 	c.Assert(err, IsNil)
@@ -582,8 +585,8 @@ func (s *validateGadgetTestSuite) TestRuleValidateHybridGadgetBrokenDupRole(c *C
         size: 750M
 `)
 
-	mod := &modelCharateristics{
-		classic: false,
+	mod := &gadgettest.ModelCharacteristics{
+		IsClassic: false,
 	}
 	giMeta, err := gadget.InfoFromGadgetYaml(brokenGadgetYaml, mod)
 	c.Assert(err, IsNil)
@@ -610,7 +613,7 @@ volumes:
 
 	ginfo, err := gadget.ReadInfo(s.dir, nil)
 	c.Assert(err, IsNil)
-	err = gadget.ValidateContent(ginfo, s.dir)
+	err = gadget.ValidateContent(ginfo, s.dir, "")
 	c.Assert(err, ErrorMatches, `invalid layout of volume "pc": cannot lay out structure #0 \("foo"\): content "foo.img": stat .*/foo.img: no such file or directory`)
 }
 
@@ -640,7 +643,7 @@ volumes:
 
 	ginfo, err := gadget.ReadInfo(s.dir, nil)
 	c.Assert(err, IsNil)
-	err = gadget.ValidateContent(ginfo, s.dir)
+	err = gadget.ValidateContent(ginfo, s.dir, "")
 	c.Assert(err, ErrorMatches, `invalid layout of volume "second": cannot lay out structure #0 \("second-foo"\): content "second.img": stat .*/second.img: no such file or directory`)
 }
 
@@ -663,13 +666,13 @@ volumes:
 
 	ginfo, err := gadget.ReadInfo(s.dir, nil)
 	c.Assert(err, IsNil)
-	err = gadget.ValidateContent(ginfo, s.dir)
+	err = gadget.ValidateContent(ginfo, s.dir, "")
 	c.Assert(err, ErrorMatches, `invalid volume "bad": structure #0 \("bad-struct"\), content source:foo/: source path does not exist`)
 
 	// make it a file, which conflicts with foo/ as 'source'
 	fooPath := filepath.Join(s.dir, "foo")
 	makeSizedFile(c, fooPath, 1, nil)
-	err = gadget.ValidateContent(ginfo, s.dir)
+	err = gadget.ValidateContent(ginfo, s.dir, "")
 	c.Assert(err, ErrorMatches, `invalid volume "bad": structure #0 \("bad-struct"\), content source:foo/: cannot specify trailing / for a source which is not a directory`)
 
 	// make it a directory
@@ -678,7 +681,7 @@ volumes:
 	err = os.Mkdir(fooPath, 0755)
 	c.Assert(err, IsNil)
 	// validate should no longer complain
-	err = gadget.ValidateContent(ginfo, s.dir)
+	err = gadget.ValidateContent(ginfo, s.dir, "")
 	c.Assert(err, IsNil)
 }
 
@@ -714,18 +717,41 @@ var gadgetYamlContentWithSave = gadgetYamlContentNoSave + `
 func (s *validateGadgetTestSuite) TestValidateEncryptionSupportErr(c *C) {
 	makeSizedFile(c, filepath.Join(s.dir, "meta/gadget.yaml"), 0, []byte(gadgetYamlContentNoSave))
 
-	mod := &modelCharateristics{systemSeed: true}
+	mod := &gadgettest.ModelCharacteristics{SystemSeed: true}
 	ginfo, err := gadget.ReadInfo(s.dir, mod)
 	c.Assert(err, IsNil)
 	err = gadget.Validate(ginfo, mod, &gadget.ValidationConstraints{
 		EncryptedData: true,
 	})
-	c.Assert(err, ErrorMatches, `gadget does not support encrypted data: volume "vol1" has no structure with system-save role`)
+	c.Assert(err, ErrorMatches, `gadget does not support encrypted data: required partition with system-save role is missing`)
 }
 
 func (s *validateGadgetTestSuite) TestValidateEncryptionSupportHappy(c *C) {
 	makeSizedFile(c, filepath.Join(s.dir, "meta/gadget.yaml"), 0, []byte(gadgetYamlContentWithSave))
-	mod := &modelCharateristics{systemSeed: true}
+	mod := &gadgettest.ModelCharacteristics{SystemSeed: true}
+	ginfo, err := gadget.ReadInfo(s.dir, mod)
+	c.Assert(err, IsNil)
+	err = gadget.Validate(ginfo, mod, &gadget.ValidationConstraints{
+		EncryptedData: true,
+	})
+	c.Assert(err, IsNil)
+}
+
+func (s *validateGadgetTestSuite) TestValidateEncryptionSupportNoUC20(c *C) {
+	makeSizedFile(c, filepath.Join(s.dir, "meta/gadget.yaml"), 0, []byte(gadgetYamlPC))
+
+	mod := &gadgettest.ModelCharacteristics{SystemSeed: false}
+	ginfo, err := gadget.ReadInfo(s.dir, mod)
+	c.Assert(err, IsNil)
+	err = gadget.Validate(ginfo, mod, &gadget.ValidationConstraints{
+		EncryptedData: true,
+	})
+	c.Assert(err, ErrorMatches, `internal error: cannot support encrypted data without requiring system-seed`)
+}
+
+func (s *validateGadgetTestSuite) TestValidateEncryptionSupportMultiVolumeHappy(c *C) {
+	makeSizedFile(c, filepath.Join(s.dir, "meta/gadget.yaml"), 0, []byte(mockMultiVolumeUC20GadgetYaml))
+	mod := &gadgettest.ModelCharacteristics{SystemSeed: true}
 	ginfo, err := gadget.ReadInfo(s.dir, mod)
 	c.Assert(err, IsNil)
 	err = gadget.Validate(ginfo, mod, &gadget.ValidationConstraints{
@@ -776,7 +802,7 @@ func (s *validateGadgetTestSuite) TestValidateContentKernelAssetsRef(c *C) {
 		makeSizedFile(c, filepath.Join(s.dir, "meta/gadget.yaml"), 0, []byte(gadgetYaml))
 		ginfo, err := gadget.ReadInfoAndValidate(s.dir, nil, nil)
 		c.Assert(err, IsNil)
-		err = gadget.ValidateContent(ginfo, s.dir)
+		err = gadget.ValidateContent(ginfo, s.dir, "")
 		if tc.good {
 			c.Check(err, IsNil, Commentf(tc.source))
 			// asset validates correctly, so let's make sure that
@@ -810,3 +836,187 @@ func (s *validateGadgetTestSuite) TestSplitKernelRefErrors(c *C) {
 		c.Check(err, ErrorMatches, tc.errStr, Commentf("kernelRef: %s", tc.kernelRef))
 	}
 }
+
+func (s *validateGadgetTestSuite) TestCanResolveOneVolumeKernelRef(c *C) {
+	lv := &gadget.LaidOutVolume{
+		Volume: &gadget.Volume{
+			Bootloader: "grub",
+			Schema:     "gpt",
+			Structure: []gadget.VolumeStructure{
+				{
+					Name:       "foo",
+					Size:       5 * quantity.SizeMiB,
+					Filesystem: "ext4",
+				},
+			},
+		},
+	}
+
+	contentNoKernelRef := []gadget.VolumeContent{
+		{UnresolvedSource: "/content", Target: "/"},
+	}
+	contentOneKernelRef := []gadget.VolumeContent{
+		{UnresolvedSource: "/content", Target: "/"},
+		{UnresolvedSource: "$kernel:ref/foo", Target: "/"},
+	}
+	contentTwoKernelRefs := []gadget.VolumeContent{
+		{UnresolvedSource: "/content", Target: "/"},
+		{UnresolvedSource: "$kernel:ref/foo", Target: "/"},
+		{UnresolvedSource: "$kernel:ref2/bar", Target: "/"},
+	}
+
+	kInfoNoRefs := &kernel.Info{}
+	kInfoOneRefButUpdateFlagFalse := &kernel.Info{
+		Assets: map[string]*kernel.Asset{
+			"ref": {
+				// note that update is false here
+				Update:  false,
+				Content: []string{"some-file"},
+			},
+		},
+	}
+	kInfoOneRef := &kernel.Info{
+		Assets: map[string]*kernel.Asset{
+			"ref": {
+				Update:  true,
+				Content: []string{"some-file"},
+			},
+		},
+	}
+	kInfoOneRefDifferentName := &kernel.Info{
+		Assets: map[string]*kernel.Asset{
+			"ref-other": {
+				Update:  true,
+				Content: []string{"some-file"},
+			},
+		},
+	}
+	kInfoTwoRefs := &kernel.Info{
+		Assets: map[string]*kernel.Asset{
+			"ref": {
+				Update:  true,
+				Content: []string{"some-file"},
+			},
+			"ref2": {
+				Update:  true,
+				Content: []string{"other-file"},
+			},
+		},
+	}
+
+	for _, tc := range []struct {
+		volumeContent  []gadget.VolumeContent
+		kinfo          *kernel.Info
+		consumed       bool
+		consumedErr    string
+		consumesOneErr string
+	}{
+		// happy case: trivial
+		{contentNoKernelRef, kInfoNoRefs, false, "", ""},
+
+		// happy case: if kernel asset has "Update: false"
+		{contentNoKernelRef, kInfoOneRefButUpdateFlagFalse, false, "", ""},
+
+		// unhappy case: kernel has one or more unresolved references in gadget
+		{contentNoKernelRef, kInfoOneRef, false, "", "gadget does not consume any of the kernel assets needing synced update \"ref\""},
+		{contentNoKernelRef, kInfoTwoRefs, false, "", "gadget does not consume any of the kernel assets needing synced update \"ref\", \"ref2\""},
+
+		// unhappy case: gadget needs different asset than kernel provides
+		{contentOneKernelRef, kInfoOneRefDifferentName, false, "", "gadget does not consume any of the kernel assets needing synced update \"ref-other\""},
+
+		// happy case: exactly one matching kernel ref
+		{contentOneKernelRef, kInfoOneRef, true, "", ""},
+		// happy case: one matching, one missing kernel ref, still considered fine
+		{contentTwoKernelRefs, kInfoTwoRefs, true, "", ""},
+	} {
+		lv.Structure[0].Content = tc.volumeContent
+		consumed, err := gadget.GadgetVolumeKernelUpdateAssetsConsumed(lv.Volume, tc.kinfo)
+		if tc.consumedErr == "" {
+			c.Check(err, IsNil, Commentf("should not fail %v", tc.volumeContent))
+			c.Check(consumed, Equals, tc.consumed)
+		} else {
+			c.Check(err, ErrorMatches, tc.consumedErr, Commentf("should fail %v", tc.volumeContent))
+		}
+
+		err = gadget.GadgetVolumeConsumesOneKernelUpdateAsset(lv.Volume, tc.kinfo)
+		if tc.consumesOneErr == "" {
+			c.Check(err, IsNil, Commentf("should not fail %v", tc.volumeContent))
+		} else {
+			c.Check(err, ErrorMatches, tc.consumesOneErr, Commentf("should fail %v", tc.volumeContent))
+		}
+	}
+}
+
+func (s *validateGadgetTestSuite) TestValidateContentKernelRefMissing(c *C) {
+	var gadgetYamlContent = `
+volumes:
+  first:
+    bootloader: grub
+    structure:
+      - name: first-foo
+        type: DA,21686148-6449-6E6F-744E-656564454649
+        size: 1M
+        content:
+          - image: first.img
+  second:
+    structure:
+      - name: second-foo
+        filesystem: ext4
+        type: 83,0FC63DAF-8483-4772-8E79-3D69D8477DE4
+        size: 128M
+        content:
+          - source: $kernel:ref/foo
+            target: /
+`
+	makeSizedFile(c, filepath.Join(s.dir, "meta/gadget.yaml"), 0, []byte(gadgetYamlContent))
+	makeSizedFile(c, filepath.Join(s.dir, "first.img"), 1, nil)
+
+	// note that there is no kernel.yaml
+	kernelUnpackDir := c.MkDir()
+
+	ginfo, err := gadget.ReadInfo(s.dir, nil)
+	c.Assert(err, IsNil)
+	err = gadget.ValidateContent(ginfo, s.dir, kernelUnpackDir)
+	c.Assert(err, ErrorMatches, `.*cannot find "ref" in kernel info.*`)
+}
+
+func (s *validateGadgetTestSuite) TestValidateContentKernelRefNotInGadget(c *C) {
+	var gadgetYamlContent = `
+volumes:
+  first:
+    bootloader: grub
+    structure:
+      - name: first-foo
+        type: DA,21686148-6449-6E6F-744E-656564454649
+        size: 1M
+        content:
+          - image: first.img
+  second:
+    structure:
+      - name: second-foo
+        filesystem: ext4
+        type: 83,0FC63DAF-8483-4772-8E79-3D69D8477DE4
+        size: 128M
+        content:
+          - source: foo
+            target: /
+`
+	makeSizedFile(c, filepath.Join(s.dir, "meta/gadget.yaml"), 0, []byte(gadgetYamlContent))
+	makeSizedFile(c, filepath.Join(s.dir, "first.img"), 1, nil)
+	makeSizedFile(c, filepath.Join(s.dir, "foo"), 1, nil)
+
+	kernelUnpackDir := c.MkDir()
+	kernelYamlContent := `
+assets:
+ ref:
+  update: true
+  content:
+   - dtbs/`
+	makeSizedFile(c, filepath.Join(kernelUnpackDir, "meta/kernel.yaml"), 0, []byte(kernelYamlContent))
+	makeSizedFile(c, filepath.Join(kernelUnpackDir, "dtbs/foo.dtb"), 0, []byte("foo.dtb content"))
+
+	ginfo, err := gadget.ReadInfo(s.dir, nil)
+	c.Assert(err, IsNil)
+	err = gadget.ValidateContent(ginfo, s.dir, kernelUnpackDir)
+	c.Assert(err, ErrorMatches, `no asset from the kernel.yaml needing synced update is consumed by the gadget at "/.*"`)
+}
