@@ -34,6 +34,11 @@ var (
 	ResolveSpecialVariable      = resolveSpecialVariable
 	ImplicitSystemPermanentSlot = implicitSystemPermanentSlot
 	ImplicitSystemConnectedSlot = implicitSystemConnectedSlot
+	LabelExpr                   = labelExpr
+	PlugAppLabelExpr            = plugAppLabelExpr
+	SlotAppLabelExpr            = slotAppLabelExpr
+	AareExclusivePatterns       = aareExclusivePatterns
+	GetDesktopFileRules         = getDesktopFileRules
 )
 
 func MprisGetName(iface interfaces.Interface, attribs map[string]interface{}) (string, error) {
