@@ -19,6 +19,10 @@
 
 package builtin
 
+import (
+	"github.com/snapcore/snapd/osutil"
+)
+
 const networkControlSummary = `allows configuring networking and network namespaces`
 
 const networkControlBaseDeclarationSlots = `
@@ -36,13 +40,13 @@ const networkControlConnectedPlugAppArmor = `
 # trusted apps.
 
 #include <abstractions/nameservice>
-/run/systemd/resolve/stub-resolv.conf r,
+/run/systemd/resolve/stub-resolv.conf rk,
 
 # systemd-resolved (not yet included in nameservice abstraction)
 #
 # Allow access to the safe members of the systemd-resolved D-Bus API:
 #
-#   https://www.freedesktop.org/wiki/Software/systemd/resolved/
+#   https://www.freedesktop.org/software/systemd/man/org.freedesktop.resolve1.html
 #
 # This API may be used directly over the D-Bus system bus or it may be used
 # indirectly via the nss-resolve plugin:
@@ -55,13 +59,53 @@ dbus send
      path="/org/freedesktop/resolve1"
      interface="org.freedesktop.resolve1.Manager"
      member="Resolve{Address,Hostname,Record,Service}"
-     peer=(name="org.freedesktop.resolve1"),
+     peer=(name="org.freedesktop.resolve1", label=unconfined),
+
+dbus (send)
+     bus=system
+     path="/org/freedesktop/resolve1"
+     interface="org.freedesktop.resolve1.Manager"
+     member="SetLink{DefaultRoute,DNSOverTLS,DNS,DNSEx,DNSSEC,DNSSECNegativeTrustAnchors,MulticastDNS,Domains,LLMNR}"
+     peer=(label=unconfined),
+
+# required by resolvectl command
+dbus (send)
+     bus=system
+     path="/org/freedesktop/resolve1"
+     interface=org.freedesktop.DBus.Properties
+     member=Get{,All}
+     peer=(label=unconfined),
+
+# required by resolvectl command
+dbus (receive)
+     bus=system
+     path="/org/freedesktop/resolve1"
+     interface=org.freedesktop.DBus.Properties
+     member=PropertiesChanged
+     peer=(label=unconfined),
+
+# required by resolvectl command
+dbus (send)
+     bus=system
+     path="/org/freedesktop/resolve1/link/*"
+     interface="org.freedesktop.DBus.Properties"
+     member=Get{,All}
+     peer=(label=unconfined),
+
+# required by resolvectl command
+dbus (receive)
+     bus=system
+     path="/org/freedesktop/resolve1/link/*"
+     interface="org.freedesktop.DBus.Properties"
+     member=PropertiesChanged
+     peer=(label=unconfined),
 
 #include <abstractions/ssl_certs>
 
 capability net_admin,
 capability net_raw,
 capability setuid, # ping
+capability net_broadcast, # openvswitchd
 
 # Allow protocols except those that we blacklist in
 # /etc/modprobe.d/blacklist-rare-network.conf
@@ -119,6 +163,7 @@ network sna,
 /{,usr/}{,s}bin/pppdump ixr,
 /{,usr/}{,s}bin/pppoe-discovery ixr,
 #/{,usr/}{,s}bin/pppstats ixr,            # needs sys_module
+/{,usr/}{,s}bin/resolvectl ixr,
 /{,usr/}{,s}bin/route ixr,
 /{,usr/}{,s}bin/routef ixr,
 /{,usr/}{,s}bin/routel ixr,
@@ -137,12 +182,16 @@ network sna,
 /sys/devices/{pci[0-9a-f]*,platform,virtual}/**/rfkill[0-9]*/{,**} r,
 /sys/devices/{pci[0-9a-f]*,platform,virtual}/**/rfkill[0-9]*/state w,
 
+# For reading the address of a particular ethernet interface
+/sys/devices/{pci[0-9a-f]*,platform,virtual}/**/net/*/address r,
+
 # arp
 network netlink dgram,
 
 # ip, et al
-/etc/iproute2/{,*} r,
+/etc/iproute2/{,**} r,
 /etc/iproute2/rt_{protos,realms,scopes,tables} w,
+/etc/iproute2/rt_{protos,tables}.d/* w,
 
 # ping - child profile would be nice but seccomp causes problems with that
 /{,usr/}{,s}bin/ping ixr,
@@ -159,13 +208,13 @@ capability setuid,
 /etc/hosts w,
 
 # resolvconf
-/sbin/resolvconf ixr,
-/run/resolvconf/{,**} r,
+/{,usr/}sbin/resolvconf ixr,
+/run/resolvconf/{,**} rk,
 /run/resolvconf/** w,
 /etc/resolvconf/{,**} r,
-/lib/resolvconf/* ix,
+/{,usr/}lib/resolvconf/* ix,
 # Required by resolvconf
-/bin/run-parts ixr,
+/{,usr/}bin/run-parts ixr,
 /etc/resolvconf/update.d/* ix,
 
 # wpa_suplicant
@@ -195,6 +244,9 @@ capability setuid,
 # are virtual and don't show up in /dev
 /dev/net/tun rw,
 
+# Access to sysfs interfaces for tun/tap/mstp/bchat device settings.
+/sys/devices/virtual/net/{tap*,mstp*,bchat*}/** rw,
+
 # access to bridge sysfs interfaces for bridge settings
 /sys/devices/virtual/net/*/bridge/* rw,
 
@@ -279,6 +331,53 @@ var networkControlConnectedPlugUDev = []string{
 	`KERNEL=="tun"`,
 }
 
+var networkControlConnectedPlugMount = []osutil.MountEntry{{
+	Name:    "/var/lib/snapd/hostfs/var/lib/dhcp",
+	Dir:     "/var/lib/dhcp",
+	Options: []string{"bind", "rw", osutil.XSnapdIgnoreMissing()},
+}}
+
+// TODO: Add a layer that derives this sort of data from mount entry, like the
+// one above, into a set of apparmor rules for snap-update-ns, like the ones
+// below.
+//
+// When setting up a mount entry, we also need corresponding
+// snap-updates-ns rules. Eg, if have:
+// []osutil.MountEntry{{
+//	Name:    "/foo/bar",
+//	Dir:     "/bar",
+//	Options: []string{"rw", "bind"},
+// }}
+// Then you can expect to need:
+// /foo/ r,
+// /foo/bar/ r,
+// mount options=(rw bind) /foo/bar/ -> /bar/,
+// umount /bar/,
+// ...
+// You'll need 'r' rules for all the directories that need to be traversed,
+// starting from the root directory all the way down to the directory being
+// mounted. This is required by the safe bind mounting trick employed by
+// snap-update-ns.
+//
+// You'll need 'rw' rules to support cases when snap-update-ns is expected to
+// create the missing directory, before performing the bind mount. Note that
+// there are two sides, one side is the host visible through
+// /var/lib/snapd/hostfs and the other side is everything else. To support
+// writes to the host side you need to coordinate with the trespassing rules
+// implemented in snap-update-ns/system.go.
+var networkControlConnectedPlugUpdateNSAppArmor = `
+/var/ r,
+/var/lib/ r,
+/var/lib/snapd/ r,
+/var/lib/snapd/hostfs/ r,
+/var/lib/snapd/hostfs/var/ r,
+/var/lib/snapd/hostfs/var/lib/ r,
+/var/lib/snapd/hostfs/var/lib/dhcp/ r,
+/var/lib/dhcp/ r,
+mount options=(rw bind) /var/lib/snapd/hostfs/var/lib/dhcp/ -> /var/lib/dhcp/,
+umount /var/lib/dhcp/,
+`
+
 func init() {
 	registerIface(&commonInterface{
 		name:                  "network-control",
@@ -289,8 +388,14 @@ func init() {
 		connectedPlugAppArmor: networkControlConnectedPlugAppArmor,
 		connectedPlugSecComp:  networkControlConnectedPlugSecComp,
 		connectedPlugUDev:     networkControlConnectedPlugUDev,
-		reservedForOS:         true,
-		suppressPtraceTrace:   true,
-	})
 
+		connectedPlugMount:            networkControlConnectedPlugMount,
+		connectedPlugUpdateNSAppArmor: networkControlConnectedPlugUpdateNSAppArmor,
+
+		suppressPtraceTrace:         true,
+		suppressSysModuleCapability: true,
+
+		// affects the plug snap because of mount backend
+		affectsPlugOnRefresh: true,
+	})
 }
