@@ -35,12 +35,17 @@ const rawusbConnectedPlugAppArmor = `
 /dev/bus/usb/[0-9][0-9][0-9]/[0-9][0-9][0-9] rw,
 
 # Allow access to all ttyUSB devices too
-/dev/tty{USB,ACM}[0-9]* rw,
+/dev/tty{USB,ACM}[0-9]* rwk,
+@{PROC}/tty/drivers r,
+
+# Allow raw access to USB printers (i.e. for receipt printers in POS systems).
+/dev/usb/lp[0-9]* rwk,
 
 # Allow detection of usb devices. Leaks plugged in USB device info
 /sys/bus/usb/devices/ r,
 /sys/devices/pci**/usb[0-9]** r,
 /sys/devices/platform/soc/*.usb/usb[0-9]** r,
+/sys/devices/platform/scb/*.pcie/pci**/usb[0-9]** r,
 
 /run/udev/data/c16[67]:[0-9] r, # ACM USB modems
 /run/udev/data/b180:*    r, # various USB block devices
@@ -48,8 +53,17 @@ const rawusbConnectedPlugAppArmor = `
 /run/udev/data/+usb:* r,
 `
 
+const rawusbConnectedPlugSecComp = `
+# Description: Allow raw access to all connected USB devices.
+# This gives privileged access to the system.
+
+# kernel uevents
+socket AF_NETLINK - NETLINK_KOBJECT_UEVENT
+`
+
 var rawusbConnectedPlugUDev = []string{
 	`SUBSYSTEM=="usb"`,
+	`SUBSYSTEM=="usbmisc"`,
 	`SUBSYSTEM=="tty", ENV{ID_BUS}=="usb"`,
 }
 
@@ -61,7 +75,7 @@ func init() {
 		implicitOnClassic:     true,
 		baseDeclarationSlots:  rawusbBaseDeclarationSlots,
 		connectedPlugAppArmor: rawusbConnectedPlugAppArmor,
+		connectedPlugSecComp:  rawusbConnectedPlugSecComp,
 		connectedPlugUDev:     rawusbConnectedPlugUDev,
-		reservedForOS:         true,
 	})
 }
