@@ -23,10 +23,12 @@ import (
 	"fmt"
 	"path/filepath"
 	"regexp"
+	"strconv"
 	"strings"
 
 	"github.com/snapcore/snapd/interfaces"
 	"github.com/snapcore/snapd/interfaces/apparmor"
+	"github.com/snapcore/snapd/interfaces/hotplug"
 	"github.com/snapcore/snapd/interfaces/udev"
 	"github.com/snapcore/snapd/snap"
 )
@@ -69,26 +71,30 @@ func (iface *serialPortInterface) String() string {
 //  - ttyXRUSBx  (Exar Corp. USB UART devices)
 //  - ttySX (UART serial ports)
 //  - ttyOX (UART serial ports on ARM)
-var serialDeviceNodePattern = regexp.MustCompile("^/dev/tty(USB|ACM|AMA|XRUSB|S|O)[0-9]+$")
+//  - ttymxcX (serial ports on i.mx6UL)
+//  - ttySCX (NXP SC16IS7xx serial devices)
+//  - ttyMSMX (Qualcomm msm7x serial devices)
+//  - ttyHSX (Qualcomm GENI based QTI serial cores)
+//  - ttyGSX (USB gadget serial devices)
+var serialDeviceNodePattern = regexp.MustCompile("^/dev/tty(mxc|USB|ACM|AMA|XRUSB|S|O|SC|MSM|HS|GS)[0-9]+$")
 
 // Pattern that is considered valid for the udev symlink to the serial device,
 // path attributes will be compared to this for validity when usb vid and pid
 // are also specified
 var serialUDevSymlinkPattern = regexp.MustCompile("^/dev/serial-port-[a-z0-9]+$")
 
-// SanitizeSlot checks validity of the defined slot
-func (iface *serialPortInterface) SanitizeSlot(slot *interfaces.Slot) error {
-	if err := sanitizeSlotReservedForOSOrGadget(iface, slot); err != nil {
-		return err
-	}
-
+// BeforePrepareSlot checks validity of the defined slot
+func (iface *serialPortInterface) BeforePrepareSlot(slot *snap.SlotInfo) error {
 	// Check slot has a path attribute identify serial device
 	path, ok := slot.Attrs["path"].(string)
 	if !ok || path == "" {
 		return fmt.Errorf("serial-port slot must have a path attribute")
 	}
 
-	// Clean the path before further checks
+	// XXX: this interface feeds the cleaned path into the regex and is
+	// left unchanged here for historical reasons. New interfaces (eg,
+	// like raw-volume) should instead use verifySlotPathAttribute() which
+	// performs additional verification.
 	path = filepath.Clean(path)
 
 	if iface.hasUsbAttrs(slot) {
@@ -129,19 +135,18 @@ func (iface *serialPortInterface) SanitizeSlot(slot *interfaces.Slot) error {
 }
 
 func (iface *serialPortInterface) UDevPermanentSlot(spec *udev.Specification, slot *snap.SlotInfo) error {
-	usbVendor, ok := slot.Attrs["usb-vendor"].(int64)
-	if !ok {
+	var usbVendor, usbProduct, usbInterfaceNumber int64
+	var path string
+	if err := slot.Attr("usb-vendor", &usbVendor); err != nil {
 		return nil
 	}
-	usbProduct, ok := slot.Attrs["usb-product"].(int64)
-	if !ok {
+	if err := slot.Attr("usb-product", &usbProduct); err != nil {
 		return nil
 	}
-	path, ok := slot.Attrs["path"].(string)
-	if !ok || path == "" {
+	if err := slot.Attr("path", &path); err != nil || path == "" {
 		return nil
 	}
-	if usbInterfaceNumber, ok := slot.Attrs["usb-interface-number"].(int64); ok {
+	if err := slot.Attr("usb-interface-number", &usbInterfaceNumber); err == nil {
 		spec.AddSnippet(fmt.Sprintf(`# serial-port
 IMPORT{builtin}="usb_id"
 SUBSYSTEM=="tty", SUBSYSTEMS=="usb", ATTRS{idVendor}=="%04x", ATTRS{idProduct}=="%04x", ENV{ID_USB_INTERFACE_NUM}=="%02x", SYMLINK+="%s"`, usbVendor, usbProduct, usbInterfaceNumber, strings.TrimPrefix(path, "/dev/")))
@@ -153,46 +158,46 @@ SUBSYSTEM=="tty", SUBSYSTEMS=="usb", ATTRS{idVendor}=="%04x", ATTRS{idProduct}==
 	return nil
 }
 
-func (iface *serialPortInterface) AppArmorConnectedPlug(spec *apparmor.Specification, plug *interfaces.Plug, plugAttrs map[string]interface{}, slot *interfaces.Slot, slotAttrs map[string]interface{}) error {
+func (iface *serialPortInterface) AppArmorConnectedPlug(spec *apparmor.Specification, plug *interfaces.ConnectedPlug, slot *interfaces.ConnectedSlot) error {
 	if iface.hasUsbAttrs(slot) {
 		// This apparmor rule is an approximation of serialDeviceNodePattern
 		// (AARE is different than regex, so we must approximate).
 		// UDev tagging and device cgroups will restrict down to the specific device
-		spec.AddSnippet("/dev/tty[A-Z]*[0-9] rw,")
+		spec.AddSnippet("/dev/tty[A-Z]*[0-9] rwk,")
 		return nil
 	}
 
 	// Path to fixed device node
-	path, pathOk := slot.Attrs["path"].(string)
-	if !pathOk {
+	var path string
+	if err := slot.Attr("path", &path); err != nil {
 		return nil
 	}
 	cleanedPath := filepath.Clean(path)
-	spec.AddSnippet(fmt.Sprintf("%s rw,", cleanedPath))
+	spec.AddSnippet(fmt.Sprintf("%s rwk,", cleanedPath))
 	return nil
 }
 
-func (iface *serialPortInterface) UDevConnectedPlug(spec *udev.Specification, plug *interfaces.Plug, plugAttrs map[string]interface{}, slot *interfaces.Slot, slotAttrs map[string]interface{}) error {
+func (iface *serialPortInterface) UDevConnectedPlug(spec *udev.Specification, plug *interfaces.ConnectedPlug, slot *interfaces.ConnectedSlot) error {
 	// For connected plugs, we use vendor and product ids if available,
 	// otherwise add the kernel device
 	hasOnlyPath := !iface.hasUsbAttrs(slot)
-	usbVendor, ok := slot.Attrs["usb-vendor"].(int64)
-	if !ok && !hasOnlyPath {
+	var usbVendor, usbProduct int64
+	var path string
+	if err := slot.Attr("usb-vendor", &usbVendor); err != nil && !hasOnlyPath {
 		return nil
 	}
-	usbProduct, ok := slot.Attrs["usb-product"].(int64)
-	if !ok && !hasOnlyPath {
+	if err := slot.Attr("usb-product", &usbProduct); err != nil && !hasOnlyPath {
 		return nil
 	}
-	path, ok := slot.Attrs["path"].(string)
-	if !ok && hasOnlyPath {
+	if err := slot.Attr("path", &path); err != nil && hasOnlyPath {
 		return nil
 	}
 
 	if hasOnlyPath {
 		spec.TagDevice(fmt.Sprintf(`SUBSYSTEM=="tty", KERNEL=="%s"`, strings.TrimPrefix(path, "/dev/")))
 	} else {
-		if usbInterfaceNumber, ok := slot.Attrs["usb-interface-number"].(int64); ok {
+		var usbInterfaceNumber int64
+		if err := slot.Attr("usb-interface-number", &usbInterfaceNumber); err == nil {
 			spec.TagDevice(fmt.Sprintf(`IMPORT{builtin}="usb_id"
 SUBSYSTEM=="tty", SUBSYSTEMS=="usb", ATTRS{idVendor}=="%04x", ATTRS{idProduct}=="%04x", ENV{ID_USB_INTERFACE_NUM}=="%02x"`, usbVendor, usbProduct, usbInterfaceNumber))
 		} else {
@@ -203,19 +208,78 @@ SUBSYSTEM=="tty", SUBSYSTEMS=="usb", ATTRS{idVendor}=="%04x", ATTRS{idProduct}==
 	return nil
 }
 
-func (iface *serialPortInterface) AutoConnect(*interfaces.Plug, *interfaces.Slot) bool {
+func (iface *serialPortInterface) AutoConnect(*snap.PlugInfo, *snap.SlotInfo) bool {
 	// allow what declarations allowed
 	return true
 }
 
-func (iface *serialPortInterface) hasUsbAttrs(slot *interfaces.Slot) bool {
-	if _, ok := slot.Attrs["usb-vendor"]; ok {
+func (iface *serialPortInterface) HotplugDeviceDetected(di *hotplug.HotplugDeviceInfo) (*hotplug.ProposedSlot, error) {
+	bus, _ := di.Attribute("ID_BUS")
+	if di.Subsystem() != "tty" || bus != "usb" || !serialDeviceNodePattern.MatchString(di.DeviceName()) {
+		return nil, nil
+	}
+
+	slot := hotplug.ProposedSlot{
+		Attrs: map[string]interface{}{
+			"path": di.DeviceName(),
+		},
+	}
+	if vendor, ok := di.Attribute("ID_VENDOR_ID"); ok {
+		slot.Attrs["usb-vendor"] = vendor
+	}
+	if product, ok := di.Attribute("ID_MODEL_ID"); ok {
+		slot.Attrs["usb-product"] = product
+	}
+	return &slot, nil
+}
+
+func slotDeviceAttrEqual(di *hotplug.HotplugDeviceInfo, devinfoAttribute string, slotAttributeValue int64) bool {
+	var attr string
+	var ok bool
+	if attr, ok = di.Attribute(devinfoAttribute); !ok {
+		return false
+	}
+	val, err := strconv.ParseInt(attr, 16, 64)
+	return err == nil && val == slotAttributeValue
+}
+
+func (iface *serialPortInterface) HandledByGadget(di *hotplug.HotplugDeviceInfo, slot *snap.SlotInfo) bool {
+	// if the slot has vendor, product and interface number set, check if they match
+	var usbVendor, usbProduct, usbInterfaceNumber int64
+	if err := slot.Attr("usb-vendor", &usbVendor); err == nil {
+		if !slotDeviceAttrEqual(di, "ID_VENDOR_ID", usbVendor) {
+			return false
+		}
+		if err := slot.Attr("usb-product", &usbProduct); err != nil {
+			return false
+		}
+		if !slotDeviceAttrEqual(di, "ID_MODEL_ID", usbProduct) {
+			return false
+		}
+		if err := slot.Attr("usb-interface-number", &usbInterfaceNumber); err == nil {
+			if !slotDeviceAttrEqual(di, "ID_USB_INTERFACE_NUM", usbInterfaceNumber) {
+				return false
+			}
+		}
 		return true
 	}
-	if _, ok := slot.Attrs["usb-product"]; ok {
+
+	var path string
+	if err := slot.Attr("path", &path); err != nil {
+		return false
+	}
+	return di.DeviceName() == path
+}
+
+func (iface *serialPortInterface) hasUsbAttrs(attrs interfaces.Attrer) bool {
+	var v int64
+	if err := attrs.Attr("usb-vendor", &v); err == nil {
 		return true
 	}
-	if _, ok := slot.Attrs["usb-interface-number"]; ok {
+	if err := attrs.Attr("usb-product", &v); err == nil {
+		return true
+	}
+	if err := attrs.Attr("usb-interface-number", &v); err == nil {
 		return true
 	}
 	return false
