@@ -62,7 +62,7 @@ apps:
   plugs: [unity8-calendar]
 `
 	s.slotInfo = &snap.SlotInfo{
-		Snap:      &snap.Info{SuggestedName: "core", Type: snap.TypeOS},
+		Snap:      &snap.Info{SuggestedName: "core", SnapType: snap.TypeOS},
 		Name:      "unity8-calendar",
 		Interface: "unity8-calendar",
 	}
