@@ -42,6 +42,7 @@ var defaultTemplate = []byte(`
 
 access
 faccessat
+faccessat2
 
 alarm
 brk
@@ -102,11 +103,17 @@ clock_gettime64
 clock_nanosleep
 clock_nanosleep_time64
 clone
+clone3
 close
+close_range
 
 # needed by ls -l
 connect
 
+# the file descriptors used here will already be mediated by apparmor,
+# the 6th argument is flags, which currently is always 0
+copy_file_range - - - - - 0
+
 chroot
 
 creat
@@ -140,6 +147,7 @@ fork
 ftime
 futex
 futex_time64
+futex_waitv
 get_mempolicy
 get_robust_list
 get_thread_area
@@ -186,7 +194,12 @@ inotify_init
 inotify_init1
 inotify_rm_watch
 
-# TIOCSTI allows for faking input (man tty_ioctl)
+# ioctl() mediation currently primarily relies on Linux capabilities as well as
+# the initial syscall for the fd to pass to ioctl(). See 'man capabilities'
+# and 'man ioctl_list'. TIOCSTI requires CAP_SYS_ADMIN but allows for faking
+# input (man tty_ioctl), so we disallow it to prevent snaps plugging interfaces
+# with 'capability sys_admin' from interfering with other snaps or the
+# unconfined user's terminal.
 # TODO: this should be scaled back even more
 ioctl - !TIOCSTI
 
@@ -246,16 +259,6 @@ mknodat - - |S_IFSOCK -
 modify_ldt
 mprotect
 
-# LP: #1448184 - these aren't currently mediated by AppArmor. Deny for now
-#mq_getsetattr
-#mq_notify
-#mq_open
-#mq_timedreceive
-#mq_timedreceive_time64
-#mq_timedsend
-#mq_timedsend_time64
-#mq_unlink
-
 mremap
 msgctl
 msgget
@@ -332,6 +335,10 @@ renameat2
 restart_syscall
 
 rmdir
+
+# glibc 2.35 unconditionally calls rseq for all threads
+rseq
+
 rt_sigaction
 rt_sigpending
 rt_sigprocmask
@@ -470,6 +477,7 @@ socket AF_NFC
 socket AF_VSOCK
 socket AF_MPLS
 socket AF_IB
+socket AF_QIPCRTR
 
 # For usrsctp, AppArmor doesn't support 'network conn,' since AF_CONN is
 # userspace and encapsulated in other domains that are mediated. As such, do
