@@ -158,10 +158,7 @@ func (b *Backend) Remove(snapName string) error {
 }
 
 func (b *Backend) deriveContent(spec *Specification, snapInfo *snap.Info) (content []string) {
-	for _, snippet := range spec.Snippets() {
-		content = append(content, snippet)
-	}
-
+	content = append(content, spec.Snippets()...)
 	return content
 }
 
