@@ -30,33 +30,43 @@ import (
 	"github.com/snapcore/snapd/dirs"
 	"github.com/snapcore/snapd/osutil"
 	"github.com/snapcore/snapd/overlord/configstate/configcore"
-	"github.com/snapcore/snapd/release"
 	"github.com/snapcore/snapd/snap"
 	"github.com/snapcore/snapd/testutil"
 )
 
 type servicesSuite struct {
 	configcoreSuite
+	serviceInstalled bool
 }
 
 var _ = Suite(&servicesSuite{})
 
 func (s *servicesSuite) SetUpTest(c *C) {
 	s.configcoreSuite.SetUpTest(c)
+	s.systemctlOutput = func(args ...string) []byte {
+		var output []byte
+		if args[0] == "show" {
+			if args[1] == "--property=ActiveState" {
+				output = []byte("ActiveState=inactive")
+			} else {
+				if s.serviceInstalled {
+					output = []byte(fmt.Sprintf("Id=%s\nType=daemon\nActiveState=inactive\nUnitFileState=enabled\nNames=%[1]s\nNeedDaemonReload=no\n", args[2]))
+				} else {
+					output = []byte(fmt.Sprintf("Id=%s\nType=\nActiveState=inactive\nUnitFileState=\nNames=%[1]s\nNeedDaemonReload=no\n", args[2]))
+				}
+			}
+		}
+		return output
+	}
+
 	c.Assert(os.MkdirAll(filepath.Join(dirs.GlobalRootDir, "etc"), 0755), IsNil)
+	s.serviceInstalled = true
 	s.systemctlArgs = nil
 	s.BaseTest.AddCleanup(snap.MockSanitizePlugsSlots(func(snapInfo *snap.Info) {}))
-
-	// mock an empty cmdline since we check the cmdline to check whether we are
-	// in install mode or not and we don't want to use the host's proc/cmdline
-	s.AddCleanup(osutil.MockProcCmdline(filepath.Join(c.MkDir(), "proc/cmdline")))
 }
 
 func (s *servicesSuite) TestConfigureServiceInvalidValue(c *C) {
-	restore := release.MockOnClassic(false)
-	defer restore()
-
-	err := configcore.Run(&mockConf{
+	err := configcore.Run(coreDev, &mockConf{
 		state: s.state,
 		changes: map[string]interface{}{
 			"service.ssh.disable": "xxx",
@@ -69,8 +79,10 @@ func (s *servicesSuite) TestConfigureServiceNotDisabled(c *C) {
 	err := configcore.SwitchDisableService("sshd.service", false, nil)
 	c.Assert(err, IsNil)
 	c.Check(s.systemctlArgs, DeepEquals, [][]string{
-		{"--root", dirs.GlobalRootDir, "unmask", "sshd.service"},
-		{"--root", dirs.GlobalRootDir, "enable", "sshd.service"},
+		{"show", "--property=Id,ActiveState,UnitFileState,Type,Names,NeedDaemonReload", "sshd.service"},
+		{"unmask", "sshd.service"},
+		{"--no-reload", "enable", "sshd.service"},
+		{"daemon-reload"},
 		{"start", "sshd.service"},
 	})
 }
@@ -79,30 +91,33 @@ func (s *servicesSuite) TestConfigureServiceDisabled(c *C) {
 	err := configcore.SwitchDisableService("sshd.service", true, nil)
 	c.Assert(err, IsNil)
 	c.Check(s.systemctlArgs, DeepEquals, [][]string{
-		{"--root", dirs.GlobalRootDir, "disable", "sshd.service"},
-		{"--root", dirs.GlobalRootDir, "mask", "sshd.service"},
+		{"show", "--property=Id,ActiveState,UnitFileState,Type,Names,NeedDaemonReload", "sshd.service"},
+		{"--no-reload", "disable", "sshd.service"},
+		{"mask", "sshd.service"},
 		{"stop", "sshd.service"},
 		{"show", "--property=ActiveState", "sshd.service"},
 	})
 }
 
 func (s *servicesSuite) TestConfigureServiceDisabledIntegration(c *C) {
-	restore := release.MockOnClassic(false)
-	defer restore()
-
 	err := os.MkdirAll(filepath.Join(dirs.GlobalRootDir, "/etc/ssh"), 0755)
 	c.Assert(err, IsNil)
 
 	for _, service := range []struct {
 		cfgName     string
 		systemdName string
+		installed   bool
 	}{
-		{"ssh", "ssh.service"},
-		{"rsyslog", "rsyslog.service"},
+		{"ssh", "ssh.service", true},  // no installed check for ssh
+		{"ssh", "ssh.service", false}, // no installed check for ssh
+		{"rsyslog", "rsyslog.service", true},
+		{"rsyslog", "rsyslog.service", false},
+		{"systemd-resolved", "systemd-resolved.service", true},
+		{"systemd-resolved", "systemd-resolved.service", false},
 	} {
 		s.systemctlArgs = nil
-
-		err := configcore.Run(&mockConf{
+		s.serviceInstalled = service.installed
+		err := configcore.Run(coreDev, &mockConf{
 			state: s.state,
 			conf: map[string]interface{}{
 				fmt.Sprintf("service.%s.disable", service.cfgName): true,
@@ -120,50 +135,48 @@ func (s *servicesSuite) TestConfigureServiceDisabledIntegration(c *C) {
 				{"show", "--property=ActiveState", srv},
 			})
 		default:
-			c.Check(s.systemctlArgs, DeepEquals, [][]string{
-				{"--root", dirs.GlobalRootDir, "disable", srv},
-				{"--root", dirs.GlobalRootDir, "mask", srv},
-				{"stop", srv},
-				{"show", "--property=ActiveState", srv},
-			})
+			if service.installed {
+				c.Check(s.systemctlArgs, DeepEquals, [][]string{
+					{"show", "--property=Id,ActiveState,UnitFileState,Type,Names,NeedDaemonReload", srv},
+					{"--no-reload", "disable", srv},
+					{"mask", srv},
+					{"stop", srv},
+					{"show", "--property=ActiveState", srv},
+				})
+			} else {
+				c.Check(s.systemctlArgs, DeepEquals, [][]string{
+					{"show", "--property=Id,ActiveState,UnitFileState,Type,Names,NeedDaemonReload", srv},
+				})
+			}
 		}
 	}
 }
 
 func (s *servicesSuite) TestConfigureConsoleConfDisableFSOnly(c *C) {
-	restore := release.MockOnClassic(false)
-	defer restore()
-
 	conf := configcore.PlainCoreConfig(map[string]interface{}{
 		"service.console-conf.disable": true,
 	})
 
 	tmpDir := c.MkDir()
-	c.Assert(configcore.FilesystemOnlyApply(tmpDir, conf, nil), IsNil)
+	c.Assert(configcore.FilesystemOnlyApply(coreDev, tmpDir, conf), IsNil)
 
 	consoleConfDisabled := filepath.Join(tmpDir, "/var/lib/console-conf/complete")
 	c.Check(consoleConfDisabled, testutil.FileEquals, "console-conf has been disabled by the snapd system configuration\n")
 }
 
 func (s *servicesSuite) TestConfigureConsoleConfEnabledFSOnly(c *C) {
-	restore := release.MockOnClassic(false)
-	defer restore()
-
 	conf := configcore.PlainCoreConfig(map[string]interface{}{
 		"service.console-conf.disable": false,
 	})
 
 	tmpDir := c.MkDir()
-	c.Assert(configcore.FilesystemOnlyApply(tmpDir, conf, nil), IsNil)
+	c.Assert(configcore.FilesystemOnlyApply(coreDev, tmpDir, conf), IsNil)
 
 	consoleConfDisabled := filepath.Join(tmpDir, "/var/lib/console-conf/complete")
 	c.Check(consoleConfDisabled, testutil.FileAbsent)
 }
 
 func (s *servicesSuite) TestConfigureConsoleConfEnableNotAtRuntime(c *C) {
-	restore := release.MockOnClassic(false)
-	defer restore()
-
 	// pretend that console-conf is disabled
 	canary := filepath.Join(dirs.GlobalRootDir, "/var/lib/console-conf/complete")
 	err := os.MkdirAll(filepath.Dir(canary), 0755)
@@ -172,7 +185,7 @@ func (s *servicesSuite) TestConfigureConsoleConfEnableNotAtRuntime(c *C) {
 	c.Assert(err, IsNil)
 
 	// now enable it
-	err = configcore.Run(&mockConf{
+	err = configcore.Run(coreDev, &mockConf{
 		state: s.state,
 		conf: map[string]interface{}{
 			"service.console-conf.disable": false,
@@ -182,14 +195,11 @@ func (s *servicesSuite) TestConfigureConsoleConfEnableNotAtRuntime(c *C) {
 }
 
 func (s *servicesSuite) TestConfigureConsoleConfDisableNotAtRuntime(c *C) {
-	restore := release.MockOnClassic(false)
-	defer restore()
-
 	// console-conf is not disabled, i.e. there is no
 	// "/var/lib/console-conf/complete" file
 
 	// now try to enable it
-	err := configcore.Run(&mockConf{
+	err := configcore.Run(coreDev, &mockConf{
 		state: s.state,
 		conf: map[string]interface{}{
 			"service.console-conf.disable": true,
@@ -199,13 +209,10 @@ func (s *servicesSuite) TestConfigureConsoleConfDisableNotAtRuntime(c *C) {
 }
 
 func (s *servicesSuite) TestConfigureConsoleConfEnableAlreadyEnabledIsFine(c *C) {
-	restore := release.MockOnClassic(false)
-	defer restore()
-
 	// Note that we have no
 	//        /var/lib/console-conf/complete
 	// file. So console-conf is already enabled
-	err := configcore.Run(&mockConf{
+	err := configcore.Run(coreDev, &mockConf{
 		state: s.state,
 		conf: map[string]interface{}{
 			"service.console-conf.disable": false,
@@ -215,9 +222,6 @@ func (s *servicesSuite) TestConfigureConsoleConfEnableAlreadyEnabledIsFine(c *C)
 }
 
 func (s *servicesSuite) TestConfigureConsoleConfDisableAlreadyDisabledIsFine(c *C) {
-	restore := release.MockOnClassic(false)
-	defer restore()
-
 	// pretend that console-conf is disabled
 	canary := filepath.Join(dirs.GlobalRootDir, "/var/lib/console-conf/complete")
 	err := os.MkdirAll(filepath.Dir(canary), 0755)
@@ -225,7 +229,7 @@ func (s *servicesSuite) TestConfigureConsoleConfDisableAlreadyDisabledIsFine(c *
 	err = ioutil.WriteFile(canary, nil, 0644)
 	c.Assert(err, IsNil)
 
-	err = configcore.Run(&mockConf{
+	err = configcore.Run(coreDev, &mockConf{
 		state: s.state,
 		conf: map[string]interface{}{
 			"service.console-conf.disable": true,
@@ -235,16 +239,13 @@ func (s *servicesSuite) TestConfigureConsoleConfDisableAlreadyDisabledIsFine(c *
 }
 
 func (s *servicesSuite) TestConfigureConsoleConfEnableDuringInstallMode(c *C) {
-	restore := release.MockOnClassic(false)
-	defer restore()
-
 	mockProcCmdline := filepath.Join(c.MkDir(), "cmdline")
 	err := ioutil.WriteFile(mockProcCmdline, []byte("snapd_recovery_mode=install snapd_recovery_system=20201212\n"), 0644)
 	c.Assert(err, IsNil)
-	restore = osutil.MockProcCmdline(mockProcCmdline)
+	restore := osutil.MockProcCmdline(mockProcCmdline)
 	defer restore()
 
-	err = configcore.Run(&mockConf{
+	err = configcore.Run(coreDev, &mockConf{
 		state: s.state,
 		conf: map[string]interface{}{
 			"service.console-conf.disable": true,
@@ -255,21 +256,24 @@ func (s *servicesSuite) TestConfigureConsoleConfEnableDuringInstallMode(c *C) {
 }
 
 func (s *servicesSuite) TestConfigureServiceEnableIntegration(c *C) {
-	restore := release.MockOnClassic(false)
-	defer restore()
-
 	err := os.MkdirAll(filepath.Join(dirs.GlobalRootDir, "/etc/ssh"), 0755)
 	c.Assert(err, IsNil)
 
 	for _, service := range []struct {
 		cfgName     string
 		systemdName string
+		installed   bool
 	}{
-		{"ssh", "ssh.service"},
-		{"rsyslog", "rsyslog.service"},
+		{"ssh", "ssh.service", true},  // no installed check for ssh
+		{"ssh", "ssh.service", false}, // no installed check for ssh
+		{"rsyslog", "rsyslog.service", true},
+		{"rsyslog", "rsyslog.service", false},
+		{"systemd-resolved", "systemd-resolved.service", true},
+		{"systemd-resolved", "systemd-resolved.service", false},
 	} {
 		s.systemctlArgs = nil
-		err := configcore.Run(&mockConf{
+		s.serviceInstalled = service.installed
+		err := configcore.Run(coreDev, &mockConf{
 			state: s.state,
 			conf: map[string]interface{}{
 				fmt.Sprintf("service.%s.disable", service.cfgName): false,
@@ -281,28 +285,33 @@ func (s *servicesSuite) TestConfigureServiceEnableIntegration(c *C) {
 		switch service.cfgName {
 		case "ssh":
 			c.Check(s.systemctlArgs, DeepEquals, [][]string{
-				{"--root", dirs.GlobalRootDir, "unmask", "sshd.service"},
-				{"--root", dirs.GlobalRootDir, "unmask", "ssh.service"},
+				{"unmask", "sshd.service"},
+				{"unmask", "ssh.service"},
 				{"start", srv},
 			})
 			sshCanary := filepath.Join(dirs.GlobalRootDir, "/etc/ssh/sshd_not_to_be_run")
 			_, err := os.Stat(sshCanary)
 			c.Assert(err, ErrorMatches, ".* no such file or directory")
 		default:
-			c.Check(s.systemctlArgs, DeepEquals, [][]string{
-				{"--root", dirs.GlobalRootDir, "unmask", srv},
-				{"--root", dirs.GlobalRootDir, "enable", srv},
-				{"start", srv},
-			})
+			if service.installed {
+				c.Check(s.systemctlArgs, DeepEquals, [][]string{
+					{"show", "--property=Id,ActiveState,UnitFileState,Type,Names,NeedDaemonReload", srv},
+					{"unmask", srv},
+					{"--no-reload", "enable", srv},
+					{"daemon-reload"},
+					{"start", srv},
+				})
+			} else {
+				c.Check(s.systemctlArgs, DeepEquals, [][]string{
+					{"show", "--property=Id,ActiveState,UnitFileState,Type,Names,NeedDaemonReload", srv},
+				})
+			}
 		}
 	}
 }
 
 func (s *servicesSuite) TestConfigureServiceUnsupportedService(c *C) {
-	restore := release.MockOnClassic(false)
-	defer restore()
-
-	err := configcore.Run(&mockConf{
+	err := configcore.Run(coreDev, &mockConf{
 		state: s.state,
 		conf: map[string]interface{}{
 			"service.snapd.disable": true,
@@ -323,7 +332,7 @@ func (s *servicesSuite) TestFilesystemOnlyApply(c *C) {
 		"service.ssh.disable":     "true",
 		"service.rsyslog.disable": "true",
 	})
-	c.Assert(configcore.FilesystemOnlyApply(tmpDir, conf, nil), IsNil)
+	c.Assert(configcore.FilesystemOnlyApply(coreDev, tmpDir, conf), IsNil)
 	c.Check(s.systemctlArgs, DeepEquals, [][]string{
 		{"--root", tmpDir, "mask", "rsyslog.service"},
 	})
