@@ -74,7 +74,8 @@ var _ = Suite(&deviceMgrSystemsSuite{})
 var _ = Suite(&deviceMgrSystemsCreateSuite{})
 
 func (s *deviceMgrSystemsBaseSuite) SetUpTest(c *C) {
-	s.deviceMgrBaseSuite.SetUpTest(c)
+	classic := false
+	s.deviceMgrBaseSuite.setupBaseTest(c, classic)
 
 	s.brands.Register("other-brand", brandPrivKey3, map[string]interface{}{
 		"display-name": "other publisher",
@@ -139,10 +140,10 @@ func (s *deviceMgrSystemsBaseSuite) SetUpTest(c *C) {
 		ModelSignKeyID: s.model.SignKeyID(),
 	}
 	err := modeenv.WriteTo("")
-	s.state.Set("seeded", true)
-
 	c.Assert(err, IsNil)
 
+	s.state.Set("seeded", true)
+
 	logbuf, restore := logger.MockLogger()
 	s.logbuf = logbuf
 	s.AddCleanup(restore)
@@ -278,6 +279,13 @@ var defaultSystemActions []devicestate.SystemAction = []devicestate.SystemAction
 var currentSystemActions []devicestate.SystemAction = []devicestate.SystemAction{
 	{Title: "Reinstall", Mode: "install"},
 	{Title: "Recover", Mode: "recover"},
+	{Title: "Factory reset", Mode: "factory-reset"},
+	{Title: "Run normally", Mode: "run"},
+}
+
+var recoverySystemActions []devicestate.SystemAction = []devicestate.SystemAction{
+	{Title: "Reinstall", Mode: "install"},
+	{Title: "Factory reset", Mode: "factory-reset"},
 	{Title: "Run normally", Mode: "run"},
 }
 
@@ -385,6 +393,59 @@ func (s *deviceMgrSystemsSuite) TestListSeedSystemsCurrentManySeeded(c *C) {
 	}})
 }
 
+func (s *deviceMgrSystemsSuite) TestListSeedSystemsCurrentInRecoveryMode(c *C) {
+	// mock recovery mode
+	modeenv := boot.Modeenv{
+		Mode:           "recover",
+		RecoverySystem: s.mockedSystemSeeds[1].label,
+
+		Model:          s.mockedSystemSeeds[1].model.Model(),
+		BrandID:        s.mockedSystemSeeds[1].brand.AccountID(),
+		Grade:          string(s.mockedSystemSeeds[1].model.Grade()),
+		ModelSignKeyID: s.mockedSystemSeeds[1].model.SignKeyID(),
+	}
+	err := modeenv.WriteTo("")
+	c.Assert(err, IsNil)
+	// update the internal mode
+	devicestate.SetSystemMode(s.mgr, "recover")
+
+	s.state.Lock()
+	s.state.Set("seeded-systems", []devicestate.SeededSystem{
+		{
+			System:  s.mockedSystemSeeds[1].label,
+			Model:   s.mockedSystemSeeds[1].model.Model(),
+			BrandID: s.mockedSystemSeeds[1].brand.AccountID(),
+		},
+	})
+	s.state.Unlock()
+
+	systems, err := s.mgr.Systems()
+	c.Assert(err, IsNil)
+	c.Assert(systems, HasLen, 3)
+	c.Check(systems, DeepEquals, []*devicestate.System{{
+		Current: false,
+		Label:   s.mockedSystemSeeds[0].label,
+		Model:   s.mockedSystemSeeds[0].model,
+		Brand:   s.mockedSystemSeeds[0].brand,
+		Actions: defaultSystemActions,
+	}, {
+		// this seed was used for installing the running system, but
+		// since we are in recovery mode, the available actions are
+		// slightly different
+		Current: true,
+		Label:   s.mockedSystemSeeds[1].label,
+		Model:   s.mockedSystemSeeds[1].model,
+		Brand:   s.mockedSystemSeeds[1].brand,
+		Actions: recoverySystemActions,
+	}, {
+		Current: false,
+		Label:   s.mockedSystemSeeds[2].label,
+		Model:   s.mockedSystemSeeds[2].model,
+		Brand:   s.mockedSystemSeeds[2].brand,
+		Actions: defaultSystemActions,
+	}})
+}
+
 func (s *deviceMgrSystemsSuite) TestBrokenSeedSystems(c *C) {
 	// break the first seed
 	err := os.Remove(filepath.Join(dirs.SnapSeedDir, "systems", s.mockedSystemSeeds[0].label, "model"))
@@ -437,7 +498,7 @@ func (s *deviceMgrSystemsSuite) TestRequestSameModeSameSystem(c *C) {
 	label := s.mockedSystemSeeds[0].label
 
 	happyModes := []string{"run"}
-	sadModes := []string{"install", "recover"}
+	sadModes := []string{"install", "recover", "factory-reset"}
 
 	for _, mode := range append(happyModes, sadModes...) {
 		s.logbuf.Reset()
@@ -488,7 +549,7 @@ func (s *deviceMgrSystemsSuite) TestRequestSeedingSameConflict(c *C) {
 	s.state.Set("seeded", nil)
 	s.state.Unlock()
 
-	for _, mode := range []string{"run", "install", "recover"} {
+	for _, mode := range []string{"run", "install", "recover", "factory-reset"} {
 		s.logbuf.Reset()
 
 		c.Logf("checking mode: %q", mode)
@@ -567,7 +628,7 @@ func (s *deviceMgrSystemsSuite) testRequestModeWithRestart(c *C, toModes []strin
 	}
 }
 
-func (s *deviceMgrSystemsSuite) TestRequestModeRunInstallForRecover(c *C) {
+func (s *deviceMgrSystemsSuite) TestRequestModeRunInstallResetForRecover(c *C) {
 	// we are in recover mode here
 	devicestate.SetSystemMode(s.mgr, "recover")
 	// non run modes use modeenv
@@ -588,7 +649,7 @@ func (s *deviceMgrSystemsSuite) TestRequestModeRunInstallForRecover(c *C) {
 	})
 	s.state.Unlock()
 
-	s.testRequestModeWithRestart(c, []string{"install", "run"}, s.mockedSystemSeeds[0].label)
+	s.testRequestModeWithRestart(c, []string{"install", "run", "factory-reset"}, s.mockedSystemSeeds[0].label)
 }
 
 func (s *deviceMgrSystemsSuite) TestRequestModeInstallRecoverForCurrent(c *C) {
@@ -610,7 +671,7 @@ func (s *deviceMgrSystemsSuite) TestRequestModeInstallRecoverForCurrent(c *C) {
 	})
 	s.state.Unlock()
 
-	s.testRequestModeWithRestart(c, []string{"install", "recover"}, s.mockedSystemSeeds[0].label)
+	s.testRequestModeWithRestart(c, []string{"install", "recover", "factory-reset"}, s.mockedSystemSeeds[0].label)
 }
 
 func (s *deviceMgrSystemsSuite) TestRequestModeErrInBoot(c *C) {
@@ -678,6 +739,8 @@ func (s *deviceMgrSystemsSuite) TestRequestModeForNonCurrent(c *C) {
 	c.Assert(err, Equals, devicestate.ErrUnsupportedAction)
 	err = s.mgr.RequestSystemAction(s.mockedSystemSeeds[1].label, devicestate.SystemAction{Mode: "recover"})
 	c.Assert(err, Equals, devicestate.ErrUnsupportedAction)
+	err = s.mgr.RequestSystemAction(s.mockedSystemSeeds[1].label, devicestate.SystemAction{Mode: "factory-reset"})
+	c.Assert(err, Equals, devicestate.ErrUnsupportedAction)
 	c.Check(s.restartRequests, HasLen, 0)
 	c.Check(s.logbuf.String(), Equals, "")
 }
@@ -781,7 +844,7 @@ func (s *deviceMgrSystemsSuite) TestRebootLabelAndModeHappy(c *C) {
 	c.Check(s.logbuf.String(), Matches, `.*: rebooting into system "20191119" in "install" mode\n`)
 }
 
-func (s *deviceMgrSystemsSuite) TestRebootModeOnlyHappy(c *C) {
+func (s *deviceMgrSystemsSuite) TestRebootFromRunOnlyHappy(c *C) {
 	s.state.Lock()
 	s.state.Set("seeded-systems", []devicestate.SeededSystem{
 		{
@@ -792,7 +855,7 @@ func (s *deviceMgrSystemsSuite) TestRebootModeOnlyHappy(c *C) {
 	})
 	s.state.Unlock()
 
-	for _, mode := range []string{"recover", "install"} {
+	for _, mode := range []string{"recover", "install", "factory-reset"} {
 		s.restartRequests = nil
 		s.bootloader.BootVars = make(map[string]string)
 		s.logbuf.Reset()
@@ -811,7 +874,7 @@ func (s *deviceMgrSystemsSuite) TestRebootModeOnlyHappy(c *C) {
 	}
 }
 
-func (s *deviceMgrSystemsSuite) TestRebootFromRecoverToRun(c *C) {
+func (s *deviceMgrSystemsSuite) TestRebootFromRecoverToOther(c *C) {
 	modeenv := boot.Modeenv{
 		Mode:           "recover",
 		RecoverySystem: s.mockedSystemSeeds[0].label,
@@ -836,17 +899,23 @@ func (s *deviceMgrSystemsSuite) TestRebootFromRecoverToRun(c *C) {
 	})
 	s.state.Unlock()
 
-	err = s.mgr.Reboot("", "run")
-	c.Assert(err, IsNil)
+	for _, mode := range []string{"run", "factory-reset"} {
+		s.restartRequests = nil
+		s.bootloader.BootVars = make(map[string]string)
+		s.logbuf.Reset()
 
-	m, err := s.bootloader.GetBootVars("snapd_recovery_mode", "snapd_recovery_system")
-	c.Assert(err, IsNil)
-	c.Check(m, DeepEquals, map[string]string{
-		"snapd_recovery_mode":   "run",
-		"snapd_recovery_system": s.mockedSystemSeeds[0].label,
-	})
-	c.Check(s.restartRequests, DeepEquals, []restart.RestartType{restart.RestartSystemNow})
-	c.Check(s.logbuf.String(), Matches, fmt.Sprintf(`.*: rebooting into system "%s" in "run" mode\n`, s.mockedSystemSeeds[0].label))
+		err = s.mgr.Reboot("", mode)
+		c.Assert(err, IsNil)
+
+		m, err := s.bootloader.GetBootVars("snapd_recovery_mode", "snapd_recovery_system")
+		c.Assert(err, IsNil)
+		c.Check(m, DeepEquals, map[string]string{
+			"snapd_recovery_mode":   mode,
+			"snapd_recovery_system": s.mockedSystemSeeds[0].label,
+		})
+		c.Check(s.restartRequests, DeepEquals, []restart.RestartType{restart.RestartSystemNow})
+		c.Check(s.logbuf.String(), Matches, fmt.Sprintf(`.*: rebooting into system "%s" in "%s" mode\n`, s.mockedSystemSeeds[0].label, mode))
+	}
 }
 
 func (s *deviceMgrSystemsSuite) TestRebootAlreadyInRunMode(c *C) {
@@ -994,7 +1063,7 @@ func (s *deviceMgrSystemsSuite) TestDeviceManagerEnsureTriedSystemMissingInModee
 	var triedSystems []string
 	s.state.Lock()
 	err = s.state.Get("tried-systems", &triedSystems)
-	c.Assert(err, Equals, state.ErrNoState)
+	c.Assert(err, testutil.ErrorIs, state.ErrNoState)
 	// also logged
 	c.Check(s.logbuf.String(), testutil.Contains, `tried recovery system outcome error: recovery system "1234" was tried, but is not present in the modeenv CurrentRecoverySystems`)
 	s.state.Unlock()
@@ -1024,7 +1093,7 @@ func (s *deviceMgrSystemsSuite) TestDeviceManagerEnsureTriedSystemBad(c *C) {
 	var triedSystems []string
 	s.state.Lock()
 	err = s.state.Get("tried-systems", &triedSystems)
-	c.Assert(err, Equals, state.ErrNoState)
+	c.Assert(err, testutil.ErrorIs, state.ErrNoState)
 	c.Check(s.logbuf.String(), testutil.Contains, `tried recovery system "1234" failed`)
 	s.state.Unlock()
 
@@ -1044,7 +1113,7 @@ func (s *deviceMgrSystemsSuite) TestDeviceManagerEnsureTriedSystemBad(c *C) {
 	s.state.Lock()
 	defer s.state.Unlock()
 	err = s.state.Get("tried-systems", &triedSystems)
-	c.Assert(err, Equals, state.ErrNoState)
+	c.Assert(err, testutil.ErrorIs, state.ErrNoState)
 	// bootenv got cleared
 	m, err = s.bootloader.GetBootVars("try_recovery_system", "recovery_system_status")
 	c.Assert(err, IsNil)
@@ -1426,7 +1495,7 @@ func (s *deviceMgrSystemsCreateSuite) TestDeviceManagerCreateRecoverySystemHappy
 
 	var triedSystemsAfterFinalize []string
 	err = s.state.Get("tried-systems", &triedSystemsAfterFinalize)
-	c.Assert(err, Equals, state.ErrNoState)
+	c.Assert(err, testutil.ErrorIs, state.ErrNoState)
 
 	modeenvAfterFinalize, err := boot.ReadModeenv("")
 	c.Assert(err, IsNil)
@@ -1454,8 +1523,8 @@ func (s *deviceMgrSystemsCreateSuite) TestDeviceManagerCreateRecoverySystemRemod
 	barSnap := snaptest.MakeTestSnapWithFiles(c, "name: bar\nversion: 1.0\nbase: core20", nil)
 	s.state.Lock()
 	// fake downloads are a nop
-	tSnapsup1 := s.state.NewTask("fake-download", "dummy task carrying snap setup")
-	tSnapsup2 := s.state.NewTask("fake-download", "dummy task carrying snap setup")
+	tSnapsup1 := s.state.NewTask("fake-download", "test task carrying snap setup")
+	tSnapsup2 := s.state.NewTask("fake-download", "test task carrying snap setup")
 	// both snaps are asserted
 	snapsupFoo := snapstate.SnapSetup{
 		SideInfo: &snap.SideInfo{RealName: "foo", SnapID: s.ss.AssertedSnapID("foo"), Revision: snap.R(99)},
@@ -1495,7 +1564,7 @@ func (s *deviceMgrSystemsCreateSuite) TestDeviceManagerCreateRecoverySystemRemod
 	})
 	tss.WaitFor(tSnapsup1)
 	tss.WaitFor(tSnapsup2)
-	// add the dummy tasks to the change
+	// add the test tasks to the change
 	chg := s.state.NewChange("create-recovery-system", "create recovery system")
 	chg.AddTask(tSnapsup1)
 	chg.AddTask(tSnapsup2)
@@ -1634,7 +1703,7 @@ func (s *deviceMgrSystemsCreateSuite) TestDeviceManagerCreateRecoverySystemRemod
 	s.state.Lock()
 	defer s.state.Unlock()
 	// fake downloads are a nop
-	tSnapsup1 := s.state.NewTask("fake-download", "dummy task carrying snap setup")
+	tSnapsup1 := s.state.NewTask("fake-download", "test task carrying snap setup")
 	// both snaps are asserted
 	snapsupFoo := snapstate.SnapSetup{
 		SideInfo: &snap.SideInfo{RealName: "foo", SnapID: s.ss.AssertedSnapID("foo"), Revision: snap.R(99)},
@@ -1659,7 +1728,7 @@ func (s *deviceMgrSystemsCreateSuite) TestDeviceManagerCreateRecoverySystemRemod
 		"snap-setup-tasks": []interface{}{tSnapsup1.ID()},
 	})
 	tss.WaitFor(tSnapsup1)
-	// add the dummy task to the change
+	// add the test task to the change
 	chg := s.state.NewChange("create-recovery-system", "create recovery system")
 	chg.AddTask(tSnapsup1)
 	chg.AddAll(tss)
@@ -1748,7 +1817,7 @@ func (s *deviceMgrSystemsCreateSuite) TestDeviceManagerCreateRecoverySystemUndo(
 	c.Assert(tskFinalize.Status(), Equals, state.DoingStatus)
 	// a reboot is expected
 	c.Check(s.restartRequests, DeepEquals, []restart.RestartType{restart.RestartSystemNow})
-	// sanity check asserted snaps location
+	// validity check asserted snaps location
 	c.Check(filepath.Join(boot.InitramfsUbuntuSeedDir, "systems/1234undo"), testutil.FilePresent)
 	p, err := filepath.Glob(filepath.Join(boot.InitramfsUbuntuSeedDir, "snaps/*"))
 	c.Assert(err, IsNil)
@@ -1804,7 +1873,7 @@ func (s *deviceMgrSystemsCreateSuite) TestDeviceManagerCreateRecoverySystemUndo(
 
 	var triedSystemsAfter []string
 	err = s.state.Get("tried-systems", &triedSystemsAfter)
-	c.Assert(err, Equals, state.ErrNoState)
+	c.Assert(err, testutil.ErrorIs, state.ErrNoState)
 
 	modeenvAfterFinalize, err := boot.ReadModeenv("")
 	c.Assert(err, IsNil)
@@ -1973,7 +2042,7 @@ func (s *deviceMgrSystemsCreateSuite) TestDeviceManagerCreateRecoverySystemErrCl
 	c.Assert(tskFinalize.Status(), Equals, state.HoldStatus)
 
 	c.Check(s.restartRequests, HasLen, 0)
-	// sanity check asserted snaps location
+	// validity check asserted snaps location
 	c.Check(filepath.Join(boot.InitramfsUbuntuSeedDir, "systems/1234error"), testutil.FileAbsent)
 	p, err := filepath.Glob(filepath.Join(boot.InitramfsUbuntuSeedDir, "snaps/*"))
 	c.Assert(err, IsNil)
