@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2015-2019 Canonical Ltd
+ * Copyright (C) 2015-2021 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -24,6 +24,7 @@ import (
 	"io/ioutil"
 	"os"
 	"path/filepath"
+	"runtime"
 	"strconv"
 	"strings"
 	"time"
@@ -38,6 +39,7 @@ import (
 	"github.com/snapcore/snapd/bootloader"
 	"github.com/snapcore/snapd/bootloader/bootloadertest"
 	"github.com/snapcore/snapd/dirs"
+	"github.com/snapcore/snapd/gadget"
 	"github.com/snapcore/snapd/osutil"
 	"github.com/snapcore/snapd/overlord"
 	"github.com/snapcore/snapd/overlord/assertstate"
@@ -48,10 +50,10 @@ import (
 	"github.com/snapcore/snapd/overlord/devicestate/devicestatetest"
 	"github.com/snapcore/snapd/overlord/hookstate"
 	"github.com/snapcore/snapd/overlord/ifacestate"
+	"github.com/snapcore/snapd/overlord/restart"
 	"github.com/snapcore/snapd/overlord/snapstate"
 	"github.com/snapcore/snapd/overlord/state"
 	"github.com/snapcore/snapd/release"
-	"github.com/snapcore/snapd/seed"
 	"github.com/snapcore/snapd/seed/seedtest"
 	"github.com/snapcore/snapd/snap"
 	"github.com/snapcore/snapd/snap/snaptest"
@@ -76,6 +78,11 @@ type firstBootBaseTest struct {
 func (t *firstBootBaseTest) setupBaseTest(c *C, s *seedtest.SeedSnaps) {
 	t.BaseTest.SetUpTest(c)
 
+	// TODO: temporary: skip due to timeouts on riscv64
+	if runtime.GOARCH == "riscv64" || os.Getenv("SNAPD_SKIP_SLOW_TESTS") != "" {
+		c.Skip("skipping slow test")
+	}
+
 	tempdir := c.MkDir()
 	dirs.SetRootDir(tempdir)
 	t.AddCleanup(func() { dirs.SetRootDir("/") })
@@ -126,7 +133,20 @@ func (t *firstBootBaseTest) startOverlord(c *C) {
 	ovld, err := overlord.New(nil)
 	c.Assert(err, IsNil)
 	ovld.InterfaceManager().DisableUDevMonitor()
+	// avoid gadget preload in the general tests cases
+	// it requires a proper seed to be available
+	devicestate.EarlyConfig = func(st *state.State, preloadGadget func() (sysconfig.Device, *gadget.Info, error)) error {
+		return nil
+	}
 	t.overlord = ovld
+	t.AddCleanup(func() {
+		devicestate.EarlyConfig = nil
+		if t.overlord == nil {
+			return
+		}
+		t.overlord.Stop()
+		t.overlord = nil
+	})
 	c.Assert(ovld.StartUp(), IsNil)
 
 	// don't actually try to talk to the store on snapstate.Ensure
@@ -158,8 +178,6 @@ func (s *firstBoot16Suite) SetUpTest(c *C) {
 	s.TestingSeed16 = &seedtest.TestingSeed16{}
 	s.setup16BaseTest(c, &s.firstBootBaseTest)
 
-	s.startOverlord(c)
-
 	s.SeedDir = dirs.SnapSeedDir
 
 	err := os.MkdirAll(filepath.Join(dirs.SnapSeedDir, "assertions"), 0755)
@@ -215,6 +233,7 @@ func (s *firstBoot16Suite) TestPopulateFromSeedOnClassicNoop(c *C) {
 	restore := release.MockOnClassic(true)
 	defer restore()
 
+	s.startOverlord(c)
 	st := s.overlord.State()
 	st.Lock()
 	defer st.Unlock()
@@ -222,6 +241,9 @@ func (s *firstBoot16Suite) TestPopulateFromSeedOnClassicNoop(c *C) {
 	err := os.Remove(filepath.Join(dirs.SnapSeedDir, "assertions"))
 	c.Assert(err, IsNil)
 
+	_, _, err = devicestate.PreloadGadget(s.overlord.DeviceManager())
+	c.Check(err, testutil.ErrorIs, state.ErrNoState)
+
 	tsAll, err := devicestate.PopulateStateFromSeedImpl(st, nil, s.perfTimings)
 	c.Assert(err, IsNil)
 	checkTrivialSeeding(c, tsAll)
@@ -250,6 +272,7 @@ func (s *firstBoot16Suite) TestPopulateFromSeedOnClassicNoSeedYaml(c *C) {
 	defer restore()
 
 	ovld, err := overlord.New(nil)
+	defer ovld.Stop()
 	c.Assert(err, IsNil)
 	st := ovld.State()
 
@@ -260,6 +283,9 @@ func (s *firstBoot16Suite) TestPopulateFromSeedOnClassicNoSeedYaml(c *C) {
 	st.Lock()
 	defer st.Unlock()
 
+	_, _, err = devicestate.PreloadGadget(ovld.DeviceManager())
+	c.Check(err, testutil.ErrorIs, state.ErrNoState)
+
 	tsAll, err := devicestate.PopulateStateFromSeedImpl(st, nil, s.perfTimings)
 	c.Assert(err, IsNil)
 	checkTrivialSeeding(c, tsAll)
@@ -275,6 +301,7 @@ func (s *firstBoot16Suite) TestPopulateFromSeedOnClassicEmptySeedYaml(c *C) {
 	defer restore()
 
 	ovld, err := overlord.New(nil)
+	defer ovld.Stop()
 	c.Assert(err, IsNil)
 	st := ovld.State()
 
@@ -301,6 +328,7 @@ func (s *firstBoot16Suite) TestPopulateFromSeedOnClassicNoSeedYamlWithCloudInsta
 	restore := release.MockOnClassic(true)
 	defer restore()
 
+	s.startOverlord(c)
 	st := s.overlord.State()
 
 	// add the model assertion and its chain
@@ -369,6 +397,7 @@ func (s *firstBoot16Suite) TestPopulateFromSeedOnClassicNoSeedYamlWithCloudInsta
 }
 
 func (s *firstBoot16Suite) TestPopulateFromSeedErrorsOnState(c *C) {
+	s.startOverlord(c)
 	st := s.overlord.State()
 	st.Lock()
 	defer st.Unlock()
@@ -539,6 +568,7 @@ func (s *firstBoot16Suite) TestPopulateFromSeedHappy(c *C) {
 	bloader.SetBootKernel("pc-kernel_1.snap")
 	bloader.SetBootBase("core_1.snap")
 
+	s.startOverlord(c)
 	st := s.overlord.State()
 	chg, model := s.makeSeedChange(c, st, nil, checkSeedTasks, checkOrder)
 	err := s.overlord.Settle(settleTimeout)
@@ -583,7 +613,7 @@ func (s *firstBoot16Suite) TestPopulateFromSeedHappy(c *C) {
 	c.Assert(err, IsNil)
 	c.Assert(info.SnapID, Equals, "foodidididididididididididididid")
 	c.Assert(info.Revision, Equals, snap.R(128))
-	c.Assert(info.Contact, Equals, "mailto:some.guy@example.com")
+	c.Assert(info.Contact(), Equals, "mailto:some.guy@example.com")
 	pubAcct, err := assertstate.Publisher(st, info.SnapID)
 	c.Assert(err, IsNil)
 	c.Check(pubAcct.AccountID(), Equals, "developerid")
@@ -630,6 +660,7 @@ func (s *firstBoot16Suite) TestPopulateFromSeedHappy(c *C) {
 }
 
 func (s *firstBoot16Suite) TestPopulateFromSeedMissingBootloader(c *C) {
+	s.startOverlord(c)
 	st0 := s.overlord.State()
 	st0.Lock()
 	db := assertstate.DB(st0)
@@ -717,6 +748,7 @@ snaps:
 	c.Assert(err, IsNil)
 
 	// run the firstboot stuff
+	s.startOverlord(c)
 	st := s.overlord.State()
 	st.Lock()
 	defer st.Unlock()
@@ -822,9 +854,19 @@ snaps:
 	c.Assert(err, IsNil)
 
 	// run the firstboot stuff
+	s.startOverlord(c)
 	st := s.overlord.State()
 	st.Lock()
 	defer st.Unlock()
+
+	dev, gi, err := devicestate.PreloadGadget(s.overlord.DeviceManager())
+	c.Assert(err, IsNil)
+	c.Check(gi.Defaults, HasLen, 4)
+	c.Check(dev.RunMode(), Equals, true)
+	c.Check(dev.Classic(), Equals, false)
+	c.Check(dev.HasModeenv(), Equals, false)
+	c.Check(dev.Kernel(), Equals, "pc-kernel")
+
 	tsAll, err := devicestate.PopulateStateFromSeedImpl(st, nil, s.perfTimings)
 	c.Assert(err, IsNil)
 
@@ -854,7 +896,7 @@ snaps:
 	defer rhk()
 
 	// ensure we have something that captures the core config
-	restore := configstate.MockConfigcoreRun(func(config.Conf) error {
+	restore := configstate.MockConfigcoreRun(func(sysconfig.Device, config.Conf) error {
 		configured = append(configured, "configcore")
 		return nil
 	})
@@ -969,6 +1011,7 @@ snaps:
 	c.Assert(err, IsNil)
 
 	// run the firstboot stuff
+	s.startOverlord(c)
 	st := s.overlord.State()
 	st.Lock()
 	defer st.Unlock()
@@ -1039,6 +1082,7 @@ func (s *firstBoot16Suite) TestImportAssertionsFromSeedClassicModelMismatch(c *C
 	defer restore()
 
 	ovld, err := overlord.New(nil)
+	defer ovld.Stop()
 	c.Assert(err, IsNil)
 	st := ovld.State()
 
@@ -1050,15 +1094,13 @@ func (s *firstBoot16Suite) TestImportAssertionsFromSeedClassicModelMismatch(c *C
 	st.Lock()
 	defer st.Unlock()
 
-	deviceSeed, err := seed.Open(dirs.SnapSeedDir, "")
-	c.Assert(err, IsNil)
-
-	_, err = devicestate.ImportAssertionsFromSeed(st, deviceSeed)
+	_, err = devicestate.ImportAssertionsFromSeed(st, "")
 	c.Assert(err, ErrorMatches, "cannot seed a classic system with an all-snaps model")
 }
 
 func (s *firstBoot16Suite) TestImportAssertionsFromSeedAllSnapsModelMismatch(c *C) {
 	ovld, err := overlord.New(nil)
+	defer ovld.Stop()
 	c.Assert(err, IsNil)
 	st := ovld.State()
 
@@ -1070,15 +1112,110 @@ func (s *firstBoot16Suite) TestImportAssertionsFromSeedAllSnapsModelMismatch(c *
 	st.Lock()
 	defer st.Unlock()
 
-	deviceSeed, err := seed.Open(dirs.SnapSeedDir, "")
+	_, err = devicestate.ImportAssertionsFromSeed(st, "")
+	c.Assert(err, ErrorMatches, "cannot seed an all-snaps system with a classic model")
+}
+
+func (s *firstBoot16Suite) TestLoadDeviceSeed(c *C) {
+	ovld, err := overlord.New(nil)
+	defer ovld.Stop()
+	c.Assert(err, IsNil)
+	st := ovld.State()
+
+	// add a bunch of assertions (model assertion and its chain)
+	assertsChain := s.makeModelAssertionChain(c, "my-model", nil)
+	for i, as := range assertsChain {
+		fname := strconv.Itoa(i)
+		if as.Type() == asserts.ModelType {
+			fname = "model"
+		}
+		s.WriteAssertions(fname, as)
+	}
+
+	// load them
+	st.Lock()
+	defer st.Unlock()
+
+	deviceSeed, err := devicestate.LoadDeviceSeed(st, "")
 	c.Assert(err, IsNil)
 
-	_, err = devicestate.ImportAssertionsFromSeed(st, deviceSeed)
-	c.Assert(err, ErrorMatches, "cannot seed an all-snaps system with a classic model")
+	c.Check(deviceSeed.Model().BrandID(), Equals, "my-brand")
+	c.Check(deviceSeed.Model().Model(), Equals, "my-model")
+
+	// verify that the model was added
+	db := assertstate.DB(st)
+	as, err := db.Find(asserts.ModelType, map[string]string{
+		"series":   "16",
+		"brand-id": "my-brand",
+		"model":    "my-model",
+	})
+	c.Assert(err, IsNil)
+	c.Check(as, DeepEquals, deviceSeed.Model())
+}
+
+func (s *firstBoot16Suite) TestLoadDeviceSeedCaching(c *C) {
+	ovld, err := overlord.New(nil)
+	defer ovld.Stop()
+	c.Assert(err, IsNil)
+	st := ovld.State()
+
+	// add a bunch of assertions (model assertion and its chain)
+	assertsChain := s.makeModelAssertionChain(c, "my-model", nil)
+	for i, as := range assertsChain {
+		fname := strconv.Itoa(i)
+		if as.Type() == asserts.ModelType {
+			fname = "model"
+		}
+		s.WriteAssertions(fname, as)
+	}
+
+	// load them
+	st.Lock()
+	defer st.Unlock()
+
+	deviceSeed, err := devicestate.LoadDeviceSeed(st, "")
+	c.Assert(err, IsNil)
+
+	c.Check(deviceSeed.Model().Model(), Equals, "my-model")
+
+	// break the seed
+	modelAway := filepath.Join(c.MkDir(), "model")
+	modelFn := filepath.Join(s.AssertsDir(), "model")
+	err = os.Rename(modelFn, modelAway)
+	c.Assert(err, IsNil)
+
+	// result is still cached
+	deviceSeed, err = devicestate.LoadDeviceSeed(st, "")
+	c.Assert(err, IsNil)
+
+	c.Check(deviceSeed.Model().Model(), Equals, "my-model")
+
+	// unload cached result
+	devicestate.UnloadDeviceSeed(st)
+
+	// error now
+	_, err1 := devicestate.LoadDeviceSeed(st, "")
+	c.Assert(err1, ErrorMatches, "seed must have a model assertion")
+
+	// refix the seed
+	err = os.Rename(modelAway, modelFn)
+	c.Assert(err, IsNil)
+
+	// error is also cached
+	_, err = devicestate.LoadDeviceSeed(st, "")
+	c.Assert(err, Equals, err1)
+
+	// unload cached error
+	devicestate.UnloadDeviceSeed(st)
+
+	// soundness check: loading works again
+	_, err = devicestate.LoadDeviceSeed(st, "")
+	c.Assert(err, IsNil)
 }
 
 func (s *firstBoot16Suite) TestImportAssertionsFromSeedHappy(c *C) {
 	ovld, err := overlord.New(nil)
+	defer ovld.Stop()
 	c.Assert(err, IsNil)
 	st := ovld.State()
 
@@ -1096,12 +1233,11 @@ func (s *firstBoot16Suite) TestImportAssertionsFromSeedHappy(c *C) {
 	st.Lock()
 	defer st.Unlock()
 
-	deviceSeed, err := seed.Open(dirs.SnapSeedDir, "")
+	deviceSeed, err := devicestate.ImportAssertionsFromSeed(st, "")
 	c.Assert(err, IsNil)
+	c.Assert(deviceSeed, NotNil)
 
-	model, err := devicestate.ImportAssertionsFromSeed(st, deviceSeed)
-	c.Assert(err, IsNil)
-	c.Assert(model, NotNil)
+	model := deviceSeed.Model()
 
 	// verify that the model was added
 	db := assertstate.DB(st)
@@ -1124,6 +1260,7 @@ func (s *firstBoot16Suite) TestImportAssertionsFromSeedHappy(c *C) {
 }
 
 func (s *firstBoot16Suite) TestImportAssertionsFromSeedMissingSig(c *C) {
+	s.startOverlord(c)
 	st := s.overlord.State()
 	st.Lock()
 	defer st.Unlock()
@@ -1137,16 +1274,14 @@ func (s *firstBoot16Suite) TestImportAssertionsFromSeedMissingSig(c *C) {
 		}
 	}
 
-	deviceSeed, err := seed.Open(dirs.SnapSeedDir, "")
-	c.Assert(err, IsNil)
-
 	// try import and verify that its rejects because other assertions are
 	// missing
-	_, err = devicestate.ImportAssertionsFromSeed(st, deviceSeed)
+	_, err := devicestate.ImportAssertionsFromSeed(st, "")
 	c.Assert(err, ErrorMatches, "cannot resolve prerequisite assertion: account-key .*")
 }
 
 func (s *firstBoot16Suite) TestImportAssertionsFromSeedTwoModelAsserts(c *C) {
+	s.startOverlord(c)
 	st := s.overlord.State()
 	st.Lock()
 	defer st.Unlock()
@@ -1158,16 +1293,14 @@ func (s *firstBoot16Suite) TestImportAssertionsFromSeedTwoModelAsserts(c *C) {
 	model2 := s.Brands.Model("my-brand", "my-second-model", modelHeaders("my-second-model"))
 	s.WriteAssertions("model2", model2)
 
-	deviceSeed, err := seed.Open(dirs.SnapSeedDir, "")
-	c.Assert(err, IsNil)
-
 	// try import and verify that its rejects because other assertions are
 	// missing
-	_, err = devicestate.ImportAssertionsFromSeed(st, deviceSeed)
+	_, err := devicestate.ImportAssertionsFromSeed(st, "")
 	c.Assert(err, ErrorMatches, "cannot have multiple model assertions in seed")
 }
 
 func (s *firstBoot16Suite) TestImportAssertionsFromSeedNoModelAsserts(c *C) {
+	s.startOverlord(c)
 	st := s.overlord.State()
 	st.Lock()
 	defer st.Unlock()
@@ -1179,12 +1312,9 @@ func (s *firstBoot16Suite) TestImportAssertionsFromSeedNoModelAsserts(c *C) {
 		}
 	}
 
-	deviceSeed, err := seed.Open(dirs.SnapSeedDir, "")
-	c.Assert(err, IsNil)
-
 	// try import and verify that its rejects because other assertions are
 	// missing
-	_, err = devicestate.ImportAssertionsFromSeed(st, deviceSeed)
+	_, err := devicestate.ImportAssertionsFromSeed(st, "")
 	c.Assert(err, ErrorMatches, "seed must have a model assertion")
 }
 
@@ -1209,7 +1339,14 @@ type: base`
 	snapdYaml := `name: snapd
 version: 1.0
 `
-	snapdFname, snapdDecl, snapdRev := s.MakeAssertedSnap(c, snapdYaml, nil, snap.R(2), "canonical")
+	// the info file is needed by the Ensure() loop of snapstate manager
+	snapdSnapFiles := [][]string{
+		{"usr/lib/snapd/info", `
+VERSION=2.54.3+git1.g479e745-dirty
+SNAPD_APPARMOR_REEXEC=1
+`},
+	}
+	snapdFname, snapdDecl, snapdRev := s.MakeAssertedSnap(c, snapdYaml, snapdSnapFiles, snap.R(2), "canonical")
 	s.WriteAssertions("snapd.asserts", snapdRev, snapdDecl)
 
 	var kernelFname string
@@ -1284,6 +1421,7 @@ snaps:
 	c.Assert(err, IsNil)
 
 	// run the firstboot stuff
+	s.startOverlord(c)
 	st := s.overlord.State()
 	st.Lock()
 	defer st.Unlock()
@@ -1315,7 +1453,7 @@ snaps:
 	// at this point the system is "restarting", pretend the restart has
 	// happened
 	c.Assert(chg.Status(), Equals, state.DoingStatus)
-	state.MockRestarting(st, state.RestartUnset)
+	restart.MockPending(st, restart.RestartUnset)
 	st.Unlock()
 	err = s.overlord.Settle(settleTimeout)
 	st.Lock()
@@ -1406,6 +1544,7 @@ snaps:
 	c.Assert(err, IsNil)
 
 	// run the firstboot stuff
+	s.startOverlord(c)
 	st := s.overlord.State()
 	st.Lock()
 	defer st.Unlock()
@@ -1442,6 +1581,7 @@ snaps:
 	c.Assert(err, IsNil)
 
 	// run the firstboot stuff
+	s.startOverlord(c)
 	st := s.overlord.State()
 	st.Lock()
 	defer st.Unlock()
@@ -1508,6 +1648,7 @@ snaps:
 	c.Assert(err, IsNil)
 
 	// run the firstboot stuff
+	s.startOverlord(c)
 	st := s.overlord.State()
 	st.Lock()
 	defer st.Unlock()
@@ -1579,6 +1720,7 @@ snaps:
 	c.Assert(ioutil.WriteFile(filepath.Join(dirs.SnapSeedDir, "seed.yaml"), content, 0644), IsNil)
 
 	// run the firstboot stuff
+	s.startOverlord(c)
 	st := s.overlord.State()
 	st.Lock()
 	defer st.Unlock()
@@ -1627,6 +1769,7 @@ snaps:
 	c.Assert(err, IsNil)
 
 	// run the firstboot stuff
+	s.startOverlord(c)
 	st := s.overlord.State()
 	st.Lock()
 	defer st.Unlock()
@@ -1658,7 +1801,7 @@ snaps:
 	// at this point the system is "restarting", pretend the restart has
 	// happened
 	c.Assert(chg.Status(), Equals, state.DoingStatus)
-	state.MockRestarting(st, state.RestartUnset)
+	restart.MockPending(st, restart.RestartUnset)
 	st.Unlock()
 	err = s.overlord.Settle(settleTimeout)
 	st.Lock()
@@ -1712,6 +1855,7 @@ snaps:
 	c.Assert(err, IsNil)
 
 	// run the firstboot stuff
+	s.startOverlord(c)
 	st := s.overlord.State()
 	st.Lock()
 	defer st.Unlock()
@@ -1766,6 +1910,7 @@ snaps:
 	c.Assert(err, IsNil)
 
 	// run the firstboot stuff
+	s.startOverlord(c)
 	st := s.overlord.State()
 	st.Lock()
 	defer st.Unlock()
@@ -1797,7 +1942,7 @@ snaps:
 	// at this point the system is "restarting", pretend the restart has
 	// happened
 	c.Assert(chg.Status(), Equals, state.DoingStatus)
-	state.MockRestarting(st, state.RestartUnset)
+	restart.MockPending(st, restart.RestartUnset)
 	st.Unlock()
 	err = s.overlord.Settle(settleTimeout)
 	st.Lock()
@@ -1836,6 +1981,7 @@ snaps:
 }
 
 func (s *firstBoot16Suite) TestCriticalTaskEdgesForPreseed(c *C) {
+	s.startOverlord(c)
 	st := s.overlord.State()
 	st.Lock()
 	defer st.Unlock()
@@ -1861,6 +2007,7 @@ func (s *firstBoot16Suite) TestCriticalTaskEdgesForPreseed(c *C) {
 }
 
 func (s *firstBoot16Suite) TestCriticalTaskEdgesForPreseedMissing(c *C) {
+	s.startOverlord(c)
 	st := s.overlord.State()
 	st.Lock()
 	defer st.Unlock()
@@ -1887,3 +2034,139 @@ func (s *firstBoot16Suite) TestCriticalTaskEdgesForPreseedMissing(c *C) {
 	_, _, _, err = devicestate.CriticalTaskEdges(ts)
 	c.Assert(err, NotNil)
 }
+
+func (s *firstBoot16Suite) TestPopulateFromSeedWithConnectHook(c *C) {
+	restore := release.MockOnClassic(true)
+	defer restore()
+
+	hooksCalled := []*hookstate.Context{}
+	restore = hookstate.MockRunHook(func(ctx *hookstate.Context, tomb *tomb.Tomb) ([]byte, error) {
+		ctx.Lock()
+		defer ctx.Unlock()
+
+		hooksCalled = append(hooksCalled, ctx)
+		return nil, nil
+	})
+	defer restore()
+
+	core18Fname, snapdFname, _, _ := s.makeCore18Snaps(c, &core18SnapsOpts{
+		classic: true,
+	})
+
+	snapFilesWithHook := [][]string{
+		{"bin/bar", ``},
+		{"meta/hooks/connect-plug-network", ``},
+	}
+
+	// put a firstboot snap into the SnapBlobDir
+	snapYaml := `name: foo
+base: core18
+version: 1.0
+plugs:
+ shared-data-plug:
+  interface: content
+  target: import
+  content: mylib
+apps:
+ bar:
+  command: bin/bar
+  plugs: [network]
+`
+	fooFname, fooDecl, fooRev := s.MakeAssertedSnap(c, snapYaml, snapFilesWithHook, snap.R(128), "developerid")
+	s.WriteAssertions("foo.asserts", s.devAcct, fooRev, fooDecl)
+
+	// put a 2nd firstboot snap into the SnapBlobDir
+	snapYaml = `name: bar
+base: core18
+version: 1.0
+slots:
+ shared-data-slot:
+  interface: content
+  content: mylib
+  read:
+   - /
+apps:
+ bar:
+  command: bin/bar
+`
+	snapFiles := [][]string{
+		{"bin/bar", ``},
+	}
+	barFname, barDecl, barRev := s.MakeAssertedSnap(c, snapYaml, snapFiles, snap.R(65), "developerid")
+	s.WriteAssertions("bar.asserts", s.devAcct, barDecl, barRev)
+
+	// add a model assertion and its chain
+	assertsChain := s.makeModelAssertionChain(c, "my-model-classic", nil)
+	s.WriteAssertions("model.asserts", assertsChain...)
+
+	// create a seed.yaml
+	content := []byte(fmt.Sprintf(`
+snaps:
+ - name: snapd
+   file: %s
+ - name: core18
+   file: %s
+ - name: foo
+   file: %s
+ - name: bar
+   file: %s
+`, snapdFname, core18Fname, fooFname, barFname))
+	err := ioutil.WriteFile(filepath.Join(dirs.SnapSeedDir, "seed.yaml"), content, 0644)
+	c.Assert(err, IsNil)
+
+	// run the firstboot stuff
+	s.startOverlord(c)
+	st := s.overlord.State()
+	st.Lock()
+	defer st.Unlock()
+
+	tsAll, err := devicestate.PopulateStateFromSeedImpl(st, nil, s.perfTimings)
+	c.Assert(err, IsNil)
+	// use the expected kind otherwise settle with start another one
+	chg := st.NewChange("seed", "run the populate from seed changes")
+	for _, ts := range tsAll {
+		chg.AddAll(ts)
+	}
+	c.Assert(st.Changes(), HasLen, 1)
+
+	checkOrder(c, tsAll, "snapd", "core18", "foo", "bar")
+
+	st.Unlock()
+	err = s.overlord.Settle(settleTimeout)
+	st.Lock()
+	c.Check(err, IsNil)
+	c.Check(chg.Err(), IsNil)
+
+	// at this point the system is "restarting", pretend the restart has
+	// happened
+	c.Assert(chg.Status(), Equals, state.DoingStatus)
+	restart.MockPending(st, restart.RestartUnset)
+	st.Unlock()
+	err = s.overlord.Settle(settleTimeout)
+	st.Lock()
+	c.Assert(err, IsNil)
+	c.Assert(chg.Status(), Equals, state.DoneStatus, Commentf("%s", chg.Err()))
+
+	c.Assert(hooksCalled, HasLen, 1)
+	c.Check(hooksCalled[0].HookName(), Equals, "connect-plug-network")
+
+	// verify that connections was made
+	var conns map[string]interface{}
+	c.Assert(st.Get("conns", &conns), IsNil)
+	c.Assert(conns, DeepEquals, map[string]interface{}{
+		"foo:network snapd:network": map[string]interface{}{
+			"auto": true, "interface": "network"},
+		"foo:shared-data-plug bar:shared-data-slot": map[string]interface{}{
+			"auto": true, "interface": "content",
+			"plug-static": map[string]interface{}{
+				"content": "mylib", "target": "import",
+			},
+			"slot-static": map[string]interface{}{
+				"content": "mylib",
+				"read": []interface{}{
+					"/",
+				},
+			},
+		},
+	})
+}
