@@ -20,7 +20,6 @@
 package ctlcmd
 
 import (
-	"errors"
 	"fmt"
 	"sort"
 	"text/tabwriter"
@@ -50,8 +49,6 @@ type servicesCommand struct {
 	} `positional-args:"yes"`
 }
 
-var errNoContextForServices = errors.New(i18n.G("cannot query services without a context"))
-
 type byApp []*snap.AppInfo
 
 func (a byApp) Len() int      { return len(a) }
@@ -61,9 +58,9 @@ func (a byApp) Less(i, j int) bool {
 }
 
 func (c *servicesCommand) Execute([]string) error {
-	context := c.context()
-	if context == nil {
-		return errNoContextForServices
+	context, err := c.ensureContext()
+	if err != nil {
+		return err
 	}
 
 	st := context.State()
@@ -91,7 +88,9 @@ func (c *servicesCommand) Execute([]string) error {
 			startup = i18n.G("enabled")
 		}
 		current := i18n.G("inactive")
-		if svc.Active {
+		if svc.DaemonScope == snap.UserDaemon {
+			current = "-"
+		} else if svc.Active {
 			current = i18n.G("active")
 		}
 		fmt.Fprintf(w, "%s.%s\t%s\t%s\t%s\n", svc.Snap, svc.Name, startup, current, clientutil.ClientAppInfoNotes(&svc))
