@@ -20,9 +20,9 @@
 package hookstate_test
 
 import (
+	"context"
 	"encoding/json"
 	"fmt"
-	"os"
 	"path/filepath"
 	"regexp"
 	"sync/atomic"
@@ -30,12 +30,14 @@ import (
 	"time"
 
 	. "gopkg.in/check.v1"
+	"gopkg.in/tomb.v2"
 
 	"github.com/snapcore/snapd/dirs"
 	"github.com/snapcore/snapd/overlord"
 	"github.com/snapcore/snapd/overlord/configstate/config"
 	"github.com/snapcore/snapd/overlord/hookstate"
 	"github.com/snapcore/snapd/overlord/hookstate/hooktest"
+	"github.com/snapcore/snapd/overlord/restart"
 	"github.com/snapcore/snapd/overlord/snapstate"
 	"github.com/snapcore/snapd/overlord/state"
 	"github.com/snapcore/snapd/snap"
@@ -59,6 +61,10 @@ type baseHookManagerSuite struct {
 	command     *testutil.MockCmd
 }
 
+var (
+	settleTimeout = testutil.HostScaledTimeout(15 * time.Second)
+)
+
 func (s *baseHookManagerSuite) commonSetUpTest(c *C) {
 	s.BaseTest.SetUpTest(c)
 
@@ -69,12 +75,16 @@ func (s *baseHookManagerSuite) commonSetUpTest(c *C) {
 	dirs.SetRootDir(c.MkDir())
 	s.o = overlord.Mock()
 	s.state = s.o.State()
+	s.state.Lock()
+	restart.Init(s.state, "boot-id-0", nil)
+	s.state.Unlock()
 	manager, err := hookstate.Manager(s.state, s.o.TaskRunner())
 	c.Assert(err, IsNil)
 	s.manager = manager
 	s.se = s.o.StateEngine()
 	s.o.AddManager(s.manager)
 	s.o.AddManager(s.o.TaskRunner())
+	c.Assert(s.o.StartUp(), IsNil)
 
 	s.AddCleanup(snap.MockSanitizePlugsSlots(func(snapInfo *snap.Info) {}))
 
@@ -172,7 +182,7 @@ func (s *hookManagerSuite) TearDownTest(c *C) {
 }
 
 func (s *hookManagerSuite) settle(c *C) {
-	err := s.o.Settle(5 * time.Second)
+	err := s.o.Settle(settleTimeout)
 	c.Assert(err, IsNil)
 }
 
@@ -263,7 +273,9 @@ func (s *hookManagerSuite) TestHookTaskEnsure(c *C) {
 
 func (s *hookManagerSuite) TestHookTaskEnsureRestarting(c *C) {
 	// we do no start new hooks runs if we are restarting
-	s.state.RequestRestart(state.RestartDaemon)
+	s.state.Lock()
+	restart.MockPending(s.state, restart.RestartDaemon)
+	s.state.Unlock()
 
 	s.se.Ensure()
 	s.se.Wait()
@@ -434,6 +446,33 @@ func (s *hookManagerSuite) TestHookTaskHandleIgnoreErrorWorks(c *C) {
 	checkTaskLogContains(c, s.task, ".*ignoring failure in hook.*")
 }
 
+func (s *hookManagerSuite) TestHookTaskHandlesHookErrorAndIgnoresIt(c *C) {
+	// tell the mock handler to return 'true' from its Error() handler,
+	// indicating to the hookmgr to ignore the original hook error.
+	s.mockHandler.IgnoreOriginalErr = true
+
+	// Simulate hook error
+	cmd := testutil.MockCommand(
+		c, "snap", ">&2 echo 'hook failed at user request'; exit 1")
+	defer cmd.Restore()
+
+	s.se.Ensure()
+	s.se.Wait()
+
+	s.state.Lock()
+	defer s.state.Unlock()
+
+	c.Check(s.mockHandler.BeforeCalled, Equals, true)
+	c.Check(s.mockHandler.DoneCalled, Equals, false)
+	c.Check(s.mockHandler.ErrorCalled, Equals, true)
+
+	c.Check(s.task.Kind(), Equals, "run-hook")
+	c.Check(s.task.Status(), Equals, state.DoneStatus)
+	c.Check(s.change.Status(), Equals, state.DoneStatus)
+
+	c.Check(s.manager.NumRunningHooks(), Equals, 0)
+}
+
 func (s *hookManagerSuite) TestHookTaskEnforcesTimeout(c *C) {
 	var hooksup hookstate.HookSetup
 
@@ -775,6 +814,40 @@ func (s *hookManagerSuite) TestHookWithoutHookOptional(c *C) {
 	s.se.Ensure()
 	s.se.Wait()
 
+	c.Check(s.mockHandler.BeforeCalled, Equals, false)
+	c.Check(s.mockHandler.DoneCalled, Equals, false)
+	c.Check(s.mockHandler.ErrorCalled, Equals, false)
+
+	c.Check(s.command.Calls(), IsNil)
+
+	s.state.Lock()
+	defer s.state.Unlock()
+
+	c.Check(s.task.Kind(), Equals, "run-hook")
+	c.Check(s.task.Status(), Equals, state.DoneStatus)
+	c.Check(s.change.Status(), Equals, state.DoneStatus)
+
+	c.Logf("Task log:\n%s\n", s.task.Log())
+}
+
+func (s *hookManagerSuite) TestHookWithoutHookAlways(c *C) {
+	s.manager.Register(regexp.MustCompile("missing-hook"), func(context *hookstate.Context) hookstate.Handler {
+		return s.mockHandler
+	})
+
+	hooksup := &hookstate.HookSetup{
+		Snap:     "test-snap",
+		Hook:     "missing-hook",
+		Optional: true,
+		Always:   true,
+	}
+	s.state.Lock()
+	s.task.Set("hook-setup", hooksup)
+	s.state.Unlock()
+
+	s.se.Ensure()
+	s.se.Wait()
+
 	c.Check(s.mockHandler.BeforeCalled, Equals, true)
 	c.Check(s.mockHandler.DoneCalled, Equals, true)
 	c.Check(s.mockHandler.ErrorCalled, Equals, false)
@@ -830,8 +903,6 @@ func checkTaskLogContains(c *C, task *state.Task, pattern string) {
 
 func (s *hookManagerSuite) TestHookTaskRunsRightSnapCmd(c *C) {
 	coreSnapCmdPath := filepath.Join(dirs.SnapMountDir, "core/12/usr/bin/snap")
-	err := os.MkdirAll(filepath.Dir(coreSnapCmdPath), 0755)
-	c.Assert(err, IsNil)
 	cmd := testutil.MockCommand(c, coreSnapCmdPath, "")
 	defer cmd.Restore()
 
@@ -982,8 +1053,8 @@ func (h *MockConcurrentHandler) Done() error {
 	return nil
 }
 
-func (h *MockConcurrentHandler) Error(err error) error {
-	return nil
+func (h *MockConcurrentHandler) Error(err error) (bool, error) {
+	return false, nil
 }
 
 func NewMockConcurrentHandler(onDone func()) *MockConcurrentHandler {
@@ -1141,6 +1212,110 @@ func (s *hookManagerSuite) TestHookHijackingNoConflict(c *C) {
 	c.Assert(err, IsNil)
 }
 
+func (s *hookManagerSuite) TestEphemeralRunHook(c *C) {
+	contextData := map[string]interface{}{
+		"key":  "value",
+		"key2": "value2",
+	}
+	s.testEphemeralRunHook(c, contextData)
+}
+
+func (s *hookManagerSuite) TestEphemeralRunHookNoContextData(c *C) {
+	var contextData map[string]interface{} = nil
+	s.testEphemeralRunHook(c, contextData)
+}
+
+func (s *hookManagerSuite) testEphemeralRunHook(c *C, contextData map[string]interface{}) {
+	var hookInvokeCalled []string
+	hookInvoke := func(ctx *hookstate.Context, tomb *tomb.Tomb) ([]byte, error) {
+		c.Check(ctx.HookName(), Equals, "configure")
+		hookInvokeCalled = append(hookInvokeCalled, ctx.HookName())
+
+		// check that context data was set correctly
+		var s string
+		ctx.Lock()
+		defer ctx.Unlock()
+		for k, v := range contextData {
+			ctx.Get(k, &s)
+			c.Check(s, Equals, v)
+		}
+		ctx.Set("key-set-from-hook", "value-set-from-hook")
+
+		return []byte("some output"), nil
+	}
+	restore := hookstate.MockRunHook(hookInvoke)
+	defer restore()
+
+	hooksup := &hookstate.HookSetup{
+		Snap:     "test-snap",
+		Revision: snap.R(1),
+		Hook:     "configure",
+	}
+	context, err := s.manager.EphemeralRunHook(context.Background(), hooksup, contextData)
+	c.Assert(err, IsNil)
+	c.Check(hookInvokeCalled, DeepEquals, []string{"configure"})
+
+	var value string
+	context.Lock()
+	context.Get("key-set-from-hook", &value)
+	context.Unlock()
+	c.Check(value, Equals, "value-set-from-hook")
+}
+
+func (s *hookManagerSuite) TestEphemeralRunHookNoSnap(c *C) {
+	hookInvoke := func(ctx *hookstate.Context, tomb *tomb.Tomb) ([]byte, error) {
+		c.Fatalf("hook should not be invoked in this test")
+		return nil, nil
+	}
+	restore := hookstate.MockRunHook(hookInvoke)
+	defer restore()
+
+	hooksup := &hookstate.HookSetup{
+		Snap:     "not-installed-snap",
+		Revision: snap.R(1),
+		Hook:     "configure",
+	}
+	contextData := map[string]interface{}{
+		"key": "value",
+	}
+	_, err := s.manager.EphemeralRunHook(context.Background(), hooksup, contextData)
+	c.Assert(err, ErrorMatches, `cannot run ephemeral hook "configure" for snap "not-installed-snap": no state entry for key`)
+}
+
+func (s *hookManagerSuite) TestEphemeralRunHookContextCanCancel(c *C) {
+	tombDying := 0
+	hookRunning := make(chan struct{})
+
+	hookInvoke := func(_ *hookstate.Context, tomb *tomb.Tomb) ([]byte, error) {
+		close(hookRunning)
+
+		select {
+		case <-tomb.Dying():
+			tombDying++
+		case <-time.After(10 * time.Second):
+			c.Fatalf("hook not canceled after 10s")
+		}
+		return nil, nil
+	}
+	restore := hookstate.MockRunHook(hookInvoke)
+	defer restore()
+
+	hooksup := &hookstate.HookSetup{
+		Snap:     "test-snap",
+		Revision: snap.R(1),
+		Hook:     "configure",
+	}
+
+	ctx, cancelFunc := context.WithCancel(context.Background())
+	go func() {
+		<-hookRunning
+		cancelFunc()
+	}()
+	_, err := s.manager.EphemeralRunHook(ctx, hooksup, nil)
+	c.Assert(err, IsNil)
+	c.Check(tombDying, Equals, 1)
+}
+
 type parallelInstancesHookManagerSuite struct {
 	baseHookManagerSuite
 }
