@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2016-2020 Canonical Ltd
+ * Copyright (C) 2016-2022 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -22,6 +22,7 @@ package ifacestate
 import (
 	"bytes"
 	"encoding/json"
+	"errors"
 	"fmt"
 	"os"
 	"sort"
@@ -37,12 +38,21 @@ import (
 	"github.com/snapcore/snapd/jsonutil"
 	"github.com/snapcore/snapd/logger"
 	"github.com/snapcore/snapd/overlord/assertstate"
+	"github.com/snapcore/snapd/overlord/ifacestate/schema"
 	"github.com/snapcore/snapd/overlord/snapstate"
 	"github.com/snapcore/snapd/overlord/state"
 	"github.com/snapcore/snapd/snap"
+	"github.com/snapcore/snapd/systemd"
 	"github.com/snapcore/snapd/timings"
 )
 
+var (
+	snapdAppArmorServiceIsDisabled = snapdAppArmorServiceIsDisabledImpl
+	profilesNeedRegeneration       = profilesNeedRegenerationImpl
+
+	writeSystemKey = interfaces.WriteSystemKey
+)
+
 func (m *InterfaceManager) selectInterfaceMapper(snaps []*snap.Info) {
 	for _, snapInfo := range snaps {
 		if snapInfo.Type() == snap.TypeSnapd {
@@ -67,7 +77,39 @@ func (m *InterfaceManager) addInterfaces(extra []interfaces.Interface) error {
 }
 
 func (m *InterfaceManager) addBackends(extra []interfaces.SecurityBackend) error {
-	opts := interfaces.SecurityBackendOptions{Preseed: m.preseed}
+	// get the snapd snap info if it is installed
+	var snapdSnap snapstate.SnapState
+	var snapdSnapInfo *snap.Info
+	err := snapstate.Get(m.state, "snapd", &snapdSnap)
+	if err != nil && !errors.Is(err, state.ErrNoState) {
+		return fmt.Errorf("cannot access snapd snap state: %v", err)
+	}
+	if err == nil {
+		snapdSnapInfo, err = snapdSnap.CurrentInfo()
+		if err != nil && err != snapstate.ErrNoCurrent {
+			return fmt.Errorf("cannot access snapd snap info: %v", err)
+		}
+	}
+
+	// get the core snap info if it is installed
+	var coreSnap snapstate.SnapState
+	var coreSnapInfo *snap.Info
+	err = snapstate.Get(m.state, "core", &coreSnap)
+	if err != nil && !errors.Is(err, state.ErrNoState) {
+		return fmt.Errorf("cannot access core snap state: %v", err)
+	}
+	if err == nil {
+		coreSnapInfo, err = coreSnap.CurrentInfo()
+		if err != nil && err != snapstate.ErrNoCurrent {
+			return fmt.Errorf("cannot access core snap info: %v", err)
+		}
+	}
+
+	opts := interfaces.SecurityBackendOptions{
+		Preseed:       m.preseed,
+		CoreSnapInfo:  coreSnapInfo,
+		SnapdSnapInfo: snapdSnapInfo,
+	}
 	for _, backend := range backends.All {
 		if err := backend.Initialize(&opts); err != nil {
 			return err
@@ -108,8 +150,13 @@ func profilesNeedRegenerationImpl() bool {
 	return mismatch
 }
 
-var profilesNeedRegeneration = profilesNeedRegenerationImpl
-var writeSystemKey = interfaces.WriteSystemKey
+// snapdAppArmorServiceIsDisabledImpl returns true if the snapd.apparmor
+// service unit exists but is disabled
+func snapdAppArmorServiceIsDisabledImpl() bool {
+	sysd := systemd.New(systemd.SystemMode, nil)
+	isEnabled, err := sysd.IsEnabled("snapd.apparmor")
+	return err == nil && !isEnabled
+}
 
 // regenerateAllSecurityProfiles will regenerate all security profiles.
 func (m *InterfaceManager) regenerateAllSecurityProfiles(tm timings.Measurer) error {
@@ -144,7 +191,11 @@ func (m *InterfaceManager) regenerateAllSecurityProfiles(tm timings.Measurer) er
 		if err := snapstate.Get(m.state, snapName, &snapst); err != nil {
 			logger.Noticef("cannot get state of snap %q: %s", snapName, err)
 		}
-		return confinementOptions(snapst.Flags)
+		opts, err := buildConfinementOptions(m.state, snapst.InstanceName(), snapst.Flags)
+		if err != nil {
+			logger.Noticef("cannot get confinement options for snap %q: %s", snapName, err)
+		}
+		return opts
 	}
 
 	// For each backend:
@@ -214,14 +265,14 @@ var removeStaleConnections = func(st *state.State) error {
 		}
 		var snapst snapstate.SnapState
 		if err := snapstate.Get(st, connRef.PlugRef.Snap, &snapst); err != nil {
-			if err != state.ErrNoState {
+			if !errors.Is(err, state.ErrNoState) {
 				return err
 			}
 			staleConns = append(staleConns, id)
 			continue
 		}
 		if err := snapstate.Get(st, connRef.SlotRef.Snap, &snapst); err != nil {
-			if err != state.ErrNoState {
+			if !errors.Is(err, state.ErrNoState) {
 				return err
 			}
 			staleConns = append(staleConns, id)
@@ -238,6 +289,22 @@ var removeStaleConnections = func(st *state.State) error {
 	return nil
 }
 
+func isBroken(st *state.State, snapName string) (bool, error) {
+	var snapst snapstate.SnapState
+	err := snapstate.Get(st, snapName, &snapst)
+	if errors.Is(err, state.ErrNoState) {
+		return false, nil
+	}
+	if err != nil {
+		return false, err
+	}
+	snapInfo, _ := snapst.CurrentInfo()
+	if snapInfo != nil && snapInfo.Broken != "" {
+		return true, nil
+	}
+	return false, nil
+}
+
 // reloadConnections reloads connections stored in the state in the repository.
 // Using non-empty snapName the operation can be scoped to connections
 // affecting a given snap.
@@ -251,6 +318,7 @@ func (m *InterfaceManager) reloadConnections(snapName string) ([]string, error)
 
 	connStateChanged := false
 	affected := make(map[string]bool)
+ConnsLoop:
 	for connId, connState := range conns {
 		// Skip entries that just mark a connection as undesired. Those don't
 		// carry attributes that can go stale. In the same spirit, skip
@@ -280,6 +348,18 @@ func (m *InterfaceManager) reloadConnections(snapName string) ([]string, error)
 			// as long as it wasn't disconnected manually; note that undesired flag is taken care of at
 			// the beginning of the loop.
 			if connState.Auto && !connState.ByGadget && connState.Interface != "core-support" {
+				// only do anything about this connection if snap isn't in a broken state, otherwise
+				// leave the connection untouched.
+				for _, snapName := range []string{connRef.PlugRef.Snap, connRef.SlotRef.Snap} {
+					broken, err := isBroken(m.state, snapName)
+					if err != nil {
+						return nil, err
+					}
+					if broken {
+						logger.Noticef("Snap %q is broken, ignored by reloadConnections", snapName)
+						continue ConnsLoop
+					}
+				}
 				delete(conns, connId)
 				connStateChanged = true
 			}
@@ -291,10 +371,12 @@ func (m *InterfaceManager) reloadConnections(snapName string) ([]string, error)
 		staticPlugAttrs := connState.StaticPlugAttrs
 		staticSlotAttrs := connState.StaticSlotAttrs
 
-		// XXX: Refresh the copy of the static connection attributes for "content" interface as long
-		// as its "content" attribute
+		// XXX: Refresh the copy of the static connection attributes for "content"
+		// and "system-files" interfaces.
 		// This is a partial and temporary solution to https://bugs.launchpad.net/snapd/+bug/1825883
-		if plugInfo.Interface == "content" {
+		// and https://bugs.launchpad.net/snapd/+bug/1942266.
+		switch plugInfo.Interface {
+		case "content":
 			var plugContent, slotContent string
 			plugInfo.Attr("content", &plugContent)
 			slotInfo.Attr("content", &slotContent)
@@ -306,6 +388,10 @@ func (m *InterfaceManager) reloadConnections(snapName string) ([]string, error)
 			} else {
 				logger.Noticef("cannot refresh static attributes of the connection %q", connId)
 			}
+		case "system-files":
+			staticPlugAttrs = utils.NormalizeInterfaceAttributes(plugInfo.Attrs).(map[string]interface{})
+			staticSlotAttrs = utils.NormalizeInterfaceAttributes(slotInfo.Attrs).(map[string]interface{})
+			updateStaticAttrs = true
 		}
 
 		// Note: reloaded connections are not checked against policy again, and also we don't call BeforeConnect* methods on them.
@@ -436,26 +522,6 @@ func addHotplugSlot(st *state.State, repo *interfaces.Repository, stateSlots map
 	return nil
 }
 
-type connState struct {
-	Auto      bool   `json:"auto,omitempty"`
-	ByGadget  bool   `json:"by-gadget,omitempty"`
-	Interface string `json:"interface,omitempty"`
-	// Undesired tracks connections that were manually disconnected after being auto-connected,
-	// so that they are not automatically reconnected again in the future.
-	Undesired        bool                   `json:"undesired,omitempty"`
-	StaticPlugAttrs  map[string]interface{} `json:"plug-static,omitempty"`
-	DynamicPlugAttrs map[string]interface{} `json:"plug-dynamic,omitempty"`
-	StaticSlotAttrs  map[string]interface{} `json:"slot-static,omitempty"`
-	DynamicSlotAttrs map[string]interface{} `json:"slot-dynamic,omitempty"`
-	// Hotplug-related attributes: HotplugGone indicates a connection that
-	// disappeared because the device was removed, but may potentially be
-	// restored in the future if we see the device again. HotplugKey is the
-	// key of the associated device; it's empty for connections of regular
-	// slots.
-	HotplugGone bool            `json:"hotplug-gone,omitempty"`
-	HotplugKey  snap.HotplugKey `json:"hotplug-key,omitempty"`
-}
-
 type gadgetConnect struct {
 	st   *state.State
 	task *state.Task
@@ -479,10 +545,10 @@ func newGadgetConnect(s *state.State, task *state.Task, repo *interfaces.Reposit
 // addGadgetConnections adds to newconns any applicable connections
 // from the gadget connections stanza.
 // conflictError is called to handle checkAutoconnectConflicts errors.
-func (gc *gadgetConnect) addGadgetConnections(newconns map[string]*interfaces.ConnRef, conns map[string]*connState, conflictError func(*state.Retry, error) error) error {
+func (gc *gadgetConnect) addGadgetConnections(newconns map[string]*interfaces.ConnRef, conns map[string]*schema.ConnState, conflictError func(*state.Retry, error) error) error {
 	var seeded bool
 	err := gc.st.Get("seeded", &seeded)
-	if err != nil && err != state.ErrNoState {
+	if err != nil && !errors.Is(err, state.ErrNoState) {
 		return err
 	}
 	// we apply gadget connections only during seeding or a remodeling
@@ -510,7 +576,7 @@ func (gc *gadgetConnect) addGadgetConnections(newconns map[string]*interfaces.Co
 
 	gconns, err := snapstate.GadgetConnections(gc.st, gc.deviceCtx)
 	if err != nil {
-		if err == state.ErrNoState {
+		if errors.Is(err, state.ErrNoState) {
 			// no gadget yet, nothing to do
 			return nil
 		}
@@ -566,7 +632,7 @@ func (gc *gadgetConnect) addGadgetConnections(newconns map[string]*interfaces.Co
 	return nil
 }
 
-func addNewConnection(st *state.State, task *state.Task, newconns map[string]*interfaces.ConnRef, conns map[string]*connState, plug *snap.PlugInfo, slot *snap.SlotInfo, conflictError func(*state.Retry, error) error) error {
+func addNewConnection(st *state.State, task *state.Task, newconns map[string]*interfaces.ConnRef, conns map[string]*schema.ConnState, plug *snap.PlugInfo, slot *snap.SlotInfo, conflictError func(*state.Retry, error) error) error {
 	connRef := interfaces.NewConnRef(plug, slot)
 	key := connRef.ID()
 	if _, ok := conns[key]; ok {
@@ -600,6 +666,10 @@ func addNewConnection(st *state.State, task *state.Task, newconns map[string]*in
 	return nil
 }
 
+// DebugAutoConnectCheck is a hook that can be set to debug auto-connection
+// candidates as they are checked.
+var DebugAutoConnectCheck func(*policy.ConnectCandidate, interfaces.SideArity, error)
+
 type autoConnectChecker struct {
 	st   *state.State
 	task *state.Task
@@ -682,6 +752,9 @@ func (c *autoConnectChecker) check(plug *interfaces.ConnectedPlug, slot *interfa
 	}
 
 	arity, err := ic.CheckAutoConnect()
+	if DebugAutoConnectCheck != nil {
+		DebugAutoConnectCheck(&ic, arity, err)
+	}
 	if err == nil {
 		return true, arity, nil
 	}
@@ -728,7 +801,7 @@ func filterUbuntuCoreSlots(candidates []*snap.SlotInfo, arities []interfaces.Sid
 // conns. cannotAutoConnectLog is called to build a log message in
 // case no applicable pair was found. conflictError is called
 // to handle checkAutoconnectConflicts errors.
-func (c *autoConnectChecker) addAutoConnections(newconns map[string]*interfaces.ConnRef, plugs []*snap.PlugInfo, filter func([]*snap.SlotInfo) []*snap.SlotInfo, conns map[string]*connState, cannotAutoConnectLog func(plug *snap.PlugInfo, candRefs []string) string, conflictError func(*state.Retry, error) error) error {
+func (c *autoConnectChecker) addAutoConnections(newconns map[string]*interfaces.ConnRef, plugs []*snap.PlugInfo, filter func([]*snap.SlotInfo) []*snap.SlotInfo, conns map[string]*schema.ConnState, cannotAutoConnectLog func(plug *snap.PlugInfo, candRefs []string) string, conflictError func(*state.Retry, error) error) error {
 	for _, plug := range plugs {
 		candSlots, arities := c.repo.AutoConnectCandidateSlots(plug.Snap.InstanceName(), plug.Name, c.check)
 
@@ -863,10 +936,10 @@ func getPlugAndSlotRefs(task *state.Task) (interfaces.PlugRef, interfaces.SlotRe
 // getConns returns information about connections from the state.
 //
 // Connections are transparently re-mapped according to remapIncomingConnRef
-func getConns(st *state.State) (conns map[string]*connState, err error) {
+func getConns(st *state.State) (conns map[string]*schema.ConnState, err error) {
 	var raw *json.RawMessage
 	err = st.Get("conns", &raw)
-	if err != nil && err != state.ErrNoState {
+	if err != nil && !errors.Is(err, state.ErrNoState) {
 		return nil, fmt.Errorf("cannot obtain raw data about existing connections: %s", err)
 	}
 	if raw != nil {
@@ -876,9 +949,9 @@ func getConns(st *state.State) (conns map[string]*connState, err error) {
 		}
 	}
 	if conns == nil {
-		conns = make(map[string]*connState)
+		conns = make(map[string]*schema.ConnState)
 	}
-	remapped := make(map[string]*connState, len(conns))
+	remapped := make(map[string]*schema.ConnState, len(conns))
 	for id, cstate := range conns {
 		cref, err := interfaces.ParseConnRef(id)
 		if err != nil {
@@ -898,8 +971,8 @@ func getConns(st *state.State) (conns map[string]*connState, err error) {
 // setConns sets information about connections in the state.
 //
 // Connections are transparently re-mapped according to remapOutgoingConnRef
-func setConns(st *state.State, conns map[string]*connState) {
-	remapped := make(map[string]*connState, len(conns))
+func setConns(st *state.State, conns map[string]*schema.ConnState) {
+	remapped := make(map[string]*schema.ConnState, len(conns))
 	for id, cstate := range conns {
 		cref, err := interfaces.ParseConnRef(id)
 		if err != nil {
@@ -914,17 +987,46 @@ func setConns(st *state.State, conns map[string]*connState) {
 }
 
 // snapsWithSecurityProfiles returns all snaps that have active
-// security profiles: these are either snaps that are active, or about
-// to be active (pending link-snap) with a done setup-profiles
+// security profiles: these are either snaps that are active,
+// inactive snaps that are being operated on, whose profile state
+// is tracked with SnapState.PendingSecurity,
+// or snap about to be active (pending link-snap) with a done
+// setup-profiles
 func snapsWithSecurityProfiles(st *state.State) ([]*snap.Info, error) {
-	infos, err := snapstate.ActiveInfos(st)
+	all, err := snapstate.All(st)
 	if err != nil {
 		return nil, err
 	}
-	seen := make(map[string]bool, len(infos))
-	for _, info := range infos {
-		seen[info.InstanceName()] = true
+	infos := make([]*snap.Info, 0, len(all))
+	seen := make(map[string]bool, len(all))
+	for instanceName, snapst := range all {
+		if snapst.Active {
+			snapInfo, err := snapst.CurrentInfo()
+			if err != nil {
+				logger.Noticef("cannot retrieve info for snap %q: %s", instanceName, err)
+				continue
+			}
+			infos = append(infos, snapInfo)
+			seen[instanceName] = true
+		} else if snapst.PendingSecurity != nil {
+			// we tracked any pending security profiles for the snap
+			seen[instanceName] = true
+			si := snapst.PendingSecurity.SideInfo
+			if si == nil {
+				// profiles removed (already)
+				continue
+			}
+			snapInfo, err := snap.ReadInfo(instanceName, si)
+			if err != nil {
+				logger.Noticef("cannot retrieve info for snap %q: %s", instanceName, err)
+				continue
+			}
+			infos = append(infos, snapInfo)
+		}
 	}
+	// look at the changes for old snapds and also
+	// the situation that are being installed, so they do not
+	// have SnapState yet
 	for _, t := range st.Tasks() {
 		if t.Kind() != "link-snap" || t.Status().Ready() {
 			continue
@@ -1160,7 +1262,7 @@ func getHotplugAttrs(task *state.Task) (ifaceName string, hotplugKey snap.Hotplu
 
 func allocHotplugSeq(st *state.State) (int, error) {
 	var seq int
-	if err := st.Get("hotplug-seq", &seq); err != nil && err != state.ErrNoState {
+	if err := st.Get("hotplug-seq", &seq); err != nil && !errors.Is(err, state.ErrNoState) {
 		return 0, fmt.Errorf("internal error: cannot allocate hotplug sequence number: %s", err)
 	}
 	seq++
@@ -1212,7 +1314,7 @@ func getHotplugSlots(st *state.State) (map[string]*HotplugSlotInfo, error) {
 	var slots map[string]*HotplugSlotInfo
 	err := st.Get("hotplug-slots", &slots)
 	if err != nil {
-		if err != state.ErrNoState {
+		if !errors.Is(err, state.ErrNoState) {
 			return nil, err
 		}
 		slots = make(map[string]*HotplugSlotInfo)
@@ -1233,7 +1335,7 @@ func findHotplugSlot(stateSlots map[string]*HotplugSlotInfo, ifaceName string, h
 	return nil
 }
 
-func findConnsForHotplugKey(conns map[string]*connState, ifaceName string, hotplugKey snap.HotplugKey) []string {
+func findConnsForHotplugKey(conns map[string]*schema.ConnState, ifaceName string, hotplugKey snap.HotplugKey) []string {
 	var connsForDevice []string
 	for id, connSt := range conns {
 		if connSt.Interface != ifaceName || connSt.HotplugKey != hotplugKey {
@@ -1244,3 +1346,16 @@ func findConnsForHotplugKey(conns map[string]*connState, ifaceName string, hotpl
 	sort.Strings(connsForDevice)
 	return connsForDevice
 }
+
+func (m *InterfaceManager) discardSecurityProfilesLate(name string, rev snap.Revision, typ snap.Type) error {
+	for _, backend := range m.repo.Backends() {
+		lateDiscardBackend, ok := backend.(interfaces.SecurityBackendDiscardingLate)
+		if !ok {
+			continue
+		}
+		if err := lateDiscardBackend.RemoveLate(name, rev, typ); err != nil {
+			return err
+		}
+	}
+	return nil
+}
