@@ -53,13 +53,14 @@ func (snapshotSuite) TestManager(c *check.C) {
 	sort.Strings(kinds)
 	c.Check(kinds, check.DeepEquals, []string{
 		"check-snapshot",
+		"cleanup-after-restore",
 		"forget-snapshot",
 		"restore-snapshot",
 		"save-snapshot",
 	})
 }
 
-func mockDummySnapshot(c *check.C) (restore func()) {
+func mockFakeSnapshot(c *check.C) (restore func()) {
 	shotfile, err := os.Create(filepath.Join(c.MkDir(), "foo.zip"))
 	c.Assert(err, check.IsNil)
 
@@ -87,7 +88,7 @@ func (snapshotSuite) TestEnsureForgetsSnapshots(c *check.C) {
 	})
 	defer restoreOsRemove()
 
-	restore := mockDummySnapshot(c)
+	restore := mockFakeSnapshot(c)
 	defer restore()
 
 	st := state.New(nil)
@@ -175,7 +176,7 @@ func (snapshotSuite) testEnsureForgetSnapshotsConflict(c *check.C, snapshotOp st
 	})
 	defer restoreOsRemove()
 
-	restore := mockDummySnapshot(c)
+	restore := mockFakeSnapshot(c)
 	defer restore()
 
 	st := state.New(nil)
@@ -215,7 +216,7 @@ func (snapshotSuite) testEnsureForgetSnapshotsConflict(c *check.C, snapshotOp st
 	c.Check(removeCalled, check.Equals, 0)
 
 	if tsk != nil {
-		// sanity check of the test setup: snapshot gets removed once conflict goes away
+		// validity check of the test setup: snapshot gets removed once conflict goes away
 		tsk.SetStatus(state.DoneStatus)
 	} else {
 		c.Check(snapshotstate.UnsetSnapshotOpInProgress(st, 1), check.Equals, snapshotOp)
@@ -278,7 +279,7 @@ func (snapshotSuite) TestDoSave(c *check.C) {
 		buf := json.RawMessage(`{"hello": "there"}`)
 		return &buf, nil
 	})()
-	defer snapshotstate.MockBackendSave(func(_ context.Context, id uint64, si *snap.Info, cfg map[string]interface{}, usernames []string) (*client.Snapshot, error) {
+	defer snapshotstate.MockBackendSave(func(_ context.Context, id uint64, si *snap.Info, cfg map[string]interface{}, usernames []string, _ *dirs.SnapDirOptions) (*client.Snapshot, error) {
 		c.Check(id, check.Equals, uint64(42))
 		c.Check(si, check.DeepEquals, &snapInfo)
 		c.Check(cfg, check.DeepEquals, map[string]interface{}{"hello": "there"})
@@ -299,12 +300,50 @@ func (snapshotSuite) TestDoSave(c *check.C) {
 	c.Assert(err, check.IsNil)
 }
 
+func (snapshotSuite) TestDoSaveGetsSnapDirOpts(c *check.C) {
+	restore := snapshotstate.MockGetSnapDirOptions(func(*state.State, string) (*dirs.SnapDirOptions, error) {
+		return &dirs.SnapDirOptions{HiddenSnapDataDir: true}, nil
+	})
+	defer restore()
+
+	snapInfo := snap.Info{
+		SideInfo: snap.SideInfo{
+			RealName: "a-snap",
+			Revision: snap.R(-1),
+		},
+		Version: "1.33",
+	}
+	defer snapshotstate.MockSnapstateCurrentInfo(func(_ *state.State, snapname string) (*snap.Info, error) {
+		c.Check(snapname, check.Equals, "a-snap")
+		return &snapInfo, nil
+	})()
+
+	var checkOpts bool
+	defer snapshotstate.MockBackendSave(func(_ context.Context, id uint64, si *snap.Info, cfg map[string]interface{}, usernames []string, opts *dirs.SnapDirOptions) (*client.Snapshot, error) {
+		c.Check(opts.HiddenSnapDataDir, check.Equals, true)
+		checkOpts = true
+		return nil, nil
+	})()
+
+	st := state.New(nil)
+	st.Lock()
+	task := st.NewTask("save-snapshot", "...")
+	task.Set("snapshot-setup", map[string]interface{}{
+		"snap": "a-snap",
+	})
+	st.Unlock()
+
+	err := snapshotstate.DoSave(task, &tomb.Tomb{})
+	c.Assert(err, check.IsNil)
+	c.Check(checkOpts, check.Equals, true)
+}
+
 func (snapshotSuite) TestDoSaveFailsWithNoSnap(c *check.C) {
 	defer snapshotstate.MockSnapstateCurrentInfo(func(*state.State, string) (*snap.Info, error) {
 		return nil, errors.New("bzzt")
 	})()
 	defer snapshotstate.MockConfigGetSnapConfig(func(*state.State, string) (*json.RawMessage, error) { return nil, nil })()
-	defer snapshotstate.MockBackendSave(func(_ context.Context, id uint64, si *snap.Info, cfg map[string]interface{}, usernames []string) (*client.Snapshot, error) {
+	defer snapshotstate.MockBackendSave(func(_ context.Context, id uint64, si *snap.Info, cfg map[string]interface{}, usernames []string, options *dirs.SnapDirOptions) (*client.Snapshot, error) {
 		return nil, nil
 	})()
 
@@ -331,7 +370,7 @@ func (snapshotSuite) TestDoSaveFailsWithNoSnapshot(c *check.C) {
 	}
 	defer snapshotstate.MockSnapstateCurrentInfo(func(*state.State, string) (*snap.Info, error) { return &snapInfo, nil })()
 	defer snapshotstate.MockConfigGetSnapConfig(func(*state.State, string) (*json.RawMessage, error) { return nil, nil })()
-	defer snapshotstate.MockBackendSave(func(_ context.Context, id uint64, si *snap.Info, cfg map[string]interface{}, usernames []string) (*client.Snapshot, error) {
+	defer snapshotstate.MockBackendSave(func(_ context.Context, id uint64, si *snap.Info, cfg map[string]interface{}, usernames []string, options *dirs.SnapDirOptions) (*client.Snapshot, error) {
 		return nil, nil
 	})()
 
@@ -355,7 +394,7 @@ func (snapshotSuite) TestDoSaveFailsBackendError(c *check.C) {
 	}
 	defer snapshotstate.MockSnapstateCurrentInfo(func(*state.State, string) (*snap.Info, error) { return &snapInfo, nil })()
 	defer snapshotstate.MockConfigGetSnapConfig(func(*state.State, string) (*json.RawMessage, error) { return nil, nil })()
-	defer snapshotstate.MockBackendSave(func(_ context.Context, id uint64, si *snap.Info, cfg map[string]interface{}, usernames []string) (*client.Snapshot, error) {
+	defer snapshotstate.MockBackendSave(func(_ context.Context, id uint64, si *snap.Info, cfg map[string]interface{}, usernames []string, options *dirs.SnapDirOptions) (*client.Snapshot, error) {
 		return nil, errors.New("bzzt")
 	})()
 
@@ -384,7 +423,7 @@ func (snapshotSuite) TestDoSaveFailsConfigError(c *check.C) {
 	defer snapshotstate.MockConfigGetSnapConfig(func(*state.State, string) (*json.RawMessage, error) {
 		return nil, errors.New("bzzt")
 	})()
-	defer snapshotstate.MockBackendSave(func(_ context.Context, id uint64, si *snap.Info, cfg map[string]interface{}, usernames []string) (*client.Snapshot, error) {
+	defer snapshotstate.MockBackendSave(func(_ context.Context, id uint64, si *snap.Info, cfg map[string]interface{}, usernames []string, options *dirs.SnapDirOptions) (*client.Snapshot, error) {
 		return nil, nil
 	})()
 
@@ -398,7 +437,7 @@ func (snapshotSuite) TestDoSaveFailsConfigError(c *check.C) {
 	})
 	st.Unlock()
 	err := snapshotstate.DoSave(task, &tomb.Tomb{})
-	c.Assert(err, check.ErrorMatches, "bzzt")
+	c.Assert(err, check.ErrorMatches, "internal error: cannot obtain current snap config: bzzt")
 }
 
 func (snapshotSuite) TestDoSaveFailsBadConfig(c *check.C) {
@@ -415,7 +454,7 @@ func (snapshotSuite) TestDoSaveFailsBadConfig(c *check.C) {
 		buf := json.RawMessage(`"hello-there"`)
 		return &buf, nil
 	})()
-	defer snapshotstate.MockBackendSave(func(_ context.Context, id uint64, si *snap.Info, cfg map[string]interface{}, usernames []string) (*client.Snapshot, error) {
+	defer snapshotstate.MockBackendSave(func(_ context.Context, id uint64, si *snap.Info, cfg map[string]interface{}, usernames []string, options *dirs.SnapDirOptions) (*client.Snapshot, error) {
 		return nil, nil
 	})()
 
@@ -448,7 +487,7 @@ func (snapshotSuite) TestDoSaveFailureRemovesStateEntry(c *check.C) {
 	defer snapshotstate.MockConfigGetSnapConfig(func(_ *state.State, snapname string) (*json.RawMessage, error) {
 		return nil, nil
 	})()
-	defer snapshotstate.MockBackendSave(func(_ context.Context, id uint64, si *snap.Info, cfg map[string]interface{}, usernames []string) (*client.Snapshot, error) {
+	defer snapshotstate.MockBackendSave(func(_ context.Context, id uint64, si *snap.Info, cfg map[string]interface{}, usernames []string, options *dirs.SnapDirOptions) (*client.Snapshot, error) {
 		var expirations map[uint64]interface{}
 		st.Lock()
 		defer st.Unlock()
@@ -518,7 +557,7 @@ func (rs *readerSuite) SetUpTest(c *check.C) {
 			rs.calls = append(rs.calls, "open")
 			return &backend.Reader{}, nil
 		}),
-		snapshotstate.MockBackendRestore(func(*backend.Reader, context.Context, snap.Revision, []string, backend.Logf) (*backend.RestoreState, error) {
+		snapshotstate.MockBackendRestore(func(*backend.Reader, context.Context, snap.Revision, []string, backend.Logf, *dirs.SnapDirOptions) (*backend.RestoreState, error) {
 			rs.calls = append(rs.calls, "restore")
 			return &backend.RestoreState{}, nil
 		}),
@@ -557,7 +596,7 @@ func (rs *readerSuite) TestDoRestore(c *check.C) {
 			Snapshot: client.Snapshot{Conf: map[string]interface{}{"hello": "there"}},
 		}, nil
 	})()
-	defer snapshotstate.MockBackendRestore(func(_ *backend.Reader, _ context.Context, _ snap.Revision, users []string, _ backend.Logf) (*backend.RestoreState, error) {
+	defer snapshotstate.MockBackendRestore(func(_ *backend.Reader, _ context.Context, _ snap.Revision, users []string, _ backend.Logf, options *dirs.SnapDirOptions) (*backend.RestoreState, error) {
 		rs.calls = append(rs.calls, "restore")
 		c.Check(users, check.DeepEquals, []string{"a-user", "b-user"})
 		return &backend.RestoreState{}, nil
@@ -581,6 +620,48 @@ func (rs *readerSuite) TestDoRestore(c *check.C) {
 	c.Check(v, check.DeepEquals, map[string]interface{}{"config": map[string]interface{}{"old": "conf"}})
 }
 
+func (rs *readerSuite) TestDoRestoreNoConfig(c *check.C) {
+	defer snapshotstate.MockConfigGetSnapConfig(func(_ *state.State, snapname string) (*json.RawMessage, error) {
+		rs.calls = append(rs.calls, "get config")
+		c.Check(snapname, check.Equals, "a-snap")
+		// simulate old config
+		raw := json.RawMessage(`{"foo": "bar"}`)
+		return &raw, nil
+	})()
+	defer snapshotstate.MockBackendOpen(func(filename string, setID uint64) (*backend.Reader, error) {
+		rs.calls = append(rs.calls, "open")
+		// set id 0 tells backend.Open to use set id from the filename
+		c.Check(setID, check.Equals, uint64(0))
+		c.Check(filename, check.Equals, "/some/1_file.zip")
+		return &backend.Reader{
+			// snapshot has no configuration to restore
+			Snapshot: client.Snapshot{Snap: "a-snap", Conf: nil},
+		}, nil
+	})()
+	defer snapshotstate.MockBackendRestore(func(_ *backend.Reader, _ context.Context, _ snap.Revision, users []string, _ backend.Logf, options *dirs.SnapDirOptions) (*backend.RestoreState, error) {
+		rs.calls = append(rs.calls, "restore")
+		c.Check(users, check.DeepEquals, []string{"a-user", "b-user"})
+		return &backend.RestoreState{}, nil
+	})()
+	defer snapshotstate.MockConfigSetSnapConfig(func(_ *state.State, snapname string, conf *json.RawMessage) error {
+		rs.calls = append(rs.calls, "set config")
+		c.Check(snapname, check.Equals, "a-snap")
+		c.Check(conf, check.IsNil)
+		return nil
+	})()
+
+	st := rs.task.State()
+	err := snapshotstate.DoRestore(rs.task, &tomb.Tomb{})
+	c.Assert(err, check.IsNil)
+	c.Check(rs.calls, check.DeepEquals, []string{"get config", "open", "restore", "set config"})
+
+	st.Lock()
+	defer st.Unlock()
+	var v map[string]interface{}
+	rs.task.Get("restore-state", &v)
+	c.Check(v, check.DeepEquals, map[string]interface{}{"config": map[string]interface{}{"foo": "bar"}})
+}
+
 func (rs *readerSuite) TestDoRestoreFailsNoTaskSnapshot(c *check.C) {
 	rs.task.State().Lock()
 	rs.task.Clear("snapshot-setup")
@@ -599,7 +680,7 @@ func (rs *readerSuite) TestDoRestoreFailsOnGetConfigError(c *check.C) {
 	})()
 
 	err := snapshotstate.DoRestore(rs.task, &tomb.Tomb{})
-	c.Assert(err, check.ErrorMatches, "internal error: cannot obtain current snap config for snapshot restore: bzzt")
+	c.Assert(err, check.ErrorMatches, "internal error: cannot obtain current snap config: bzzt")
 	c.Check(rs.calls, check.DeepEquals, []string{"get config"})
 }
 
@@ -640,7 +721,7 @@ func (rs *readerSuite) TestDoRestoreFailsUnserialisableSnapshotConfigError(c *ch
 }
 
 func (rs *readerSuite) TestDoRestoreFailsOnRestoreError(c *check.C) {
-	defer snapshotstate.MockBackendRestore(func(*backend.Reader, context.Context, snap.Revision, []string, backend.Logf) (*backend.RestoreState, error) {
+	defer snapshotstate.MockBackendRestore(func(*backend.Reader, context.Context, snap.Revision, []string, backend.Logf, *dirs.SnapDirOptions) (*backend.RestoreState, error) {
 		rs.calls = append(rs.calls, "restore")
 		return nil, errors.New("bzzt")
 	})()
@@ -662,6 +743,30 @@ func (rs *readerSuite) TestDoRestoreFailsAndRevertsOnSetConfigError(c *check.C)
 }
 
 func (rs *readerSuite) TestUndoRestore(c *check.C) {
+	defer snapshotstate.MockConfigSetSnapConfig(func(st *state.State, snapName string, raw *json.RawMessage) error {
+		rs.calls = append(rs.calls, "set config")
+		c.Check(string(*raw), check.Equals, `{"foo":"bar"}`)
+		return nil
+	})()
+
+	st := rs.task.State()
+	st.Lock()
+	v := map[string]interface{}{"config": map[string]interface{}{"foo": "bar"}}
+	rs.task.Set("restore-state", &v)
+	st.Unlock()
+
+	err := snapshotstate.UndoRestore(rs.task, &tomb.Tomb{})
+	c.Assert(err, check.IsNil)
+	c.Check(rs.calls, check.DeepEquals, []string{"set config", "revert"})
+}
+
+func (rs *readerSuite) TestUndoRestoreNoConfig(c *check.C) {
+	defer snapshotstate.MockConfigSetSnapConfig(func(st *state.State, snapName string, raw *json.RawMessage) error {
+		rs.calls = append(rs.calls, "set config")
+		c.Check(raw, check.IsNil)
+		return nil
+	})()
+
 	st := rs.task.State()
 	st.Lock()
 	var v map[string]interface{}
