@@ -31,6 +31,7 @@ import (
 	"github.com/snapcore/snapd/overlord/state"
 	"github.com/snapcore/snapd/snap"
 	"github.com/snapcore/snapd/snap/snaptest"
+	"github.com/snapcore/snapd/testutil"
 )
 
 func target(at *snapstate.AliasTarget) string {
@@ -151,7 +152,6 @@ func (s *snapmgrTestSuite) TestAutoAliasesDelta(c *C) {
 			"alias2": "cmd2",
 			"alias4": "cmd4",
 			"alias5": "cmd5",
-			"alias6": "cmd6b",
 		}, nil
 	}
 
@@ -178,11 +178,10 @@ func (s *snapmgrTestSuite) TestAutoAliasesDelta(c *C) {
 	c.Check(changed, HasLen, 1)
 	which := changed["alias-snap"]
 	sort.Strings(which)
-	c.Check(which, DeepEquals, []string{"alias4", "alias5", "alias6"})
+	c.Check(which, DeepEquals, []string{"alias4", "alias5"})
 
-	c.Check(dropped, DeepEquals, map[string][]string{
-		"alias-snap": {"alias3"},
-	})
+	c.Check(dropped, HasLen, 1)
+	c.Check(dropped["alias-snap"], testutil.DeepUnsortedMatches, []string{"alias3", "alias6"})
 }
 
 func (s *snapmgrTestSuite) TestAutoAliasesDeltaAll(c *C) {
@@ -197,6 +196,11 @@ func (s *snapmgrTestSuite) TestAutoAliasesDeltaAll(c *C) {
 				"alias5": "cmd5",
 			}, nil
 		}
+		if info.InstanceName() == "alias-snap-apps-go-away" {
+			return map[string]string{
+				"alias-remains": "cmd1",
+			}, nil
+		}
 		return nil, nil
 	}
 
@@ -217,6 +221,17 @@ func (s *snapmgrTestSuite) TestAutoAliasesDeltaAll(c *C) {
 		Current: snap.R(2),
 		Active:  true,
 	})
+	snapstate.Set(s.state, "alias-snap-apps-go-away", &snapstate.SnapState{
+		Sequence: []*snap.SideInfo{
+			{RealName: "alias-snap-apps-go-away", Revision: snap.R(1)},
+		},
+		Current: snap.R(1),
+		Active:  true,
+		Aliases: map[string]*snapstate.AliasTarget{
+			"alias-remains":   {Auto: "cmd1"},
+			"alias-goes-away": {Auto: "cmd2"},
+		},
+	})
 
 	changed, dropped, err := snapstate.AutoAliasesDelta(s.state, nil)
 	c.Assert(err, IsNil)
@@ -226,12 +241,15 @@ func (s *snapmgrTestSuite) TestAutoAliasesDeltaAll(c *C) {
 	sort.Strings(which)
 	c.Check(which, DeepEquals, []string{"alias1", "alias2", "alias4", "alias5"})
 
-	c.Check(dropped, HasLen, 0)
+	c.Check(dropped, DeepEquals, map[string][]string{
+		"alias-snap-apps-go-away": {"alias-goes-away"},
+	})
 
 	c.Check(seen, DeepEquals, map[string]bool{
-		"core":       true,
-		"alias-snap": true,
-		"other-snap": true,
+		"core":                    true,
+		"alias-snap":              true,
+		"other-snap":              true,
+		"alias-snap-apps-go-away": true,
 	})
 }
 
@@ -279,7 +297,6 @@ func (s *snapmgrTestSuite) TestRefreshAliases(c *C) {
 			"alias1": "cmd1",
 			"alias2": "cmd2",
 			"alias4": "cmd4",
-			"alias5": "cmd5",
 		}, nil
 	}
 
@@ -576,10 +593,8 @@ func (s *snapmgrTestSuite) TestAliasRunThrough(c *C) {
 	c.Assert(err, IsNil)
 	chg.AddAll(ts)
 
-	s.state.Unlock()
 	defer s.se.Stop()
 	s.settle(c)
-	s.state.Lock()
 
 	c.Assert(chg.Status(), Equals, state.DoneStatus, Commentf("%v", chg.Err()))
 	expected := fakeOps{
@@ -635,10 +650,8 @@ func (s *snapmgrTestSuite) TestParallelInstanceAliasRunThrough(c *C) {
 	c.Assert(err, IsNil)
 	chg.AddAll(ts)
 
-	s.state.Unlock()
 	defer s.se.Stop()
 	s.settle(c)
-	s.state.Lock()
 
 	c.Assert(chg.Status(), Equals, state.DoneStatus, Commentf("%v", chg.Err()))
 	expected := fakeOps{
@@ -755,10 +768,8 @@ func (s *snapmgrTestSuite) TestAliasOverAutoRunThrough(c *C) {
 	c.Assert(err, IsNil)
 	chg.AddAll(ts)
 
-	s.state.Unlock()
 	defer s.se.Stop()
 	s.settle(c)
-	s.state.Lock()
 
 	c.Assert(chg.Status(), Equals, state.DoneStatus, Commentf("%v", chg.Err()))
 	expected := fakeOps{
@@ -800,7 +811,7 @@ func (s *snapmgrTestSuite) TestAliasUpdateChangeConflict(c *C) {
 	// need a change to make the tasks visible
 	s.state.NewChange("alias", "...").AddAll(ts)
 
-	_, err = snapstate.Update(s.state, "some-snap", "some-channel", snap.R(0), s.user.ID, snapstate.Flags{})
+	_, err = snapstate.Update(s.state, "some-snap", nil, s.user.ID, snapstate.Flags{})
 	c.Assert(err, ErrorMatches, `snap "some-snap" has "alias" change in progress`)
 }
 
@@ -815,7 +826,7 @@ func (s *snapmgrTestSuite) TestUpdateAliasChangeConflict(c *C) {
 		SnapType: "app",
 	})
 
-	ts, err := snapstate.Update(s.state, "some-snap", "some-channel", snap.R(0), s.user.ID, snapstate.Flags{})
+	ts, err := snapstate.Update(s.state, "some-snap", nil, s.user.ID, snapstate.Flags{})
 	c.Assert(err, IsNil)
 	// need a change to make the tasks visible
 	s.state.NewChange("update", "...").AddAll(ts)
@@ -958,10 +969,8 @@ func (s *snapmgrTestSuite) TestDisableAllAliasesRunThrough(c *C) {
 	c.Assert(err, IsNil)
 	chg.AddAll(ts)
 
-	s.state.Unlock()
 	defer s.se.Stop()
 	s.settle(c)
-	s.state.Lock()
 
 	c.Assert(chg.Status(), Equals, state.DoneStatus, Commentf("%v", chg.Err()))
 	expected := fakeOps{
@@ -1030,10 +1039,8 @@ func (s *snapmgrTestSuite) TestParallelInstanceDisableAllAliasesRunThrough(c *C)
 	c.Assert(err, IsNil)
 	chg.AddAll(ts)
 
-	s.state.Unlock()
 	defer s.se.Stop()
 	s.settle(c)
-	s.state.Lock()
 
 	c.Assert(chg.Status(), Equals, state.DoneStatus, Commentf("%v", chg.Err()))
 	expected := fakeOps{
@@ -1127,10 +1134,8 @@ func (s *snapmgrTestSuite) TestRemoveManualAliasRunThrough(c *C) {
 	c.Assert(err, IsNil)
 	chg.AddAll(ts)
 
-	s.state.Unlock()
 	defer s.se.Stop()
 	s.settle(c)
-	s.state.Lock()
 
 	c.Assert(chg.Status(), Equals, state.DoneStatus, Commentf("%v", chg.Err()))
 	expected := fakeOps{
@@ -1179,10 +1184,8 @@ func (s *snapmgrTestSuite) TestRemoveManualAliasOverAutoRunThrough(c *C) {
 	c.Assert(err, IsNil)
 	chg.AddAll(ts)
 
-	s.state.Unlock()
 	defer s.se.Stop()
 	s.settle(c)
-	s.state.Lock()
 
 	c.Assert(chg.Status(), Equals, state.DoneStatus, Commentf("%v", chg.Err()))
 	expected := fakeOps{
@@ -1228,10 +1231,8 @@ func (s *snapmgrTestSuite) TestParallelInstanceRemoveManualAliasRunThrough(c *C)
 	c.Assert(err, IsNil)
 	chg.AddAll(ts)
 
-	s.state.Unlock()
 	defer s.se.Stop()
 	s.settle(c)
-	s.state.Lock()
 
 	c.Assert(chg.Status(), Equals, state.DoneStatus, Commentf("%v", chg.Err()))
 	expected := fakeOps{
@@ -1278,7 +1279,7 @@ func (s *snapmgrTestSuite) TestUpdateDisableAllAliasesChangeConflict(c *C) {
 		},
 	})
 
-	ts, err := snapstate.Update(s.state, "some-snap", "some-channel", snap.R(0), s.user.ID, snapstate.Flags{})
+	ts, err := snapstate.Update(s.state, "some-snap", nil, s.user.ID, snapstate.Flags{})
 	c.Assert(err, IsNil)
 	// need a change to make the tasks visible
 	s.state.NewChange("update", "...").AddAll(ts)
@@ -1301,7 +1302,7 @@ func (s *snapmgrTestSuite) TestUpdateRemoveManualAliasChangeConflict(c *C) {
 		},
 	})
 
-	ts, err := snapstate.Update(s.state, "some-snap", "some-channel", snap.R(0), s.user.ID, snapstate.Flags{})
+	ts, err := snapstate.Update(s.state, "some-snap", nil, s.user.ID, snapstate.Flags{})
 	c.Assert(err, IsNil)
 	// need a change to make the tasks visible
 	s.state.NewChange("update", "...").AddAll(ts)
@@ -1326,7 +1327,7 @@ func (s *snapmgrTestSuite) TestDisableAllAliasesUpdateChangeConflict(c *C) {
 	// need a change to make the tasks visible
 	s.state.NewChange("alias", "...").AddAll(ts)
 
-	_, err = snapstate.Update(s.state, "some-snap", "some-channel", snap.R(0), s.user.ID, snapstate.Flags{})
+	_, err = snapstate.Update(s.state, "some-snap", nil, s.user.ID, snapstate.Flags{})
 	c.Assert(err, ErrorMatches, `snap "some-snap" has "alias" change in progress`)
 }
 
@@ -1349,7 +1350,7 @@ func (s *snapmgrTestSuite) TestRemoveManualAliasUpdateChangeConflict(c *C) {
 	// need a change to make the tasks visible
 	s.state.NewChange("unalias", "...").AddAll(ts)
 
-	_, err = snapstate.Update(s.state, "some-snap", "some-channel", snap.R(0), s.user.ID, snapstate.Flags{})
+	_, err = snapstate.Update(s.state, "some-snap", nil, s.user.ID, snapstate.Flags{})
 	c.Assert(err, ErrorMatches, `snap "some-snap" has "unalias" change in progress`)
 }
 
@@ -1396,10 +1397,8 @@ func (s *snapmgrTestSuite) TestPreferRunThrough(c *C) {
 	c.Assert(err, IsNil)
 	chg.AddAll(ts)
 
-	s.state.Unlock()
 	defer s.se.Stop()
 	s.settle(c)
-	s.state.Lock()
 
 	c.Assert(chg.Status(), Equals, state.DoneStatus, Commentf("%v", chg.Err()))
 	expected := fakeOps{
@@ -1460,10 +1459,8 @@ func (s *snapmgrTestSuite) TestParallelInstancePreferRunThrough(c *C) {
 	c.Assert(err, IsNil)
 	chg.AddAll(ts)
 
-	s.state.Unlock()
 	defer s.se.Stop()
 	s.settle(c)
-	s.state.Lock()
 
 	c.Assert(chg.Status(), Equals, state.DoneStatus, Commentf("%v", chg.Err()))
 	expected := fakeOps{
@@ -1523,7 +1520,7 @@ func (s *snapmgrTestSuite) TestUpdatePreferChangeConflict(c *C) {
 		},
 	})
 
-	ts, err := snapstate.Update(s.state, "some-snap", "some-channel", snap.R(0), s.user.ID, snapstate.Flags{})
+	ts, err := snapstate.Update(s.state, "some-snap", nil, s.user.ID, snapstate.Flags{})
 	c.Assert(err, IsNil)
 	// need a change to make the tasks visible
 	s.state.NewChange("update", "...").AddAll(ts)
@@ -1552,6 +1549,6 @@ func (s *snapmgrTestSuite) TestPreferUpdateChangeConflict(c *C) {
 	// need a change to make the tasks visible
 	s.state.NewChange("prefer", "...").AddAll(ts)
 
-	_, err = snapstate.Update(s.state, "some-snap", "some-channel", snap.R(0), s.user.ID, snapstate.Flags{})
+	_, err = snapstate.Update(s.state, "some-snap", nil, s.user.ID, snapstate.Flags{})
 	c.Assert(err, ErrorMatches, `snap "some-snap" has "prefer" change in progress`)
 }
