@@ -20,6 +20,8 @@
 package snapstate
 
 import (
+	"context"
+	"errors"
 	"fmt"
 	"os"
 	"time"
@@ -35,16 +37,23 @@ import (
 	"github.com/snapcore/snapd/strutil"
 	"github.com/snapcore/snapd/timeutil"
 	"github.com/snapcore/snapd/timings"
+	userclient "github.com/snapcore/snapd/usersession/client"
 )
 
 // the default refresh pattern
 const defaultRefreshSchedule = "00:00~24:00/4"
 
 // cannot keep without refreshing for more than maxPostponement
-const maxPostponement = 60 * 24 * time.Hour
+const maxPostponement = 95 * 24 * time.Hour
 
-// cannot inhibit refreshes for more than maxInhibition
-const maxInhibition = 7 * 24 * time.Hour
+// buffer for maxPostponement when holding snaps with auto-refresh gating
+const maxPostponementBuffer = 5 * 24 * time.Hour
+
+// cannot inhibit refreshes for more than maxInhibition;
+// deduct 1s so it doesn't look confusing initially when two notifications
+// get displayed in short period of time and it immediately goes from "14 days"
+// to "13 days" left.
+const maxInhibition = 14*24*time.Hour - time.Second
 
 // hooks setup by devicestate
 var (
@@ -56,6 +65,44 @@ var (
 // refreshRetryDelay specified the minimum time to retry failed refreshes
 var refreshRetryDelay = 20 * time.Minute
 
+// refreshCandidate carries information about a single snap to update as part
+// of auto-refresh.
+type refreshCandidate struct {
+	SnapSetup
+	Version string `json:"version,omitempty"`
+}
+
+func (rc *refreshCandidate) Type() snap.Type {
+	return rc.SnapSetup.Type
+}
+
+func (rc *refreshCandidate) SnapBase() string {
+	return rc.SnapSetup.Base
+}
+
+func (rc *refreshCandidate) DownloadSize() int64 {
+	return rc.DownloadInfo.Size
+}
+
+func (rc *refreshCandidate) InstanceName() string {
+	return rc.SnapSetup.InstanceName()
+}
+
+func (rc *refreshCandidate) Prereq(st *state.State) []string {
+	return rc.SnapSetup.Prereq
+}
+
+func (rc *refreshCandidate) SnapSetupForUpdate(st *state.State, params updateParamsFunc, userID int, globalFlags *Flags) (*SnapSetup, *SnapState, error) {
+	var snapst SnapState
+	if err := Get(st, rc.InstanceName(), &snapst); err != nil {
+		return nil, nil, err
+	}
+	return &rc.SnapSetup, &snapst, nil
+}
+
+// soundness check
+var _ readyUpdateInfo = (*refreshCandidate)(nil)
+
 // autoRefresh will ensure that snaps are refreshed automatically
 // according to the refresh schedule.
 type autoRefresh struct {
@@ -64,6 +111,7 @@ type autoRefresh struct {
 	lastRefreshSchedule string
 	nextRefresh         time.Time
 	lastRefreshAttempt  time.Time
+	managedDeniedLogged bool
 }
 
 func newAutoRefresh(st *state.State) *autoRefresh {
@@ -127,6 +175,31 @@ func (m *autoRefresh) EffectiveRefreshHold() (time.Time, error) {
 	return holdTime, nil
 }
 
+func (m *autoRefresh) ensureRefreshHoldAtLeast(duration time.Duration) error {
+	now := time.Now()
+
+	// get the effective refresh hold and check if it is sooner than the
+	// specified duration in the future
+	effective, err := m.EffectiveRefreshHold()
+	if err != nil {
+		return err
+	}
+
+	if effective.IsZero() || effective.Sub(now) < duration {
+		// the effective refresh hold is sooner than the desired delay, so
+		// move it out to the specified duration
+		holdTime := now.Add(duration)
+		tr := config.NewTransaction(m.state)
+		err := tr.Set("core", "refresh.hold", &holdTime)
+		if err != nil && !config.IsNoOption(err) {
+			return err
+		}
+		tr.Commit()
+	}
+
+	return nil
+}
+
 // clearRefreshHold clears refresh.hold configuration.
 func (m *autoRefresh) clearRefreshHold() {
 	tr := config.NewTransaction(m.state)
@@ -159,7 +232,7 @@ func canRefreshOnMeteredConnection(st *state.State) (bool, error) {
 	tr := config.NewTransaction(st)
 	var onMetered string
 	err := tr.GetMaybe("core", "refresh.metered", &onMetered)
-	if err != nil && err != state.ErrNoState {
+	if err != nil && !errors.Is(err, state.ErrNoState) {
 		return false, err
 	}
 
@@ -253,54 +326,74 @@ func (m *autoRefresh) Ensure() error {
 		logger.Debugf("Next refresh scheduled for %s.", m.nextRefresh.Format(time.RFC3339))
 	}
 
-	// should we hold back refreshes?
-	holdTime, err := m.EffectiveRefreshHold()
+	held, holdTime, err := m.isRefreshHeld()
 	if err != nil {
 		return err
 	}
-	if holdTime.After(now) {
-		return nil
-	}
-	if !holdTime.IsZero() {
-		// expired hold case
-		m.clearRefreshHold()
-		if m.nextRefresh.Before(holdTime) {
-			// next refresh is obsolete, compute the next one
-			delta := timeutil.Next(refreshSchedule, holdTime, maxPostponement)
-			now = time.Now()
-			m.nextRefresh = now.Add(delta)
-		}
-	}
 
 	// do refresh attempt (if needed)
-	if !m.nextRefresh.After(now) {
-		var can bool
-		can, err = m.canRefreshRespectingMetered(now, lastRefresh)
-		if err != nil {
-			return err
-		}
-		if !can {
-			// clear nextRefresh so that another refresh time is calculated
-			m.nextRefresh = time.Time{}
-			return nil
+	if !held {
+		if !holdTime.IsZero() {
+			// expired hold case
+			m.clearRefreshHold()
+			if m.nextRefresh.Before(holdTime) {
+				// next refresh is obsolete, compute the next one
+				delta := timeutil.Next(refreshSchedule, holdTime, maxPostponement)
+				now = time.Now()
+				m.nextRefresh = now.Add(delta)
+			}
 		}
 
-		// Check that we have reasonable delays between attempts.
-		// If the store is under stress we need to make sure we do not
-		// hammer it too often
-		if !m.lastRefreshAttempt.IsZero() && m.lastRefreshAttempt.Add(refreshRetryDelay).After(time.Now()) {
-			return nil
-		}
+		// refresh is also "held" if the next time is in the future
+		// note that the two times here could be exactly equal, so we use
+		// !After() because that is true in the case that the next refresh is
+		// before now, and the next refresh is equal to now without requiring an
+		// or operation
+		if !m.nextRefresh.After(now) {
+			var can bool
+			can, err = m.canRefreshRespectingMetered(now, lastRefresh)
+			if err != nil {
+				return err
+			}
+			if !can {
+				// clear nextRefresh so that another refresh time is calculated
+				m.nextRefresh = time.Time{}
+				return nil
+			}
 
-		err = m.launchAutoRefresh()
-		if _, ok := err.(*httputil.PerstistentNetworkError); !ok {
-			m.nextRefresh = time.Time{}
-		} // else - refresh will be retried after refreshRetryDelay
+			// Check that we have reasonable delays between attempts.
+			// If the store is under stress we need to make sure we do not
+			// hammer it too often
+			if !m.lastRefreshAttempt.IsZero() && m.lastRefreshAttempt.Add(refreshRetryDelay).After(time.Now()) {
+				return nil
+			}
+
+			err = m.launchAutoRefresh()
+			if _, ok := err.(*httputil.PersistentNetworkError); !ok {
+				m.nextRefresh = time.Time{}
+			} // else - refresh will be retried after refreshRetryDelay
+		}
 	}
 
 	return err
 }
 
+// isRefreshHeld returns whether an auto-refresh is currently held back or not,
+// as indicated by m.EffectiveRefreshHold().
+func (m *autoRefresh) isRefreshHeld() (bool, time.Time, error) {
+	now := time.Now()
+	// should we hold back refreshes?
+	holdTime, err := m.EffectiveRefreshHold()
+	if err != nil {
+		return false, time.Time{}, err
+	}
+	if holdTime.After(now) {
+		return true, holdTime, nil
+	}
+
+	return false, holdTime, nil
+}
+
 func (m *autoRefresh) ensureLastRefreshAnchor() {
 	seedTime, _ := getTime(m.state, "seed-time")
 	if !seedTime.IsZero() {
@@ -329,12 +422,24 @@ func (m *autoRefresh) ensureLastRefreshAnchor() {
 // TODO: we can remove the refreshSchedule reset because we have validation
 //       of the schedule now.
 func (m *autoRefresh) refreshScheduleWithDefaultsFallback() (ts []*timeutil.Schedule, scheduleAsStr string, legacy bool, err error) {
-	if managed, legacy := refreshScheduleManaged(m.state); managed {
+	managed, requested, legacy := refreshScheduleManaged(m.state)
+	if managed {
 		if m.lastRefreshSchedule != "managed" {
 			logger.Noticef("refresh is managed via the snapd-control interface")
 			m.lastRefreshSchedule = "managed"
 		}
+		m.managedDeniedLogged = false
 		return nil, "managed", legacy, nil
+	} else if requested {
+		// managed refresh schedule was denied
+		if !m.managedDeniedLogged {
+			logger.Noticef("managed refresh schedule denied, no properly configured snapd-control")
+			m.managedDeniedLogged = true
+		}
+		// fallback to default schedule
+		return refreshScheduleDefault()
+	} else {
+		m.managedDeniedLogged = false
 	}
 
 	tr := config.NewTransaction(m.state)
@@ -370,6 +475,23 @@ func (m *autoRefresh) refreshScheduleWithDefaultsFallback() (ts []*timeutil.Sche
 	return refreshScheduleDefault()
 }
 
+func autoRefreshSummary(updated []string) string {
+	var msg string
+	switch len(updated) {
+	case 0:
+		return ""
+	case 1:
+		msg = fmt.Sprintf(i18n.G("Auto-refresh snap %q"), updated[0])
+	case 2, 3:
+		quoted := strutil.Quoted(updated)
+		// TRANSLATORS: the %s is a comma-separated list of quoted snap names
+		msg = fmt.Sprintf(i18n.G("Auto-refresh snaps %s"), quoted)
+	default:
+		msg = fmt.Sprintf(i18n.G("Auto-refresh %d snaps"), len(updated))
+	}
+	return msg
+}
+
 // launchAutoRefresh creates the auto-refresh taskset and a change for it.
 func (m *autoRefresh) launchAutoRefresh() error {
 	perfTimings := timings.New(map[string]string{"ensure": "auto-refresh"})
@@ -380,8 +502,30 @@ func (m *autoRefresh) launchAutoRefresh() error {
 	}()
 
 	m.lastRefreshAttempt = time.Now()
+
+	// NOTE: this will unlock and re-lock state for network ops
 	updated, tasksets, err := AutoRefresh(auth.EnsureContextTODO(), m.state)
-	if _, ok := err.(*httputil.PerstistentNetworkError); ok {
+
+	// TODO: we should have some way to lock just creating and starting changes,
+	//       as that would alleviate this race condition we are guarding against
+	//       with this check and probably would eliminate other similar race
+	//       conditions elsewhere
+
+	// re-check if the refresh is held because it could have been re-held and
+	// pushed back, in which case we need to abort the auto-refresh and wait
+	held, _, holdErr := m.isRefreshHeld()
+	if holdErr != nil {
+		return holdErr
+	}
+
+	if held {
+		// then a request came in that pushed the refresh out, so we will need
+		// to try again later
+		logger.Noticef("Auto-refresh was delayed mid-way through launching, aborting to try again later")
+		return nil
+	}
+
+	if _, ok := err.(*httputil.PersistentNetworkError); ok {
 		logger.Noticef("Cannot prepare auto-refresh change due to a permanent network error: %s", err)
 		return err
 	}
@@ -391,19 +535,10 @@ func (m *autoRefresh) launchAutoRefresh() error {
 		return err
 	}
 
-	var msg string
-	switch len(updated) {
-	case 0:
+	msg := autoRefreshSummary(updated)
+	if msg == "" {
 		logger.Noticef(i18n.G("auto-refresh: all snaps are up-to-date"))
 		return nil
-	case 1:
-		msg = fmt.Sprintf(i18n.G("Auto-refresh snap %q"), updated[0])
-	case 2, 3:
-		quoted := strutil.Quoted(updated)
-		// TRANSLATORS: the %s is a comma-separated list of quoted snap names
-		msg = fmt.Sprintf(i18n.G("Auto-refresh snaps %s"), quoted)
-	default:
-		msg = fmt.Sprintf(i18n.G("Auto-refresh %d snaps"), len(updated))
 	}
 
 	chg := m.state.NewChange("auto-refresh", msg)
@@ -437,65 +572,113 @@ func autoRefreshInFlight(st *state.State) bool {
 
 // refreshScheduleManaged returns true if the refresh schedule of the
 // device is managed by an external snap
-func refreshScheduleManaged(st *state.State) (managed bool, legacy bool) {
+func refreshScheduleManaged(st *state.State) (managed, requested, legacy bool) {
 	var confStr string
 
 	// this will only be "nil" if running in tests
 	if CanManageRefreshes == nil {
-		return false, legacy
+		return false, false, legacy
 	}
 
 	// check new style timer first
 	tr := config.NewTransaction(st)
 	err := tr.Get("core", "refresh.timer", &confStr)
 	if err != nil && !config.IsNoOption(err) {
-		return false, legacy
+		return false, false, legacy
 	}
 	// if not set, fallback to refresh.schedule
 	if confStr == "" {
 		if err := tr.Get("core", "refresh.schedule", &confStr); err != nil {
-			return false, legacy
+			return false, false, legacy
 		}
 		legacy = true
 	}
 
 	if confStr != "managed" {
-		return false, legacy
+		return false, false, legacy
 	}
-	return CanManageRefreshes(st), legacy
+
+	return CanManageRefreshes(st), true, legacy
 }
 
 // getTime retrieves a time from a state value.
 func getTime(st *state.State, timeKey string) (time.Time, error) {
 	var t1 time.Time
 	err := st.Get(timeKey, &t1)
-	if err != nil && err != state.ErrNoState {
+	if err != nil && !errors.Is(err, state.ErrNoState) {
 		return time.Time{}, err
 	}
 	return t1, nil
 }
 
+// asyncPendingRefreshNotification broadcasts desktop notification in a goroutine.
+//
+// This allows the, possibly slow, communication with each snapd session agent,
+// to be performed without holding the snap state lock.
+var asyncPendingRefreshNotification = func(context context.Context, client *userclient.Client, refreshInfo *userclient.PendingSnapRefreshInfo) {
+	go func() {
+		if err := client.PendingRefreshNotification(context, refreshInfo); err != nil {
+			logger.Noticef("Cannot send notification about pending refresh: %v", err)
+		}
+	}()
+}
+
 // inhibitRefresh returns an error if refresh is inhibited by running apps.
 //
 // Internally the snap state is updated to remember when the inhibition first
 // took place. Apps can inhibit refreshes for up to "maxInhibition", beyond
 // that period the refresh will go ahead despite application activity.
 func inhibitRefresh(st *state.State, snapst *SnapState, info *snap.Info, checker func(*snap.Info) error) error {
-	if err := checker(info); err != nil {
-		now := time.Now()
-		if snapst.RefreshInhibitedTime == nil {
-			// Store the instant when the snap was first inhibited.
-			// This is reset to nil on successful refresh.
-			snapst.RefreshInhibitedTime = &now
-			Set(st, info.InstanceName(), snapst)
-			return err
-		}
+	checkerErr := checker(info)
+	if checkerErr == nil {
+		return nil
+	}
 
-		if now.Sub(*snapst.RefreshInhibitedTime) < maxInhibition {
-			// If we are still in the allowed window then just return
-			// the error but don't change the snap state again.
-			return err
+	// Get pending refresh information from compatible errors or synthesize a new one.
+	var refreshInfo *userclient.PendingSnapRefreshInfo
+	if err, ok := checkerErr.(*BusySnapError); ok {
+		refreshInfo = err.PendingSnapRefreshInfo()
+	} else {
+		refreshInfo = &userclient.PendingSnapRefreshInfo{
+			InstanceName: info.InstanceName(),
 		}
 	}
-	return nil
+
+	// Decide on what to do depending on the state of the snap and the remaining
+	// inhibition time.
+	now := time.Now()
+	switch {
+	case snapst.RefreshInhibitedTime == nil:
+		// If the snap did not have inhibited refresh yet then commence a new
+		// window, during which refreshes are postponed, by storing the current
+		// time in the snap state's RefreshInhibitedTime field. This field is
+		// reset to nil on successful refresh.
+		snapst.RefreshInhibitedTime = &now
+		refreshInfo.TimeRemaining = (maxInhibition - now.Sub(*snapst.RefreshInhibitedTime)).Truncate(time.Second)
+		Set(st, info.InstanceName(), snapst)
+	case now.Sub(*snapst.RefreshInhibitedTime) < maxInhibition:
+		// If we are still in the allowed window then just return the error but
+		// don't change the snap state again.
+		// TODO: as time left shrinks, send additional notifications with
+		// increasing frequency, allowing the user to understand the urgency.
+		refreshInfo.TimeRemaining = (maxInhibition - now.Sub(*snapst.RefreshInhibitedTime)).Truncate(time.Second)
+	default:
+		// If we run out of time then consume the error that would normally
+		// inhibit refresh and notify the user that the snap is refreshing right
+		// now, by not setting the TimeRemaining field of the refresh
+		// notification message.
+		checkerErr = nil
+	}
+
+	// Send the notification asynchronously to avoid holding the state lock.
+	asyncPendingRefreshNotification(context.TODO(), userclient.New(), refreshInfo)
+	return checkerErr
+}
+
+// for testing outside of snapstate
+func MockRefreshCandidate(snapSetup *SnapSetup, version string) interface{} {
+	return &refreshCandidate{
+		SnapSetup: *snapSetup,
+		Version:   version,
+	}
 }
