@@ -29,6 +29,7 @@ import (
 	. "gopkg.in/check.v1"
 
 	"github.com/snapcore/snapd/overlord/state"
+	"github.com/snapcore/snapd/testutil"
 )
 
 func TestState(t *testing.T) { TestingT(t) }
@@ -76,6 +77,23 @@ func (ss *stateSuite) TestGetAndSet(c *C) {
 	c.Check(&mSt2B, DeepEquals, mSt2)
 }
 
+func (ss *stateSuite) TestStrayTaskWithNoChange(c *C) {
+	st := state.New(nil)
+	st.Lock()
+	defer st.Unlock()
+
+	chg := st.NewChange("change", "...")
+	t1 := st.NewTask("foo", "...")
+	chg.AddTask(t1)
+	_ = st.NewTask("bar", "...")
+
+	// only the task with associate change is returned
+	c.Assert(st.Tasks(), HasLen, 1)
+	c.Assert(st.Tasks()[0].ID(), Equals, t1.ID())
+	// but count includes all tasks
+	c.Assert(st.TaskCount(), Equals, 2)
+}
+
 func (ss *stateSuite) TestSetPanic(c *C) {
 	st := state.New(nil)
 	st.Lock()
@@ -94,7 +112,7 @@ func (ss *stateSuite) TestGetNoState(c *C) {
 
 	var mSt1B mgrState1
 	err := st.Get("mgr9", &mSt1B)
-	c.Check(err, Equals, state.ErrNoState)
+	c.Check(err, testutil.ErrorIs, state.ErrNoState)
 }
 
 func (ss *stateSuite) TestSetToNilDeletes(c *C) {
@@ -112,7 +130,7 @@ func (ss *stateSuite) TestSetToNilDeletes(c *C) {
 
 	var v1 map[string]int
 	err = st.Get("a", &v1)
-	c.Check(err, Equals, state.ErrNoState)
+	c.Check(err, testutil.ErrorIs, state.ErrNoState)
 	c.Check(v1, HasLen, 0)
 }
 
@@ -126,7 +144,7 @@ func (ss *stateSuite) TestNullMeansNoState(c *C) {
 
 	var v1 map[string]int
 	err = st.Get("a", &v1)
-	c.Check(err, Equals, state.ErrNoState)
+	c.Check(err, testutil.ErrorIs, state.ErrNoState)
 	c.Check(v1, HasLen, 0)
 }
 
@@ -168,10 +186,9 @@ func (ss *stateSuite) TestCache(c *C) {
 }
 
 type fakeStateBackend struct {
-	checkpoints      [][]byte
-	error            func() error
-	ensureBefore     time.Duration
-	restartRequested bool
+	checkpoints  [][]byte
+	error        func() error
+	ensureBefore time.Duration
 }
 
 func (b *fakeStateBackend) Checkpoint(data []byte) error {
@@ -186,10 +203,6 @@ func (b *fakeStateBackend) EnsureBefore(d time.Duration) {
 	b.ensureBefore = d
 }
 
-func (b *fakeStateBackend) RequestRestart(t state.RestartType) {
-	b.restartRequested = true
-}
-
 func (ss *stateSuite) TestImplicitCheckpointAndRead(c *C) {
 	b := new(fakeStateBackend)
 	st := state.New(b)
@@ -914,7 +927,7 @@ func (ss *stateSuite) TestPruneMaxChangesHonored(c *C) {
 	c.Assert(st.Changes(), HasLen, 10)
 
 	// one extra change that just now entered ready state
-	chg := st.NewChange(fmt.Sprintf("chg99"), "so-ready")
+	chg := st.NewChange("chg99", "so-ready")
 	t := st.NewTask("foo", "so-ready")
 	when := 1 * time.Second
 	state.MockChangeTimes(chg, time.Now().Add(-when), time.Now().Add(-when))
@@ -963,66 +976,6 @@ func (ss *stateSuite) TestPruneHonorsStartOperationTime(c *C) {
 	c.Check(chg.Status(), Equals, state.HoldStatus)
 }
 
-func (ss *stateSuite) TestRequestRestart(c *C) {
-	b := new(fakeStateBackend)
-	st := state.New(b)
-
-	ok, t := st.Restarting()
-	c.Check(ok, Equals, false)
-	c.Check(t, Equals, state.RestartUnset)
-
-	st.RequestRestart(state.RestartDaemon)
-
-	c.Check(b.restartRequested, Equals, true)
-
-	ok, t = st.Restarting()
-	c.Check(ok, Equals, true)
-	c.Check(t, Equals, state.RestartDaemon)
-}
-
-func (ss *stateSuite) TestRequestRestartSystemAndVerifyReboot(c *C) {
-	b := new(fakeStateBackend)
-	st := state.New(b)
-
-	st.Lock()
-	err := st.VerifyReboot("boot-id-1")
-	st.Unlock()
-	c.Assert(err, IsNil)
-
-	ok, t := st.Restarting()
-	c.Check(ok, Equals, false)
-	c.Check(t, Equals, state.RestartUnset)
-
-	st.Lock()
-	st.RequestRestart(state.RestartSystem)
-	st.Unlock()
-
-	c.Check(b.restartRequested, Equals, true)
-
-	ok, t = st.Restarting()
-	c.Check(ok, Equals, true)
-	c.Check(t, Equals, state.RestartSystem)
-
-	var fromBootID string
-	st.Lock()
-	c.Check(st.Get("system-restart-from-boot-id", &fromBootID), IsNil)
-	st.Unlock()
-	c.Check(fromBootID, Equals, "boot-id-1")
-
-	st.Lock()
-	err = st.VerifyReboot("boot-id-1")
-	st.Unlock()
-	c.Check(err, Equals, state.ErrExpectedReboot)
-
-	st.Lock()
-	err = st.VerifyReboot("boot-id-2")
-	st.Unlock()
-	c.Assert(err, IsNil)
-	st.Lock()
-	c.Check(st.Get("system-restart-from-boot-id", &fromBootID), Equals, state.ErrNoState)
-	st.Unlock()
-}
-
 func (ss *stateSuite) TestReadStateInitsCache(c *C) {
 	st, err := state.ReadState(nil, bytes.NewBufferString("{}"))
 	c.Assert(err, IsNil)
@@ -1050,3 +1003,17 @@ func (ss *stateSuite) TestTimingsSupport(c *C) {
 	c.Assert(err, IsNil)
 	c.Check(tims, DeepEquals, []int{1, 2, 3})
 }
+
+func (ss *stateSuite) TestNoStateErrorIs(c *C) {
+	err := &state.NoStateError{Key: "foo"}
+	c.Assert(err, testutil.ErrorIs, &state.NoStateError{})
+	c.Assert(err, testutil.ErrorIs, &state.NoStateError{Key: "bar"})
+	c.Assert(err, testutil.ErrorIs, state.ErrNoState)
+}
+
+func (ss *stateSuite) TestNoStateErrorString(c *C) {
+	err := &state.NoStateError{}
+	c.Assert(err.Error(), Equals, `no state entry for key`)
+	err.Key = "foo"
+	c.Assert(err.Error(), Equals, `no state entry for key "foo"`)
+}
