@@ -760,6 +760,115 @@ func (s *writerSuite) TestDownloadedCheckTypeCore(c *C) {
 	c.Check(err, ErrorMatches, `core snap has unexpected type: base`)
 }
 
+func (s *writerSuite) TestSeedSnapsWriteMetaCore16(c *C) {
+	model := s.Brands.Model("my-brand", "my-model", map[string]interface{}{
+		"display-name": "my model",
+		"architecture": "amd64",
+		"gadget":       "pc",
+		"kernel":       "pc-kernel",
+	})
+
+	// the minimum set of snaps
+	s.makeSnap(c, "core", "")
+	s.makeSnap(c, "pc-kernel", "")
+	s.makeSnap(c, "pc", "")
+
+	w, err := seedwriter.New(model, s.opts)
+	c.Assert(err, IsNil)
+
+	err = w.SetOptionsSnaps([]*seedwriter.OptionsSnap{{Name: "pc", Channel: "edge"}})
+	c.Assert(err, IsNil)
+
+	_, err = w.Start(s.db, s.newFetcher)
+	c.Assert(err, IsNil)
+
+	snaps, err := w.SnapsToDownload()
+	c.Assert(err, IsNil)
+	c.Check(snaps, HasLen, 3)
+
+	for _, sn := range snaps {
+		s.fillDownloadedSnap(c, w, sn)
+	}
+
+	complete, err := w.Downloaded()
+	c.Assert(err, IsNil)
+	c.Check(complete, Equals, true)
+
+	err = w.SeedSnaps(nil)
+	c.Assert(err, IsNil)
+
+	err = w.WriteMeta()
+	c.Assert(err, IsNil)
+
+	// check seed
+	seedYaml, err := seedwriter.InternalReadSeedYaml(filepath.Join(s.opts.SeedDir, "seed.yaml"))
+	c.Assert(err, IsNil)
+
+	c.Check(seedYaml.Snaps, HasLen, 3)
+
+	// check the files are in place
+	for i, name := range []string{"core", "pc-kernel", "pc"} {
+		info := s.AssertedSnapInfo(name)
+
+		fn := info.Filename()
+		p := filepath.Join(s.opts.SeedDir, "snaps", fn)
+		c.Check(p, testutil.FilePresent)
+
+		channel := "stable"
+		if name == "pc" {
+			channel = "edge"
+		}
+
+		c.Check(seedYaml.Snaps[i], DeepEquals, &seedwriter.InternalSnap16{
+			Name:    info.SnapName(),
+			SnapID:  info.SnapID,
+			Channel: channel,
+			File:    fn,
+			Contact: info.Contact(),
+		})
+	}
+
+	l, err := ioutil.ReadDir(filepath.Join(s.opts.SeedDir, "snaps"))
+	c.Assert(err, IsNil)
+	c.Check(l, HasLen, 3)
+
+	// check assertions
+	seedAssertsDir := filepath.Join(s.opts.SeedDir, "assertions")
+	storeAccountKeyPK := s.StoreSigning.StoreAccountKey("").PublicKeyID()
+	brandAcctKeyPK := s.Brands.AccountKey("my-brand").PublicKeyID()
+
+	for _, fn := range []string{"model", brandAcctKeyPK + ".account-key", "my-brand.account", storeAccountKeyPK + ".account-key"} {
+		p := filepath.Join(seedAssertsDir, fn)
+		c.Check(p, testutil.FilePresent)
+	}
+
+	c.Check(filepath.Join(seedAssertsDir, "model"), testutil.FileEquals, asserts.Encode(model))
+
+	acct := seedtest.ReadAssertions(c, filepath.Join(seedAssertsDir, "my-brand.account"))
+	c.Assert(acct, HasLen, 1)
+	c.Check(acct[0].Type(), Equals, asserts.AccountType)
+	c.Check(acct[0].HeaderString("account-id"), Equals, "my-brand")
+
+	// check the snap assertions are also in place
+	for _, snapName := range []string{"pc-kernel", "core", "pc"} {
+		p := filepath.Join(seedAssertsDir, fmt.Sprintf("16,%s.snap-declaration", s.AssertedSnapID(snapName)))
+		decl := seedtest.ReadAssertions(c, p)
+		c.Assert(decl, HasLen, 1)
+		c.Check(decl[0].Type(), Equals, asserts.SnapDeclarationType)
+		c.Check(decl[0].HeaderString("snap-name"), Equals, snapName)
+		p = filepath.Join(seedAssertsDir, fmt.Sprintf("%s.snap-revision", s.AssertedSnapRevision(snapName).SnapSHA3_384()))
+		rev := seedtest.ReadAssertions(c, p)
+		c.Assert(rev, HasLen, 1)
+		c.Check(rev[0].Type(), Equals, asserts.SnapRevisionType)
+		c.Check(rev[0].HeaderString("snap-id"), Equals, s.AssertedSnapID(snapName))
+	}
+
+	// validity check of seedtest helper
+	const usesSnapd = false
+	// core seeds do not use system labels
+	seedtest.ValidateSeed(c, s.opts.SeedDir, "", usesSnapd, s.StoreSigning.Trusted)
+}
+
 func (s *writerSuite) TestSeedSnapsWriteMetaCore18(c *C) {
 	model := s.Brands.Model("my-brand", "my-model", map[string]interface{}{
 		"display-name":   "my model",
@@ -790,7 +899,7 @@ func (s *writerSuite) TestSeedSnapsWriteMetaCore18(c *C) {
 	c.Assert(err, IsNil)
 	c.Check(snaps, HasLen, 6)
 
-	s.AssertedSnapInfo("cont-producer").Contact = "author@cont-producer.net"
+	s.AssertedSnapInfo("cont-producer").EditedContact = "mailto:author@cont-producer.net"
 	for _, sn := range snaps {
 		s.fillDownloadedSnap(c, w, sn)
 	}
@@ -832,7 +941,7 @@ func (s *writerSuite) TestSeedSnapsWriteMetaCore18(c *C) {
 			SnapID:  info.SnapID,
 			Channel: channel,
 			File:    fn,
-			Contact: info.Contact,
+			Contact: info.Contact(),
 		})
 	}
 
@@ -870,6 +979,11 @@ func (s *writerSuite) TestSeedSnapsWriteMetaCore18(c *C) {
 		c.Check(rev[0].Type(), Equals, asserts.SnapRevisionType)
 		c.Check(rev[0].HeaderString("snap-id"), Equals, s.AssertedSnapID(snapName))
 	}
+
+	// validity check of seedtest helper
+	const usesSnapd = true
+	// core18 seeds do not use system labels
+	seedtest.ValidateSeed(c, s.opts.SeedDir, "", usesSnapd, s.StoreSigning.Trusted)
 }
 
 func (s *writerSuite) TestSeedSnapsWriteMetaCore18StoreAssertion(c *C) {
@@ -991,7 +1105,7 @@ func (s *writerSuite) TestLocalSnapsCore18FullUse(c *C) {
 	c.Assert(localSnaps, HasLen, 5)
 
 	for _, sn := range localSnaps {
-		si, aRefs, err := seedwriter.DeriveSideInfo(sn.Path, tf, s.db)
+		si, aRefs, err := seedwriter.DeriveSideInfo(sn.Path, model, tf, s.db)
 		if !asserts.IsNotFound(err) {
 			c.Assert(err, IsNil)
 		}
@@ -1412,7 +1526,7 @@ func (s *writerSuite) TestSeedSnapsWriteMetaClassicWithCore(c *C) {
 			SnapID:  info.SnapID,
 			Channel: "stable",
 			File:    fn,
-			Contact: info.Contact,
+			Contact: info.Contact(),
 		})
 	}
 }
@@ -1468,7 +1582,7 @@ func (s *writerSuite) TestSeedSnapsWriteMetaClassicSnapdOnly(c *C) {
 			SnapID:  info.SnapID,
 			Channel: "stable",
 			File:    fn,
-			Contact: info.Contact,
+			Contact: info.Contact(),
 		})
 	}
 }
@@ -1522,7 +1636,7 @@ func (s *writerSuite) TestSeedSnapsWriteMetaExtraSnaps(c *C) {
 	c.Assert(err, IsNil)
 	c.Assert(snaps, HasLen, 6)
 
-	s.AssertedSnapInfo("cont-producer").Contact = "author@cont-producer.net"
+	s.AssertedSnapInfo("cont-producer").EditedContact = "mailto:author@cont-producer.net"
 	for _, sn := range snaps {
 		s.fillDownloadedSnap(c, w, sn)
 	}
@@ -1585,7 +1699,7 @@ func (s *writerSuite) TestSeedSnapsWriteMetaExtraSnaps(c *C) {
 			SnapID:  info.SnapID,
 			Channel: channel,
 			File:    fn,
-			Contact: info.Contact,
+			Contact: info.Contact(),
 		})
 	}
 
@@ -1646,7 +1760,7 @@ func (s *writerSuite) TestSeedSnapsWriteMetaLocalExtraSnaps(c *C) {
 	c.Assert(localSnaps, HasLen, 1)
 
 	for _, sn := range localSnaps {
-		si, aRefs, err := seedwriter.DeriveSideInfo(sn.Path, tf, s.db)
+		si, aRefs, err := seedwriter.DeriveSideInfo(sn.Path, model, tf, s.db)
 		if !asserts.IsNotFound(err) {
 			c.Assert(err, IsNil)
 		}
@@ -1665,7 +1779,7 @@ func (s *writerSuite) TestSeedSnapsWriteMetaLocalExtraSnaps(c *C) {
 	c.Assert(err, IsNil)
 	c.Assert(snaps, HasLen, 6)
 
-	s.AssertedSnapInfo("cont-producer").Contact = "author@cont-producer.net"
+	s.AssertedSnapInfo("cont-producer").EditedContact = "mailto:author@cont-producer.net"
 	for _, sn := range snaps {
 		s.fillDownloadedSnap(c, w, sn)
 	}
@@ -1739,7 +1853,7 @@ func (s *writerSuite) TestSeedSnapsWriteMetaLocalExtraSnaps(c *C) {
 			SnapID:     info.SnapID,
 			Channel:    channel,
 			File:       fn,
-			Contact:    info.Contact,
+			Contact:    info.Contact(),
 			Unasserted: unasserted,
 		})
 	}
@@ -1799,7 +1913,7 @@ func (s *writerSuite) TestSeedSnapsWriteMetaCore20(c *C) {
 		},
 	})
 
-	// sanity
+	// validity
 	c.Assert(model.Grade(), Equals, asserts.ModelSigned)
 
 	s.makeSnap(c, "snapd", "")
@@ -1821,7 +1935,7 @@ func (s *writerSuite) TestSeedSnapsWriteMetaCore20(c *C) {
 	c.Assert(err, IsNil)
 	c.Check(snaps, HasLen, 7)
 
-	s.AssertedSnapInfo("cont-producer").Contact = "author@cont-producer.net"
+	s.AssertedSnapInfo("cont-producer").EditedContact = "mailto:author@cont-producer.net"
 	s.AssertedSnapInfo("cont-consumer").Private = true
 	for _, sn := range snaps {
 		// check the used channel at this level because in the
@@ -1943,11 +2057,16 @@ func (s *writerSuite) TestSeedSnapsWriteMetaCore20(c *C) {
 			"private": true,
 		},
 		s.AssertedSnapID("cont-producer"): {
-			"contact": "author@cont-producer.net",
+			"contact": "mailto:author@cont-producer.net",
 		},
 	})
 
 	c.Check(filepath.Join(systemDir, "options.yaml"), testutil.FileAbsent)
+
+	// validity check of seedtest helper
+	const usesSnapd = true
+	seedtest.ValidateSeed(c, s.opts.SeedDir, s.opts.Label, usesSnapd,
+		s.StoreSigning.Trusted)
 }
 
 func (s *writerSuite) TestCore20InvalidLabel(c *C) {
@@ -2012,7 +2131,7 @@ func (s *writerSuite) TestDownloadedCore20CheckBase(c *C) {
 		},
 	})
 
-	// sanity
+	// validity
 	c.Assert(model.Grade(), Equals, asserts.ModelSigned)
 
 	s.makeSnap(c, "snapd", "")
@@ -2059,7 +2178,7 @@ func (s *writerSuite) TestDownloadedCore20CheckBaseModes(c *C) {
 		},
 	})
 
-	// sanity
+	// validity
 	c.Assert(model.Grade(), Equals, asserts.ModelSigned)
 
 	s.makeSnap(c, "snapd", "")
@@ -2107,7 +2226,7 @@ func (s *writerSuite) TestDownloadedCore20CheckBaseEphemeralOK(c *C) {
 		},
 	})
 
-	// sanity
+	// validity
 	c.Assert(model.Grade(), Equals, asserts.ModelSigned)
 
 	s.makeSnap(c, "snapd", "")
@@ -2156,8 +2275,9 @@ func (s *writerSuite) TestDownloadedCore20CheckBaseCoreXX(c *C) {
 		{[]interface{}{requiredBaseCore16Ent}, `cannot add snap "required-base-core16" without also adding its base "core16" \(or "core"\) explicitly`},
 	}
 
-	s.opts.Label = "20191003"
-	for _, t := range tests {
+	baseLabel := "20191003"
+	for idx, t := range tests {
+		s.opts.Label = fmt.Sprintf("%s%d", baseLabel, idx)
 		snaps := []interface{}{
 			map[string]interface{}{
 				"name":            "pc-kernel",
@@ -2228,7 +2348,7 @@ func (s *writerSuite) TestDownloadedCore20MissingDefaultProviderModes(c *C) {
 		},
 	})
 
-	// sanity
+	// validity
 	c.Assert(model.Grade(), Equals, asserts.ModelSigned)
 
 	s.makeSnap(c, "snapd", "")
@@ -2276,8 +2396,6 @@ func (s *writerSuite) TestCore20NonDangerousDisallowedDevmodeSnaps(c *C) {
 
 	s.opts.Label = "20191107"
 
-	const expectedErr = `cannot override channels, add local snaps or extra snaps with a model of grade higher than dangerous`
-
 	w, err := seedwriter.New(model, s.opts)
 	c.Assert(err, IsNil)
 
@@ -2326,7 +2444,7 @@ func (s *writerSuite) TestCore20NonDangerousDisallowedOptionsSnaps(c *C) {
 
 	pcFn := s.makeLocalSnap(c, "pc")
 
-	s.opts.Label = "20191107"
+	baseLabel := "20191107"
 
 	tests := []struct {
 		optSnap *seedwriter.OptionsSnap
@@ -2338,7 +2456,8 @@ func (s *writerSuite) TestCore20NonDangerousDisallowedOptionsSnaps(c *C) {
 
 	const expectedErr = `cannot override channels, add devmode snaps, local snaps, or extra snaps with a model of grade higher than dangerous`
 
-	for _, t := range tests {
+	for idx, t := range tests {
+		s.opts.Label = fmt.Sprintf("%s%d", baseLabel, idx)
 		w, err := seedwriter.New(model, s.opts)
 		c.Assert(err, IsNil)
 
@@ -2357,7 +2476,7 @@ func (s *writerSuite) TestCore20NonDangerousDisallowedOptionsSnaps(c *C) {
 			c.Assert(localSnaps, HasLen, 1)
 
 			for _, sn := range localSnaps {
-				si, aRefs, err := seedwriter.DeriveSideInfo(sn.Path, tf, s.db)
+				si, aRefs, err := seedwriter.DeriveSideInfo(sn.Path, model, tf, s.db)
 				if !asserts.IsNotFound(err) {
 					c.Assert(err, IsNil)
 				}
@@ -2445,7 +2564,7 @@ func (s *writerSuite) TestSeedSnapsWriteMetaCore20LocalSnaps(c *C) {
 		},
 	})
 
-	// sanity
+	// validity
 	c.Assert(model.Grade(), Equals, asserts.ModelDangerous)
 
 	s.makeSnap(c, "snapd", "")
@@ -2469,7 +2588,7 @@ func (s *writerSuite) TestSeedSnapsWriteMetaCore20LocalSnaps(c *C) {
 	c.Assert(localSnaps, HasLen, 1)
 
 	for _, sn := range localSnaps {
-		_, _, err := seedwriter.DeriveSideInfo(sn.Path, tf, s.db)
+		_, _, err := seedwriter.DeriveSideInfo(sn.Path, model, tf, s.db)
 		c.Assert(asserts.IsNotFound(err), Equals, true)
 		f, err := snapfile.Open(sn.Path)
 		c.Assert(err, IsNil)
@@ -2572,7 +2691,7 @@ func (s *writerSuite) TestSeedSnapsWriteMetaCore20ChannelOverrides(c *C) {
 		},
 	})
 
-	// sanity
+	// validity
 	c.Assert(model.Grade(), Equals, asserts.ModelDangerous)
 
 	s.makeSnap(c, "snapd", "")
@@ -2676,7 +2795,7 @@ func (s *writerSuite) TestSeedSnapsWriteMetaCore20ModelOverrideSnapd(c *C) {
 			}},
 	})
 
-	// sanity
+	// validity
 	c.Assert(model.Grade(), Equals, asserts.ModelSigned)
 
 	s.makeSnap(c, "snapd", "")
@@ -2776,7 +2895,7 @@ func (s *writerSuite) TestSnapsToDownloadCore20OptionalSnaps(c *C) {
 			}},
 	})
 
-	// sanity
+	// validity
 	c.Assert(model.Grade(), Equals, asserts.ModelSigned)
 
 	s.makeSnap(c, "snapd", "")
@@ -2824,7 +2943,7 @@ func (s *writerSuite) TestSeedSnapsWriteMetaCore20ExtraSnaps(c *C) {
 			}},
 	})
 
-	// sanity
+	// validity
 	c.Assert(model.Grade(), Equals, asserts.ModelDangerous)
 
 	s.makeSnap(c, "snapd", "")
@@ -2850,7 +2969,7 @@ func (s *writerSuite) TestSeedSnapsWriteMetaCore20ExtraSnaps(c *C) {
 	c.Assert(localSnaps, HasLen, 1)
 
 	for _, sn := range localSnaps {
-		_, _, err := seedwriter.DeriveSideInfo(sn.Path, tf, s.db)
+		_, _, err := seedwriter.DeriveSideInfo(sn.Path, model, tf, s.db)
 		c.Assert(asserts.IsNotFound(err), Equals, true)
 		f, err := snapfile.Open(sn.Path)
 		c.Assert(err, IsNil)
@@ -2998,7 +3117,7 @@ func (s *writerSuite) TestSeedSnapsWriteMetaCore20LocalAssertedSnaps(c *C) {
 			}},
 	})
 
-	// sanity
+	// validity
 	c.Assert(model.Grade(), Equals, asserts.ModelDangerous)
 
 	s.makeSnap(c, "snapd", "")
@@ -3022,7 +3141,7 @@ func (s *writerSuite) TestSeedSnapsWriteMetaCore20LocalAssertedSnaps(c *C) {
 	c.Assert(localSnaps, HasLen, 2)
 
 	for _, sn := range localSnaps {
-		si, aRefs, err := seedwriter.DeriveSideInfo(sn.Path, tf, s.db)
+		si, aRefs, err := seedwriter.DeriveSideInfo(sn.Path, model, tf, s.db)
 		c.Assert(err, IsNil)
 		f, err := snapfile.Open(sn.Path)
 		c.Assert(err, IsNil)
@@ -3146,7 +3265,7 @@ func (s *writerSuite) TestSeedSnapsWriteMetaCore20SignedLocalAssertedSnaps(c *C)
 	c.Assert(localSnaps, HasLen, 1)
 
 	for _, sn := range localSnaps {
-		si, aRefs, err := seedwriter.DeriveSideInfo(sn.Path, tf, s.db)
+		si, aRefs, err := seedwriter.DeriveSideInfo(sn.Path, model, tf, s.db)
 		c.Assert(err, IsNil)
 		f, err := snapfile.Open(sn.Path)
 		c.Assert(err, IsNil)
@@ -3201,3 +3320,36 @@ func (s *writerSuite) TestSeedSnapsWriteMetaCore20SignedLocalAssertedSnaps(c *C)
 	// no options file was created
 	c.Check(filepath.Join(systemDir, "options.yaml"), testutil.FileAbsent)
 }
+
+func (s *writerSuite) TestSeedSnapsWriteCore20ErrWhenDirExists(c *C) {
+	model := s.Brands.Model("my-brand", "my-model", map[string]interface{}{
+		"display-name": "my model",
+		"architecture": "amd64",
+		"base":         "core20",
+		"grade":        "signed",
+		"snaps": []interface{}{
+			map[string]interface{}{
+				"name":            "pc-kernel",
+				"id":              s.AssertedSnapID("pc-kernel"),
+				"type":            "kernel",
+				"default-channel": "20",
+			},
+			map[string]interface{}{
+				"name":            "pc",
+				"id":              s.AssertedSnapID("pc"),
+				"type":            "gadget",
+				"default-channel": "20",
+			}},
+	})
+
+	err := os.MkdirAll(filepath.Join(s.opts.SeedDir, "systems", "1234"), 0755)
+	c.Assert(err, IsNil)
+	s.opts.Label = "1234"
+	w, err := seedwriter.New(model, s.opts)
+	c.Assert(err, IsNil)
+	c.Assert(w, NotNil)
+
+	_, err = w.Start(s.db, s.newFetcher)
+	c.Assert(err, ErrorMatches, `system "1234" already exists`)
+	c.Assert(seedwriter.IsSytemDirectoryExistsError(err), Equals, true)
+}
