@@ -199,14 +199,17 @@ distro_install_package() {
         ubuntu-*|debian-*)
             # shellcheck disable=SC2086
             quiet eatmydata apt-get install $APT_FLAGS -y "${pkg_names[@]}"
+            retval=$?
             ;;
         amazon-*|centos-7-*)
             # shellcheck disable=SC2086
             quiet yum -y install $YUM_FLAGS "${pkg_names[@]}"
+            retval=$?
             ;;
         fedora-*|centos-*)
             # shellcheck disable=SC2086
             quiet dnf -y --refresh install $DNF_FLAGS "${pkg_names[@]}"
+            retval=$?
             ;;
         opensuse-*)
             # packages may be downgraded in the repositories, which would be
@@ -219,10 +222,12 @@ distro_install_package() {
 
             # shellcheck disable=SC2086
             quiet zypper install -y --allow-downgrade --force-resolution $ZYPPER_FLAGS "${pkg_names[@]}"
+            retval=$?
             ;;
         arch-*)
             # shellcheck disable=SC2086
             pacman -Suq --needed --noconfirm "${pkg_names[@]}"
+            retval=$?
             ;;
         *)
             echo "ERROR: Unsupported distribution $SPREAD_SYSTEM"
@@ -230,6 +235,10 @@ distro_install_package() {
             ;;
     esac
     test "$orig_xtrace" = on && set -x
+    # pass any errors up
+    if [ "$retval" != "0" ]; then
+        return $retval
+    fi
 }
 
 distro_purge_package() {
@@ -383,6 +392,14 @@ distro_install_build_snapd(){
         if os.query is-trusty && [ "$SPREAD_REBOOT" = 0 ]; then
             REBOOT
         fi
+    elif [ -n "$PPA_GPG_KEY" ] && [ -n "$PPA_SOURCE_LINE" ]; then
+        echo "$PPA_GPG_KEY" | apt-key add -
+        echo "${PPA_SOURCE_LINE//"YOUR_UBUNTU_VERSION_HERE"/"$(lsb_release -c -s)"}" >> /etc/apt/sources.list
+        apt update
+        apt install -y snapd
+
+        # Double check that it really comes from the PPA
+        apt show snapd | MATCH "APT-Sources: http.*private-ppa\.launchpad(content)?\.net"
     elif [ -n "$PPA_VALIDATION_NAME" ]; then
         apt install -y snapd
         add-apt-repository -y "$PPA_VALIDATION_NAME"
@@ -392,7 +409,7 @@ distro_install_build_snapd(){
         apt update
 
         # Double check that it really comes from the PPA
-        apt show snapd | grep "APT-Sources: http.*ppa.launchpad.net"
+        apt show snapd | MATCH "APT-Sources: http.*ppa\.launchpad(content)?\.net"
     else
         packages=
         case "$SPREAD_SYSTEM" in
@@ -435,10 +452,21 @@ distro_install_build_snapd(){
                 # to create sockets with incorrect context, this installation of
                 # socket activated snaps fails, see:
                 # https://bugzilla.redhat.com/show_bug.cgi?id=1660141
+                # https://bugzilla.redhat.com/show_bug.cgi?id=1197886
                 # https://github.com/systemd/systemd/issues/9997
                 systemctl daemon-reexec
                 ;;
         esac
+        case "$SPREAD_SYSTEM" in
+            fedora-*)
+                # the problem with SELinux policy also affects the user instance
+                # in 248, see:
+                # https://bugzilla.redhat.com/show_bug.cgi?id=1960576
+                # note, this fixes it for the root user only, the test user
+                # session is created dynamically as needed
+                systemctl --user daemon-reexec
+                ;;
+        esac
 
         if os.query is-arch-linux; then
             # Arch policy does not allow calling daemon-reloads in package
@@ -451,7 +479,7 @@ distro_install_build_snapd(){
             fi
         fi
 
-        if [[ "$SPREAD_SYSTEM" == opensuse-tumbleweed-* ]]; then
+        if os.query is-opensuse || os.query is-arch-linux; then
             # Package installation applies vendor presets only, which leaves
             # snapd.apparmor disabled.
             systemctl enable --now snapd.apparmor.service
@@ -485,10 +513,12 @@ distro_get_package_extension() {
 
 pkg_dependencies_ubuntu_generic(){
     echo "
+        python3
         autoconf
         automake
         autotools-dev
         build-essential
+        ca-certificates
         clang
         curl
         devscripts
@@ -504,6 +534,7 @@ pkg_dependencies_ubuntu_generic(){
         libseccomp-dev
         libudev-dev
         man
+        mtools
         netcat-openbsd
         pkg-config
         python3-docutils
@@ -518,7 +549,7 @@ pkg_dependencies_ubuntu_classic(){
     echo "
         avahi-daemon
         cups
-        dbus-x11
+        fish
         fontconfig
         gnome-keyring
         jq
@@ -537,18 +568,7 @@ pkg_dependencies_ubuntu_classic(){
 
     case "$SPREAD_SYSTEM" in
         ubuntu-14.04-*)
-                pkg_linux_image_extra
-            ;;
-        ubuntu-16.04-32)
-            echo "
-                dbus-user-session
-                gccgo-6
-                evolution-data-server
-                fwupd
-                gnome-online-accounts
-                packagekit
-                "
-                pkg_linux_image_extra
+            pkg_linux_image_extra
             ;;
         ubuntu-16.04-64)
             echo "
@@ -566,10 +586,22 @@ pkg_dependencies_ubuntu_classic(){
                 "
                 pkg_linux_image_extra
             ;;
+        ubuntu-18.04-32)
+            echo "
+                dbus-user-session
+                gccgo-6
+                evolution-data-server
+                fwupd
+                gnome-online-accounts
+                packagekit
+                "
+                pkg_linux_image_extra
+            ;;
         ubuntu-18.04-64)
             echo "
                 dbus-user-session
                 gccgo-8
+                gperf
                 evolution-data-server
                 fwupd
                 packagekit
@@ -578,15 +610,27 @@ pkg_dependencies_ubuntu_classic(){
             ;;
         ubuntu-20.04-64)
             echo "
+                dbus-user-session
                 evolution-data-server
+                fwupd
                 gccgo-9
+                libvirt-daemon-system
                 packagekit
+                qemu-kvm
                 qemu-utils
                 shellcheck
                 "
             ;;
-        ubuntu-20.10-64)
+        ubuntu-22.04-64)
+            # bpftool is part of linux-tools package
             echo "
+                dbus-user-session
+                fwupd
+                golang
+                libvirt-daemon-system
+                linux-tools-$(uname -r)
+                lz4
+                qemu-kvm
                 qemu-utils
                 "
             ;;
@@ -599,7 +643,6 @@ pkg_dependencies_ubuntu_classic(){
             echo "
                 autopkgtest
                 debootstrap
-                dbus-user-session
                 eatmydata
                 evolution-data-server
                 fwupd
@@ -613,6 +656,15 @@ pkg_dependencies_ubuntu_classic(){
                 "
             ;;
     esac
+    case "$SPREAD_SYSTEM" in
+        debian-11-*|debian-sid-*)
+            echo "
+                 bpftool
+                 strace
+                 systemd-timesyncd
+                 "
+            ;;
+    esac
 }
 
 pkg_linux_image_extra (){
@@ -635,8 +687,10 @@ pkg_dependencies_ubuntu_core(){
         pkg_linux_image_extra
 }
 
-pkg_dependencies_fedora(){
+pkg_dependencies_fedora_centos_common(){
     echo "
+        python3
+        bpftool
         clang
         curl
         dbus-x11
@@ -653,10 +707,10 @@ pkg_dependencies_fedora(){
         nmap-ncat
         nfs-utils
         PackageKit
+        polkit
         python3-yaml
         python3-dbus
         python3-gobject
-        redhat-lsb-core
         rpm-build
         udisks2
         upower
@@ -665,13 +719,27 @@ pkg_dependencies_fedora(){
         strace
         zsh
         "
+    if ! os.query is-centos 9; then
+        echo "
+            fish
+            redhat-lsb-core
+        "
+    fi
+}
+
+pkg_dependencies_fedora(){
+    echo "
+         libcap-static
+        "
 }
 
 pkg_dependencies_amazon(){
     echo "
+        python3
         curl
         dbus-x11
         expect
+        fish
         fontconfig
         fwupd
         git
@@ -679,6 +747,7 @@ pkg_dependencies_amazon(){
         grub2-tools
         jq
         iptables-services
+        libcap-static
         man
         nc
         net-tools
@@ -696,21 +765,31 @@ pkg_dependencies_amazon(){
 
 pkg_dependencies_opensuse(){
     echo "
+        python3
         apparmor-profiles
+        audit
+        bash-completion
+        bpftool
         clang
         curl
+        dbus-1-python3
         evolution-data-server
         expect
+        fish
         fontconfig
         fwupd
         git
         golang-packaging
+        iptables
         jq
         lsb-release
         man
+        man-pages
         nfs-kernel-server
+        nss-mdns
         PackageKit
         python3-yaml
+        strace
         netcat-openbsd
         osc
         udisks2
@@ -724,12 +803,15 @@ pkg_dependencies_opensuse(){
 
 pkg_dependencies_arch(){
     echo "
+    apparmor
     base-devel
     bash-completion
+    bpf
     clang
     curl
     evolution-data-server
     expect
+    fish
     fontconfig
     fwupd
     git
@@ -757,7 +839,6 @@ pkg_dependencies_arch(){
     xdg-user-dirs
     xdg-utils
     xfsprogs
-    apparmor
     zsh
     "
 }
@@ -775,7 +856,11 @@ pkg_dependencies(){
         amazon-*|centos-7-*)
             pkg_dependencies_amazon
             ;;
-        fedora-*|centos-*)
+        centos-*)
+            pkg_dependencies_fedora_centos_common
+            ;;
+        fedora-*)
+            pkg_dependencies_fedora_centos_common
             pkg_dependencies_fedora
             ;;
         opensuse-*)
