@@ -2,8 +2,6 @@
 
 set -eux
 
-# shellcheck source=tests/lib/dirs.sh
-. "$TESTSLIB/dirs.sh"
 # shellcheck source=tests/lib/snaps.sh
 . "$TESTSLIB/snaps.sh"
 # shellcheck source=tests/lib/pkgdb.sh
@@ -57,6 +55,9 @@ ensure_jq() {
     elif os.query is-core20; then
         snap install --devmode --edge jq-core20
         snap alias jq-core20.jq jq
+    elif os.query is-core22; then
+        snap install --devmode --edge jq-core22
+        snap alias jq-core22.jq jq
     else
         snap install --devmode jq
     fi
@@ -68,8 +69,7 @@ disable_refreshes() {
 
     echo "Modify state to make it look like the last refresh just happened"
     systemctl stop snapd.socket snapd.service
-    jq ".data[\"last-refresh\"] = \"$(date +%Y-%m-%dT%H:%M:%S%:z)\"" /var/lib/snapd/state.json > /var/lib/snapd/state.json.new
-    mv /var/lib/snapd/state.json.new /var/lib/snapd/state.json
+    "$TESTSTOOLS"/snapd-state prevent-autorefresh
     systemctl start snapd.socket snapd.service
 
     echo "Minimize risk of hitting refresh schedule"
@@ -80,6 +80,7 @@ disable_refreshes() {
     snap remove --purge jq
     snap remove --purge jq-core18
     snap remove --purge jq-core20
+    snap remove --purge jq-core22
 }
 
 setup_systemd_snapd_overrides() {
@@ -99,6 +100,20 @@ EOF
     systemctl start snapd.service
 }
 
+# setup_experimental_features enables experimental snapd features passed
+# via optional EXPERIMENTAL_FEATURES environment variable. The features must be
+# separated by commas and "experimental." prefixes should be omitted.
+setup_experimental_features() {
+    if [ -n "$EXPERIMENTAL_FEATURES" ]; then
+        echo "$EXPERIMENTAL_FEATURES" | while IFS="," read -r FEATURE; do
+            echo "Enabling feature experimental.$FEATURE"
+            snap set system "experimental.$FEATURE"=true
+        done
+    else
+        echo "There are no experimental snapd features to enable"
+    fi
+}
+
 update_core_snap_for_classic_reexec() {
     # it is possible to disable this to test that snapd (the deb) works
     # fine with whatever is in the core snap
@@ -111,6 +126,8 @@ update_core_snap_for_classic_reexec() {
     # we re-exec by default.
     # To accomplish that, we'll just unpack the core we just grabbed,
     # shove the new snap-exec and snapctl in there, and repack it.
+    SNAP_MOUNT_DIR="$(os.paths snap-mount-dir)"
+    LIBEXEC_DIR="$(os.paths libexec-dir)"
 
     # First of all, unmount the core
     core="$(readlink -f "$SNAP_MOUNT_DIR/core/current" || readlink -f "$SNAP_MOUNT_DIR/ubuntu-core/current")"
@@ -122,7 +139,7 @@ update_core_snap_for_classic_reexec() {
     # clean the old snapd binaries, just in case
     rm squashfs-root/usr/lib/snapd/* squashfs-root/usr/bin/snap
     # and copy in the current libexec
-    cp -a "$LIBEXECDIR"/snapd/* squashfs-root/usr/lib/snapd/
+    cp -a "$LIBEXEC_DIR"/snapd/* squashfs-root/usr/lib/snapd/
     # also the binaries themselves
     cp -a /usr/bin/snap squashfs-root/usr/bin/
     # make sure bin/snapctl is a symlink to lib/
@@ -185,7 +202,7 @@ update_core_snap_for_classic_reexec() {
     }
 
     # Make sure we're running with the correct copied bits
-    for p in "$LIBEXECDIR/snapd/snap-exec" "$LIBEXECDIR/snapd/snap-confine" "$LIBEXECDIR/snapd/snap-discard-ns" "$LIBEXECDIR/snapd/snapd" "$LIBEXECDIR/snapd/snap-update-ns"; do
+    for p in "$LIBEXEC_DIR/snapd/snap-exec" "$LIBEXEC_DIR/snapd/snap-confine" "$LIBEXEC_DIR/snapd/snap-discard-ns" "$LIBEXEC_DIR/snapd/snapd" "$LIBEXEC_DIR/snapd/snap-update-ns"; do
         check_file "$p" "$core/usr/lib/snapd/$(basename "$p")"
     done
     for p in /usr/bin/snapctl /usr/bin/snap; do
@@ -193,6 +210,56 @@ update_core_snap_for_classic_reexec() {
     done
 }
 
+prepare_memory_limit_override() {
+    # set up memory limits for snapd bu default unless explicit requested not to
+    # or the system is known to be problematic
+    local set_limit=1
+
+    case "$SPREAD_SYSTEM" in
+        ubuntu-core-16-*|ubuntu-core-18-*|ubuntu-16.04-*|ubuntu-18.04-*)
+            # the tests on UC16, UC18 and correspondingly 16.04 and 18.04 have
+            # demonstrated that the memory limit state claimed by systemd may be
+            # out of sync with actual memory controller setting for the
+            # snapd.service cgroup
+            set_limit=0
+            ;;
+        amazon-linux-*)
+            # similar issues have been observed on Amazon Linux 2
+            set_limit=0
+            ;;
+        *)
+            if [ -n "${SNAPD_NO_MEMORY_LIMIT:-}" ]; then
+                set_limit=0
+            fi
+            ;;
+    esac
+
+    if [ "$set_limit" = "0" ]; then
+        # make sure the file does not exist then
+        rm -f /etc/systemd/system/snapd.service.d/memory-max.conf
+    else
+        mkdir -p /etc/systemd/system/snapd.service.d
+        # Use MemoryMax to set the memory limit for snapd.service, that is the
+        # main snapd process and its subprocesses executing within the same
+        # cgroup. If snapd hits the memory limit, it will get killed by
+        # oom-killer which will be caught in restore_project_each in
+        # prepare-restore.sh.
+        #
+        # This ought to set MemoryMax, but on systems with older systemd we need to
+        # use MemoryLimit, which is deprecated and replaced by MemoryMax now, but
+        # systemd is backwards compatible so the limit is still set.
+        cat <<EOF > /etc/systemd/system/snapd.service.d/memory-max.conf
+[Service]
+# mvo: disabled because of many failures in restore, e.g. in PR#11014
+#MemoryLimit=100M
+EOF
+    fi
+    # the service setting may have changed in the service so we need
+    # to ensure snapd is reloaded
+    systemctl daemon-reload
+    systemctl restart snapd
+}
+
 prepare_each_classic() {
     mkdir -p /etc/systemd/system/snapd.service.d
     if [ -z "${SNAP_REEXEC:-}" ]; then
@@ -276,11 +343,9 @@ prepare_classic() {
 
         # Cache snaps
         # shellcheck disable=SC2086
-        cache_snaps core ${PRE_CACHE_SNAPS}
-
-        echo "Cache the snaps profiler snap"
-        if [ "$PROFILE_SNAPS" = 1 ]; then
-            cache_snaps test-snapd-profiler
+        cache_snaps core core18 ${PRE_CACHE_SNAPS}
+        if os.query is-pc-amd64; then
+            cache_snaps core20
         fi
 
         # now use parameterized core channel (defaults to edge) instead
@@ -300,15 +365,19 @@ prepare_classic() {
 
         disable_refreshes
 
-        echo "Ensure that the bootloader environment output does not contain any of the snap_* variables on classic"
-        # shellcheck disable=SC2119
-        output=$("$TESTSTOOLS"/boot-state bootenv show)
-        if echo "$output" | MATCH snap_ ; then
-            echo "Expected bootloader environment without snap_*, got:"
-            echo "$output"
-            exit 1
+        # Check bootloader environment output in architectures different to s390x which uses zIPL
+        if ! [ "$(uname  -m)" = "s390x" ]; then
+            echo "Ensure that the bootloader environment output does not contain any of the snap_* variables on classic"
+            # shellcheck disable=SC2119
+            output=$("$TESTSTOOLS"/boot-state bootenv show)
+            if echo "$output" | MATCH snap_ ; then
+                echo "Expected bootloader environment without snap_*, got:"
+                echo "$output"
+                exit 1
+            fi
         fi
 
+        setup_experimental_features
         systemctl stop snapd.{service,socket}
         save_snapd_state
         systemctl start snapd.socket
@@ -363,11 +432,34 @@ StandardOutput=journal+console
 StandardError=journal+console
 EOF
 
+    cp "${SPREAD_PATH}"/data/completion/bash/complete.sh "${UNPACK_DIR}"/usr/lib/snapd/complete.sh
+
     snap pack --filename="$TARGET" "$UNPACK_DIR"
 
     rm -rf "$UNPACK_DIR"
 }
 
+repack_kernel_snap() {
+    local TARGET=$1
+    local VERSION
+    local UNPACK_DIR
+    local CHANNEL
+
+    VERSION=$(nested_get_version)
+    if [ "$VERSION" = 16 ]; then
+        CHANNEL=latest
+    else
+        CHANNEL=$VERSION
+    fi
+
+    echo "Repacking kernel snap"
+    UNPACK_DIR=/tmp/kernel-unpack
+    snap download --basename=pc-kernel --channel="$CHANNEL/edge" pc-kernel
+    unsquashfs -no-progress -d "$UNPACK_DIR" pc-kernel.snap
+    snap pack --filename="$TARGET" "$UNPACK_DIR"
+
+    rm -rf pc-kernel.snap "$UNPACK_DIR"
+}
 
 repack_snapd_snap_with_deb_content_and_run_mode_firstboot_tweaks() {
     local TARGET="$1"
@@ -375,6 +467,10 @@ repack_snapd_snap_with_deb_content_and_run_mode_firstboot_tweaks() {
     local UNPACK_DIR="/tmp/snapd-unpack"
     unsquashfs -no-progress -d "$UNPACK_DIR" snapd_*.snap
 
+    # data/preseed.json is not included in the deb, use the latest
+    # version from source tree to replace the one in the re-packed snapd snap.
+    cp "$PROJECT_PATH/data/preseed.json" "$UNPACK_DIR"/usr/lib/snapd
+
     # clean snap apparmor.d to ensure we put the right snap-confine apparmor
     # file in place. Its called usr.lib.snapd.snap-confine on 14.04 but
     # usr.lib.snapd.snap-confine.real everywhere else
@@ -451,34 +547,91 @@ touch /root/spread-setup-done
 EOF
     chmod 0755 "$UNPACK_DIR"/usr/lib/snapd/snapd.spread-tests-run-mode-tweaks.sh
 
+    cp "${SPREAD_PATH}"/data/completion/bash/complete.sh "${UNPACK_DIR}"/usr/lib/snapd/complete.sh
+
     snap pack "$UNPACK_DIR" "$TARGET"
     rm -rf "$UNPACK_DIR"
 }
 
+# Builds kernel snap with bad kernel.efi, in different ways
+# $1: snap we will modify
+# $2: target folder for the new snap
+# $3: argument, type of corruption we want for kernel.efi
+uc20_build_corrupt_kernel_snap() {
+    local ORIG_SNAP="$1"
+    local TARGET_DIR="$2"
+    local optArg=${3:-}
+
+    # kernel snap is huge, unpacking to current dir
+    local REPACKED_DIR=repacked-kernel
+    local KERNEL_EFI_PATH=$REPACKED_DIR/kernel.efi
+    unsquashfs -d "$REPACKED_DIR" "$ORIG_SNAP"
+
+    case "$optArg" in
+        --empty)
+            printf "" > "$KERNEL_EFI_PATH"
+            ;;
+        --zeros)
+            dd if=/dev/zero of="$KERNEL_EFI_PATH" count=1
+            ;;
+        --bad-*)
+            section=${optArg#--bad-}
+            # Get the file offset for the section, put zeros at the beginning of it
+            sectOffset=$(objdump -w -h "$KERNEL_EFI_PATH" | grep "$section" |
+                             awk '{print $6}')
+            dd if=/dev/zero of="$KERNEL_EFI_PATH" \
+               bs=1 seek=$((0x$sectOffset)) count=512 conv=notrunc
+            ;;
+    esac
+
+    # Make snap smaller, we don't need the fw with qemu
+    rm -rf "$REPACKED_DIR"/firmware/*
+    snap pack "$REPACKED_DIR" "$TARGET_DIR"
+    rm -rf "$REPACKED_DIR"
+}
 
 uc20_build_initramfs_kernel_snap() {
     # carries ubuntu-core-initframfs
     add-apt-repository ppa:snappy-dev/image -y
-    apt install ubuntu-core-initramfs -y
+    # On focal, lvm2 does not reinstall properly after being removed.
+    # So we need to clean up in case the VM has been re-used.
+    if os.query is-focal; then
+        systemctl unmask lvm2-lvmpolld.socket
+    fi
+    # TODO: install the linux-firmware as the current version of
+    # ubuntu-core-initramfs does not depend on it, but nonetheless requires it
+    # to build the initrd
+    apt install ubuntu-core-initramfs linux-firmware -y
 
     local ORIG_SNAP="$1"
     local TARGET="$2"
 
+    # TODO proper option support here would be nice but bash is hard and this is
+    # easier, and likely we won't need to both inject a panic and set the epoch
+    # bump simultaneously
     local injectKernelPanic=false
-    injectKernelPanicArg=${3:-}
-    if [ "$injectKernelPanicArg" = "--inject-kernel-panic-in-initramfs" ]; then
-        injectKernelPanic=true
-    fi
+    local initramfsEpochBumpTime
+    initramfsEpochBumpTime=$(date '+%s')
+    optArg=${3:-}
+    case "$optArg" in
+        --inject-kernel-panic-in-initramfs)
+            injectKernelPanic=true
+            ;;
+        --epoch-bump-time=*)
+            # this strips the option and just gives us the value
+            initramfsEpochBumpTime="${optArg#--epoch-bump-time=}"
+            ;;
+    esac
     
     # kernel snap is huge, unpacking to current dir
     unsquashfs -d repacked-kernel "$ORIG_SNAP"
 
-
     # repack initrd magic, beware
     # assumptions: initrd is compressed with LZ4, cpio block size 512, microcode
     # at the beginning of initrd image
     (
         cd repacked-kernel
+        unpackeddir="$PWD"
         #shellcheck disable=SC2010
         kver=$(ls "config"-* | grep -Po 'config-\K.*')
 
@@ -497,19 +650,40 @@ uc20_build_initramfs_kernel_snap() {
         cp -ar unpacked-initrd skeleton
         # all the skeleton edits go to a local copy of distro directory
         skeletondir=$PWD/skeleton
-        cp -a /usr/lib/snapd/snap-bootstrap "$skeletondir/main/usr/lib/snapd/snap-bootstrap"
-        # modify the-tool to verify that our version is used when booting - this
-        # is verified in the tests/core/basic20 spread test
-        sed -i -e 's/set -e/set -ex/' "$skeletondir/main/usr/lib/the-tool"
-        echo "" >> "$skeletondir/main/usr/lib/the-tool"
-        echo "if test -d /run/mnt/data/system-data; then touch /run/mnt/data/system-data/the-tool-ran; fi" >> \
-            "$skeletondir/main/usr/lib/the-tool"
+        cp -a /usr/lib/snapd/snap-bootstrap "$skeletondir/main/usr/lib/snapd/snap-bootstrap.real"
+        cat <<'EOF' | sed -E "s/^ {8}//" >"$skeletondir/main/usr/lib/snapd/snap-bootstrap"
+        #!/bin/sh
+        set -eux
+        if [ "$1" != initramfs-mounts ]; then
+            exec /usr/lib/snapd/snap-bootstrap.real "$@"
+        fi
+        beforeDate="$(date --utc '+%s')"
+        /usr/lib/snapd/snap-bootstrap.real "$@"
+        if [ -d /run/mnt/data/system-data ]; then
+            touch /run/mnt/data/system-data/the-tool-ran
+        fi
+        # also copy the time for the clock-epoch to system-data, this is
+        # used by a specific test but doesn't hurt anything to do this for
+        # all tests
+        mode="$(grep -Eo 'snapd_recovery_mode=([a-z]+)' /proc/cmdline)"
+        mode=${mode##snapd_recovery_mode=}
+        mkdir -p /run/mnt/ubuntu-seed/test
+        stat -c '%Y' /usr/lib/clock-epoch >> /run/mnt/ubuntu-seed/test/${mode}-clock-epoch
+        echo "$beforeDate" > /run/mnt/ubuntu-seed/test/${mode}-before-snap-bootstrap-date
+        date --utc '+%s' > /run/mnt/ubuntu-seed/test/${mode}-after-snap-bootstrap-date
+EOF
+
+        chmod +x "$skeletondir/main/usr/lib/snapd/snap-bootstrap"
 
         if [ "$injectKernelPanic" = "true" ]; then
             # add a kernel panic to the end of the-tool execution
-            echo "echo 'forcibly panicing'; echo c > /proc/sysrq-trigger" >> "$skeletondir/main/usr/lib/the-tool"
+            echo "echo 'forcibly panicing'; echo c > /proc/sysrq-trigger" >> "$skeletondir/main/usr/lib/snapd/snap-bootstrap"
         fi
 
+        # bump the epoch time file timestamp, converting unix timestamp to 
+        # touch's date format
+        touch -t "$(date --utc "--date=@$initramfsEpochBumpTime" '+%Y%m%d%H%M')" "$skeletondir/main/usr/lib/clock-epoch"
+
         # copy any extra files to the same location inside the initrd
         if [ -d ../extra-initrd/ ]; then
             cp -a ../extra-initrd/* "$skeletondir"/main
@@ -522,10 +696,14 @@ uc20_build_initramfs_kernel_snap() {
             # accommodate assumptions about tree layout, use the unpacked initrd
             # to pick up the right modules
             cd unpacked-initrd/main
+            # XXX: pass feature 'main' and u-c-i picks up any directory named
+            # after feature inside skeletondir and uses that a template
             ubuntu-core-initramfs create-initrd \
                                   --kernelver "$kver" \
                                   --skeleton "$skeletondir" \
-                                  --kerneldir "lib/modules" \
+                                  --kerneldir "lib/modules/$kver" \
+                                  --firmwaredir "$unpackeddir/firmware" \
+                                  --feature 'main' \
                                   --output ../../repacked-initrd
         )
 
@@ -729,12 +907,14 @@ setup_reflash_magic() {
         snap download "--channel=${SNAPD_CHANNEL}" snapd
     fi
 
-    # we cannot use "names.sh" here because no snaps are installed yet
+    # we cannot use "snaps.names tool" here because no snaps are installed yet
     core_name="core"
     if os.query is-core18; then
         core_name="core18"
     elif os.query is-core20; then
         core_name="core20"
+    elif os.query is-core22; then
+        core_name="core22"
     fi
     # XXX: we get "error: too early for operation, device not yet
     # seeded or device model not acknowledged" here sometimes. To
@@ -745,17 +925,23 @@ setup_reflash_magic() {
     snap model --verbose
     # remove the above debug lines once the mentioned bug is fixed
     snap install "--channel=${CORE_CHANNEL}" "$core_name"
-    if os.query is-core16 || os.query is-core18; then
-        UNPACK_DIR="/tmp/$core_name-snap"
-        unsquashfs -no-progress -d "$UNPACK_DIR" /var/lib/snapd/snaps/${core_name}_*.snap
-    fi
+    UNPACK_DIR="/tmp/$core_name-snap"
+    unsquashfs -no-progress -d "$UNPACK_DIR" /var/lib/snapd/snaps/${core_name}_*.snap
 
-    # install ubuntu-image
-    snap install --classic --edge ubuntu-image
+    if os.query is-core16; then
+        # the new ubuntu-image expects mkfs to support -d option, which was not
+        # supported yet by the version of mkfs that shipped with Ubuntu 16.04
+        snap install ubuntu-image --channel="$UBUNTU_IMAGE_SNAP_CHANNEL" --classic
+    else
+        # shellcheck source=tests/lib/image.sh
+        . "$TESTSLIB/image.sh"
+        get_ubuntu_image
+    fi
 
     # needs to be under /home because ubuntu-device-flash
     # uses snap-confine and that will hide parts of the hostfs
     IMAGE_HOME=/home/image
+    IMAGE=pc.img
     mkdir -p "$IMAGE_HOME"
 
     # ensure that ubuntu-image is using our test-build of snapd with the
@@ -769,11 +955,12 @@ setup_reflash_magic() {
         repack_snapd_snap_with_deb_content "$IMAGE_HOME"
         # FIXME: fetch directly once its in the assertion service
         cp "$TESTSLIB/assertions/ubuntu-core-18-amd64.model" "$IMAGE_HOME/pc.model"
-        IMAGE=core18-amd64.img
     elif os.query is-core20; then
         repack_snapd_snap_with_deb_content_and_run_mode_firstboot_tweaks "$IMAGE_HOME"
         cp "$TESTSLIB/assertions/ubuntu-core-20-amd64.model" "$IMAGE_HOME/pc.model"
-        IMAGE=core20-amd64.img
+    elif os.query is-core22; then
+        repack_snapd_snap_with_deb_content_and_run_mode_firstboot_tweaks "$IMAGE_HOME"
+        cp "$TESTSLIB/assertions/ubuntu-core-22-amd64.model" "$IMAGE_HOME/pc.model"
     else
         # FIXME: install would be better but we don't have dpkg on
         #        the image
@@ -809,9 +996,6 @@ EOF
 
         # FIXME: fetch directly once its in the assertion service
         cp "$TESTSLIB/assertions/pc-${REMOTE_STORE}.model" "$IMAGE_HOME/pc.model"
-
-        # FIXME: how to test store updated of ubuntu-core with sideloaded snap?
-        IMAGE=all-snap-amd64.img
     fi
 
     EXTRA_FUNDAMENTAL=
@@ -824,17 +1008,68 @@ EOF
         # need to download it
         snap download --channel="$KERNEL_CHANNEL" pc-kernel
 
-        EXTRA_FUNDAMENTAL="--extra-snaps $PWD/pc-kernel_*.snap"
+        EXTRA_FUNDAMENTAL="--snap $PWD/pc-kernel_*.snap"
         IMAGE_CHANNEL="$GADGET_CHANNEL"
     fi
 
-    if os.query is-core20; then
-        snap download --basename=pc-kernel --channel="20/$KERNEL_CHANNEL" pc-kernel
+    if os.query is-core20 || os.query is-core22; then
+        if os.query is-core20; then
+            BRANCH=20
+        elif os.query is-core22; then
+            BRANCH=22
+        fi
+        snap download --basename=pc-kernel --channel="${BRANCH}/${KERNEL_CHANNEL}" pc-kernel
         # make sure we have the snap
         test -e pc-kernel.snap
         # build the initramfs with our snapd assets into the kernel snap
         uc20_build_initramfs_kernel_snap "$PWD/pc-kernel.snap" "$IMAGE_HOME"
-        EXTRA_FUNDAMENTAL="--extra-snaps $IMAGE_HOME/pc-kernel_*.snap"
+        EXTRA_FUNDAMENTAL="--snap $IMAGE_HOME/pc-kernel_*.snap"
+
+        # also add debug command line parameters to the kernel command line via
+        # the gadget in case things go side ways and we need to debug
+        snap download --basename=pc --channel="${BRANCH}/${KERNEL_CHANNEL}" pc
+        test -e pc.snap
+        unsquashfs -d pc-gadget pc.snap
+        
+        # TODO: it would be desirable when we need to do in-depth debugging of
+        # UC20 runs in google to have snapd.debug=1 always on the kernel command
+        # line, but we can't do this universally because the logic for the env
+        # variable SNAPD_DEBUG=0|false does not overwrite the turning on of 
+        # debug messages in some places when the kernel command line is set, so
+        # we get failing tests since there is extra stuff on stderr than 
+        # expected in the test when SNAPD_DEBUG is turned off
+        # so for now, don't include snapd.debug=1, but eventually it would be
+        # nice to have this on
+
+        if [ "$SPREAD_BACKEND" = "google" ]; then
+            # the default console settings for snapd aren't super useful in GCE,
+            # instead it's more useful to have all console go to ttyS0 which we 
+            # can read more easily than tty1 for example
+            for cmd in "console=ttyS0" "dangerous" "systemd.journald.forward_to_console=1" "rd.systemd.journald.forward_to_console=1" "panic=-1"; do
+                echo "$cmd" >> pc-gadget/cmdline.full
+            done
+        else
+            # but for other backends, just add the additional debugging things
+            # on top of whatever the gadget currently is configured to use
+            for cmd in "dangerous" "systemd.journald.forward_to_console=1" "rd.systemd.journald.forward_to_console=1"; do
+                echo "$cmd" >> pc-gadget/cmdline.extra
+            done
+        fi
+
+        # TODO: this probably means it's time to move this helper out of 
+        # nested.sh to somewhere more general
+        
+        #shellcheck source=tests/lib/nested.sh
+        . "$TESTSLIB/nested.sh"
+        KEY_NAME=$(nested_get_snakeoil_key)
+
+        SNAKEOIL_KEY="$PWD/$KEY_NAME.key"
+        SNAKEOIL_CERT="$PWD/$KEY_NAME.pem"
+
+        nested_secboot_sign_gadget pc-gadget "$SNAKEOIL_KEY" "$SNAKEOIL_CERT"
+        snap pack --filename=pc-repacked.snap pc-gadget 
+        mv pc-repacked.snap $IMAGE_HOME/pc-repacked.snap
+        EXTRA_FUNDAMENTAL="$EXTRA_FUNDAMENTAL --snap $IMAGE_HOME/pc-repacked.snap"
     fi
 
     # 'snap pack' creates snaps 0644, and ubuntu-image just copies those in
@@ -843,7 +1078,7 @@ EOF
 
     # on core18 we need to use the modified snapd snap and on core16
     # it is the modified core that contains our freshly build snapd
-    if os.query is-core18 || os.query is-core20; then
+    if os.query is-core18 || os.query is-core20 || os.query is-core22; then
         extra_snap=("$IMAGE_HOME"/snapd_*.snap)
     else
         extra_snap=("$IMAGE_HOME"/core_*.snap)
@@ -855,14 +1090,55 @@ EOF
         exit 1
     fi
 
-    /snap/bin/ubuntu-image -w "$IMAGE_HOME" "$IMAGE_HOME/pc.model" \
-                           --channel "$IMAGE_CHANNEL" \
-                           "$EXTRA_FUNDAMENTAL" \
-                           --extra-snaps "${extra_snap[0]}" \
-                           --output "$IMAGE_HOME/$IMAGE"
-    rm -f ./pc-kernel_*.{snap,assert} ./pc_*.{snap,assert} ./snapd_*.{snap,assert}
+    # download the core20 snap manually from the specified channel for UC20
+    if os.query is-core20 || os.query is-core22; then
+        if os.query is-core20; then
+            BASE=core20
+        elif os.query is-core22; then
+            BASE=core22
+        fi
+        snap download "${BASE}" --channel="$BASE_CHANNEL" --basename="${BASE}"
+        
+        # we want to download the specific channel referenced by $BASE_CHANNEL, 
+        # but if we just seed that revision and $BASE_CHANNEL != $IMAGE_CHANNEL,
+        # then immediately on booting, snapd will refresh from the revision that
+        # is seeded via $BASE_CHANNEL to the revision that is in $IMAGE_CHANNEL,
+        # so to prevent that from happening (since that automatic refresh will 
+        # confuse spread and make tests fail in awkward, confusing ways), we
+        # unpack the snap and re-pack it so that it is not asserted and thus 
+        # won't be automatically refreshed
+        # note that this means that when $IMAGE_CHANNEL != $BASE_CHANNEL, we
+        # will have unasserted snaps for all snaps on UC20 in GCE spread:
+        # * snapd (to test the branch)
+        # * pc-kernel (to test snap-bootstrap from the branch)
+        # * pc (to aid in debugging by modifying the kernel command line)
+        # * core20 (to avoid the automatic refresh issue)
+        if [ "$IMAGE_CHANNEL" != "$BASE_CHANNEL" ]; then
+            unsquashfs -d "${BASE}-snap" "${BASE}.snap"
+            snap pack --filename="${BASE}-repacked.snap" "${BASE}-snap"
+            rm -r "${BASE}-snap"
+            mv "${BASE}-repacked.snap" "${IMAGE_HOME}/${BASE}.snap"
+        else 
+            mv "${BASE}.snap" "${IMAGE_HOME}/${BASE}.snap"
+        fi
+        
+        EXTRA_FUNDAMENTAL="$EXTRA_FUNDAMENTAL --snap ${IMAGE_HOME}/${BASE}.snap"
+    fi
+    local UBUNTU_IMAGE="$GOHOME"/bin/ubuntu-image
+    if os.query is-core16; then
+        # ubuntu-image on 16.04 needs to be installed from a snap
+        UBUNTU_IMAGE=/snap/bin/ubuntu-image
+    fi
+    # shellcheck disable=SC2086
+    "$UBUNTU_IMAGE" snap \
+                    -w "$IMAGE_HOME" "$IMAGE_HOME/pc.model" \
+                    --channel "$IMAGE_CHANNEL" \
+                    $EXTRA_FUNDAMENTAL \
+                    --snap "${extra_snap[0]}" \
+                    --output-dir "$IMAGE_HOME"
+    rm -f ./pc-kernel_*.{snap,assert} ./pc-kernel.{snap,assert} ./pc_*.{snap,assert} ./snapd_*.{snap,assert} ./core{20,22}.{snap,assert}
 
-    if os.query is-core20; then
+    if os.query is-core20 || os.query is-core22; then
         # (ab)use ubuntu-seed
         LOOP_PARTITION=2
     else
@@ -872,7 +1148,7 @@ EOF
     # expand the uc16 and uc18 images a little bit (400M) as it currently will
     # run out of space easily from local spread runs if there are extra files in
     # the project not included in the git ignore and spread ignore, etc.
-    if ! os.query is-core20; then
+    if ! (os.query is-core20 || os.query is-core22); then
         # grow the image by 400M
         truncate --size=+400M "$IMAGE_HOME/$IMAGE"
         # fix the GPT table because old versions of parted complain about this 
@@ -896,20 +1172,20 @@ EOF
     dev=$(basename "$devloop")
 
     # resize the 2nd partition from that loop device to fix the size
-    if ! os.query is-core20; then
+    if ! (os.query is-core20 || os.query is-core22); then
         resize2fs -p "/dev/mapper/${dev}p${LOOP_PARTITION}"
     fi
 
     # mount it so we can use it now
     mount "/dev/mapper/${dev}p${LOOP_PARTITION}" /mnt
 
-    mkdir -p /mnt/user-data/
     # copy over everything from gopath to user-data, exclude:
     # - VCS files
     # - built debs
     # - golang archive files and built packages dir
     # - govendor .cache directory and the binary,
     if os.query is-core16 || os.query is-core18; then
+        mkdir -p /mnt/user-data/
         # we need to include "core" here because -C option says to ignore 
         # files the way CVS(?!) does, so it ignores files named "core" which
         # are core dumps, but we have a test suite named "core", so including 
@@ -923,15 +1199,21 @@ EOF
           --exclude /gopath/pkg/ \
           --include core/ \
           /home/gopath /mnt/user-data/
-    elif os.query is-core20; then
+    elif os.query is-core20 || os.query is-core22; then
         # prepare passwd for run-mode-overlay-data
+
+        # use /etc/{group,passwd,shadow,gshadow} from the core20 snap, merged
+        # with some bits from our current system - we don't want to use the
+        # /etc/group from the current system as classic and core gids and uids
+        # don't match, but we still need the same test/ubuntu/root user info
+        # in core as we currently have in classic
         mkdir -p /root/test-etc
         mkdir -p /var/lib/extrausers
         touch /var/lib/extrausers/sub{uid,gid}
         for f in group gshadow passwd shadow; do
-            grep -v "^root:" /etc/"$f" > /root/test-etc/"$f"
+            grep -v "^root:" "$UNPACK_DIR/etc/$f" > /root/test-etc/"$f"
             grep "^root:" /etc/"$f" >> /root/test-etc/"$f"
-            chgrp --reference /etc/"$f" /root/test-etc/"$f"
+            chgrp --reference "$UNPACK_DIR/etc/$f" /root/test-etc/"$f"
             # create /var/lib/extrausers/$f
             # append ubuntu, test user for the testing
             grep "^test:" /etc/"$f" >> /var/lib/extrausers/"$f"
@@ -940,8 +1222,9 @@ EOF
             MATCH "^test:" </var/lib/extrausers/"$f"
             MATCH "^ubuntu:" </var/lib/extrausers/"$f"
         done
-        # Make sure systemd-journal group has the "test" user as a member. Due to the way we copy that from the host
-        # and merge it from the core snap this is done explicitly as a second step.
+        # Make sure systemd-journal group has the "test" user as a member. Due
+        # to the way we copy that from the host and merge it from the core snap
+        # this is done explicitly as a second step.
         sed -r -i -e 's/^systemd-journal:x:([0-9]+):$/systemd-journal:x:\1:test/' /root/test-etc/group
         tar -c -z \
           --exclude '*.a' \
@@ -967,25 +1250,39 @@ EOF
     # the reflash magic
     # FIXME: ideally in initrd, but this is good enough for now
     cat > "$IMAGE_HOME/reflash.sh" << EOF
-#!/bin/sh -ex
-mount -t tmpfs none /tmp
-cp /bin/busybox /tmp
-cp $IMAGE_HOME/$IMAGE /tmp
-sync
+#!/tmp/busybox sh
+set -e
+set -x
+
 # blow away everything
 OF=/dev/sda
 if [ -e /dev/vda ]; then
     OF=/dev/vda
 fi
-/tmp/busybox dd if=/tmp/$IMAGE of=\$OF bs=4M
+dd if=/tmp/$IMAGE of=\$OF bs=4M
 # and reboot
-/tmp/busybox sync
-/tmp/busybox echo b > /proc/sysrq-trigger
+sync
+echo b > /proc/sysrq-trigger
+
+EOF
+
+    cat > "$IMAGE_HOME/prep-reflash.sh" << EOF
+#!/bin/sh -ex
+mount -t tmpfs none /tmp
+cp /bin/busybox /tmp
+cp $IMAGE_HOME/reflash.sh /tmp
+cp $IMAGE_HOME/$IMAGE /tmp
+sync
+
+# re-exec using busybox from /tmp
+exec /tmp/reflash.sh
+
 EOF
     chmod +x "$IMAGE_HOME/reflash.sh"
+    chmod +x "$IMAGE_HOME/prep-reflash.sh"
 
     DEVPREFIX=""
-    if os.query is-core20; then
+    if os.query is-core20 || os.query is-core22; then
         DEVPREFIX="/boot"
     fi
     # extract ROOT from /proc/cmdline
@@ -994,7 +1291,7 @@ EOF
 set default=0
 set timeout=2
 menuentry 'flash-all-snaps' {
-linux $DEVPREFIX/vmlinuz root=$ROOT ro init=$IMAGE_HOME/reflash.sh console=ttyS0
+linux $DEVPREFIX/vmlinuz root=$ROOT ro init=$IMAGE_HOME/prep-reflash.sh console=tty1 console=ttyS0
 initrd $DEVPREFIX/initrd.img
 }
 EOF
@@ -1030,9 +1327,7 @@ prepare_ubuntu_core() {
     snap wait system seed.loaded
 
     echo "Ensure fundamental snaps are still present"
-    # shellcheck source=tests/lib/names.sh
-    . "$TESTSLIB/names.sh"
-    for name in "$gadget_name" "$kernel_name" "$core_name"; do
+    for name in "$(snaps.name gadget)" "$(snaps.name kernel)" "$(snaps.name core)"; do
         if ! snap list "$name"; then
             echo "Not all fundamental snaps are available, all-snap image not valid"
             echo "Currently installed snaps"
@@ -1042,7 +1337,7 @@ prepare_ubuntu_core() {
     done
 
     echo "Ensure the snapd snap is available"
-    if os.query is-core18 || os.query is-core20; then
+    if os.query is-core18 || os.query is-core20 || os.query is-core22; then
         if ! snap list snapd; then
             echo "snapd snap on core18 is missing"
             snap list
@@ -1057,6 +1352,8 @@ prepare_ubuntu_core() {
             rsync_snap="test-snapd-rsync-core18"
         elif os.query is-core20; then
             rsync_snap="test-snapd-rsync-core20"
+        elif os.query is-core22; then
+            rsync_snap="test-snapd-rsync-core22"
         fi
         snap install --devmode --edge "$rsync_snap"
         snap alias "$rsync_snap".rsync rsync
@@ -1068,7 +1365,7 @@ prepare_ubuntu_core() {
 
     echo "Ensure the core snap is cached"
     # Cache snaps
-    if os.query is-core18 || os.query is-core20; then
+    if os.query is-core18 || os.query is-core20 || os.query is-core22; then
         if snap list core >& /dev/null; then
             echo "core snap on core18 should not be installed yet"
             snap list
@@ -1078,14 +1375,11 @@ prepare_ubuntu_core() {
         if os.query is-core18; then
             cache_snaps test-snapd-sh-core18
         fi
-    fi
-
-    echo "Cache the snaps profiler snap"
-    if [ "$PROFILE_SNAPS" = 1 ]; then
-        if os.query is-core18; then
-            cache_snaps test-snapd-profiler-core18
-        else
-            cache_snaps test-snapd-profiler
+        if os.query is-core20; then
+            cache_snaps test-snapd-sh-core20
+        fi
+        if os.query is-core22; then
+            cache_snaps test-snapd-sh-core22
         fi
     fi
 
@@ -1094,6 +1388,11 @@ prepare_ubuntu_core() {
 
     # Snapshot the fresh state (including boot/bootenv)
     if ! is_snapd_state_saved; then
+
+        # important to remove disabled snaps before calling save_snapd_state
+        # or restore will break
+        remove_disabled_snaps
+        setup_experimental_features
         systemctl stop snapd.service snapd.socket
         save_snapd_state
         systemctl start snapd.socket
